
/* PilotDoc.class
Copyright (C) 2002  Alexander Gross

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package superwaba.ext.palm.io.doc;
import waba.io.*;
import waba.sys.*;
import waba.util.*;

/**
  * This class is intended for use on a Palm device to access documents in
  * the common zDoc format. Many DOC Readers supports this format. This is
  * a standart implementation made with the zDoc format description available
  * a www.pyrite.org .
  * Be aware to use a document utility like Pyrite which makes documents without
  * any extensions. There are also some document utitlities which does not set
  * the document headers correct. I recommend the use of Pyrite.
  *
  * Use of the class
  *
  * <pre>
  * Catalog catalog = new Catalog("TestDoc.REAd.TEXt", Catalog.READ_WRITE);
  * PilotDoc    doc = new PilotDoc(catalog);
  * // read the header of the doc
  * doc.readHeader();
  * // now decode all the records into a version 1 document
  * for (int i = 1; i < doc.getNumRecords(); i++) {
  * 	doc.readRecord(i);
  *	doc.setVersion(1);
  * 	doc.writeRecord(i);
  *	doc.setVersion(2);
  * }
  *
  * doc.setVersion(1);
  * doc.writeHeader();
  * </pre>
  *
  * @author Alexander Gro <a.gross@graumueller.com>
  * @version 0.3
  */

public class PilotDoc
{
   public static final int DOC_READ_WRITE_ERROR = -1;
   public static final int DOC_VERSION_ERROR = -2;
   public static final int DOC_NO_CONTENT_ERROR = -3;
   public static final int DOC_BUFFER_ERROR = -4;
   public static final int DOC_UNEXPECTED_ERROR = -9;
   private PilotDocHeader header = new PilotDocHeader();
   private PilotDocRecord record = new PilotDocRecord();

   /**
     * Hastable which contains the bookmark records after a
     * readBookmark() call. You can alter this Hashtable direct
     * and write it back with writeBookmark().
     */
   public waba.util.Hashtable bookmark = new Hashtable(1);
   private Catalog catalog;
   private ResizeStream rstream;
   private DataStream dstream;
   private String content = new String();

   /**
     * Constructs a PilotDoc Object
     * @param c a open catalog to the database
     */
   public PilotDoc(Catalog c)
   {
      catalog = c;
      rstream = new ResizeStream(c, 4096);
      dstream = new DataStream(rstream);
   }

   /**
     * Get the value of version
     * @return version
     */
   public int getVersion()
   {
      return header.version;
   }

   /**
     * Get the value of spare
     * @return spare
     */
   public int getSpare()
   {
      return header.spare;
   }

   /**
     * Get the value of storyLen
     * @return storyLen
     */
   public int getStoryLen()
   {
      return header.storyLen;
   }

   /**
     * Get the value of recordSize
     * @return recordSize
     */
   public int getRecordSize()
   {
      return header.recordSize;
   }

   /**
     * Get the value of numRecords
     * @return numRecords
     */
   public int getNumRecords()
   {
      return header.numRecords;
   }

   /**
     * Get the value of position
     * @return position
     */
   public int getPosition()
   {
      return header.position;
   }

   /**
     * Set the value of version
     * @param version
     */
   public void setVersion(int param)
   {
      header.version = param;
   }

   /**
     * Set the value of spare
     * @param spare
     */
   public void setSpare(int param)
   {
      header.spare = param;
   }

   /**
     * Set the value of storyLen
     * @param storyLen
     */
   public void setStoryLen(int param)
   {
      header.storyLen = param;
   }

   /**
     * Set the value of numRecords
     * @param numRecords
     */
   public void setNumRecords(int param)
   {
      header.numRecords = param;
   }

   /**
     * Sets the value of recordSize
     * @param recordSize
     */
   public void setRecordSize(int param)
   {
      header.recordSize = param;
   }

   /**
     * Set the value of position
     * @param position
     */
   public void setPosition(int param)
   {
      header.position = param;
   }

   /**
     * Reads the header of the PilotDoc
     * @return the bytes read or, if negative, a read error
     *         PilotDoc.DOC_XXXX
     */
   public int readHeader()
   {
      catalog.setRecordPos(0);
      header.read(dstream);
      if((header.version < 1) || (header.version > 2))
         return DOC_VERSION_ERROR;
      else
         return 16;
   }

   /**
     * Write the header of the PilotDoc
     * @return the bytes written or, if negative, a read error
     *         PilotDoc.DOC_XXXX
     */
   public int writeHeader()
   {
      if(catalog.setRecordPos(0))
         catalog.setRecordPos(0);
      else
      {
         catalog.addRecord(16, 0);
         catalog.setRecordPos(0);
      }
      int written = header.write(dstream);
      if(written == 0)
         return DOC_READ_WRITE_ERROR;
      else
         return written;
   }

   /**
     * Reads the content of a record
     * @return the bytes read or, if negative, a read error
     *         PilotDoc.DOC_XXXX
     */
   public int readRecord(int rec)
   {
      byte[] buf;
      catalog.setRecordPos(rec);
      int recSize = catalog.getRecordSize();
      buf = new byte[recSize];
      dstream.readBytes(buf, 0, recSize);
      record.assign(buf, recSize);
      if(header.version == 2)
         record.decompress();
      if(buf.length > 0)
      {
         setContent(new String(record.getBuffer()));
         return record.getBuffer().length;
      }
      else
         if(buf.length == 0)
            return DOC_NO_CONTENT_ERROR;
         else
            return DOC_UNEXPECTED_ERROR;
   }

   /**
     * Writes the content of a record
     * @return the bytes written or, if negative, a write error
     *         PilotDoc.DOC_XXXX
     */
   public int writeRecord(int rec)
   {
      int recSize = header.recordSize;
      byte[] buf = content.getBytes();
      record.assign(buf, buf.length);
      if(header.version == 2)
         record.compress();
      buf = record.getBuffer();
      int written = 0;
      int recLen = buf.length;
      if(catalog.setRecordPos(rec))
      {
         catalog.setRecordPos(rec);
         catalog.resizeRecord(recLen);
         written = dstream.writeBytes(buf, 0, recLen);
      }
      else
      {
         catalog.addRecord(recLen, rec);
         catalog.setRecordPos(rec);
         written = dstream.writeBytes(buf, 0, recLen);
      }
      if(written > 0)
         return written;
      else
         if(written == 0)
            return DOC_NO_CONTENT_ERROR;
         else
            return DOC_UNEXPECTED_ERROR;
   }

   /**
     * Gets the content of the record as String
     * @return content
     */
   public String getContent()
   {
      return new String(record.getBuffer());
   }

   /**
     * Sets the content of the record as String
     * @param content
     * @return bytes in buffer or DOC_BUFFER_ERROR on fail
     *         (maybe if String bigger that recordSize (4096 bytes))
     */
   public int setContent(String s)
   {
      if(s.getBytes().length > header.recordSize)
         return DOC_BUFFER_ERROR;
      else
      {
         content = s;
         byte[] buf = s.getBytes();
         record.assign(buf, buf.length);
         return buf.length;
      }
   }

   /**
     * Gets the content of the record as byte array
     * @return buffer
     */
   public byte[] getBuffer()
   {
      return record.getBuffer();
   }

   /**
     * Reads the bookmark records of the document
     * @return number of bookmarks read
     */
   public int readBookmarks()
   {

      // get the indices of the database
      int index = header.numRecords + 1;
      int last = catalog.getRecordCount();

      // step through the records and read the bookmarks
      if(last > header.numRecords + 1)
      {
         for(int i = index;i < last;i++)
         {
            catalog.setRecordPos(i);
            String nam = dstream.readCString();
            byte[] tmp = new byte[20];
            catalog.setRecordPos(i);
            catalog.readBytes(tmp, 0, 20);

            // compute int from the 4 byte array
            int pos = (tmp[16] * 0x01000000) + (tmp[17] * 0x00010000) + (tmp[18] * 0x00000100) + tmp[19];

            // put name and position in Hashtable
            bookmark.put(nam, Convert.toString(pos));
         }
      }

      // return how many bookmarks read
      return last - index;
   }

   /**
     * Writes the bookmark recordstream of the document
     * @param bookmarks as Hashtable
     * @return bookmarks as Hashtable
     */
   public int writeBookmarks()
   {
      // check if catalog is writable, if not return eroor
      if(catalog.getAttributes() == Catalog.DB_ATTR_READ_ONLY)
         return DOC_READ_WRITE_ERROR;
      else
      {
         // get the indices of the catalog
         int index = header.numRecords + 1;
         int last = catalog.getRecordCount();

         // delete all old records
         if(last > header.numRecords + 1)
         {
            for(int i = index;i < last;i++)
            {
               catalog.setRecordPos(i);
               catalog.deleteRecord();
            }
         }

         // init written bytes integer
         int written = 0;

         // get the keys from the bookmark Hash
         Vector keys = bookmark.getKeys();
         int kcnt = keys.getCount();

         // if there are entries in the Hash
         if(kcnt > 0)
         {

            // step through the records
            for(int i = index;i < index + kcnt;i++)
            {

               // cut the name to 15 bytes
               String nam = (String)keys.get(i - index);
               if(nam.length() > 15)
                  nam = nam.substring(0, 15);

               // init bytes array with null and copy the name
               byte[] tmp = new byte[]
               {
                  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                  0, 0, 0
               };
               Vm.copyArray(nam.getBytes(), 0, tmp, 0, nam.length());

               // set the bytes for position
               tmp[16] = (byte)((Convert.toInt((String)bookmark.get(keys.get(i - index))) >>> 24) & 0x0ff);
               tmp[17] = (byte)((Convert.toInt((String)bookmark.get(keys.get(i - index))) >>> 16) & 0x0ff);
               tmp[18] = (byte)((Convert.toInt((String)bookmark.get(keys.get(i - index))) >>> 8) & 0x0ff);
               tmp[19] = (byte)Convert.toInt((String)bookmark.get(keys.get(i - index)));

               // add a new record and write data
               catalog.addRecord(20, i);
               catalog.setRecordPos(i);
               written = dstream.writeBytes(tmp, 0, 20);
            }

            // return written byte - must be 20
            return written;
         }
         else

            // return 0 if nothing written
            return 0;
      }
   }
}
