package superwaba.ext.xplat.game;

import waba.fx.*;
import waba.ui.*;

/**
 * The animated button control. <br>
 * <pre>
 * This control displays an animated button which can take 'S' different states
 * and each state is fades in or out in 'F' frames. 'S' and 'F' represent the
 * two first constructor arguments. The frames of this special animation have to be
 * ordered to be supported by this class. The states are numbered from 0 to 'S'-1
 * and the frames order is the following depending on the layoutType value:<br>
 * FADE_OUT_LAYOUT    :  S0F0,S0F1,S0F2,S1F0,S1F1,S1F2,S2F0,S2F1,S2F2<br>
 * FADE_IN_LAYOUT     :  S0F2,S0F1,S0F0,S1F2,S1F1,S1F0,S2F2,S2F1,S2F0<br>
 * FADE_OUT_IN_LAYOUT :  S0F0,S0F1,S1F1,S1F0,S1F1,S2F1,S2F0,S2F1,S0F1<br>
 * where S stands for state, F for frame and where S0F0, S1F0 and S2F0
 * are the full states and the others are transition frames.
 * Open the 'onOff.bmp' in the PingGame sample and you will understand ;-)
 * </pre>
 * @author Frank Diebolt
 * @author Guilherme Campos Hazan
 * @version 1.1
 */
public class AnimatedButton extends Animation
{
  /**
   * Defines the frames animation order. In the case of a "S" states button of "F"
   * frames per state, FADE_OUT_LAYOUT means that the frames are a "S" set of
   * "F" frames that are fading out the state, that means the first frame of
   * of each set is the full state image.
   * In the FADE_IN_LAYOUT layout, it's the opposite, namely the last frame of
   * each set represents the state ending position.
   * Finaly the FADE_OUT_IN_LAYOUT is a mix of the two others, because inter-frames
   * represent successively fading out from one state to fading in to next state.
   * @see the game tutorial for samples.
   */
  public static final int FADE_OUT_LAYOUT   = 0;

  /** Frames fading in mode.  @see FADE_OUT_LAYOUT */
  public static final int FADE_IN_LAYOUT    = 1;

  /** Frames fading out then fading in mode. @see FADE_OUT_LAYOUT */
  public static final int FADE_OUT_IN_LAYOUT = 2;

  /** current animated button state */
  protected int state;

  protected int layoutType;
  protected int fadeInState;
  protected int framesPerState;
  protected int maxStates;
  protected int statesIndexes[];

  private final static int IDLE = -1;

  /**
   * Animated button constructor.
   * @param states number of states of the button.
   * @param framesPerState number of frames for each state.
   * @param layoutType FADE_OUT_LAYOUT, FADE_IN_LAYOUT or FADE_OUT_IN_LAYOUT.
   * @param transColor the transparency color
   * @param framePeriod delay in millisecconds between two frames
   * @see FADE_OUT_LAYOUT
   * @see FADE_IN_LAYOUT
   * @see FADE_OUT_IN_LAYOUT
   * @deprecated will be removed in next version
   */
  public AnimatedButton(int states,int framesPerState,int layoutType,
                        Color transColor,int framePeriod)
  {
    super(states*framesPerState,transColor,framePeriod);

    this.framesPerState=framesPerState;
    this.layoutType=layoutType;

    setCommon(states);
  }

  /**
   * Animated button constructor.
   * @param frames button different states frames in multi-frame BMP format.
   * @param states number of states of the button.
   * @param framesPerState number of frames for each state.
   * @param layoutType FADE_OUT_LAYOUT, FADE_IN_LAYOUT or FADE_OUT_IN_LAYOUT.
   * @param transColor the transparency color
   * @param framePeriod delay in millisecconds between two frames
   * @see FADE_OUT_LAYOUT
   * @see FADE_IN_LAYOUT
   * @see FADE_OUT_IN_LAYOUT
   */
  public AnimatedButton(Image frames,int states,int framesPerState,int layoutType,
                        Color transColor,int framePeriod) // fdie@341_2
  {
    super(frames,frames.getWidth()/(states*framesPerState),transColor,framePeriod);

    this.framesPerState=framesPerState;
    this.layoutType=layoutType;

    setCommon(states);
  }

  private void setCommon(int states)
  {
    maxStates=states;
    statesIndexes=new int[states];

    for (int s=0; s<states; s++)
      statesIndexes[s]=(layoutType==FADE_IN_LAYOUT) ? ((s+1)*framesPerState)-1
                                                    : s*framesPerState;

    curFrame=statesIndexes[state=0];
    fadeInState=IDLE;
  }

  /**
   * Set the animated button state.
   * @param state value between 0 and states-1
   */
  public void setState(int state)
  {
    if (isPlaying)
    {
      stop();
      fadeInState=IDLE;
    }

    this.state=state;
    curFrame=statesIndexes[state];
    repaintNow();
  }

  /**
   * Get the animated button state.
   * @return value between 0 and states-1
   */
  public int getState()
  {
    return state;
  }

  /**
   * Animated button event handler.
   * @param event
   */
  public void onEvent(Event event)
  {
    if (event.type == PenEvent.PEN_DOWN && fadeInState==IDLE)
       inc(((PenEvent)event).x >= (width>>1));
    else
    if (event.type == AnimationEvent.FINISH)
    {
      if (fadeInState!=IDLE)
      {
        state=fadeInState;
        fadeInState=IDLE;
        if (layoutType==FADE_OUT_IN_LAYOUT)
        {
          postEvent(new ControlEvent(ControlEvent.PRESSED,this));
          return;
        }
        int dest=statesIndexes[state];
        if (layoutType!=FADE_IN_LAYOUT)
           start(dest+framesPerState-1,dest,-1,LOOPS_ONE_SHOT);
        else
           start(dest-framesPerState+1,dest,1,LOOPS_ONE_SHOT);
      }
      else postEvent(new ControlEvent(ControlEvent.PRESSED,this));
    }
  }

  /**
   * Increase/decrease the animated button state.
   * @param up boolean with a true value to increase the value, decrease otherwise
   */
  protected void inc(boolean up)
  {
    int dir=up ? 1:-1;
    int dest=(state+maxStates+dir) % maxStates;
    int src=statesIndexes[state];
    if (layoutType==FADE_OUT_IN_LAYOUT)
       start(src,statesIndexes[dest],dir,LOOPS_ONE_SHOT);
    else
    if (layoutType!=FADE_IN_LAYOUT)
       start(src,src+framesPerState-1,1,LOOPS_ONE_SHOT);
    else
       start(src,src-framesPerState+1,-1,LOOPS_ONE_SHOT);
    fadeInState=dest;
  }
}