package superwaba.ext.xplat.game;

import waba.fx.*;
import waba.sys.*;

/**
 * Class implementing a game animated sprite. <br>
 * <pre>
 * This is a sprite that supports multiple images corresponding to a sprite animation
 * or sprite's different states.
 * </pre>
 * @see Sprite
 * @author Frank Diebolt
 * @author Guilherme Campos Hazan
 * @version 1.1
 */
public class AnimatedSprite extends Sprite
{
  /** Images width/height dimensions. <b>READ-ONLY</b> attributes. */
  public int widths[],heights[];

  /** Images half width/height dimensions. <b>READ-ONLY</b> attributes.*/
  protected int halfWidths[],halfHeights[];

  /** Different sprite's frames, if several Image frames have been provided. */
  protected Image frames[];

  /** Current selected frame. */
  protected int curFrame; // fdie@341_1

  /**
   * AnimatedSprite constructor. <br>
   * @param frames sprite different state images.
   * @param transColor sprite's transparency color or null if none<br>
   * (needed in DRAW_SPRITE mode to keep the current background).
   * @param saveBckgd true if the background should be saved each time the sprite
   * is drawn to restore it once the sprite moves.
   * @param region defines the sprite valid area.<br>If null, a default region
   * is set to prevent the largest sprite frame to leave even partially the screen.
   */
  protected AnimatedSprite(Image frames[], Color transColor, boolean saveBckgd, Rect region)
  {
    super(frames[0],transColor,saveBckgd,region);

    int nFrames=frames.length;
    this.frames=frames;

    widths=new int[nFrames];
    heights=new int[nFrames];
    halfWidths=new int[nFrames];
    halfHeights=new int[nFrames];

    boolean sameSizes=true;

    for (int i=nFrames-1; i>=0; i--)
    {
      image=frames[i];

      int w=image.getWidth();
      int h=image.getHeight();
      if (sameSizes && (w!=width || h!=height)) sameSizes=false;

      if (w > width)  width=w;
      if (h > height) height=h;

      if (w<=0 || h<=0) throw new GameEngineException("bad Sprite bitmap");

      halfWidths[i]  = (w>>1);
      halfHeights[i] = (h>>1);
    }
    if (sameSizes)
    {
      // all images have the same size, no need to handle various sizes
      widths=heights=null;
      halfWidths=halfHeights=null;

      halfWidth=(width>>1);
      halfHeight=(height>>1);
    }

    setReg(region);
  }

  /**
   * AnimatedSprite constructor. <br>
   * @param frames sprite different state frames in multi-frame BMP format.
   * @param frameWidth width in pixels of one frame.
   * @param transColor sprite's transparency color or null if none<br>
   * (needed in DRAW_SPRITE mode to keep the current background).
   * @param saveBckgd true if the background should be saved each time the sprite
   * is drawn to restore it once the sprite moves.
   * @param region defines the sprite valid area.<br>If null, a default region
   * is set to prevent the largest sprite frame to leave even partially the screen.
   * @see Gif2Bmps
   */
  protected AnimatedSprite(Image frames, int frameWidth, Color transColor, boolean saveBckgd, Rect region) // fdie@341_1
  {
    // the Sprite constructor set the image, width, height, half... attributes
    super(frames,transColor,saveBckgd,region);

    // check the number of frames
    if (frameWidth<=0 || (width/frameWidth)<=0)
      throw new GameEngineException("No frames");

    // fix the real frame width & halfwidth, height is ok
    width=frameWidth;
    halfWidth=width>>1;

    setReg(region);
  }

  private void setReg(Rect region)
  {
    if (region==null)
    {
      regionMinx=(width>>1);
      regionMiny=(height>>1);
      regionMaxx=Settings.screenWidth-width+regionMinx;
      regionMaxy=Settings.screenHeight-height+regionMiny;
    }
    else setRegion(region);
  }

  /**
   * Draw the sprite's current frame at it's current
   * position using the defined drawOp. <br>
   * If the Sprite has been created with enabled background saving
   * (usefull only when the screen is not cleared before the new frame display)
   * the previously stored background will be restored first and the
   * sprite is displayed in a second step.<br>
   * @see Sprite
   */
  public void show() // fdie@341_1 : added to support direct multi-frame image support
  {
    // are we in background restoring mode ?
    if (background != null)
    {
      // Is it not the first paint ?
      if (bgX!=INVALID)
      {
        // position didn't change, what are we doing here ?
        if (!screenErased && bgX==centerX && bgY==centerY)
           return;
        // buffer -> screen
        gfx.copyRect(background,0,0,width,height,bgX-halfWidth,bgY-halfHeight);
      }
      // screen -> buffer
      bgX=centerX;
      bgY=centerY;
      bgGfx.copyRect(surface,bgX-halfWidth,bgY-halfHeight,width,height,0,0);
    }
    // copy the sprite image to the graphic context

    if (frames!=null) // distinguished frames or mutli-frames BMP format
    {
      gfx.drawImage(image,centerX-halfWidth,centerY-halfHeight,drawOp,transColor,doClip);
    }
    else
    {
      // @todo Guich may add a copyRect with drawOp, transpColor & clipping, meanwhile we have to it explicitly
      gfx.setDrawOp(drawOp);
      if (transColor!=null) gfx.setBackColor(transColor);
      gfx.copyRect(image,width*curFrame,0,width,height,centerX-halfWidth,centerY-halfHeight);
    }

  }

  /**
   * Set the sprite's current frame.
   * @param frame number to use for next sprite processing
   */
  public void setFrame(int frame)
  {
    if (widths!=null)
    {
      /*
        Frames have different sizes, update the different sprite's internal state
        to reflect the current one.
        The Sprite region is not updated, it's set using the sprite max frame size.
      */
      width=widths[frame];
      height=heights[frame];
      halfWidth=halfWidths[frame];
      halfHeight=halfHeights[frame];
    }
    curFrame=frame;
    if (frames!=null) // fdie@341_1 : if no direct multi-frame image, then switch the image
      image=frames[frame];
  }
}
