package superwaba.ext.xplat.game;

import waba.fx.*;
import waba.sys.*;
import waba.util.Vector;
import waba.sys.Thread;
import waba.ui.*;

/**
 * The Animation control class. <br>
 * <pre>
 * This control displays an animation that can be loaded from indexed BMP files
 * (one frame per image) or by a multi-frames BMP. This kind of BMP file contains a
 * list of images having all the same size and that lay side by side.
 * You may use the Gif2Bmps tool in \superwaba\apps\xplat\Gif2Bmps folder to
 * convert animated gif into the convenient format.
 *
 * warning: multi-frames BMP files can't be larger than 64Kb (the palmOS record limit)
 * </pre>
 * @see Gif2Bmps
 * @author Frank Diebolt
 * @author Guilherme Campos Hazan
 * @version 1.1
 */
public class Animation extends Control
{
  /**
   * drawOp drawing operation used when the animation images are copied to the
   * screen buffer.
   * @see waba.fx.Graphics
   */
  public int drawOp;

  /** Delay between two frames. */
  public int framePeriod;

  /** Current frame of the animation. */
  protected int curFrame;

  /** Reflects the animation play state. */
  public boolean isPlaying;

  /** Reflects the animation pause state. */
  public boolean isPaused;

  /** Transparency color. */
  protected Color transColor;

  /** Maximum number of frames. */
  protected int maxFrames;

  /** Frames buffer, all frames must have the same size. */
  protected Image framesBuffer[];

  // true if the animation has been constructed with a multi-frame BMP
  private boolean multiFramesImage; // fdie@341_2 : direct multi-frame image ?

  // Insert count
  private int insert;

  int frameSequence[];

  private static MainWindow mainWin;

  private Thread animThread;
  private int startFrame,endFrame;

  /** Should each frame drawing be notified ? */
  protected boolean notifyEachFrame; // fdie@341_3 deprecated

  /** Event notification mask,
   * default value is eventFinish, that means that an event is posted
   * only when the animation finishes.
   **/
  protected int eventsMask = eventFinish; // fdie@341_3 new notification mechanism

  /** no notify at all */
  public static final int eventNone      = 0x0;
  /** notify animation endings */
  public static final int eventFinish    = 0x1;
  /** notify animation loops */
  public static final int eventLoop      = 0x2;
  /** notify animation frames */
  public static final int eventFrame     = 0x4;

  // fdie@341_6+
  /** start function "loops" argument special value to loop endlessly  */
  public static final int LOOPS_UNLIMITED 	= -10000;
  /** start function "loops" argument special value to do one loop */
  public static final int LOOPS_ONE_SHOT 	= 1;

  // number of remaining loops
  int loopCount; // fdie@341_6-


  /**
   * Background image.
   */
  protected Image background; // fdie@400_51 : save animation background

  /**
   * Background graphic context.
   */
  protected Graphics bgGfx; // fdie@400_51



  /**
   * Animation constructor.
   * @param maxImages maximum number of frames
   * @param transColor the transparency color
   * @param framePeriod delay in millisecconds between two frames
   *
   * @deprecated will be removed in next release
   */
  public Animation(int maxFrames,Color transColor,int framePeriod)
  {
    super();

    this.transColor=transColor;
    if (transColor!=null) drawOp=Graphics.DRAW_SPRITE;
    this.framePeriod=framePeriod;

    if (mainWin==null) mainWin=MainWindow.getMainWindow();

    isPlaying=isPaused=false;

    renew(maxFrames);
  }

  /**
   * Animation constructor.
   * @param frames single image containing all frames in mlti-frames BMP format.
   * @param frameWidth width in pixels of one frame.
   * @param transColor the transparency color
   * @param framePeriod delay in millisecconds between two frames
   * @see Gif2Bmps
   */
  public Animation(Image frames,int frameWidth,Color transColor,int framePeriod) // fdie@341_2 : direct multi-frame image constructor
  {
    super();
    multiFramesImage=true;

    this.transColor=transColor;
    if (transColor!=null) drawOp=Graphics.DRAW_SPRITE;
    this.framePeriod=framePeriod;

    if (mainWin==null) mainWin=MainWindow.getMainWindow();

    isPlaying=isPaused=false;

    renew(1);
    framesBuffer[0]=frames;

    width=frames.getWidth();
    height=frames.getHeight();

    // check the number of frames and store the frame array management properties
    if (frameWidth<=0 || ((insert=maxFrames=width/frameWidth))<=0)
      throw new GameEngineException("No frames");

    width=frameWidth;

  }

  /**
   * Clear the animation to add new frames.
   */
  public void renew(int maxFrames) throws GameEngineException
  {
    if (isPlaying) throw new GameEngineException("is running");

    framesBuffer=new Image[this.maxFrames=maxFrames];
    insert=curFrame=0;
  }

  /**
   * Number of frames in the animation.
   * @return frames amount
   */
  public int size()
  {
    return (frameSequence != null) ? frameSequence.length : insert;
  }

  /**
   * Add a frame to the animation.
   *
   * @param image/frame to add
   * @return 1 if the frame has been added, else 0
   *
   * @throws GameEngineException
   * @deprecated will be removed in next release
   */
  public int addFrame(Image frame) throws GameEngineException
  {
    if (insert >= maxFrames) return 0;
    if (insert <= 0)
    {
      width=frame.getWidth();
      height=frame.getHeight();
    }
    else
    if (width != frame.getWidth() || height != frame.getHeight())
      throw new GameEngineException(DifferentImagesExceptionText);

    framesBuffer[insert]=frame;
    insert++;

    return 1;
  }

  /**
   * Add several frames to the animation.
   * This function adds a list of images loaded from indexed image files.
   * For instance a collection of files "myFrame0.bmp, myFrame1.bmp ... myFrameN.bmp"
   *
   * @param basename of the image filenames (ie. "myFrame")
   * @return number of frames added
   *
   * @throws GameEngineException
   * @deprecated will be removed in next release
   *
   * @see Gif2Bmps. This tool can convert animated gif files to
   * such a numbered collection of BMP files.
   */
  public int addMultiFrames(String basename) throws GameEngineException
  {
    int count=0;
    try
    {
      while (addFrame(new Image(basename+count+".bmp"))>0)
         count++;
    }
    catch (Exception ex)
    {
      // reached the end of the images to load
    }
    return count;
  }

  /**
   * Add several frames to the animation.
   * This function adds a list of images contained in a single BMP file.
   * The different frames have the same width and lay side by side, this function
   * extracts the frames and builds the animation.
   *
   * @param image multi-frames BMP file ( < 64K palmOS record limit)
   * @param width of one frame
   * @return number of frames added
   *
   * @throws GameEngineException
   * @deprecated will be removed in next release
   *
   * @see Gif2Bmps. This tool can convert animated gif files to
   * such a numbered collection of BMP files.
   */
  public int addMultiFrames(Image image,int width) throws GameEngineException
  {
    if (insert>=maxFrames) return 0;
    int h=image.getHeight();

    int totalWidth = image.getWidth();

    int count=0;
    for (int offset=0; offset<totalWidth; offset+=width)
    {
      if (insert>=maxFrames) return count;
      Image ni=new Image(width,h);
      ni.getGraphics().copyRect(image,offset,0,width,h,0,0);
      if (addFrame(ni)>0) count++;
      else break;
    }

    return count;
  }

  /**
   * Add several frames to the animation.
   * @param frames to add
   * @return number of frames added
   *
   * @throws GameEngineException
   * @deprecated will be removed in next release
   */
  public int setFrames(Image frames[]) throws GameEngineException
  {
    if (frames.length>maxFrames) throw new GameEngineException("Too many images");

    for (int i=0; i<maxFrames; i++)
    {
      Image img=frames[i];
      int w=img.getWidth();
      int h=img.getHeight();

      if (i==0)
      {
        width=w;
        height=h;
      }
      else
      if (w != width || h != height)
        throw new GameEngineException(DifferentImagesExceptionText);
    }
    framesBuffer=frames;

    return insert=maxFrames=frames.length;
  }

  /** Returns the prefered width of this control. */
  public int getPreferredWidth()
  {
     return width;
  }

  /** Returns the prefered height of this control. */
  public int getPreferredHeight()
  {
    return height;
  }

  /** Display a given frame.
   * @param frame index to display.
   **/
  public void drawFrame(int frame)
  {
    if (!isPlaying)
    {
      int backup=curFrame;
      curFrame=frame;
      repaintNow();
      curFrame=backup;
    }
  }

  /** Called by the system to draw the animation. **/
  public void onPaint(Graphics gfx)
  {
    if (drawOp == Graphics.DRAW_SPRITE) // Guich proposed optimization
    {
       // fdie@400_51 : save animation background
       if (background == null)
       {
         background=new Image(width,height);
         bgGfx=background.getGraphics();
         // screen -> buffer
         bgGfx.copyRect(parent/*this*/,x,y,width,height,0,0);
       }
       else
       {
         // buffer -> screen
         gfx.copyRect(background,0,0,width,height,0,0);
       }
    }

    // frame lookup table, for special animations
    int index=curFrame;
    if (frameSequence != null) index=frameSequence[index];

    if (multiFramesImage)  // fdie@341_2 : direct multi-frame image ?
      gfx.copyImageRect(framesBuffer[0],width*index,0,width,height,drawOp,transColor == null ? Color.WHITE:transColor,false);
    else
      gfx.drawImage(framesBuffer[index],0,0,drawOp,(transColor == null ? Color.WHITE:transColor),false);
  }

  /**
   * Enable the posting of frame displayed events.
   * @return the previous state
   * @deprecated is replaced by notifyMask
   */
  public boolean notifyEachFrame(boolean enable) // fdie@341_3 deprecated
  {
    boolean p=notifyEachFrame;
    notifyEachFrame=enable;
    return p;
  }

  /**
   * Enable the posting of events.
   * By default the posting of events are disabled.
   */
  public void enableEvents(int mask) // fdie@341_3
  {
    eventsMask=mask;
  }

  /**
   * Pauses a running animation.
   * if the application is not playing, this call has no effect.
   */
  public void pause()
  {
    if (isPlaying) isPaused=true;
  }

  /**
   * Resumes a paused animation.
   * if the application is not playing, this call has no effect.
   */
  public void resume()
  {
    if (isPlaying) isPaused=false;
  }

  /** Stops the animation.
   * if the application is not playing, this call has no effect.
   **/
  public void stop()
  {
    if (isPlaying)
    {
      isPlaying=false;
      isPaused=false;
      mainWin.removeThread(animThread);
      animThread=null;
    }
  }

  /**
   * Starts the animation.
   * This function starts the animation with or without looping when the end is reached.
   * If the application is playing, this call has no effect.
   *
   * @param loop indicator to restart the animation automaticaly when it ends.
   * @deprecated replaced by : public void start(int loops)
   */
  public void start(boolean loop)
  {
    start(0,insert-1,1,loop);
  }

  /**
   * Starts the animation with a frame range.
   * This function starts an animation by specifying the frame range and a loop flag.
   * If the application is playing, this call has no effect.
   *
   * @param sFrame start frame
   * @param eFrame end frame
   * @param loop true means that the animations restarts automaticaly
   *
   * @deprecated replaced by : public void start(final int sFrame,final int eFrame,final int step,int loops)
   */
  public void start(final int sFrame,final int eFrame,final int step,final boolean loop)
  {
    if (isPlaying) return;

    if (animThread!=null) return;

    final int nSteps;

    if (sFrame>=0)
    {
      frameSequence=null;
      startFrame=sFrame;
      endFrame=eFrame;
      nSteps=insert;
    }
    else
    {
      startFrame=0;
      nSteps=frameSequence.length;
      endFrame=nSteps-1;
    }

    curFrame=startFrame;

    /**
     * this is the main part of the animation.
     * the use of a thread to handle the animation display makes the animation
     * non blocking and other UI events can be handled meanwhile for instance
     * other animations ;-)
     */

    animThread=new Thread()
    {
      public void run()
      {
        if (!isPlaying || isPaused) return;

        if (lastFrameTime<0) lastFrameTime=Vm.getTimeStamp();

        int now = Vm.getTimeStamp();
        if (Vm.getTimeElapsed(now,lastFrameTime) < framePeriod) return;
        else lastFrameTime=now;

        repaintNow();
        if (notifyEachFrame || ((eventsMask & eventFrame) != 0))
          postEvent(new AnimationEvent(AnimationEvent.FRAME,this));

        if (curFrame != endFrame)
        {
          curFrame+=step;
          if (curFrame<0 || curFrame>=insert)
             curFrame=(curFrame+nSteps)%nSteps;
        }
        else if (loop)
        {
          curFrame=startFrame;
          if ((eventsMask & eventLoop) != 0)
            postEvent(new AnimationEvent(AnimationEvent.LOOP,this));
        }
        else if (loopCount == LOOPS_UNLIMITED || --loopCount > 0) // fdie@341_6 will be the only loop management once the above boolean loop is removed
        {
          curFrame=startFrame;
          if ((eventsMask & eventLoop) != 0)
            postEvent(new AnimationEvent(AnimationEvent.LOOP,this));
        }
        else
        {
          isPlaying=false;
          mainWin.removeThread(animThread);
          animThread=null;
          if ((eventsMask & eventFinish) != 0)
            postEvent(new AnimationEvent(AnimationEvent.FINISH,this));
        }
      }
      public void started() {}
      public void stopped() {}

      private int lastFrameTime=-1;
    };
    mainWin.addThread(animThread,false);
    isPlaying=true;
    isPaused=false;
  }

  /**
   * Starts the animation with a non standard frame sequence.
   * You can provide a different frame sequence than the standard 0,1,2,3,4..n frame order.
   * Use this function when you have an animation that has some identical frames.
   * In such case, just keep one unique frame and recreate the original frame sequence
   * by referencing the originaly identical frames several times.
   * If the application is playing, this call has no effect.
   * @param frameSequence frame numbers array which defines the frames sequence
   * @param loop indicator to restart the animation automaticaly when it ends.
   *
   * @deprecated replaced by : public void start(int frameSequence[],int loops)
   **/
  public void start(int frameSequence[],boolean loop)
  {
    if (isPlaying) return;

    this.frameSequence=frameSequence;
    start(-1,0,1,loop);
  }

  // fdie@341_6+
  /**
   * Starts the animation.
   * This function starts the animation and loops the specified amount of time.
   * If the application is playing, this call has no effect.
   *
   * @param loops integer value specifying the number of loops or LOOPS_UNLIMITED.
   * @see LOOPS_UNLIMITED
   */
  public void start(int loops)
  {
    start(0,insert-1,1,loops);
  }

  /**
   * Starts the animation with a non standard frame sequence.
   * You can provide a different frame sequence than the standard 0,1,2,3,4..n frame order.
   * Use this function when you have an animation that has some identical frames.
   * In such case, just keep one unique frame and recreate the original frame sequence
   * by referencing the originaly identical frames several times.
   * If the application is playing, this call has no effect.
   * @param frameSequence frame numbers array which defines the frames sequence
   * @param loops integer value specifying the number of loops or LOOPS_UNLIMITED.
   * @see LOOPS_UNLIMITED
   **/
  public void start(int frameSequence[],int loops)
  {
    if (isPlaying) return;

    this.frameSequence=frameSequence;
    start(-1,0,1,loops);
  }

  /**
   * Starts the animation with a frame range.
   * This function starts an animation by specifying the frame range and a loop flag.
   * If the application is playing, this call has no effect.
   *
   * @param sFrame start frame
   * @param eFrame end frame
   * @param loops integer value specifying the number of loops or LOOPS_UNLIMITED.
   * @see LOOPS_UNLIMITED
   */
  public void start(final int sFrame,final int eFrame,final int step,final int loops)
  {
    this.loopCount=loops;
    start(sFrame,eFrame,step,false);
  }
  // fdie@341_6-

  private static final String DifferentImagesExceptionText = "Does not support different image sizes";

  /**
   * This is a static helper function that divides a multi-frames BMP
   * into distinguished Images.<br>
   * The original image size should be a multiple of the frame width.
   * @param original multi-frames image.
   * @param frameWidth the width of a single frame.
   * @return
   */
  public static Image []splitImage(Image original,int frameWidth)
  {
    Vector list=new Vector(10);

    int totalWidth  = original.getWidth();
    int frameHeight = original.getHeight();

    int offset=0;
    while (offset<totalWidth)
    {
      Image ni=new Image(frameWidth,frameHeight);
      Graphics src=original.getGraphics();
      Graphics dst=ni.getGraphics();
      dst.copyRect(original,offset,0,frameWidth,frameHeight,0,0);
      list.add(ni);

      offset += frameWidth;
    }

    int size=list.size();
    if (size<=0) return null;

    Image result[]=new Image[size];
    for (int i=0; i<size; i++)
      result[i]=(Image)list.items[i];
    return result;
  }
}
