package superwaba.ext.xplat.game;

import waba.fx.*;
import waba.ui.*;

/**
 * Game Main Window of the GameEngine.
 * Note: you can't access this class directly. You must use the methods
 * provided by the GameEngine instead.
 * @author Frank Diebolt
 * @author Guilherme Campos Hazan
 * @version 1.1
 */
class GameEngineMainWindow extends MainWindow
{
   /**
    * Superwaba game API version.
    */
   private static final GameEngineException notInitializedException = new GameEngineException("GameEngine not initialized");
   static GameEngine engine;

   private Options options;
   private HighScores hscores;

   private boolean isRunning;
   private Graphics graf,g0;
   private ISurface isurf;

   private Timer gameTimer;

   /** Do nothing. All attributes must be set by the Game implementor */
   protected GameEngineMainWindow()
   {
   }

   /** Called just after the GameEngine is constructed */
   protected void setGameEngine(GameEngine engine)
   {
      this.engine = engine;
   }

   /** Returns a Graphics instance for the drawing surface.
     * A new instance is always returned */
   static final Graphics getGraphics()
   {
      if (engine==null) throw notInitializedException;
      return engine.gameIsDoubleBuffered ? engine.getOffScreen().getGraphics() : engine.createGraphics();
   }

   /** Returns the drawing surface */
   static final ISurface getSurface()
   {
      if (engine==null) throw notInitializedException;
      return engine.gameIsDoubleBuffered ? (ISurface)engine.getOffScreen() : (ISurface)engine;
   }
   /**
    * Replace SW onStart() handling to initialize the engine.<br>
    * <B>Could not be overloaded.</B>
    * Notifies onGameStart().<br>
    */
   public final void onStart()
   {
     if (engine.gameIsDoubleBuffered)
     {
       // use double buffer to draw in an offscreen buffer that is copied
       // to the screen once drawing is complete.
       setDoubleBuffer(true);
     }

     flicker = engine.gameDoClearScreen;
     if (!engine.gameHasUI) // can greately improve drawing performance?
     {
        isurf = getSurface();
        graf = getGraphics();
        g0 = createGraphics();
     }

     engine.gameIsRunning = isRunning = false;

     engine.onGameInit();
   }

   /**
    * Replace SW onExit() handling to shutdown the engine.<br>
    * <B>Could not be overloaded.</B>
    * Notifies onGameExit().
    */
   public final void onExit()
   {
     engine.onGameExit();
     shutdown();
   }

   /**
    * Get a new instance of the game highscores.
    * @return HighScores.
    */
   HighScores getHighScores()
   {
      if (hscores != null) return hscores;
      if (engine.gameHighscoresSize <= 0) return null;
      return hscores = new HighScores(engine);
   }

   /**
    * Get a new instance of the game options.
    * @return Options.
    */
   Options getOptions()
   {
      return options!=null ? options : (options=new Options(engine));
   }

   /**
    * Create a new TextRenderer.
    * A TextRenderer performs a fast String display with an optional integer value.
    * @param font to display with.
    * @param foreColor text color, may be null.
    * @param text to render.
    * @param maxDigits digits to display (with leading zeros).
    * @return a new TextRenderer.
    * @see TextRenderer TextRenderer for more information
    */
   TextRenderer createTextRenderer (Font font,Color foreColor, String text,int maxDigits)
   {
     return new TextRenderer(this,font,foreColor,backColor,text,maxDigits);
   }

   /*
    * Shutdown the GameEngine and releases its resources.
    */
   private final void shutdown()
   {
      engine.gameIsRunning = isRunning = false;

      if (hscores!=null)
         hscores.close();
      if (options!=null)
         options.close();
      if (gameTimer!=null)
         removeTimer(gameTimer);
   }

   /*
    * Stops the game. <br>
    * Arms a timer, because onGameStop() events that may start a UI we have to be in
    * the event handling.<br>
    * The game will be interrupted at next timer event.
    */
   void stop ()
   {
      engine.gameIsRunning = isRunning = false;
      if (engine.gameRefreshPeriod <= 0)
         gameTimer=super.addTimer(gameTimerControl,10);
   }

  /**
   * Start the game mainloop. <br>
   * NOTE: If the game is time based (arcade game), this function arms the mainloop
   * timer which causes scheduled onPaint() calls to let the game redraw a new frame.<br>
   * For non time based games, you have to call explicitly the refresh() function to
   * redraw a frame (on pen events for instance or any other kind of events...)
   * @see #stop
   */
   void start()
   {
      // first let's clean the screen, maybe we can also set the background color
      super.eraseBackgroundNow=true;
      //refresh();

      engine.onGameStart();
      engine.gameIsRunning = isRunning = true;

      if (engine.gameRefreshPeriod > 0)
         gameTimer=super.addTimer(gameTimerControl,engine.gameRefreshPeriod);

      // start first game display
      refresh();
   }

   /** Refreshes the screen. If there are UI elements on the screen (GameEngine.hasUI = true)
     * calls the slower _doPaint. Otherwise, just prepares the buffer and repaint it.
     */
   void refresh()
   {
      if (!engine.gameHasUI)
      {
         if (flicker)
         {
            graf.setBackColor(backColor);
            graf.clearScreen();
         }
         engine.onPaint(graf);
         g0.copyScreen(isurf, 0,0,this.height);
      }
      else _doPaint(0,0,width,height);
   }

   /**
    * Replace SW onEvent handling to identify and notify some usefull game events.<br>
    * <B>Could not be overloaded.</B>
    * @param evt event that occurs.
    */
   public final void onEvent(Event evt)
   {
      switch (evt.type)
      {
         case ControlEvent.TIMER:   engine.onTimer      ((ControlEvent)evt); break;
         case KeyEvent.KEY_PRESS:   engine.onKey        ((KeyEvent)evt);     break;
         case PenEvent.PEN_DOWN:    engine.onPenDown    ((PenEvent)evt);     break;
         case PenEvent.PEN_DRAG:    engine.onPenDrag    ((PenEvent)evt);     break;
         case PenEvent.PEN_UP:      engine.onPenUp      ((PenEvent)evt);     break;
         default:                   engine.onOtherEvent (evt);
      }
   }

   Control gameTimerControl=new Control() {
     public final void onEvent(Event e)
     {
       if (isRunning)
         refresh();
       else
       {
          removeTimer(gameTimer);
          gameTimer=null;
          // erase screen before user may display UI
          eraseBackgroundNow=true;
          _doPaint(0,0,width,height);

          // Notifies the game run mode quit by calling onGameStop().
          engine.onGameStop();
       }
     }
   };
}
