package superwaba.ext.xplat.game;

import waba.io.*;

/**
 * The game highscores management class. <br>
 * <br>
 * The highscores are stored in a waba Catalog object which is linked to the
 * application through the software creatorID. This causes the highscores
 * database deletion when the game with the same creatorID is erased.
 * <p>
 * The complete database name is composed by the game name 'appl' and its creatorID, both
 * provided during the engine setup.<br>
 * The highscores database name is : ${appl}_HSC.${creatorID}.DATA
 * <p>
 * You can find a complete game API sample named 'Ping' in the SW examples folder.<br>
 * Here is some sample code:
 *
 * <pre>
 * import superwaba.ext.xplat.game.*;
 * import superwaba.ext.xplat.util.props.*;
 * ...
 *
 * <i>public class Ping extends <U>GameEngine</U></i> {
 *
 * // constructor
 * public Ping()
 * {
 *   waba.sys.Settings.setPalmOSStyle(true);
 *
 *   // define the game API setup attributes
 *
 *   gameName             = "Ping";
 *
 *   // when not run on device, appCreatorId does not always return the same value.
 *
 *   gameCreatorID        = Settings.onDevice ? waba.sys.Settings.appCreatorId:"PiNg";
 *
 *   gameVersion          = 100;   // v1.00
 *   gameHighscoresSize   = 7;     // store the best 7 highscores
 *   gameRefreshPeriod    = 75;    // 75 ms refresh periods
 *   gameIsDoubleBuffered = true;  // used double buffering to prevent flashing display
 *   gameDoClearScreen    = true;  // screen is cleared before each frame display
 *   gameHasUI            = false; // no UI elements, frame displays are optimized
 *   ...
 * }
 *
 * public void addUserScore(Property.Str user,int score) {
 *
 *   <B>HighScores</B> hs=<U>getHighScores</U>();
 *
 *   // try to add a new score to the highscores
 *
 *   HighScoreEntry insEntry=<B>hs.add</B>(score,user);
 *
 *   if (insEntry!=null) {    // is score a highscore ?
 *
 *     // get access to all highscores
 *     HighScoreEntry[] entries=<B>hs.getEntries()</B>;
 *
 *     for (int n=0; n&lt;<B>hs.size</B>(); n++) {
 *
 *       // access the nth highscore entry
 *       HighScoreEntry entry=entries[n];
 *
 *       Vm.debug("name: "+entry.name+" score="+Convert.toString(entry.score));
 *     }
 *
 *   }
 *
 * }
 * </pre>
 * @author Frank Diebolt
 * @version 1.0
 */

public class HighScores extends Catalog
{
   private final static String dbName_suffix = "_HSC.";
   private final static String dbType        = ".DATA";
   private final static String NO_NAME       = "no name";

   HighScoreEntry entries[]=null;
   private boolean dirty;
   private int validEntries; // number of valid highscore entries

   /** This class must be instantiated using the GameEngineMainWindow.getHighScores */
   protected HighScores(GameEngine engine)
   {
      super(engine.gameName+dbName_suffix+engine.gameCreatorID+dbType,Catalog.CREATE);
      if (!isOpen())
         throw new GameEngineException("HS:access error");

      entries=new HighScoreEntry[engine.gameHighscoresSize];
      for (int i=0; i<entries.length; i++)
         entries[i] = new HighScoreEntry(this);
      dirty=false;

      if (!setRecordPos(0)) return;

      DataStream ds=new DataStream(this);

      int version=ds.readInt();
      /* this test is not needed until a database format change occurs...
      if (version>GameEngine.GAME_ENGINE_VERSION)
        throw new GameEngineException("HS:game API "+GameEngine.GAME_ENGINE_VERSION+" out of date error");
      */

      validEntries=ds.readInt();
      for (int i=0; i<entries.length; i++)
      {
         if (i < validEntries)
         {
            entries[i].name=ds.readString();
            entries[i].score=ds.readInt();
         }
         else
         {
            entries[i].name=null;
            entries[i].score=0;
         }
      }
   }

   /**
    * Amount of highscore entries.
    * @return the number of entries in the highscore table.
    */
   public final int size()
   {
      return entries.length;
   }

   /**
    * Retrieve the highscore entries.
    * @return an array to the highscore entries.
    */
   public final HighScoreEntry[] getEntries()
   {
      return entries;
   }

   /**
    * add a new score to the highscore table.
    * @param score to add.
    * @param name of the performer, may be null.
    * @see HighScoreEntry
    * @return added entry or null
    */
   public HighScoreEntry add(int score,String name)
   {
      int insert;

      // lookup the position to insert the score
      for (insert=0; insert<entries.length; insert++)
         if (entries[insert].name==null || score > entries[insert].score)
            break;

      if (insert>=entries.length) return null; // didn't found a lower score to replace

      // reuse the last highscore entry, sorry but you are ejected
      HighScoreEntry free=entries[entries.length-1];

      // create a free entry by moving back the least scores
      for (int m=entries.length-2; m>=insert; m--)
        entries[m+1]=entries[m];

      // store the new score
      free.name=name;
      free.score=score;
      entries[insert]=free;
      if (validEntries<entries.length) validEntries++;

      dirty=true;
      return entries[insert];
   }

   /**
    * Stores the highscores database.
    * @return false if no changes to save or an error occurs
    */
   public boolean save()
   {
      if (!dirty) return false;

      // use this nice object to resize the highscore record
      ResizeStream rs=new ResizeStream(this,100);

      if (!rs.restartRecord(0))
         throw new GameEngineException("Cannot setRecordPos");

      DataStream ds=new DataStream(rs);
      ds.writeInt(GameEngine.GAME_ENGINE_VERSION);
      ds.writeInt(validEntries);
      for (int i=0; i < validEntries; i++)
      {
        ds.writeString(entries[i].name);
        ds.writeInt(entries[i].score);
      }
      rs.endRecord();
      dirty=false;

      return true;
   }

   /**
    * closes the highscores database.
    * @return true if succeeded.
    */
   public boolean close()
   {
      save();
      return super.close();
   }
}
