package superwaba.ext.xplat.game;

import waba.fx.*;
import waba.util.*;

/**
 * A game specific, fast text renderer (no memory allocs during runtime). <br>
 * Important! The text displayed must not reach the screen limits!<br>
 * <br>
 * It renders the text and the digits in image buffers and the display
 * functions of this class just copies the image buffers back to the
 * graphic context.<br>
 * <br>
 * The text buffer is copied first, then the configured amount of digits
 * of the optional integer are copied next to the text with leading zeros.<br>
 * <br>
 * This class can be used to display game information such as scores, levels, etc.
 * <br>
 * <p>
 * You can find a complete game API sample named 'Ping' in the SW examples folder.<br>
 * Here is some sample code:
 *
 * <pre>
 * import superwaba.ext.xplat.game.*;
 * import superwaba.ext.xplat.util.props.*;
 * ...
 *
 * <i>public class Ping extends <U>GameEngine</U></i> {
 *
 *   // 2 text renderers to quickly display level and score values
 *
 *   private TextRenderer levelRenderer;
 *   private TextRenderer scoreRenderer;
 *
 * //---------------------------------------------------------
 * // overload the API's game init event.
 * // this function is called when the game is launched.
 * //---------------------------------------------------------
 *
 * <i>public void onGameInit()</i> {
 *
 *     // create two text renderers, one for the 'level' and one for the 'score'
 *     // we use the current font, the text should be black on the game background
 *
 *     // level display has 2 digits: max value is 99
 *
 *     levelRenderer=createTextRenderer(getFont(),Color.BLACK,"level:",2);
 *
 *     // score display has 5 digits: max value is 99999
 *
 *     scoreRenderer=createTextRenderer(getFont(),Color.BLACK,"score:",5);
 *   }
 *
 * //---------------------------------------------------------
 * // overload SuperWaba's onPaint function.
 * //---------------------------------------------------------
 *
 * <i>public void onPaint(Graphics gfx)</i> {
 *
 *   ...
 *   if (gameIsRunning) {  // game engine's running state
 *
 *     ...
 *     // display the game level & score, both enable transparency
 *
 *     levelRenderer.display(15,2,level,true);
 *     scoreRenderer.display(80,2,score,true);
 *     ...
 *   }
 *   ...
 * }
 * @author Frank Diebolt
 * @author Guilherme Campos Hazan
 * @version 1.1
 */

public class TextRenderer
{
  /*
   * share all digit renderings by font.
   */
  private static Hashtable byFont=new Hashtable(3);
  private static String strDigits[]={"0","1","2","3","4","5","6","7","8","9"};

  private Image []digits;
  private Image textImg;
  private int w0;

  private int maxDigits;
  private Color backColor;
  private Graphics gfx;

  /**
   * height in pixels of the drawing (= font metrics height).
   */
  protected int fmH;
  /**
   * width in pixels of the text.
   */
  protected int textWidth;

  /** Creates a text renderer with the given parameters.
   * @param surface to display the text on
   * @param font to use
   * @param foreColor text foreground color
   * @param backColor text background color
   * @param text to be displayed before the digits (or null if none)
   * @param maxDigits the number of digits to display. E.g.: 4 means the max value shown will be 9999.
   */
  protected TextRenderer(ISurface surface,Font font,Color foreColor,
                         Color backColor,String text,int maxDigits)
  {
    this.maxDigits = maxDigits;
    this.backColor = backColor;

    gfx = GameEngineMainWindow.getGraphics();
    gfx.setFont(font);
    gfx.setBackColor(backColor);
    gfx.setForeColor(foreColor);

    FontMetrics fm=new FontMetrics(font,surface);

    this.textWidth = text != null ? fm.getTextWidth(text) : 0;
    this.w0 = fm.getCharWidth('0');
    this.fmH = fm.getHeight();

    // lookup the digits in the shared space or build a new digits rendering
    Object o=byFont.get(font);
    if (o==null)
    {
       // store the new digits renderings
       o = buildDigits(font, backColor, foreColor);
       byFont.put(font,o);
    }
    digits=(Image[])o;
    if (text != null)
       textImg = render(text, textWidth, font, backColor, foreColor); // guich@340_10
  }

  /**
   * Display a text rendering.
   * This function just draws the text to the
   * GameEngineMainWindow at the specified x,y position.
   * @param x position.
   * @param y position.
   * @param transparent should the background be preserved.
   */

  public void display(int x, int y, boolean transparent)
  {
    if (textImg!=null)
       gfx.drawImage(textImg,x,y,transparent?Graphics.DRAW_SPRITE:Graphics.DRAW_PAINT,backColor,false);
  }

  /**
   * Display a text rendering.
   * This function just copies back the image buffer of the text to the
   * GameEngineMainWindow at the specified x,y position and displays the
   * integer value next to the text by using pre-rendered digit image buffers.
   * @param x position.
   * @param y position.
   * @param value positive integer value to draw next to the text.
   * @param transparent should the background be preserved.
   */

  public void display(int x,int y,int value,boolean transparent)
  {
    if (value < 0) return;
    if (textImg != null)
       gfx.drawImage(textImg,x,y,transparent?Graphics.DRAW_SPRITE:Graphics.DRAW_PAINT,backColor,false);

    if (maxDigits < 1) return;

    // compute how many digits intValue has so we can left justify the text
    int numDigits = 0;
    int tempValue = value;
    while (true)
    {
      numDigits++;
      tempValue /= 10;
      if (tempValue == 0)
         break;
    }

    x += textWidth + numDigits * w0;

    // most significant digit
    int drawOp = transparent?Graphics.DRAW_SPRITE:Graphics.DRAW_PAINT;
    while (true)
    {
       x -= w0;
       int d = value % 10;
       gfx.drawImage(digits[d],x,y,drawOp,backColor,false);
       value /= 10;
       if (value == 0)
          break;
    }
  }

  private Image[] buildDigits(Font font, Color backColor, Color foreColor)
  {
    Image array[]=new Image[10];
    for (int i=0; i<10; i++)
       array[i]=render(strDigits[i], w0, font, backColor, foreColor);
    return array;
  }

  private Image render(String text, int w, Font font, Color backColor, Color foreColor)
  {
    Image image = new Image(w,fmH);
    Graphics gfx=image.getGraphics();
    if (backColor!=null)
      gfx.setBackColor(backColor);
    gfx.clearScreen();
    gfx.setForeColor(foreColor!=null?foreColor:Color.WHITE);
    gfx.setFont(font);
    gfx.drawText(text,0,0);
    gfx.free();
    return image;
  }

  /** Returns the maximum width of the text plus the number of digits */
  public int getWidth()
  {
     return textWidth + (maxDigits*w0);
  }
  /** Returns the height of the current font */
  public int getHeight()
  {
     return fmH;
  }
}
