package superwaba.ext.xplat.html;

import waba.sys.*;
import waba.util.*;

/**
 * The <code>Attributes</code> class defines the current set of attributes to be
 * applied to text when it is displayed.  The HTML parser keeps a
 * stack containing instances of this attributes class.  As the parser
 * parses the HTML, new tags (e.g. &lt;CENTER&gt;) cause an entry to be pushed
 * onto the attribute stack with a new setting (e.g. alignment=CENTER).
 * When the parser see &lt;/CENTER&gt;, the stack is popped.
 *
 * @version $Header: $
 * @author Jim Guistwite
 */

public class Attributes
{
  /**
   * The integer version of the tag that caused this attribute entry to be created.*/
  int tag;

  /**
   * Set of boolean attributes that are stored at the bits in this
   * integer.  These attributes are ITALIC, SEPARATE_LINE, BOLD, PRE, etc.
   */
  int battr;

  /**
   * Border to be placed above the text associated with this attribute set.
   */
  char topborder;

  /**
   * Border to be placed below the text associated with this attribute set.
   */
  char bottomborder;

  /**
   * Size of the font to use to renderer the text.
   */
  char fontSize;

  /**
   * Alignment to specify the location of the text.
   */
  char alignment;

  /**
   * Current indent length for the text.
   */
  char indent;

  /**
   * If this text is part of a link, what is the href for this link
   */
  String href;

  /**
   * If this text is part of a form, what form?
   */
  Form  form;


  /**
   * If this text is an element in a list, save what list it is.
   */
  ListAttr list;

  /**
   * Used to align headings at the left.
   */
  public static final int ALIGN_LEFT = 0;

  /**
   * Used to align headings in the center.
   */
  public static final int ALIGN_CENTER = 1;

  /**
   * Used to align headings at the right.
   */
  public static final int ALIGN_RIGHT = 2;


  /**
   * Attribute for boolean attribute mask.
   */
  public static final int ITALIC = 1;
  /**
   * Attribute for boolean attribute mask.
   */
  public static final int SEPARATE_LINE = 2;
  /**
   * Attribute for boolean attribute mask.
   */
  public static final int BOLD = 4;
  /**
   * Attribute for boolean attribute mask.
   */
  public static final int PRE = 8;
  /**
   * Attribute for boolean attribute mask.
   */
  public static final int LISTITEM = 16;
  /**
   * Attribute for boolean attribute mask.
   */
  public static final int STRIKETHROUGH = 32;
  /**
   * Attribute for boolean attribute mask.
   */
  public static final int SUBSCRIPT = 64;
  /**
   * Attribute for boolean attribute mask.
   */
  public static final int SUPERSCRIPT = 128;
  /**
   * Attribute for boolean attribute mask.
   */
  public static final int UNDERLINE = 256;

  /**
   * Construct a new attributes object.
   * The font size defaults to 12.
   */
  public Attributes()
  {
    fontSize = 12;
  }

  /*
   * Copy this attributes object.
   * @return new attributes object identical to this one.
   */
  public Object copy()
  {
    Attributes rv = new Attributes();

    rv.battr = battr;
    rv.href = href;
    rv.form = form;
    rv.alignment = alignment;
    rv.fontSize = fontSize;
    rv.topborder = topborder;
    rv.bottomborder = bottomborder;
    rv.indent = indent;
    rv.list = list;

    return rv;
  }

  /**
   * Update this attributes object according to the tag
   * and attributes provided.
   * @param tag HtmlTag integer corresponding to the tag encounter in the input.
   * @param attr attributes associated with the HTML tag.
   */
  public void configure(int tag, Hashtable attr)
  {
     String s;
     switch (tag)// guich@330_31: changed dozens of if/else to a beautiful switch/case
     {
         case HtmlTag.H1:
         case HtmlTag.H2:
         case HtmlTag.H3:
         case HtmlTag.H4:
         case HtmlTag.H5:
         case HtmlTag.H6:
         {
            battr |= BOLD;
            topborder = 5;
            bottomborder = 5;
            int []sizes = {15,13,11,10,9,8}; // guich@330_31: optimized lots of ifs
            fontSize = (char)sizes[tag-HtmlTag.H1];

            battr |= SEPARATE_LINE;
            //separateLine = true;

            if ((s = (String)attr.get("align")) != null)
            {
            	if (s.equalsIgnoreCase("center"))
            	  alignment = ALIGN_CENTER;
            	else if (s.equalsIgnoreCase("left"))
            	  alignment = ALIGN_LEFT;
            	else
            	  alignment = ALIGN_RIGHT;
            }
            break;
         }
         case HtmlTag.PRE:
            battr |= PRE;
            break;
         case HtmlTag.B:
            battr |= BOLD;
            break;
         case HtmlTag.BIG:
            fontSize += 1;
            break;
         case HtmlTag.CODE:
            fontSize -= 1;
            break;
         case HtmlTag.ADDRESS:
            battr |= ITALIC;
            break;
         case HtmlTag.BLOCKQUOTE:
            topborder = 7;
            bottomborder = 7;
            indent = (char)5;
            battr |= SEPARATE_LINE;
            break;
         case HtmlTag.A:
            href = (String)(attr.get("href"));
            battr |= UNDERLINE;
            break;
         case HtmlTag.CENTER: alignment = ALIGN_CENTER; break;
         case HtmlTag.DIR: list = new ListAttr(tag, attr); break;
         case HtmlTag.DT:
            if (indent < 5)
	            indent = 0;
            else
	            indent -= 5;
            alignment = ALIGN_LEFT;
            break;
         case HtmlTag.DD:
            battr |= SEPARATE_LINE;
            alignment = ALIGN_LEFT;
            indent += (char)5;
            break;
         case HtmlTag.LI:
            battr |= LISTITEM;
            battr |= SEPARATE_LINE;
            //separateLine = true;
            alignment = ALIGN_LEFT;
            indent += (char)5;
            break;
         case HtmlTag.EM:
            battr |= BOLD;
            battr |= ITALIC;
            break;
         case HtmlTag.I: battr |= ITALIC; break;
         case HtmlTag.FONT:
            if ((s = (String)(attr.get("size"))) != null)
            	fontSize = (char)Convert.toInt(s);
            break;
         case HtmlTag.Q:
            indent += 5;
            battr |= SEPARATE_LINE;
            alignment = ALIGN_LEFT;
            break;
         case HtmlTag.S:
            battr |= STRIKETHROUGH;
            break;
         case HtmlTag.STRIKE:
            battr |= STRIKETHROUGH;
            break;
         case HtmlTag.SAMP:
            fontSize = 10;
            break;
         case HtmlTag.SMALL:
            fontSize -= 1;
            break;
         case HtmlTag.STRONG:
            battr |= BOLD;
            break;
         case HtmlTag.SUB:
            battr |= SUBSCRIPT;
            fontSize = 9;
            break;
         case HtmlTag.SUP:
            battr |= SUPERSCRIPT;
            fontSize = 9;
            break;
         case HtmlTag.U:
            battr |= UNDERLINE;
            break;
         case HtmlTag.OL:
            list = new ListAttr(tag, attr);
            break;
         case HtmlTag.UL:
            list = new ListAttr(tag, attr);
            break;
      }
   }
}
