package superwaba.ext.xplat.html;

import waba.util.*;
import waba.sys.*;

/**
 * The <code>HtmlParser</code> is responsible for turning the stream
 * of tokens returned by the HtmlTokenizer into a Document object.
 * The Document can then be used by the HtmlRenderer for display.
 *
 * @version $Header: $
 * @author Jim Guistwite
 */

public class HtmlParser
{
  Document doc;
  HtmlTokenizer tok;
  /**
   * Bitmap of tags that simply updates attributes and does
   * not create an Element to be added to the document.
   */
  private static IntVector attributeTags; // guich@330_31: optimized a big test in attributeTag to a bit test
  static
  {
    attributeTags = new IntVector();
    attributeTags.ensureBit((short)HtmlTag.TOTALTAGS);
    attributeTags.setBit((short)HtmlTag.A,true);
    attributeTags.setBit((short)HtmlTag.ADDRESS,true);
    attributeTags.setBit((short)HtmlTag.B,true);
    attributeTags.setBit((short)HtmlTag.BIG,true);
    attributeTags.setBit((short)HtmlTag.CENTER,true);
    attributeTags.setBit((short)HtmlTag.CODE,true);
    attributeTags.setBit((short)HtmlTag.DIR,true);
    attributeTags.setBit((short)HtmlTag.DD,true);
    attributeTags.setBit((short)HtmlTag.DT,true);
    attributeTags.setBit((short)HtmlTag.EM,true);
    attributeTags.setBit((short)HtmlTag.H1,true);
    attributeTags.setBit((short)HtmlTag.H2,true);
    attributeTags.setBit((short)HtmlTag.H3,true);
    attributeTags.setBit((short)HtmlTag.H4,true);
    attributeTags.setBit((short)HtmlTag.H5,true);
    attributeTags.setBit((short)HtmlTag.H6,true);
    attributeTags.setBit((short)HtmlTag.I,true);
    attributeTags.setBit((short)HtmlTag.LI,true);
    attributeTags.setBit((short)HtmlTag.PRE,true);
    attributeTags.setBit((short)HtmlTag.OL,true);
    attributeTags.setBit((short)HtmlTag.FONT,true);
    attributeTags.setBit((short)HtmlTag.Q,true);
    attributeTags.setBit((short)HtmlTag.S,true);
    attributeTags.setBit((short)HtmlTag.SAMP,true);
    attributeTags.setBit((short)HtmlTag.STRIKE,true);
    attributeTags.setBit((short)HtmlTag.SMALL,true);
    attributeTags.setBit((short)HtmlTag.STRONG,true);
    attributeTags.setBit((short)HtmlTag.SUB,true);
    attributeTags.setBit((short)HtmlTag.SUP,true);
    attributeTags.setBit((short)HtmlTag.U,true);
  }
  /**
   * Construct a new HtmlParser.  Note that a single parser cannot
   * be parsing the two different documents at a time since the tokenizer
   * and document are class members.
   */
  public HtmlParser()
  {
  }

  /**
   * Parse the text provided and return a Document instance.
   * @param txt HTML text to be parsed.
   * @return Document instance for this HTML.
   */
  public Document parse(String txt)
  {
    tok = new HtmlTokenizer();
    tok.listener = new RootHandler(this);
    tok.parse(txt);
    return doc;
  }

  /**
   * The <code>RootHandler</code> class parses the beginning of the document.
   */
  class RootHandler extends HtmlHandler
  {
    boolean inTitle;

    public RootHandler(HtmlParser parser)
    {
      super(parser);
    }

    public void comment(String data)
    {
      if (!Settings.onDevice)
         Vm.debug("comment: " + data);
    }

    public void startElement(int tag, Hashtable p)
    {
      switch (tag)
      {
         case HtmlTag.HTML:
            doc = new Document();
            break;
         case HtmlTag.TITLE:
            inTitle = true;
            break;
         case HtmlTag.BASE:
            String href = (String)p.get("href");
            if (href != null)
               doc.base = href;
            break;
         case HtmlTag.BODY:
            ElementHandler eh = new ElementHandler(parser);
            eh.collection = doc;
            push(eh);
            break;
      }
    }

    public void endElement(int tag)
    {
      if (tag == HtmlTag.TITLE)
         inTitle = false;
    }
    public void characters(String s)
    {
      if (inTitle)
         doc.title = s;
    }
  }

  /**
   * The <code>ElementHandler</code> class parses the content of the document.
   */
  class ElementHandler extends HtmlHandler
  {

    Vector attributeStack;
    Attributes current;
    ElementCollection collection;

    public ElementHandler(HtmlParser p)
    {
      super(p);
      attributeStack = new Vector(20);
      current = new Attributes();
      current.tag = -1;
    }

    public void comment(String data)
    {
    }

    public void startElement(int tag, Hashtable p)
    {
      // check for an auto-pop tag.
      if ((autoPop(current.tag)) && causesAutoPop(current.tag, tag))
         popAttr();

      if (attributeTags.isBitSet((short)tag))
      {
         pushAttr();
         current.tag = tag;
         current.configure(tag, p);
      }

      switch (tag)
      {
         case HtmlTag.PRE:
            tok.setNewlineSignificant(true);
            break;
         case HtmlTag.P:
         case HtmlTag.BR:
            collection.add(new Paragraph(tag, p, current));
            break;
         case HtmlTag.MENU:
         case HtmlTag.DIR:
            break;
         case HtmlTag.HR:
            collection.add(new Hr(tag, p, current));
            break;
         case HtmlTag.FORM:
            Form form = new Form(tag, p, current);
            collection.add(form);
            pushAttr();
            current.tag = tag;
            current.form = form;
            break;
         case HtmlTag.INPUT:
            collection.add(new Input(tag, p, current));
            break;
         case HtmlTag.HtmlTable:
            // push down another handler.
            HtmlTable t = new HtmlTable(tag, p, current);
            collection.add(t);
            ElementHandler eh = new TableHandler(parser, t);
            eh.collection = t;
            push(eh);
            break;
      }
    }

    public void endElement(int tag)
    {
      if (tag == HtmlTag.BODY || tag == HtmlTag.HtmlTable)
         pop();
      else
      if (attributeTags.isBitSet((short)tag))
      {
         if (current.tag != tag)
         {
            boolean mismatch = true;
            while (autoPop(current.tag))
            {
               popAttr();
               if (current.tag == tag)
               {
                  mismatch = false;
                  break;
               }
            }
            if (mismatch)
            {
               if (!Settings.onDevice)
               {
                  Vm.debug("tag mismatch " + tag);
                  Vm.debug(attributeStack.toString());
               }
            }
         }
         else
            popAttr();
      }
    }

    public void characters(String s)
    {
      TextElement te = new TextElement(s, current);
      collection.add(te);
    }

    private void pushAttr()
    {
      Attributes tmp = (Attributes)(current.copy());
      attributeStack.push(current);
      current = tmp;
    }

    private void popAttr()
    {
      Attributes tmp = (Attributes)(attributeStack.pop());
      if (tmp != null)
         current = tmp;
    }
  }

  /**
   * The <code>TableHandler</code> parses tables.
   */
  class TableHandler extends ElementHandler
  {
    HtmlTable HtmlTable;

    public TableHandler(HtmlParser p, HtmlTable t)
    {
      super(p);
      HtmlTable = t;
    }

    public void startElement(int tag, Hashtable p)
    {
      if (tag == HtmlTag.TR)
         HtmlTable.startRow(p, current);
      else
      if (tag == HtmlTag.TD)
         HtmlTable.startCell(p);
      else
         super.startElement(tag, p);
    }

    public void endElement(int tag)
    {
      if (tag == HtmlTag.TR)
         HtmlTable.endRow();
      else
      if (tag == HtmlTag.TD)
         HtmlTable.endCell();
      else
         super.endElement(tag);
    }
  }

  /**
   * Return true if this tag causes a previous tag to be automatacally popped.
   */
  private boolean causesAutoPop(int autoPopTag, int tag)
  {
    return !(tag == HtmlTag.P || tag == HtmlTag.BR);
  }

  /**
   * Return true if this tag does not require an end tag.
   */
  private boolean autoPop(int tag)
  {
    return tag == HtmlTag.DD || tag == HtmlTag.LI;
  }
}



