package superwaba.ext.xplat.html;

import waba.util.*;
import waba.ui.*;
import waba.fx.*;

/**
 * The <code>TextElement</code> class is the main class responsible
 * for renderering the text of the document.  The text is broken up
 * into one or more segments in an attempt to display the text without
 * a horizontal scrollbar.
 *
 * @version $Header: $
 * @author Jim Guistwite
 */

public class TextElement extends Element
{
  static Image bullet = new Image("dot.bmp");
  Vector segments;

  String text;
  String heading;

  Font font;
  FontMetrics fm;

  /**
   * Construct a new TextElement with the text and current attributes
   * provided.
   * @param text text to be displayed in the renderer.
   * @param attr current attributes to be associated with the text.
   */
  public TextElement(String text, Attributes attr)
  {
    super(attr);
    this.text = text;
  }

  /**
   * Implement the layout method to break text text into segments and
   * position those segments updating the layout context along the way.
   * @param renderer HtmlRenderer instance to draw in.
   * @param lc layout context to update.
   */
  public void layout(HtmlRenderer renderer, LayoutContext lc)
  {
    //System.out.println("TextElement::text - " + text);
    //System.out.println("TextElement::layout context - " + lc);
    //System.out.println("TextElement::attr - " + attr);

    segments = new Vector(2);
    region = new Region();

    if (font == null)
    {
      font = new Font("Helvetica", (((attr.battr & Attributes.BOLD) != 0) ? Font.BOLD : Font.PLAIN), attr.fontSize);
      fm = renderer.getFontMetrics(font);
    }

    //System.out.println("control: " + control);
    //System.out.println("font: " + font);
    //System.out.println("font size: " + fontSize);

    // special hack - no border required at top of document.
    if (lc.y == 0)
      attr.topborder = 0;

    if ( (attr.battr & Attributes.SEPARATE_LINE) != 0 && lc.textOnLine)
    {
      // first finish off last line.
      breakLine(lc);
    }

    //System.out.println(text + " lc: " + lc + "  attr: " + attr);
    if (lc.textOnLine)
    {
      // add some space.
      lc.x += fm.getTextWidth(" ");
    }
    // figure out how much text can fit on this line and break it into
    // segments.
    if ((attr.battr & Attributes.PRE) != 0)
      newlineSegmentize(fm, lc);
    else
      segmentize(fm, lc);
  }

  /**
   * Internal class use to hold a segement of text
   */
  class Segment
  {
    String txt;
    Rect rect;

    public Segment(String txt)
    {
      this.txt = txt;
      rect = new Rect(0, 0, 0, 0);
    }
  }

  /**
   * Private method to break up the text into segments that can fit in
   * the renderer (hopefully) without a horizontal scrollbar.
   * @param fm font metrics to use to perform text size calculations
   * @param lc layout context to update
   */
  private void segmentize(FontMetrics fm, LayoutContext lc)
  {
    StringTokenizer tok = new StringTokenizer(text, " \n");

    String soFar = "";
    int    soFarWidth = 0;

    String next;
    int    nextWidth;

    String textWithNext;
    int    textWithNextWidth = 0;

    // for now the width is what's left on this line
    int remainingWidth = 0;

    int spaceWidth = fm.getTextWidth(" ");

    boolean firstSegment = true;

    // leave space for bullet.
    if ((attr.battr & Attributes.LISTITEM) != 0)
    {
      soFarWidth = 10;

      heading = attr.list.getAndIncrement();
    }

    while(tok.hasMoreTokens())
    {
      next = tok.nextToken();
      nextWidth = fm.getTextWidth(next);

      remainingWidth = (lc.rendererWidth - lc.x);

      if (soFar.length() == 0)
      {
         textWithNext = next;
         textWithNextWidth = nextWidth;//fm.getTextWidth(next);
      }
      else
      {
         textWithNext = soFar + " " + next;
         textWithNextWidth = soFarWidth + spaceWidth + nextWidth;
      }

      if (textWithNextWidth > remainingWidth)
      {
         // if this is such a long line that can't be wrapped.
         if ((lc.x == 0) && (soFar.length() == 0))
         {
           // do it anyway
           //System.out.println("makeSegment: 1");
           makeSegment(next, lc, textWithNextWidth, true, firstSegment, tok.hasMoreTokens(), fm);
           firstSegment = false;
           soFar = "";
           soFarWidth = 0;
         }
         else
         {
           //System.out.println("makeSegment: 2");
           makeSegment(soFar, lc, soFarWidth, true, firstSegment, true, fm);
           firstSegment = false;
           soFar = next;
           soFarWidth = nextWidth;
         }
      }
      else
      {
         if (tok.getLastDelimeter() == '\n')
         {
           // break here!
           //System.out.println("makeSegment: 3");
           makeSegment(textWithNext, lc, textWithNextWidth, true, firstSegment, tok.hasMoreTokens(), fm);
           firstSegment = false;
           soFar = "";
           soFarWidth = 0;
         }
         else
         {
           soFar = textWithNext;
           soFarWidth = textWithNextWidth;
         }
      }
    }
    if (soFar.length() != 0)
    {
      //System.out.println("makeSegment: 4");
      makeSegment(soFar, lc, soFarWidth, (soFarWidth > remainingWidth), firstSegment, false, fm);
      firstSegment = false;
    }
  }

  /**
   * This method is called under special circumstances where the text
   * should be broken only when a newline is encountered in the input.
   * @param fm font metrics to use to perform text size calculations
   * @param lc layout context to update
   */
  public void newlineSegmentize(FontMetrics fm, LayoutContext lc)
  {
    //System.out.println("newline tokenize: '" + text + "'");
    StringTokenizer tok = new StringTokenizer(text, "\n");

    int width;
    String next;
    boolean firstSegment = true;
    while(tok.hasMoreTokens())
    {
      next = tok.nextToken();
      if (next.length() != 0)
      {
         width = fm.getTextWidth(next);
         makeSegment(next, lc, width, tok.getLastDelimeter() == '\n', firstSegment, tok.hasMoreTokens(), fm);
         firstSegment = false;
      }
    }
  }

  /**
   * Private method to add a segment to the list of segments.
   */
  private void makeSegment(String txt, LayoutContext lc, int width, boolean lineComplete, boolean firstSegment, boolean moreSegments, FontMetrics fm)
  {
    int height = fm.getHeight();

    if (firstSegment)
      height += attr.topborder;

    if (!moreSegments)
      height += attr.bottomborder;

    if (height > lc.lineHeight)
      lc.lineHeight = height;

    if (attr.alignment == Attributes.ALIGN_CENTER)
    {
      int diff = (lc.rendererWidth-width);
      if (diff >= 0)
         lc.x = diff/2;
    }

    if (txt.length() != 0)
    {
      Segment s = new Segment(txt);

      if ((attr.battr & Attributes.LISTITEM) != 0)
         width += 12;


      // before adding text, make sure it is indented correctly.
      if (lc.textOnLine == false)
         lc.x = (int)(attr.indent);


      s.rect.x = lc.x;
      s.rect.y = lc.y;
      if (firstSegment)
         s.rect.y +=  attr.topborder;
      s.rect.width = width;
      s.rect.height = fm.getHeight();

      segments.add(s);
      region.add(s.rect);
    }

    if (width > lc.preferredWidth)
    {
      lc.preferredWidth = width;
      //System.out.println("max width became " + lc.preferredWidth);
    }

    // end this line
    if (((attr.battr & Attributes.SEPARATE_LINE) != 0) || (lineComplete))
    {
      breakLine(lc, true);
    }
    else
    {
      lc.textOnLine = true;
      lc.x += width;
    }
  }

  /**
   * Implement the onPaint method to draw the text into the renderer.
   * @param r renderer to draw into.
   * @param g graphics context to draw with.
   */
  public void onPaint(HtmlRenderer r, Graphics g)
  {
    g.setForeColor(Color.BLACK);
    g.setFont(font);
    //FontMetrics fm = r.getFontMetrics(font);

    int subSuperOffset = 0;
    if ((attr.battr & Attributes.SUBSCRIPT) != 0)
      subSuperOffset = 3;
    else
    if ((attr.battr & Attributes.SUPERSCRIPT) != 0)
      subSuperOffset = -3;

    int n = segments.size();
    int offset = 0;//(int)(fm.getHeight()*vborder);
    for (int i =0; i < n; i++)
    {
      Segment s = (Segment)segments.items[i];
      //g.drawText(s.txt, -1*hOffset, ypos);
      int xpos = s.rect.x;
      if ((attr.battr & Attributes.LISTITEM) != 0)
      {
         if (attr.list.listType == ListAttr.LIST_BULLET)
         {
            //System.out.println("rect size: " + s.rect.height);
            int ypos = subSuperOffset + s.rect.y + offset + (s.rect.height - 5)/2;

            //System.out.println("drawing bullet at " + s.rect.x + " " + ypos);
            g.drawImage(bullet, s.rect.x, ypos);
         }
         else
         {
            int ypos = subSuperOffset + s.rect.y + offset;
            //System.out.println("drawing bullet at " + s.rect.x + " " + ypos);
            g.drawText(heading, s.rect.x, ypos);
         }
         xpos += 12;
      }
      //System.out.println("drawing " + s.txt + " at " + xpos + " " + String.valueOf(s.rect.y + offset));
      g.drawText(s.txt, xpos, s.rect.y+offset+subSuperOffset);
      if ((attr.battr & Attributes.UNDERLINE) != 0)
      {
         int ypos = s.rect.y + fm.getAscent() + subSuperOffset;
         g.drawLine(xpos, ypos,
         xpos + s.rect.width,
         ypos);
      }

      if ((attr.battr & Attributes.STRIKETHROUGH) != 0)
      {
         int ypos = s.rect.y + fm.getHeight()/2 + subSuperOffset;
         g.drawLine(xpos, ypos,
         xpos + s.rect.width,
         ypos);
      }
    }
  }

  /**
   * Implement the mouseEnter method to set a message in the browser
   * when the "mouse" is over a link.
   * @param renderer html renderer to set message in.
   * @param event mouse event
   */
  public void mouseEnter(HtmlRenderer renderer, Event event)
  {
    if (attr.href != null)
      renderer.setMessage(attr.href);
  }

  /**
   * Implement the mouseExit method to reset the message.
   * @param renderer html renderer to set message in.
   * @param event mouse event
   */
  public void mouseExit(HtmlRenderer renderer, Event event)
  {
    if (attr.href != null)
      renderer.setMessage("");
  }

  /**
   * Implement the mousePress method to follow a link.
   * @param renderer html renderer to set message in.
   * @param event pen event that caused this action to be performed.
   */
  public void mousePress(HtmlRenderer renderer, PenEvent event)
  {
    if (attr.href != null)
      renderer.linkSelection(attr.href);
  }
}
