/*****************************************************************************
 *                                Waba Extras
 *
 * Version History
 * Date                Version  Programmer
 * ----------  -------  -------  ------------------------------------------
 * 17/10/1999  New      1.0.0    Rob Nielsen  Class created
 * 10-Jun-2001          2.0.0    Mr. Tines    Subclass waba.io.DataStream
 *
 ****************************************************************************/

package superwaba.ext.xplat.io;

import waba.sys.*;

/**
 * DataStream is a wrapper you can place around any Stream such as a
 * SerialPort, Catalog, or BufferStream which lets you read and write
 * standard Waba data types like ints, floats, and Strings in a simple
 * manner.  Here's an example
 * <pre>
 *   SerialPort port=new SerialPort(9600,0);
 *   DataStream ds=new DataStream(port);
 *   ds.writeString("Hello");
 *   int status=ds.readUnsignedByte();
 *   if (status==1)
 *   {
 *     ds.writeString("Pi");
 *     ds.writeFloat(3.14);
 *   }
 *   port.close();
 * </pre>
 */
public class DataStream extends waba.io.DataStream
{
  /**
   * Constructs a new DataStream which sits upon the given stream using
   * the big endian notation for multibyte values.
   * @param stream the base stream
   */
  public DataStream(waba.io.Stream stream)
  {
    super(stream);
  }

  /**
   * Reads a C-style string from the stream.  This is a NUL (0) terminated
   * series of characters.  This format is commonly used by other Palm
   * applications.
   * @return the loaded String
   */
  public String readCString()
  {
    // if we've got a BufferStream under us, we can optimise the read.
    if (stream instanceof BufferStream)
    {
      byte[] b=((BufferStream)stream).readBytesUntil((byte)0);
      if (b==null)
        return "";
      readByte(); // clear the zero from the stream
      int size=b.length;
      char[] c=new char[size];
      for(int i=0;i<size;i++)
        c[i]=(char)b[i];
      b=null;
      return new String(c);
    }

    // otherwise we have to read one character at a time.
    int cnt=0;
    char[] c=new char[100];
    while (true)
    {
      stream.readBytes(b,0,1);
      if (b[0]==0)
        break;
      if (cnt>=c.length)
      {
        char[] temp=new char[c.length*2];
        Vm.copyArray(c,0,temp,0,cnt);
        c=temp;
      }
      c[cnt++]=(char)b[0];
    }
    return new String(c,0,cnt);
  }

  /**
   * Reads a fixed length string from the stream.  The given number of
   * characters are read and converted to a String.
   * @param length the number of characters to read
   * @return the loaded string
   */
  public String readFixedString(int length)
  {
    char[] c=new char[length];
    byte[] b=new byte[length];
    stream.readBytes(b,0,length);
    for(int i=0;i<length;i++)
      c[i]=(char)b[i];
    return new String(c);
  }

  /**
   * Writes a fixed length string to the stream.  If the given string is
   * longer than the given length, it will be truncated and if it is shorter,
   * it will be padded with spaces.
   * @param s the string to write
   * @param length the length of the fixed string
   */
  public void writeFixedString(String s,int length)
  {
    writeFixedString(s,length,' ');
  }

  /**
   * Writes a fixed length string to the stream.  If the given string is
   * longer than the given length, it will be truncated and if it is shorter,
   * it will be padded the given pad character.
   * @param s the string to write
   * @param length the length of the fixed string
   * @param the character to pad if the string is shorter than the length
   */
  public void writeFixedString(String s,int length,char pad)
  {
    if (length==0)
      return;
    byte[] b=new byte[length];
    int slen=0;
    if (s!=null)
    {
      char[] c=s.toCharArray();
      slen=c.length;
      for(int i=0;i<slen&&i<length;i++)
        b[i]=(byte)c[i];
    }
    for(int i=slen;i<length;i++)
      b[i]=(byte)pad; // pad the rest
    stream.writeBytes(b,0,length);
  }
}