package superwaba.ext.xplat.io;

import waba.io.*;
import waba.sys.*;

/**
 * This is the client-side class. It implements a pdbconduit-through-a-serial-port.
 * Created by Guich from the SerialSocket class created by Rick Grehan.
 * See examples/io/PDBConduit.
 */
public class SerialPDB
{
   public static final byte PUT_CATALOG = 1;
   public static final byte GET_CATALOG = 2;
   public static final byte PUT_TXT = 3;
   public static final byte GET_TXT = 4;
   public static final byte DELETE = 5;
   public static final byte SHUTDOWN = 99;

   /** 10 sec timeout in millisecs for default */
   private int timeout = 10000;

   /** Serial port object */
   protected SerialPort sport;

   /** DataStream object */
   protected DataStream ds;

   /** temp buffer */
   protected byte []buf = new byte[16];

   /** Create and open the Serial connection */
   public SerialPDB(int baud, int timeout)
   {
      // Open serial port 0
      sport = new SerialPort(0, baud);
      if(!sport.isOpen())
      {
         sport.close();
         sport = null;
      }
      else
      {
         this.timeout = timeout;
         sport.setReadTimeout(timeout);
         ds = new DataStream(sport);

         // Flush the port
         flush();
      }
   }

   /** Return true if we're connected */
   public boolean isOpen()
   {
      return sport != null;
   }

   /** Send the specified catalog to the server. If onlyDirty is true,
     * only dirty records will be sent. You may also specify resetDirtyAttr to reset the
     * flag in all record sync'ed. startRecord and endRecord can
     * be -1 to send all Catalog.
     * You may specify in the name the full path to the catalog.
     */
   public boolean put(Catalog cat2send, String name, int startRecord,
       int endRecord, boolean onlyDirty, boolean resetDirtyAttr)
   {
      // check parameters
      if (startRecord < 0) startRecord = 0;
      if (endRecord <= 0) endRecord = cat2send.getRecordCount();
      if (!cat2send.isOpen() || endRecord <= 0) // closed or empty catalog?
         return false;
      flush();
      cat2send.setRecordPos(-1); // unlock any locked records
      // start communication with server
      ds.writeByte(PUT_CATALOG);
      ds.writeString(name);
      // get ack for parameters
      if (!ds.readBoolean())
         return false;
      // loop through the records and send each one.
      for (int i = startRecord; i < endRecord; i++)
         if (cat2send.setRecordPos(i))
         {
            byte attr = cat2send.getRecordAttributes();
            if (onlyDirty && (attr & Catalog.REC_ATTR_DIRTY) == 0)
               continue;
            int size = cat2send.getRecordSize();
            // need to grow the buffer?
            if (size > buf.length)
               buf = new byte[size];
            cat2send.readBytes(buf,0,size);
            // send the bytes to the server
            ds.writeInt(size);
            ds.writeBytes(buf,0,size);
            if (resetDirtyAttr)
               cat2send.setRecordAttributes((byte)(attr & ~Catalog.REC_ATTR_DIRTY));
         }
      // tell the server we're done
      ds.writeInt(-1);
      return ds.readBoolean();
   }

   /** Requests the server to send the specified catalog, placing the records
     * in the given fillIn catalog. startRecord and endRecord can be -1 to bring
     * the full Catalog. If onlyDirty is true, only dirty records will be received.
     * You may also specify resetDirtyAttr to reset the flag in all record sync'ed.
     * You may specify in the name parameter the full path to the file.
     */
   public boolean get(Catalog fillIn, String name, int startRecord, int endRecord,
       boolean onlyDirty, boolean resetDirtyAttr)
   {
      flush();
      // start communication with server
      ds.writeByte(GET_CATALOG);
      ds.writeString(name);
      ds.writeInt(startRecord);
      ds.writeInt(endRecord);
      ds.writeBoolean(onlyDirty);
      ds.writeBoolean(resetDirtyAttr);
      // get ack for parameters
      if (!ds.readBoolean())
         return false;
      for (int i = 0; ; i++)
      {
         int size = ds.readInt();
         //Vm.debug("record #"+i+" size: "+size);
         if (size == -1) // end?
            break;
         if (size > buf.length) buf = new byte[size];
         int read = ds.readBytes(buf,0,size);
         if (read == 0)
         {
            Vm.debug("read 0 because "+sport.lastError);
            ds.writeInt(-1);
            return false;
         }
         if (fillIn.addRecord(size) == -1) // could not add record?
            return false;
         fillIn.writeBytes(buf,0,size);
         ds.writeBoolean(true); // flow control - without this a further readInt will probably fail
      }
      return true;
   }

   /** Send the following String array to the server, placing it in the given txt filename.
     * You may specify the full path to the fileName.
     */
   public boolean put(String fileName, boolean deleteIfExists, String []strings)
   {
      flush();
      // start communication with server
      ds.writeByte(PUT_TXT);
      ds.writeString(fileName);
      ds.writeBoolean(deleteIfExists);
      // get ack for parameters
      if (!ds.readBoolean())
         return false;
      ds.writeStringArray(strings);
      return ds.readBoolean();
   }

   /** Get a String array from the server from the specified text file.
     * You may specify the full path to the fileName.
     */
   public String[] get(String fileName)
   {
      flush();
      // start communication with server
      ds.writeByte(GET_TXT);
      ds.writeString(fileName);
      // get ack for parameters
      if (!ds.readBoolean())
         return null;
      return ds.readStringArray();
   }

   /** Delete a catalog at the server. This method does NOT delete a txt (txt files are
     * always overwritten)! */
   public void delete(String fileName)
   {
      flush();
      // start communication with server
      ds.writeByte(DELETE);
      ds.writeString(fileName);
      // get ack
      ds.readBoolean();
   }

   /** Shutdown the server. */
   public void shutdownServer()
   {
      flush();
      ds.writeByte(SHUTDOWN);
   }


   // Flush any input
   protected void flush()
   {
      // Set timeout for instant return
      sport.setReadTimeout(0);

      while(sport.readCheck()!=0)
         sport.readBytes(buf,0,1);

      // Return timeout to normal
      sport.setReadTimeout(timeout);
   }
}
