package superwaba.ext.xplat.io;

import waba.io.*;
import waba.sys.*;
import waba.util.*;

/** Proxy for Serial PDB. You may extend this class to provide any other
  * functionalities you may want.
  * See examples/io/PDBConduit.
  */

public class SerialPDBServer
{
   public static final byte PUT_CATALOG = 1;
   public static final byte GET_CATALOG = 2;
   public static final byte PUT_TXT = 3;
   public static final byte GET_TXT = 4;
   public static final byte DELETE = 5;
   public static final byte SHUTDOWN = 99;

   SerialPort sport;
   DataStream ds;
   byte []buf = new byte[16];

   /** Construct a serial server to listen at <code>port</code> using <code>baud</code>
     * rate */
   public SerialPDBServer(int port, int baud)
   {
      // Open the serial port
      sport = new SerialPort(port,baud);
      sport.setReadTimeout(100000000);
      // attach a DataStream to it to make things easier
      ds = new DataStream(sport);
   }

   public void putCatalog()
   {
      // read parameters
      String name = ds.readString();
      Catalog cat = new Catalog(name,Catalog.CREATE);
      // send ack for parameters
      if (!cat.isOpen())
      {
         ds.writeBoolean(false);
         return;
      }
      ds.writeBoolean(true);
      int errCount = 0;

      println("Putting catalog "+name);
      int received = 0;

      for (int i=0;; i++)
      {
         int size = ds.readInt();
         if (size == -1)
            break;
         //println("record size "+size+" being created");
         if (buf.length < size) buf = new byte[size];
         ds.readBytes(buf,0,size);
         received++;
         //println("Received record #"+i+" with "+size+" bytes");
         if (cat.addRecord(size) != -1)
            cat.writeBytes(buf,0,size);
         else
            errCount++;
      }
      cat.close();
      println("Received "+received+" records. Success? "+(errCount==0));
      ds.writeBoolean(errCount == 0);
      flush();
   }

   public void getCatalog()
   {
      // read parameters
      String name = ds.readString();
      int startRecord = ds.readInt();
      int endRecord = ds.readInt();
      boolean onlyDirty = ds.readBoolean();
      boolean resetDirtyAttr = ds.readBoolean();
      Catalog cat = new Catalog(name, Catalog.READ_ONLY);
      // send ack for parameters
      if (!cat.isOpen())
      {
         ds.writeBoolean(false);
         return;
      }
      ds.writeBoolean(true);

      println("Getting catalog "+name);
      // check parameter limits
      int n = cat.getRecordCount();
      if (startRecord < 0) startRecord = 0;
      if (endRecord > n || endRecord <= 0) endRecord = n;
      int sent = 0;

      for (int i = startRecord; i < endRecord; i++)
         if (cat.setRecordPos(i))
         {
            byte attr = cat.getRecordAttributes();
            if (onlyDirty && (attr & Catalog.REC_ATTR_DIRTY) == 0)
               continue;
            // read from catalog
            int size = cat.getRecordSize();
            if (size > buf.length) buf = new byte[size];
            cat.readBytes(buf,0,size);

            // and write to the serial
            ds.writeInt(size);
            ds.writeBytes(buf,0,size);
            if (resetDirtyAttr)
               cat.setRecordAttributes((byte)(attr & ~Catalog.REC_ATTR_DIRTY));
            sent++;
            ds.readBoolean(); // flow control - note: this causes the transfer to be very much slow
         }
      println("Sent "+sent+" records");
      cat.close();
      ds.writeInt(-1);
      flush();
   }

   public void putTxt()
   {
      String name = ds.readString();
      boolean deleteIfExists = ds.readBoolean();
      java.io.FileOutputStream fos = null;
      try
      {
         java.io.File f = new java.io.File(name);
         if (f.exists() && deleteIfExists)
            f.delete();
         fos = new java.io.FileOutputStream(f);
      }
      catch (Exception e)
      {
         // send ack for parameters
         ds.writeBoolean(false);
         return;
      }

      println("Putting txt "+name);
      // send ack for parameters
      ds.writeBoolean(true);
      // write the txt here...
      java.io.PrintWriter pw = new java.io.PrintWriter(fos);
      String []strings = ds.readStringArray();
      for (int i = 0; i < strings.length; i++)
         pw.println(strings[i]);
      pw.close();
      ds.writeBoolean(true);
      flush();
   }

   public void getTxt()
   {
      // start communication with server
      String name = ds.readString();
      java.io.FileInputStream fis = null;
      try
      {
         fis = new java.io.FileInputStream(name);
      }
      catch (Exception e)
      {
         // send ack for parameters
         ds.writeBoolean(false);
         return;
      }

      Vector v = new Vector(100);
      try
      {
         readFile(v, fis);
      } catch (Exception e) {v.removeAllElements();}
      // empty file? return nack
      int n = v.size();
      if (n == 0)
      {
         ds.writeBoolean(false);
         return;
      }

      println("Getting txt "+name);
      // send ack for parameters
      ds.writeBoolean(true);

      String []strings = (String [])v.toObjectArray();
      ds.writeStringArray(strings);
      flush();
   }

   public void delete()
   {
      // read parameters
      String name = ds.readString();
      Catalog cat = new Catalog(name,Catalog.READ_ONLY);
      if (cat.isOpen())
      {
         println("Deleting "+name);
         cat.delete();
      }

      // send ack
      ds.writeBoolean(true);
      flush();
   }


   // reads a text file
   public static void readFile(Vector v, java.io.InputStream is) throws Exception
   {
      v.removeAllElements();
      try
      {
         int b = is.read();
         char c;
         String s = "";
         while (b >= 0)
         {
            c = (char)b;
            if (c == '\n')
            {
               v.addElement(s);
               s = "";
            } else s += c;
            b = is.read();
            if ((char)b == '\r') b = is.read();
         }
         if (s.length() > 0)
            v.addElement(s);
      }
      finally
      {
         is.close();
      }
   }


   // Flush any input
   private void flush()
   {
      byte []tbuf = new byte[2];
      // Set timeout for instant return
      sport.setReadTimeout(0);

      while(sport.readCheck()!=0)
         sport.readBytes(tbuf,0,1);

      // Return timeout to normal
      sport.setReadTimeout(100000000);
   }

   public void listen()
   {
      waba.applet.JavaBridge.setNonGUIApp();

      println("Idle");
      flush();
      // See if there's anything to be done
      while (true)
      {
         int command = ds.readByte();

         // Get the command and act accordingly
         switch(command)
         {
            case PUT_CATALOG: putCatalog(); break;
            case GET_CATALOG: getCatalog(); break;
            case PUT_TXT:     putTxt();     break;
            case GET_TXT:     getTxt();     break;
            case DELETE:      delete();     break;
            case SHUTDOWN:    println("Server being shutdown by the client..."); return;
            default:
               println("Unknown command received: "+command+". Flushing stream...");
               flush();
         }
      }
   }

   private static boolean showDebug=true;
   private static void println(String s)
   {
      if (showDebug) System.out.println(s);
   }

   public static void main(String args[])
   {
      if (args.length == 0)
         Vm.debug("Syntax: SerialPDBServer port baud </q>\n/q is optional and will hide all messages");
      else
      {
         int port = Convert.toInt(args[0]);
         if (port == 0) port = 1;
         int baud = Convert.toInt(args[1]);
         if (baud == 0) baud = 19200;
         if (args.length == 3)
            showDebug = false;
         new SerialPDBServer(port,baud).listen();
      }
   }
}
