/*********************************************************************************
 *  Copyright (C) 2002 Guilherme Campos Hazan <guich@superwaba.org.>             *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library is NOT free software.                                           *
 *  Once you paid for it, you can use it in how many devices you want.           *
 *                                                                               *
 *********************************************************************************/

package superwaba.ext.xplat.io.print;

import superwaba.ext.xplat.io.print.PrinterException;

/** This class represents a print manager, which is extended by the drivers to
  * provide a common way to handle the printer. Please be careful when using
  * these methods; any error may freeze the device.
  */
public abstract class PrintManager
{
   /** The Graphics instance that may be assigned to this print manager */
	protected Graphics g;
	/** The TypeWriter instance that may be assigned to this print manager */
	protected TypeWriter tw;
	/** The Settings instance that may be assigned to this print manager */
	protected Settings set;

   /** This class cannot be instantiated by its constructor */
	protected PrintManager()
	{
	}

   /** This method must be extended and must set the fields Graphics, TypeWriter and Settings accordingly.
     * Not all fields are required to be used.
     */
	public abstract void open(String printerName, String portName, int baudRate, boolean textMode) throws PrinterException;

   /** May return the Graphics instance if this print manager was opened using textMode = true. Will throw NotSupportedException otherwise. */
	public Graphics getGraphics() throws NotSupportedException
	{
		if (g == null) throw new NotSupportedException("Either you have openned this print manager in text mode or the driver is not capable to render graphics");
		return g;
	}

   /** May return the TypeWritter instance if this print manager was opened using textMode = false. Will throw NotSupportedException otherwise. */
	public TypeWriter getTypeWriter() throws NotSupportedException
	{
		if (tw == null) throw new NotSupportedException("Either you have openned this print manager in graphics mode or the driver is not capable to act as a typewriter");
		return tw;
	}

	/** Will return the Settings instance assigned to this print manager */
	public Settings getSettings()
	{
		return set;
	}

   /** Gets a FontMetrics object for the given font */
	public abstract FontMetrics getFontMetrics(Font f);

	/** Closes this print manager. */
	public abstract void close() throws PrinterException;

	/** Do a form feed. You must always call formFeed before calling finishPage. */
	public abstract void formFeed() throws PrinterException, NotSupportedException;

	/** Start a new page. You must always start a new page before writting to the
	  * printer. */
	public abstract void startPage() throws PrinterException;

	/** Ends a page, flushing the contents. You must always call formFeed before calling finishPage. */
	public abstract void finishPage() throws PrinterException;

	/** Send bytes to the printer, usualy scape codes and contents */
	public abstract int writeBytes(byte []buf, int start, int len) throws PrinterException, ArrayIndexOutOfBoundsException, NotSupportedException;

   /** Given the PrinterException, returns a detailed error message for the
     * problem.
     */
   public abstract String getDetailedErrorMessage(PrinterException e);
}