/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2001-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  Copyright (C) 2003 Dave Slaughter <dslaughter@safeway.co.uk>                 *
 *  Copyright (C) 2000 Tom Cuthill                                               *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package superwaba.ext.xplat.io.scanner;

import waba.sys.Settings;
import waba.ui.*;

/**
 * Scanner accesses the Symbol barcode scanner.
 * <p>
 * Scanner access is only available when running under a native SuperWaba VM, it
 * is not supported when running under Java.
 * <p>
 * Since there is only one scanner per device, the scanner class is static.
 * Activating the scanner causes the physical scanner to be powered, and
 * enables the trigger.  Deactivating the scanner removes power from the
 * scanner and prevents scans from taking place.  The scanner should always
 * be deactivated at the end of processing to prevent excessive battery drain.
 * <p>
 * When the scanner is activated, scan events will appear in the MainWindow's
 * onEvent method.  The scan events will contain a String describing either
 * the item scanned or a battery error. Deactivating the scanner prevents
 * scan events from being delivered.
 * <p>
 * You can change this behaviour by calling setContinuousScanning(false),
 * which will cause each call to activate() to only allow a single scan.
 * <p>
 * Since barcodes can have many formats, this class includes a method to
 * register the barcode types with the scanner.  These types must each be
 * individually set as parameters to the scanner.  After all the parameter
 * types have been set, the commitBarcodeParams method must be called to
 * register the parameters with the scanner.  Once this is done the scanner
 * is able to decode the specified barcode types.
 * <p>
 * A typical processing sequence is given below:
 * <pre>
 * if (!Scanner.active())
 *     return;
 * if (!Scanner.setBarcodeParam(Scanner.BARCODABAR, true))
 *     return;
 * if (!Scanner.setBarcodeParam(Scanner.BARBOOKLAND_EAN, true))
 *     return;
 * if (!Scanner.commitBarcodeParams())
 *     return;
 * ...
 * if (!Scanner.deactivate())
 *      return;
 *
 * </pre>
 */

public class Scanner
{
   /**
   * Barcode types
   */

   /** To be used in the setBarcodeParam method */
   public final static int BARCODE39           = 0;
   /** To be used in the setBarcodeParam method */
   public final static int BARUPCA             = 1;
   /** To be used in the setBarcodeParam method */
   public final static int BARUPCE             = 2;
   /** To be used in the setBarcodeParam method */
   public final static int BAREAN13            = 3;
   /** To be used in the setBarcodeParam method */
   public final static int BAREAN8             = 4;
   /** To be used in the setBarcodeParam method */
   public final static int BARD2OF5            = 5;
   /** To be used in the setBarcodeParam method */
   public final static int BARI2OF5            = 6;
   /** To be used in the setBarcodeParam method */
   public final static int BARCODABAR          = 7;
   /** To be used in the setBarcodeParam method */
   public final static int BARCODE128          = 8;
   /** To be used in the setBarcodeParam method */
   public final static int BARCODE93           = 9;
   /** To be used in the setBarcodeParam method */
   public final static int BARMSI_PLESSEY      = 11;
   /** To be used in the setBarcodeParam method */
   public final static int BARUPCE1            = 12;
   /** To be used in the setBarcodeParam method */
   public final static int BARTRIOPTICCODE39   = 13;
   /** To be used in the setBarcodeParam method */
   public final static int BARUCC_EAN128       = 14;
   /** To be used in the setBarcodeParam method */
   public final static int BARBOOKLAND_EAN     = 83;
   /** To be used in the setBarcodeParam method */
   public final static int BARISBT128          = 84;
   /** To be used in the setBarcodeParam method */
   public final static int BARUPCEANCOUPONCODE = 85;

   /** To be used in the setParam method */
   public final static int TRIGGERING_PARAM             = 1;
   /** To be used in the setParam method */
   public final static int LINEAR_SECURITY_LEVEL_PARAM  = 2;
   /** To be used in the setParam method */
   public final static int SUPPLEMENTALS_PARAM          = 3;
   /** To be used in the setParam method. Needs barcodeType. */
   public final static int TRANSMIT_CHECKDIGIT_PARAM    = 4;
   /** To be used in the setParam method. Needs barcodeType. */
   public final static int PREAMBLE_PARAM               = 5;
   /** To be used in the setParam method */
   public final static int MSI_PLESSEY_CHECKDIGIT_PARAM = 7;
   /** To be used in the setParam method */
   public final static int MSI_PLESSEY_OPTIONS_PARAM    = 8;
   /** To be used in the setParam method */
   public final static int MSI_PLESSEY_ALGORITHMS_PARAM = 9;
   /** To be used in the setParam method */
   public final static int TRANSMIT_CODEID_PARAM        = 10;
   /** To be used in the setParam method */
   public final static int SCAN_DATA_TRANSMISSION_PARAM = 12;
   /** To be used in the setParam method */
   public final static int SCAN_ANGLE_PARAM             = 13;

   public static boolean isActive = false; // guich@200b4: avoid activate the scanner again

   /**
    * Flag which can be set to false so activate() only performs one scan
    * <p>
    * Note: the Symbol Scanner API differs on Palm and PocketPC (the Java
    * emulation works the same as PocketPC).  Under Palm activating the scanner
    * will allow the user to scan any number of barcodes until the scanner is
    * deactivated.  Under Java/PocketPC activating the Scanner only schedules
    * a single scan.
    * <p>
    * We obviously need this class act the same on both platforms.  Rather than
    * forcing one choice on the user, the continuousScanning flag allows the user
    * to choose which mode they want.
    */
   private static boolean continousScanning = true;

   public static waba.ui.Control listener; // maybe theres another better way to do this...

   /**
   * Activate the scanner. Return true if the scanner could be activated,
   * false otherwise.  (A false condition may arise if batteries have run
   * too low, or if the unit does not have the scan library installed.
   */
   public static boolean activate()
   {
      isActive = NativeMethods4JDK.scannerActivate();
      return isActive;
   }

   /**
   * Set a scanner parameter defining whether a barcode format will be used
   * in scanning.  Choose from one of the barcode types given above.
   * @param barcodeType the type of barcode under consideration
   * @param enable a flag to enable or disable decoding barcodes of this type.
   */
   public static boolean setBarcodeParam(int barcodeType, boolean enable)
   {
      return NativeMethods4JDK.scannerSetBarcodeParam(barcodeType, enable);
   }
   /**
   * Set whether activate() will continuously scan (the default),
   * or stop after a single scan.
   * @param continuous true if activate() continuously scans, false if it will
   * only schedule a single scan
   */
   public static void setContinuousScanning(boolean continuous)
   {
	  continousScanning = continuous;
   }

   /**
     * Set a parameter for the barcode. You can use the xxx_PARAM constants to
     * know which parameter to set, and also the values below for their values.
     * Be careful because some functions may require also the barcode type.
     * <pre>
     *    // triggering modes
     *    #define LEVEL                                       0x00
     *    #define PULSE                                       0x02
     *    #define HOST                                        0x08
     *
     *    // Linear code type security
     *    #define SECURITY_LEVEL0                             0x00
     *    #define SECURITY_LEVEL1                             0x01
     *    #define SECURITY_LEVEL2                             0x02
     *    #define SECURITY_LEVEL3                             0x03
     *    #define SECURITY_LEVEL4                             0x04
     *
     *    // UPC/EAN Supplementals
     *    #define IGNORE_SUPPLEMENTALS                        0x00
     *    #define DECODE_SUPPLEMENTALS                        0x01
     *    #define AUTODISCRIMINATE_SUPPLEMENTALS              0x02
     *
     *    // Transmit Check Digit options
     *    #define DO_NOT_TRANSMIT_CHECK_DIGIT                 0x00
     *    #define TRANSMIT_CHECK_DIGIT                        0x01
     *
     *    // Preamble options
     *    #define NO_PREAMBLE                                 0x00
     *    #define SYSTEM_CHARACTER                            0x01
     *    #define SYSTEM_CHARACTER_COUNTRY_CODE               0x02
     *
     *    // CheckDigit verification options
     *    #define DISABLE_CHECK_DIGIT 					           0x00
     *    #define USS_CHECK_DIGIT                             0x01
     *    #define OPCC_CHECK_DIGIT                            0x02
     *
     *    // MSI Plessey checkdigit options
     *    #define ONE_CHECK_DIGIT                             0x00
     *    #define TWO_CHECK_DIGITS                            0x01
     *
     *    // MSI Plessey check digit algorithms
     *    #define MOD10_MOD11                                 0x00
     *    #define MOD10_MOD10                                 0x01
     *
     *    // Transmit Code ID Character options
     *    #define AIM_CODE_ID_CHARACTER                       0x01
     *    #define SYMBOL_CODE_ID_CHARACTER                    0x02
     *
     *    // Scan data transmission formats
     *    #define DATA_AS_IS                                  0x00
     *    #define DATA_SUFFIX1                                0x01
     *    #define DATA_SUFFIX2                                0x02
     *    #define DATA_SUFFIX1_SUFFIX2                        0x03
     *    #define PREFIX_DATA                                 0x04
     *    #define PREFIX_DATA_SUFFIX1                         0x05
     *    #define PREFIX_DATA_SUFFIX2                         0x06
     *    #define PREFIX_DATA_SUFFIX1_SUFFIX2                 0x07
     *
     *    // Scan angle options
     *    #define SCAN_ANGLE_WIDE 					              0xB6
     *    #define SCAN_ANGLE_NARROW					              0xB5
     * </pre>
     */
   public static boolean setParam(int type, int barcodeType, int value) // guich@330_43
   {
      return NativeMethods4JDK.scannerSetParam(type, barcodeType, value);
   }

   /**
   * Commit the barcode parameters to the scanner. Returns true if the
   * operation is successful and false otherwise.
   *
   */
   public static boolean commitBarcodeParams()
   {
      return NativeMethods4JDK.scannerCommitBarcodeParams();
   }

   /**  Set the length of a barcode. The lengthType must be one of the following values:
     * <pre>
     *    // Length types for the barcode SetLengths calls
     *    #define ANY_LENGTH                                  0x00
     *    #define ONE_DISCRETE_LENGTH                         0x01
     *    #define TWO_DISCRETE_LENGTHS                        0x02
     *    #define LENGTH_WITHIN_RANGE                         0x03
     * </pre>
     */
   public static boolean setBarcodeLength(int barcodeType, int lengthType, int length1, int length2 ) // guich@330_43
   {
      return NativeMethods4JDK.scannerSetBarcodeLength(barcodeType, lengthType, length1, length2);
   }

   /**
   * Get the decoded string of what has been scanned.  If an error occurs a
   * null String will be returned.
   */
   public static String getData()
   {
      return NativeMethods4JDK.scannerGetData();
   }

   /**
   * Get the scan manager version as a String. If an error occurs or if this method
   * is called before the Scanner is initialized, a null String will be returned.
   */
   public static String getScanManagerVersion()
   {
      return !isActive?null:NativeMethods4JDK.scannerGetScanManagerVersion();
   }

   /**
   * Get the Scanner Port Driver version as a String. If an error occurs or if this method
   * is called before the Scanner is initialized, a null String will be returned.
   */
   public static String getScanPortDriverVersion()
   {
      return !isActive?null:NativeMethods4JDK.scannerGetScanPortDriverVersion();
   }

   /**
   * Deactivate the scanner. Returns true if the operation is successful
   * and false otherwise.
   */
   public static boolean deactivate()
   {
      if (isActive)
         isActive = !NativeMethods4JDK.scannerDeactivate();
      return !isActive;
   }

   /** Dispatch the event to the current listener (associated through the
   * <code>listener</code> public member. Called from the native library.
   * If there are no assigned listeners, the event is sent to the top most window,
   * and the window is validated.
   */
   protected static void _onEvent(int type)
   {
      Control dest = listener;
      if (dest == null)
         dest = Window.getTopMost();
      // convert from the type to the internal event id
      int id = type == 2 ? ScanEvent.BATTERY_ERROR :
               type == 1 ? ScanEvent.SCANNED
                  /* 0 */ : ScanEvent.TRIGGERED;
      // dispatch to the listener
      dest.onEvent(new ScanEvent(id));
      Window win = dest.getParentWindow(); // guich@400_44
      if (win != null)
         win.validate();
      if ("PalmOS".equals(Settings.platform))
      {
         // If we are running on a Palm unit then calling activate() on the Symbol Scanner
         // API will cause the scanner to allow any number of scans until deactivate is
         // called. So, if continuousScanning is set to false then we need to disable the
         // scanner
         if (!continousScanning)
            deactivate();
      } else
      {
         // If we are running on a PocketPC unit, or under Java emulation, then calling
         // activate() on the Symbol Scanner API only schedules a single scan.
         // So, if continuousScanning is true we need to schedule another scan.
         if (continousScanning)
            activate();
         else
            isActive = false;
      }
   }
}