/*********************************************************************************
 *  Copyright (C) 2003 Guilherme Campos Hazan <support@superwaba.com.br>         *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library is NOT free software.                                           *
 *  Once you paid for it, you can use it in how many devices you want.           *
 *                                                                               *
 *********************************************************************************/

package superwaba.ext.xplat.io.search;

/** With this class, you can do searches in a Catalog 5 to 10 times faster than in Java.
  * This class is stored in the SWCatSearch.pdb and is used with the SWCatSearch.prc file.
  * <p>
  * Here is an example:
  * <pre>
  * if (!Settings.onDevice || Vm.attachNativeLibrary("SWCatSearch"))
  * {
  *    Catalog cat = new Catalog("TestDB."+Settings.appCreatorId+".DATA", Catalog.READ_ONLY);
  *    CatalogSearch catSearch = new CatalogSearch(cat);
  *    String text = "a9999"; // string to search
  *    int offset = 0; // offset in record
  *    int startRec = 10; // start searching from record 10
  *    cat.setRecordPos(startRec);
  *
  *    // now we convert the String to a byte array
  *    // Note that we're searching for the exact string,
  *    // case sensitive and whole word only.
  *    int length = 0;
  *    ByteArrayStream bas = new ByteArrayStream(16);
  *    DataStream ds = new DataStream(bas);
  *    length += ds.writeString(text);
  *    byte []toSearch = bas.getBuffer(); // the buffer may be greater than length
  *
  *    int start = Vm.getTimeStamp();
  *    int recIndex = catSearch.searchBytes(toSearch, length, offset);
  *    int end = Vm.getTimeStamp();
  *    Vm.debug("Returned record: "+recIndex);
  *    Vm.debug("Elapsed: "+Vm.getTimeElapsed(end,start)+" ms");
  * }
  * </pre>
  * Some notes:<br>
  * . The library SWCatSearch.pdb is automaticaly loaded.<br>
  * . The recordOffset is useful when you write something before of
  *   what you wanna search. For example, if your record contains an
  *   int before a string, you may set recordOffset to 4 (the int is
  *   stored in the first 4 bytes).<br>
  * . Be careful: to search for a substring inside a String, you must
  *   remember to skip the first two bytes where the String length is
  *   stored.<br>
  * . The search starts from the current positioned record in the Catalog.<br>
  * . This class can be used in desktop.
  * <p>
  * This is not free software. This library can be purchased at the <a href="http://www.superwaba.com.br/add.asp">SuperWaba site</a>.
  */

public class CatalogSearch
{
	/** The underlying catalog */
   protected waba.io.Catalog cat;

   /** Creates a CatalogSearch object with the given parameter. Note
     * that the catalog <b>must</b> be opened. If you want to start
     * the search from a specific record index, just do a setRecordPos
     * in the catalog before calling <code>searchBytes</code>.
     */
   public CatalogSearch(waba.io.Catalog anopenedAndPositionedCatalog)
   {
      this.cat = anopenedAndPositionedCatalog;
   }

   /** Searches the underlying Catalog for the given byte array. Calls the
     * native method in SWCatSearch.prc (or a java code if running in desktop).
     * @param toSearch The byte array used to compare with the contents
     * of each record.
     * @param length How many bytes will be searched inside <code>toSearch</code>
     * (may be smaller than <code>oSearch.length</code>)
     * @param offsetInRec How many bytes to skip from the record start.
     * @return the record index if found, or -1 if toSearch is null, or
     * the Catalog is null or closed or empty, or length <= 0, or
     * length > toSearch.length.
     */
   public int searchBytes(byte []toSearch, int length, int offsetInRec)
   {
      return superwaba.ext.xplat.io.search.NativeMethods4JDK.catalogSearchSearchBytes(this, toSearch, length, offsetInRec);
   }
}