
/* Blob.java -- Access a SQL Binary Large OBject.
Copyright (C) 1999, 2000, 2002 Free Software Foundation, Inc.

This file is part of GNU Classpath.

GNU Classpath is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU Classpath is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Classpath; see the file COPYING.  If not, write to the
Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA.

Linking this library statically or dynamically with other modules is
making a combined work based on this library.  Thus, the terms and
conditions of the GNU General Public License cover the whole
combination.

As a special exception, the copyright holders of this library give you
permission to link this library with independent modules to produce an
executable, regardless of the license terms of these independent
modules, and to copy and distribute the resulting executable under
terms of your choice, provided that you also meet, for each linked
independent module, the terms and conditions of the license of that
module.  An independent module is a module which is not derived from
or based on this library.  If you modify this library, you may extend
this exception to your version of the library, but you are not
obligated to do so.  If you do not wish to do so, delete this
exception statement from your version. */
package superwaba.ext.xplat.sql;

/**
  * This interface specified methods for accessing a SQL BLOB (Binary
  * Large OBject) type.
  * This object has been greatly simplified for SW.
  * WARNING: In SuperWaba, it is allowable for a ResultSet to optimize access
  * by reusing the same instance of the Blob object across calls to getBlob(..)
  * This means that the content of a Blob object is valid only until getBlob(..)
  * is called again.  It is up to the developer to copy this data if it is needed
  * across multiple calls to getBlob(..)
  *
  * @author Aaron M. Renn (arenn@urbanophile.com)
  * @since 1.2
  */

public abstract class Blob
{
   
   /**
     * This method returns the number of bytes in the BLOB.
     *
     * @return The number of bytes in the BLOB.
     * @exception SQLException If an error occurs.
     */
   public abstract long length() throws SQLException;
   
   /**
     * This method returns up to the requested bytes of this BLOB as a
     * <code>byte</code> array.
     *
     * @param pos The index into the BLOB to start returning bytes from.
     * @param length The requested number of bytes to return.
     * @return The requested bytes from the BLOB.
     * @exception SQLException If an error occurs.
     */
   public abstract byte[] getBytes(long pos, int length) throws SQLException;
   
   /**
     * This method returns a byte array that may be larger than the actual
     * data retrieved.  The caller of this method must be sure to limit their
     * access to only the portion of the byte array that was actually retrieved.
     * That is, the element indices of 0 to Blob.length() - 1.
     * This method is unique to SuperWaba and was added to allow for column
     * binding optimization (to avoid creating a new byte array as getBytes must
     * usually do).
     * There is no promise given that this will be more effecient.  It is
     * dependent on the driver implementation, but it should never be less effecient
     * than calling getBytes(0, blob.getLength()).
     *
     * @return The internal byte buffer used for the BLOB.
     * @exception SQLException if an error occurs.
     */
   public abstract byte[] getByteBuffer() throws SQLException;
   
   /**
     * This method returns a stream that will read the bytes of the BLOB.
     * NOTE: removed from SW.
     * @return A stream that will read the bytes of the BLOB.
     * @exception SQLException If an error occurs.
     */
   
   //  public InputStream getBinaryStream() throws SQLException;
   
   /**
     * This method returns the index into the BLOB at which the first instance
     * of the specified bytes occur.  The searching starts at the specified
     * index into the BLOB.
     * NOTE: removed from SW
     * @param pattern The byte pattern to search for.
     * @param offset The index into the BLOB to starting searching for the pattern.
     * @return The offset at which the pattern is first found, or -1 if the
     *         pattern is not found.
     * @exception SQLException If an error occurs.
     */
   
   //  public long position(byte[] pattern, long start) throws SQLException;
   
   /**
     * This method returns the index into the BLOB at which the first instance
     * of the specified pattern occurs.  The searching starts at the specified
     * index into this BLOB.  The bytes in the specified <code>Blob</code> are
     * used as the search pattern.
     * NOTE: removed from SW
     * @param pattern The <code>Blob</code> containing the byte pattern to
     *        search for.
     * @param offset The index into the BLOB to starting searching for the pattern.
     * @return The offset at which the pattern is first found, or -1 if the
     *         pattern is not found.
     * @exception SQLException If an error occurs.
     */
   
   //  public long position(Blob pattern, long start) throws SQLException;
   
   /**
     * @exception SQLException If an error occurs.
     * @since 1.4
     */
   public abstract int setBytes(long pos, byte[] bytes) throws SQLException;
   
   /**
     * @exception SQLException If an error occurs.
     * @since 1.4
     */
   public abstract int setBytes(long pos, byte[] bytes, int offset, int len) throws SQLException;

/**
  * NOTE: removed from SW
  * @exception SQLException If an error occurs.
  * @since 1.4
  */

//  public OutputStream setBinaryStream(long pos) throws SQLException;

/**
  * NOTE: removed from SW
  * @exception SQLException If an error occurs.
  * @since 1.4
  */

//  public void truncate(long len) throws SQLException;
}
