
/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2003 Guilherme Campos Hazan <support@superwaba.com.br>         *
 *                 and Andy DePue <sir_anodos@yahoo.com>                         *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/
package superwaba.ext.xplat.sql;

/**
  * This class allows the driver to identify the value as a SQL Timestamp.
  * Please note that the JDBC specifies support for a 'nano' field, and
  * although the constructor forces you to pass in a nano field, this field is
  * completely ignored in SW.  The constructor exists as it is for compatibility
  * with JDBC code, and that is all.
  */

public class Timestamp extends Time
{

   /**
     * This method initializes a new instance of this class with the
     * specified year, month, day, hour, minute, second.
     * NOTE: in SW, nanos is completely ignored.
     * While this method is deprecated in the
     * latest JDBC, it is the only way to do it for now in SW, as SW does not
     * have a 'long' representation of time, and the conversion overhead is
     * considered too great for small devices (this may change in the future).
     *
     * @param year The year for this Timestamp (year - 1900)
     * @param month The month for this Timestamp (0-11)
     * @param day The day for this Timestamp (1-31)
     * @param hour The hour for this Timestamp (0-23)
     * @param minute The minute for this Timestamp (0-59)
     * @param second The second for this Timestamp (0-59)
     * @param nanos The nanosecond value for this Timestamp (0 to 999,999,9999)
     */
   public Timestamp(int year, int month, int day, int hour, int minute, int second, int nanos)
   {
      super(hour, minute, second);
      _date.year = year;
      _date.month = month;
      _date.day = day;
   }
   protected Timestamp(waba.sys.Time time)
   {
      super(time);
   }

   /**
     * This method returns this date in JDBC format.
     *
     * @return This date as a string.
     */
   public String toString()
   {
      if(_toString == null)
      {
         StringBuffer buf = new StringBuffer();
         buf.append(convertDateToString());
         buf.append(' ');
         buf.append(convertTimeToString());
         _toString = buf.toString();
      }
      return _toString;
   }

   /**
     * This methods tests whether this object is earlier than the specified
     * object.
     *
     * @param ts The other <code>Timestamp</code> to test against.
     * @return <code>true</code> if this object is earlier than the other object,
     *         <code>false</code> otherwise.
     */
   public boolean before(Timestamp ts)
   {
      waba.sys.Time t = _date;
      int myDate = (t.year * 10000) + (t.month * 100) + t.day;
      int hisDate = (ts.getYear() * 10000) + (ts.getMonth() * 100) + ts.getDate();
      return myDate < hisDate || (myDate == hisDate && ((t.hour * 10000) + (t.minute * 100) + t.second) < (ts.getHour() * 10000) + (ts.getMinute() * 100) + ts.getSecond());
   }

   /**
     * This methods tests whether this object is later than the specified
     * object.
     *
     * @param ts The other <code>Timestamp</code> to test against.
     *
     * @return <code>true</code> if this object is later than the other object,
     * <code>false</code> otherwise.
     */
   public boolean after(Timestamp ts)
   {
      waba.sys.Time t = _date;
      int myDate = (t.year * 10000) + (t.month * 100) + t.day;
      int hisDate = (ts.getYear() * 10000) + (ts.getMonth() * 100) + ts.getDate();
      return myDate > hisDate || (myDate == hisDate && ((t.hour * 10000) + (t.minute * 100) + t.second) > (ts.getHour() * 10000) + (ts.getMinute() * 100) + ts.getSecond());
   }

   /**
     * This method these the specified <code>Object</code> for equality
     * against this object.  This will be true if an only if the specified
     * object is an instance of <code>Timestamp</code> and has the same
     * time value fields.
     *
     * @param obj The object to test against for equality.
     *
     * @return <code>true</code> if the specified object is equal to this
     * object, <code>false</code> otherwise.
     */
   public boolean equals(Object obj)
   {
      if(obj == null)
         return false;
      if(!(obj instanceof Timestamp))
         return false;
      return equals((Timestamp)obj);
   }

   /**
     * This method tests the specified timestamp for equality against this
     * object.  This will be true if and only if the specified object is
     * not <code>null</code> and contains all the same time value fields
     * as this object.
     *
     * @param ts The <code>Timestamp</code> to test against for equality.
     *
     * @return <code>true</code> if the specified object is equal to this
     * object, <code>false</code> otherwise.
     */
   public boolean equals(Timestamp ts)
   {
      if(ts == null)
         return false;
      waba.sys.Time t = _date;
      return t.year == ts.getYear() && t.month == ts.getMonth() && t.day == ts.getDate() && t.hour == ts.getHour() && t.minute == ts.getMinute() && t.second == ts.getSecond();
   }

   /**
     * This method returns a new instance of this class by parsing a
     * date in JDBC format.  JDBC format is yyyy-MM-dd HH:mm:ss.
     *
     * @param str The string to parse.
     * @return The resulting <code>java.sql.Timestamp</code> value.
     */
   public static Date valueOf(String str)
   {
      int pos = str.indexOf(' ');
      if(pos != -1)
      {
         waba.util.Date date = new waba.util.Date(str.substring(0, pos), waba.sys.Settings.DATE_YMD);
         waba.sys.Time time = parseTime(str, pos + 1);
         time.year = date.getYear() - 1900;
         time.month = date.getMonth();
         time.day = date.getDay();
         return new Timestamp(time);
      }
      return new Timestamp(new waba.sys.Time());
   }
}
