
/*********************************************************************************
  *  SuperWaba Virtual Machine, version 4                                         *
  *  Copyright (C) 2003 Guilherme Campos Hazan <support@superwaba.com.br>         *
  *                 and Andy DePue <sir_anodos@yahoo.com>                         *
  *  All Rights Reserved                                                          *
  *                                                                               *
  *  This library and virtual machine is free software; you can redistribute      *
  *  it and/or modify it under the terms of the Amended GNU Lesser General        *
  *  Public License distributed with this software.                               *
  *                                                                               *
  *  This library and virtual machine is distributed in the hope that it will     *
  *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
  *                                                                               *
  *  All software that links or interacts with this library and virtual machine,  *
  *  or any derivative works, require the software to prominently display the     *
  *  following notice:                                                            *
  *                                                                               *
  *                   Created with SuperWaba                                      *
  *                  http://www.superwaba.org                                     *
  *                                                                               *
  *  Please see the software license located at SuperWabaSDK/license.txt          *
  *  for more details.                                                            *
  *                                                                               *
  *  You should have received a copy of the License along with this software;     *
  *  if not, write to                                                             *
  *                                                                               *
  *     Guilherme Campos Hazan                                                    *
  *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
  *     Rio de Janeiro / RJ - Brazil                                              *
  *     Cep: 22050-000                                                            *
  *     E-mail: support@superwaba.com.br                                          *
  *                                                                               *
  *********************************************************************************/
package superwaba.ext.xplat.sql.db2e;
import superwaba.ext.xplat.sql.*;
import superwaba.ext.xplat.sql.db2e.db2ex.*;
import waba.util.Hashtable;
import waba.util.Vector;
import waba.sys.Vm;

/**
  There are currently no example apps ready to show how to use the DB2e WDBC
  driver.  If you have experience with JDBC, then it should be fairly easy
  to get up to speed with DB2e on SW.  Be sure to read this entire file for
  information regarding WDBC driver capabilities.
  <p>
  Here is a short code example that shows how to load the WDBC driver for DB2e:
  <pre>
  try
  {
  Class.forName("superwaba.ext.xplat.sql.db2e.Driver");
  }
  catch(ClassNotFoundException cnfe)
  {
  // Code to handle if driver not found or unable to load...
  }
  </pre>

  Connect to the DB2e driver using a URL of the form:
  <pre>
  wdbc:db2e:[regular db2e connection string]
  </pre>
  DB2e currently connects only to the local database, so does not need any additional connection information (such as server, user, password, etc).  For now, just use a dummy string for the [regular db2e connection string] portion.  Here is example code that obtains a DB2e connection:

  <pre>
  import superwaba.ext.xplat.sql.*;
  import superwaba.ext.xplat.sql.db2e.*;

  ...

  Connection connection = null;

  try
  {
  connection = DriverManager.getConnection("wdbc:db2e:server");
  }
  catch(SQLException sqlError)
  {
  // Code to handle connection error...
  }
  </pre>

  These are just a few tips concerning Db2e WDBC driver:<br>
  1. To keep the driver as simple as possible, we had to impose a few limitations:<br>
  a. Access fields consistently in relation to the data type you use to retrieve the field's value.  In other words, don't use "getString(...)" on a column and then "getInt(...)" on the same column.  An exception will be thrown.  This is because the driver caches the objects it uses to access the column so that it can minimize the creation of new instances while retrieving the value.  This holds across rows too.  If you use getString(...) on a column and then reposition the result set to a new row, you must still use getString(...) on the same column.  In fact, this even goes across separate ResultSets returned from a prepared statement.  If you re-execute the prepared statement, it will reuse the same column types, so you must still use the same access method you used before.<br>
  b. This is a DB2e limitation: you can have only one ResultSet per (Prepared)Statement.<br>
  c. Other DB2e limitations (as of V8.1): Max connections: 1.  Max open (Prepared)Statements: 20.  Max transactions: 1 (transaction is on a Connection basis).<br>
  d. To save space, we did not implement updatable ResultSets (CONCUR_UPDATABLE).  You will have to manually update rows/data by using SQL: INSERT, UPDATE, DELETE, etc.  DB2e does not provide any kind of "updatable" interface, so we would have had to implement it by dynamically building SQL statement anyway, which would have just added more overhead to the whole process.<br>
  e. We have not implements ResultSets that can fetch row sets.  We fetch one row at a time from Db2e.<br>
  f. We have not implemented FETCH_REVERSE, to save having to check a "direction" flag everytime a "next()" is performed.  If you need to go in reverse, do it manually using previous(), relative(), or absolute() (which are all implemented and fully functional).<br>
  g. To the best of our knowledge, DB2e supports only TYPE_FORWARD_ONLY and TYPE_SCROLL_SENSITIVE.<br>
  h. To the best of our knowledge, DB2e is TRANSACTION_READ_COMMITTED only.<br>
  i. We have not implemented any of the DatabaseMetaData methods that would return a "ResultSet".  If you need any of this information, you currently have to query it manually using SQL tailored specifically for DB2e's system table structure.<br>
  j. The content of the data within a Blob instance returned from getBlob(...) is temporary!  This means that it will not survive a call to any of the position methods (next, previous, relative, absolute).  When you call a position method, it will overwrite the Blob byte buffer with new data.  This is because the driver reuses the exact same Blob instance for that column throughout the life of a Statement to minimize object creation. If you want to keep the data around, you will have to manually copy it out of the Blob byte[] buffer (Vm.copyArray, anyone?).<br>
  k. We don't attempt data type conversion.  We let DB2e handle that for us.  So if you are trying to getX(...) on a field that isn't X, then it will work so long as DB2e will automatically convert to that type for you.  Oh, and once you call getX(...) on a field that isn't X, you must always call getX(...) and cannot use any other get*(...) methods on that field.<br>
  l. There is no Long data type in this implementation.  getLong(...) is currently implemented as 'return (long)getInt(...)'<br>
  m. We have not yet implemented float or double types!!  I know this is important to some people, but we didn't need it and didn't spend the time to implement it... maybe someone could volunteer?<br>
  n. We have not implemented getObject yet.  It will throw an exception.  Would you like to volunteer?<br>
  <br>
  2. Tips for optimizing your usage of the DB2e WDBC driver:<br>
  a. Believe it or not, getBlob(...) is much more effecient than getBytes(...).  This is because getBytes() must allocate a new byte[] array every time it is called, while getBlob(...) can reuse the same Blob instance for a particular column.  getBytes() just calls getBlob() and returns a copy of the byte array.<br>
  b. A SuperWaba only method (not found in the JDBC Blob interface) is Blob.getByteBuffer().  This was added because it can be made so much more effecient than Blob.getBytes(long pos, int length) in certain scenarios.  The DB2e driver allocates a byte[] buffer large enough to hold the largest Blob when a Blob column is first accessed.  This buffer is returned when you call getByteBuffer, which is more effecient because a new byte[] array instance does not need to be created, and also means that the buffer could actually be larger than the Blob data for this row, so you must be careful to read only Blob.length() bytes from the buffer.<br>
  c. Try to avoid using the getX(String columnName) methods, as they impose a relatively heavy hit to look up the column name and then call their getX(int columnIndex) brother.<br>
  d. We have tried to optimize PreparedStatement so that you take a light hit when you reexecute the statement.  Basically, the ResultSet instance is reused, avoiding an object allocation and allows the query information to be cached (such as column info and types).<br>
  e. Same for ResultSetMetaData.<br>
  f. Take advantage of setMaxFieldSize() when you can.  It will save memory and CPU cycles if your max size is less than the sizes of your fields.<br>
  g. Use common sense.  Reuse as much as possible.  We try to stick by this in our code (though we aren't perfect).<br>
  h. getLong() is currently implemented as 'return (long)getInt(...)'.  But don't worry, because ints in SW are 32-bit, so getInt actually uses DB2e's notion of LONG.<br>
  i. We have not yet optimized Time/Date/Timestamp functions for reuse, so a new Time/Date/Timestamp instance will be created every time a 'get' method is called for one of these.<br>
  j. We have not implemented getRef.  This may be pulled from SW anyway.<br>
  k. getInt(...) is faster than getShort() and getLong().<br>
  l. Try to call getString(...) just once per column per row, for you shall pay the penalty of a new String instance every time you call.<br>
  */

public class Driver extends superwaba.ext.xplat.sql.Driver
{
   static boolean _libraryLoaded = false;
   static Db2e _db;
   static int _envHandle;
   static Vector _connections = new Vector();
   static Hashtable _dummyUserPass = new Hashtable(2, 1.0f);
   static
   {

      //    Vm.debug("In static initializer (superwaba.ext.xplat.db2e.Driver).");
      _dummyUserPass.put("user", "");
      _dummyUserPass.put("password", "");

      //    Vm.debug("Finished 'putting'.");
      try
      {

         //      Vm.debug("Calling DriverManager.");
         DriverManager.registerDriver(new Driver());

      //      Vm.debug("Finished calling DriverManager.");
      }
      catch(SQLException error)
      {
         Vm.debug("Unable to register superwaba.ext.xplat.sql.db2e.Driver: " + error);
         throw new RuntimeException("Unable to register DB2e driver: " + error);
      }

   //    Vm.debug("Leaving static initializer.");
   }
   private DriverPropertyInfo[] _propertyInfo = new DriverPropertyInfo[1];
   public Driver()
   {

   }

   /** This method tests whether or not the driver believes it can connect to
     * the specified database.  The driver should only test whether it
     * understands and accepts the URL. It should not necessarily attempt to
     * probe the database for a connection.
     *
     * @param The database URL string.
     * @return <code>true</code> if the drivers can connect to the database,
     *         <code>false</code> otherwise.
     * @exception SQLException If an error occurs.
     *
     */
   public boolean acceptsURL(String url) throws SQLException
   {

      //    Vm.debug("In acceptsURL('" + url + "')...");
      int i = url.indexOf(':');
      if(i != -1 && url.substring(0, i).equals("wdbc"))
      {

         //      Vm.debug("Found 'wdbc' portion");
         int prev = i;
         i = url.indexOf(':', i + 1);
         if(i != -1 && url.substring(prev + 1, i).equals("db2e"))

            //        Vm.debug("Found 'db2e' portion.  Returning true.");
            return true;
      }

      //    Vm.debug("Returning false.");
      return false;
   }

   /** This method connects to the specified database using the connection
     * properties supplied.  If the driver does not understand the database
     * URL, it should return <code>null</code> instead of throwing an
     * exception since the <code>DriverManager</code> will probe a driver
     * in this manner.
     *
     * @param url The URL string for this connection.
     * @param properties The list of database connection properties.
     * @return A <code>Connection</code> object for the newly established
     *         connection, or <code>null</code> if the URL is not understood.
     * @exception SQLException If an error occurs.
     *
     */
   public Connection connect(String url, Hashtable info) throws SQLException
   {

      //    Vm.debug("In connect('" + url + "'), calling acceptsURL().");
      if(!acceptsURL(url))
         return null;

      //    Vm.debug("Calling checkEnvironment().");
      checkEnvironment();
      int i = url.indexOf(':');
      i = url.indexOf(':', i + 1);
      String db2eURL = url.substring(i + 1);

      //    Vm.debug("Got db2eURL: " + db2eURL);
      Db2eConnection con = new Db2eConnection(this, _db, _envHandle, db2eURL, info != null ? info : _dummyUserPass);

      //    Vm.debug("Created new Db2eConnection.");
      return con;
   }
   protected void checkEnvironment() throws SQLException
   {
      if(!_libraryLoaded)
      {
         if(!Db2e.isNativeAttached())
         {
            if(!Db2e.attachNative())
               throw new SQLException("Failed to load SWDB2e native library.");
            else
            {
               _db = new Db2e();
               _libraryLoaded = true;
            }
         }
         else
         {

            //        Vm.debug("Db2e natives already loaded.");
            _db = new Db2e();
            _libraryLoaded = true;
         }
      }

      // Make sure we have an environment.
      if(_envHandle == 0)
      {
         try
         {

            //        Vm.debug("Creating new environment.");
            _envHandle = _db.allocHandle(Db2e.SQL_HANDLE_ENV, 0);

         //        Vm.debug("Created new environment.");
         }
         catch(DB2Exception error)
         {
            handleError(error);
         }
      }
   }
   static void handleError(DB2Exception error) throws SQLException
   {

      /*    Vm.debug("DB2 state: " + (new String(error.getSQLState())));
      Vm.debug("DB2 native: " + error.getNativeErr());
      Vm.debug("DB2 SQL Error: " + error.getSQLErr()); */
      throw new SQLException(error.getMessage(), (new String(error.getSQLState())), error.getNativeErr());
   }
   static void registerConnection(Db2eConnection con) throws SQLException
   {
      _connections.addElement(con);
   }
   static void deregisterConnection(Db2eConnection con) throws SQLException
   {
      _connections.removeElement(con);

      // If the final connection closes, then free our environment.
      if(_connections.size() == 0 && _envHandle != 0)
      {
         try
         {
            _db.freeHandle(Db2e.SQL_HANDLE_ENV, _envHandle);
            _envHandle = 0;
         }
         catch(DB2Exception error)
         {
            handleError(error);
         }
      }
   }

   /** This method returns the major version number of the driver.
     *
     * @return The major version number of the driver.
     *
     */
   public int getMajorVersion()
   {
      return 0;
   }

   /** This method returns the minor version number of the driver.
     *
     * @return The minor version number of the driver.
     *
     */
   public int getMinorVersion()
   {
      return 5;
   }

   /** This method returns an array of possible properties that could be
     * used to connect to the specified database.
     *
     * @param url The URL string of the database to connect to.
     * @param properties The list of properties the caller is planning to use
     *        to connect to the database.
     * @return A list of possible additional properties for a connection to this
     *         database.  This list may be empty.
     * @exception SQLException If an error occurs.
     *
     */
   public DriverPropertyInfo[] getPropertyInfo(String url, Hashtable properties) throws SQLException
   {
      return _propertyInfo;
   }

   /** This method tests whether or not the driver is JDBC compliant.  This
     * method should only return <code>true</code> if the driver has been
     * certified as JDBC compliant.
     *
     * @return <code>true</code> if the driver has been certified JDBC compliant,
     *         <code>false</code> otherwise.
     *
     */
   public boolean jdbcCompliant()
   {
      return false;
   }
}
