package superwaba.ext.xplat.ui;

/**
  * The GridModel interface specifies the methods the Grid will use to interrogate a tabular data model.<br><br>
  * The Grid can be set up to display any data model which implements the GridModel interface with a couple of lines of code:<br>
  * <pre>
  * GridModel myData = new MyTableModel();
  * Grid table = new Grid(myData);
  *</pre>
  *
  * @author Stein Rustad (stein@politikontroll.org)
  */

public interface GridModel
{
    /**
      * Returns the number of columns in the model.
      * A Grid uses this method to determine how many columns it should display.
      *
      * @return the number of columns in the model
      */

    public int getColumnCount();

   /**
    * Returns the number of items this model contains
    * @return the number of items
    */
   public int getItemCount();

    /**
      * Returns the preferred width of the column with index columnIndex.
      *
      * @param columnIndex the column which is queried for its width
      * @return the preferred width for this column. If the Grid should use best-fit for the current font/text, return -1
      */

    public int getPreferredColumnWidth(int columnIndex);
    /**
      * Returns the preferred height of the row with index rowIndex.
      *
      * @param rowIndex the row which is queried for its height
      * @return the preferred height for this row. If the Grid should use best-fit for the current font, return -1
      */

    public int getPreferredRowHeight(int rowIndex);
    /**
      * Returns the number of rows in the model.
      * A Grid uses this method to determine how many rows it should display.
      *
      * @return the number of rows in the model
      */

    public int getRowCount();
    /**
      * Returns the value at the current cell
      *
      * @param rowIndex the row whose value is to be queried
      * @param columnIndex the column whose value is to be queried
      * @return the value that is associated with this cell. The implementation decides how this is associated.
      */

    public Object getSelected(int rowIndex, int colIndex);
  /**
    * Returns the chars that are allowed to be entered into this cell
    * Only makes sense for cells that are editable
    *
    * @param rowIndex the row whose valid chars are requested
    * @param colIndex the column whose valid chars are requested
    * @return the valid chars for this cell
    */

  public String getValidChars(int rowIndex, int colIndex);
    /**
      * Returns the value for the cell at columnIndex and rowIndex.
      * A Grid uses this method to determine how many rows it should display.
      *
      * @param rowIndex the row whose value is to be queried
      * @param columnIndex the column whose value is to be queried
      * @return the value Object at the specified cell
      */

    public Object getValueAt(int rowIndex, int columnIndex);
    /**
      * Returns true if the cell at rowIndex and columnIndex is editable.
      * Otherwise, setValueAt on the cell will not change the value of that cell.
      *
      * @param rowIndex the row whose value is to be queried
      * @param columnIndex the column whose value is to be queried
      * @return true if the cell is editable
      */

    public boolean isCellEditable(int rowIndex, int columnIndex);
    /**
      * Returns if the row is a header.
      * A Grid uses this method to determine if the row should be displayed diffrently from others (bold font).
      *
      * @return true if the row is a header
      */

    public boolean isHeader(int rowIndex);
    /**
      * Sets the value in the cell at columnIndex and rowIndex to aValue.
      *
      * @param aValue the new value
      * @param rowIndex the row whose value is to be changed
      * @param columnIndex the column whose value is to be changed
      * @return the value Object at the specified cell
      */

    public void setValueAt(Object aValue, int rowIndex, int columnIndex);
}

