/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package superwaba.ext.xplat.ui;

import waba.ui.*;
import waba.util.*;

/** used to make a bit easy to place controls in a container. The controls must be inserted in horizontal order.
    isnt necessary to fill all the specified number of cols.
    example:
    <pre>
      JustifiedContainer jc = new JustifiedContainer();
      jc.setColumnCount(2);
      jc.setGaps(0,6,0); // 6 pixels of horizontal gap
      jc.setJustify(0,RIGHT,TOP); // justify column 0 at right
      jc.add(0,new Label("Please fill the fields",Label.CENTER),0,0); // container width and preferred height
      jc.add(0,new Label("Name:"),70,0);
      jc.add(1,new Edit(),0,0); // fill column 1 to the end
      jc.add(0,new Label("Address:"),70,0);
      jc.add(1,new Edit(),0,0);
    </pre>
    Note: You can use the PREFERRED+delta as width/height, but you can't use delta in the justified parameters.
*/
public class JustifiedContainer extends Container
{
   static final int RANGE = 4000;
   private class Column
   {
      /** gaps horizontal e vertical */
      public int gapX=2,gapY=3;
      /** justificacao para a coluna */
      public int justX=LEFT,justY=TOP;
      /** ultima coordenadas inseridas */
      public int lastX=0,lastY=-1,lastW=0,lastH=0;
      /** vetor de controles desta coluna */
      public Vector controls = new Vector(3);
   }
   /** numero de colunas do container */
   private int columnCount=1;
   /** colunas */
   private Column columns[];
   //////////////////////////////////////////////////////////////////////////
   public JustifiedContainer()
   {
   }
   //////////////////////////////////////////////////////////////////////////
   /** initialize the columns */
   public void setColumnCount(int numcols)
   {
      if (numcols == 0) numcols = 1;
      this.columnCount = numcols;
      this.columns = new Column[columnCount];
      for (int i =0; i < columnCount; i++)
         columns[i] = new Column();
   }
   //////////////////////////////////////////////////////////////////////////
   /** set the justifies of the column <col>; can be LEFT, RIGHT, TOP, BOTTOM, CENTER */
   public void setJustify(int col, int justX, int justY)
   {
      columns[col].justX = justX;
      columns[col].justY = justY;
   }
   //////////////////////////////////////////////////////////////////////////
   /** set the gaps for the column <col> */
   public void setGaps(int col, int gapX, int gapY)
   {
      columns[col].gapX = gapX;
      columns[col].gapY = gapY;
   }
   //////////////////////////////////////////////////////////////////////////
   /** adds the control.
   @param control the specified control.
   @param col which column to add
   @param controlW if greater than 0: width of the control; if 0: fill to the end of the container width; if PREFERRED: preferred control width
   @param controlH if greater than 0: height of the control; if 0: default line height; if PREFERRED: preferred control height
   */
   public void add(int col, Control control, int controlW, int controlH)
   {
      super.add(control);
      if (columns == null) setColumnCount(1);
      Column c = columns[col];
      c.controls.add(control);

      int leftColumnX = col==0?0:(columns[col-1].lastX+columns[col-1].lastW);
      if (controlW == 0 || controlW == FILL) controlW = col==0?width:(width-leftColumnX-c.gapX); else
      if ((PREFERRED-RANGE) <= controlW && controlW <= (PREFERRED+RANGE)) controlW = control.getPreferredWidth() + controlW-PREFERRED;

      if (controlH == 0) controlH = fm.getHeight(); else
      if ((PREFERRED-RANGE) <= controlH && controlH <= (PREFERRED+RANGE)) controlH = control.getPreferredHeight() + controlH-PREFERRED;

      int controlX=0,controlY=0;
      if (col == 0)
         switch (c.justY)
         {
            case TOP    : controlY = c.lastY==-1?0:(c.lastY+c.lastH+c.gapY); break;
            case BOTTOM : controlY = (c.lastY==-1?height:(c.lastY-c.gapY))-controlH; break;
            case CENTER : controlY = (c.lastY+c.lastH)/2+(height-controlH)/2; break;
         }
      else
      {
         controlY = columns[0].lastY;
         if (controlH < columns[0].lastH) // if smalled than the line, centers the control
            controlY += (columns[0].lastH-controlH)/2;
      }
      if (col > 0 && controlY != c.lastY) c.lastX = 0; // changed line, zero it.
      switch (c.justX)
      {
         case LEFT  : controlX = leftColumnX!=0?(leftColumnX+c.gapX):0; break;
         case RIGHT : controlX = (c.lastX==0?width:(c.lastX-c.gapX)) - controlW; break;
         case CENTER: controlX = (width - leftColumnX - controlW)/2; break;
      }
      control.setRect(controlX,controlY,controlW,controlH);
      c.lastX = controlX;
      c.lastY = controlY;
      c.lastW = controlW;
      c.lastH = controlH;
   }
   //////////////////////////////////////////////////////////////////////////
}