/*
 * The contents of this file are subject to the "Lesser GNU Public License"
 * (LGPL); you may not use this file except in compliance with the License.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License.
 *
 */

package superwaba.ext.xplat.util.xml;

import waba.sys.Vm;
import waba.io.Stream;

/** Listener that must be extended to process the events thrown by HttpLiteParser.
 * parses a HTTP header. State 0 indicates that the header has been read and
 * the buffer may be left pointing at any content.
  */
public class HttpLiteParser extends LiteParser
{
    protected HttpListener httpListener;
    protected byte eol=0;

    /** Constructor.
     * @param listener
     */
    public HttpLiteParser(HttpListener listener)
    {
        httpListener=listener;
    }

    /**
     * This method parses the bytes from the passed stream.
     * @param input Input stream.
     * @param bufferSize The bufer size to use. No tag or content can be
     * larger than the buffer.
     * @exception ParserException
     */
    public void parse(Stream input,int bufferSize) throws ParserException
    {
       // Create buffer
       StreamBuffer sbuf = new StreamBuffer(input,bufferSize);
       parse(sbuf);
    }

    /** Parse a Stream Buffer
     * @param sbuff A buffer that can be refilled from a stream
     * @exception ParserException
     */
    public void parse(StreamBuffer sbuf) throws ParserException
    {
       /* Initialize global variables. */
       pos = 0;
       httpListener.len = 0;
       state = -8;
       eol=0;

       // Parse any data passed in
       if (sbuf.pos<sbuf.size)
       {
           pos=sbuf.pos;
           httpListener.start=sbuf.mark;
           parseBuffer(sbuf.buffer,sbuf.size);
           sbuf.pos=pos;
           sbuf.mark=httpListener.start;
       }

       // Parse stream
       while(state<0 && sbuf.fill()>=0)
       {
           pos=sbuf.pos;
           httpListener.start=sbuf.mark;
           parseBuffer(sbuf.buffer,sbuf.size);
           sbuf.pos=pos;
           sbuf.mark=httpListener.start;
       }

       // Exception if we are not at a finishing state.
       if (state != 0)
       {
           Vm.debug("state="+state);
           throw new ParserException("Unexpected end of stream!");
       }
    }

    /**
     * This method parses the given array of bytes.
     * The algorithm is based on DFA (Deterministic Finite Automata) theory and
     * Regular Expression theory.
     * @param input The input byte array.
     * @param size The size of bytes to be parsed. Normally should be <code>input.length</code>
     */
    public void parse(byte []input, int size) throws ParserException
    {
        /* Initialize global variables. */
        pos = 0;
        state = -8;
        httpListener.len = 0;
        httpListener.start = -1;
        eol=0;

        parseBuffer(input,size);

        if (state != 0)
        {
            Vm.debug("state="+state);
            throw new ParserException("Unexpected end of stream!");
        }
    }

    /**
     * @param size
     * @exception ParserException
     */
    protected void parseBuffer(byte []input,  int size) throws ParserException
    {
//         int s=18;
//         if (s>size)
//             s=size;
//         Vm.debug("HP:"+size+" pos="+pos+
//                  "\n  "+new String(input,0,s)+"..."+
//                  ((pos>0 && (pos+s)<size)?("\n..."+new String(input,pos,s)+"..."):"")+
//                  "\n..."+new String(input,size-s-1,s));

        httpListener.buffer = input;

        int last=-99;
        for (; pos < size; pos++)
        {
            byte ch = input[pos];

//             if (state!=last)
//             {
//                 Vm.debug("State="+state);
//                 last=state;
//             }
//             waba.sys.Vm.debug("hc='"+(char)ch+"'");

            /* The DFA is hardcoded in this sequence of states. */
            switch (state)
            {
              case -8:
                  if (ch <= SPACE)
                  {
                      //state = -8;
                  }
                  else if (ch=='h' | ch=='H')
                  {
                      httpListener.len = 1;
                      httpListener.start = pos;
                      state = -7;
                  }
                  else
                      throw new ParserException("S-8: pos="+pos);
                  break;

              case -7:
                  if (ch==SPACE)
                  {
                      httpListener.foundVersion();
                      httpListener.len=0;
                      httpListener.start=-1;
                      state=-6;
                  }
                  else if (ch < SPACE)
                      throw new ParserException("S-7: pos="+pos);
                  else
                  {
                      //state = -7;
                      httpListener.len++;
                  }
                  break;

              case -6:
                  if (ch==SPACE)
                  {
                      httpListener.foundCode();
                      httpListener.len=0;
                      httpListener.start=-1;
                      state=-5;
                  }
                  else if (ch < SPACE)
                      throw new ParserException("S-6: pos="+pos);
                  else
                  {
                      //state = -6;
                      if (httpListener.start == -1) httpListener.start = pos;
                      httpListener.len++;
                  }
                  break;

              case -5:
                  if (ch==CARRIAGE_RETURN || ch==LINE_FEED)
                  {
                      httpListener.foundReason();
                      httpListener.len=0;
                      httpListener.start=-1;
                      eol=ch;
                      state=-1;
                  }
                  else
                  {
                      //state = -5;
                      if (httpListener.start == -1) httpListener.start = pos;
                      httpListener.len++;
                  }
                  break;


              case -4: // Get header name
                  if (ch==CARRIAGE_RETURN || ch==LINE_FEED)
                  {
                      if (httpListener.start>0)
                          httpListener.foundHttpHeader();
                      eol=ch;
                      state=-1;
                  }
                  if (ch==COLON)
                  {
                      if (httpListener.start>0)
                          httpListener.foundHttpHeader();
                      httpListener.len=0;
                      httpListener.start=-1;
                      state=-3;
                  }
                  else if (ch!=SPACE && ch!=TAB)
                  {
                      if (httpListener.start == -1) httpListener.start = pos;
                      httpListener.len=pos-httpListener.start+1;
                  }
                  break;

              case -3: // Get header value
                  if (ch==CARRIAGE_RETURN || ch==LINE_FEED)
                  {
                      if (httpListener.start>0)
                          httpListener.foundHttpValue();
                      eol=ch;
                      state=-1;
                  }
                  else if (ch!=SPACE && ch!=TAB)
                  {
                      if (httpListener.start == -1) httpListener.start = pos;
                      httpListener.len=pos-httpListener.start+1;
                  }
                  break;

              case -1: // Start of header line
                  if (ch==CARRIAGE_RETURN || ch==LINE_FEED)
                  {
                      if (ch==LINE_FEED && eol==CARRIAGE_RETURN)
                          eol=0;
                      else
                      {
                          state=0;
                          // Parsed header so return.
                          return;
                      }
                  }
                  else if (ch==COLON || ch==SPACE || ch==TAB)
                  {
                      httpListener.len=1;
                      httpListener.start=-1;
                      state=-3;
                  }
                  else
                  {
                      httpListener.len=1;
                      httpListener.start=pos;
                      state=-4;
                  }
                  break;
            }
        }
    }
}
