/*
 * The contents of this file are subject to the "Lesser GNU Public License"
 * (LGPL); you may not use this file except in compliance with the License.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License.
 *
 */

package superwaba.ext.xplat.util.xml;

import waba.sys.Vm;
import waba.io.Stream;

/**
 * A buffer that can be filled by a Stream.
 * This class is used by the LiteParser derivations for handling the
 * buffering from from a stream. Eventually it should be merged with the
 * superwaba.ext.xplat.io.BufferStream.
 * <p>
 * The buffer maintains a size, pos and mark indexes where
 * mark<=pos<=size<=buffer.length and 0<=pos.
 * <p>
 * Data in the buffer after the pos or after mark (if set) is preserved in
 * during a fill and moved to the start of the buffer.
 */
public class StreamBuffer
{
    /** The stream to be buffered */
    public Stream stream;
    /** The data buffer for bytes read from the stream */
    public byte[] buffer;

    /** Current length of the content in the buffer */
    public int size;

    /** Current parsing position within the buffer */
    public int pos;

    /** Starting position of unconsumed data in the buffer */
    public int mark;

    public StreamBuffer(Stream stream,int size)
    {
        this.stream=stream;
        buffer=new byte[size];
        size=0;
        pos=0;
        mark=-1;
    }

    /** Save unconsumed data and fill the remainder of the buffer.
     * @return The number of bytes written.
     */
    public int fill()
    {
        int count=0;
        // save data after the mark.
        if (mark>=0 && mark<size)
        {
            Vm.copyArray(buffer,mark,buffer,0,size-mark);
            size=size-mark;
            pos=pos-mark;
            mark=0;
        }
        // save data after the pos
        else if (size>0 && pos<size)
        {
            Vm.copyArray(buffer,pos,buffer,0,size-pos);
            size=size-pos;
            pos=0;
            mark=-1;
        }
        else
        {
            size=0;
            pos=0;
            mark=-1;
        }

        // read some more data
        Vm.debug("bl-s="+(buffer.length-size));
        count=stream.readBytes(buffer,size,buffer.length-size);
        if (count>0)
            size+=count;
        else if (count==0)
        {
            Vm.sleep(2000);
            count=stream.readBytes(buffer,size,buffer.length-size);
            if (count>0)
                size+=count;
        }
        return count;
    }

}
