/*
 * XmlLiteParser version 1.0
 * Created on September 26, 2002
 *
 * Copyright (C) 2002
 * Giovanni Remigi
 * via Procche n.28 36030 Zugliano (Vicenza),
 * Italy. All Rights Reserved.
 *
 * The contents of this file are subject to the "Lesser GNU Public License"
 * (LGPL); you may not use this file except in compliance with the License.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License.
 *
 * Thanks to Fabiano Dal Castello for providing various fixes and hints.
 *
 * Reestructured and optimized by Guich for SuperWaba 3.3.
 */

package superwaba.ext.xplat.util.xml;

import waba.sys.*;

/**
 * <p>This is a fast and simple parser for a lite version of XML format. You can
 * use this parser with SuperWaba VM to parse data from an open Stream.</p>
 * <p>The parser looks for:</p>
 * <ul>
 * <li>start-tag name,</li>
 * <li>end-tag name,</li>
 * <li>attribute name,</li>
 * <li>attribute value,</li>
 * <li>content between a start-tag and an end-tag.</li>
 * </ul>
 *
 * <p>When the parser finds these elements it calls respectively:</p>
 * <ul>
 * <li><tt>foundStartTagName();</tt></li>
 * <li><tt>foundEndTagName();</tt></li>
 * <li><tt>foundAttributeName();</tt></li>
 * <li><tt>foundAttributeValue();</tt></li>
 * <li><tt>foundContent();</tt></li>
 * </ul>
 *
 * <p>The parsing algorithm is based on DFA (Deterministic Finite Automata)
 * theory and Regular Expression theory. See the following schema: <img src="../../../../../xmldfa.gif">
 * </p>
 *
 * <p>In order to use this parser you need to extend the XmlListener class and
 * override the above methods. To start parsing, call the method
 * parse(byte inputBytes, int length). </p>
 * See also samples/util/xml.
 *
 * @author Giovanni Remigi
 * @email g.remigi@tiscalinet.it
 * @author Greg Wilkins
 * @version 0.02
 */

 // guich@330: incorporated into superwaba.ext and optimized the code.

public class XmlLiteParser extends LiteParser
{
   /* Return true if ch is the Most Significant char of a Tag or Attribute
    * name. MSChar is the first char of a Tag or Attribute name.
    */
   private static boolean []isMSChar = new boolean[256];
   /* Return true if ch is the Less Significant char of a Tag or Attribute
    * name. LSChar is any char (except the first) of a Tag or Attribute name.
    */
   private static boolean []isLSChar = new boolean[256];

   /** Build boolean arrays */
   static
   {
      // isMS
      isMSChar[UNDERSCORE] = isMSChar[COLON] = true;
      for (int i = 65; i <= 90; i++)
         isLSChar[i] = isMSChar[i] = true;
      for (int i = 97; i <= 122; i++)
         isLSChar[i] = isMSChar[i] = true;

      // isLS
      isLSChar[UNDERSCORE] = isLSChar[COLON] = true;
      isLSChar[HYPHEN] = isLSChar[FULL_STOP] = true; // ramonchiara@400_92: changed == to =
      for (int i = 48; i <= 57; i++)
         isLSChar[i] = true;
   }

   /** Listener of this parser */
   protected XmlListener listener;


   /* Creates a new instance of XmlLiteParser. */
   public XmlLiteParser(XmlListener listener)
   {
      this.listener = listener;
   }

   /**
    * This method parses the given array of bytes.
    * The algorithm is based on DFA (Deterministic Finite Automata) theory and
    * Regular Expression theory.
    * @param input The input byte array.
    * @param size The size of bytes to be parsed. Normally should be <code>input.length</code>
    */
   public void parse(byte []input, int size) throws ParserException
   {
       /* Initialize global variables. */
       pos = 0;
       state = 0;
       listener.len = 0;
       listener.start = -1;

       parseBuffer(input,size);

       if (state != 0)
       {
           Vm.debug("state="+state);
           throw new ParserException("Unexpected end of stream!");
       }
   }

    /** Parse a Stream Buffer
     * @param sbuff A buffer that can be refilled from a stream
     * @exception ParserException
     */
    public void parse(StreamBuffer sbuf) throws ParserException
    {
       /* Initialize global variables. */
       state = 0;
       listener.len = 0;

       // Parse any data passed in
       if (sbuf.pos<sbuf.size)
       {
           pos=sbuf.pos;
           listener.start=sbuf.mark;
           parseBuffer(sbuf.buffer,sbuf.size);
           sbuf.pos=pos;
           sbuf.mark=listener.start;
       }

       // Parse remainder of the stream
       while(sbuf.pos<sbuf.size || sbuf.fill()>=0)
       {
           pos=sbuf.pos;
           listener.start=sbuf.mark;
           parseBuffer(sbuf.buffer,sbuf.size);
           sbuf.pos=pos;
           sbuf.mark=listener.start;
       }

       // Exception if we are not at a finishing state.
       if (state != 0)
       {
           Vm.debug("state="+state);
           throw new ParserException("Unexpected end of stream!");
       }
    }

    /**
     * This method does the actual parsing and call backs working for the
     * current values of pos, len, start and state.  It is called by
     * parse(byte[],int) and may be called by derived XmlLiteParsers.
     *atn
     * @param size
     * @exception ParserException
     */
    protected void parseBuffer(byte []input, int size) throws ParserException
   {
//        int s=18;
//        if (s>size)
//            s=size;
//        Vm.debug("XP:"+size+" pos="+pos+
//                 "\n  "+new String(input,0,s)+"..."+
//                 ((pos>0 && (pos+s)<size)?("\n..."+new String(input,pos,s)+"..."):"")+
//                 "\n..."+new String(input,size-s-1,s));

      listener.buffer = input;

      // some useful aliases
      boolean []isMSChar = this.isMSChar;
      boolean []isLSChar = this.isLSChar;

      int last=-99;
      for (; pos < size; pos++)
      {
          byte ch = input[pos];

//           if (state!=last)
//           {
//               Vm.debug("State="+state);
//               last=state;
//           }
//           waba.sys.Vm.debug("ch:'"+(char)ch+"'");

         /* The DFA is hardcoded in this sequence of states. */
         switch (state)
         {
            case 0:
               if (ch <= SPACE)
               {
                  //state = 0;
               }
               else
               if (ch == LESS_THAN)
               {
                  listener.start = -1;
                  listener.len = 0;
                  state = 1;
               }
               else
                  throw new ParserException("S0: pos="+pos);
               break;
            case 1:
               if (isMSChar[ch])
               {
                  if (listener.start == -1) listener.start = pos;
                  listener.len++;
                  state = 2;
               }
               else
               if (ch == FRACTION_SLASH)
                  state = 8;
               else
                  throw new ParserException("S1: pos="+pos);
               break;
            case 2:
               if (isLSChar[ch])
               {
                  if (listener.start == -1) listener.start = pos;
                  listener.len++;
                  //state = 2;
               }
               else
               if (ch == FRACTION_SLASH || ch == GREATER_THAN || ch <= SPACE)
               {
                  listener.foundStartTagName();
                  listener.len = 0;
                  listener.start = -1;
                  state = (ch == FRACTION_SLASH)?4:(ch == GREATER_THAN)?5:3;
               }
               else throw new ParserException("S2: pos="+pos);
               break;
            case 3:
               if (ch <= SPACE)
               {
                  //state = 3;
               }
               else
               if (ch == GREATER_THAN)
                  state = 5;
               else
               if (ch == FRACTION_SLASH)
                  state = 4;
               else
               if (isMSChar[ch])
               {
                  if (listener.start == -1) listener.start = pos;
                  listener.len++;
                  state = 6;
               }
               else throw new ParserException("S3: pos="+pos);
               break;
            case 4:
               if (ch == GREATER_THAN)
                  state = 0;
               else
                  throw new ParserException("S4: pos="+pos);
               break;
            case 5:
               if (ch <= SPACE)
               {
                  //state = 5;
               }
               else
               if (ch == LESS_THAN)
                  state = 1;
               else
               {
                  if (listener.start == -1) listener.start = pos;
                  listener.len++;
                  state = 15;
               }
               break;
            case 6:
               if (isLSChar[ch])
               {
                  if (listener.start == -1) listener.start = pos;
                  listener.len++;
                  //state = 6;
               }
               else
               if (ch <= SPACE || ch == EQUAL_SIGN)
               {
                  listener.foundAttributeName();
                  listener.len = 0;
                  listener.start = -1;
                  state = (ch <= SPACE)?11:12;
               }
               else throw new ParserException("S6: pos="+pos);
               break;
            case 7:
               if (ch == FRACTION_SLASH)
                  state = 8;
               else
                  throw new ParserException("S7: pos="+pos);
               break;
            case 8:
               if (isMSChar[ch])
               {
                  if (listener.start == -1) listener.start = pos;
                  listener.len++;
                  state = 9;
               }
               else throw new ParserException("S8: pos="+pos);
               break;
            case 9:
               if (isLSChar[ch])
               {
                  if (listener.start == -1) listener.start = pos;
                  listener.len++;
                  //state = 9;
               }
               else
               if (ch <= SPACE || ch == GREATER_THAN)
               {
                  listener.foundEndTagName();
                  listener.len = 0;
                  listener.start = -1;
                  state = (ch == GREATER_THAN)?0:10;
               }
               else
                  throw new ParserException("S9: pos="+pos);
               break;
            case 10:
               if (ch <= SPACE)
               {
                  //state = 10;
               }
               else
               if (ch == GREATER_THAN)
                  state = 0;
               else
                  throw new ParserException("S10: pos="+pos);
               break;
            case 11:
               if (ch <= SPACE)
               {
                  //state = 11;
               }
               else
               if (ch == EQUAL_SIGN)
                  state = 12;
               else
                  throw new ParserException("S11: pos="+pos);
               break;
            case 12:
               if (ch <= SPACE)
               {
                  //state = 12;
               }
               else
               if (ch == QUOTATION_MARK)
                  state = 13;
               else
                  throw new ParserException("S12: pos="+pos);
               break;
            case 13:
               if (ch == QUOTATION_MARK)
               {
                  listener.foundAttributeValue();
                  listener.len = 0;
                  listener.start = -1;
                  state = 14;
               }
               else
               {
                  if (listener.start == -1) listener.start = pos;
                  listener.len++;
                  //state = 13;
               }
               break;
            case 14:
               if (ch <= SPACE)
                  state = 3;
               else
               if (ch == GREATER_THAN)
                  state = 5;
               else
               if (ch == FRACTION_SLASH)
                  state = 4;
               else
                   throw new ParserException("S14: pos="+pos);
               break;
            case 15:
               if (ch == LESS_THAN)
               {
                  listener.foundContent();
                  listener.len = 0;
                  listener.start = -1;
                  state = 7;
               }
               else
               {
                  if (listener.start == -1) listener.start = pos;
                  listener.len++;
                  //state = 15;
               }
               break;
         } // end case
      }
   }
}
