package waba.applet;

/** Implements a simple FIFO queue of unlimited size.  This queue is thread safe,
 * including functionality that allows one thread to wait for another to push an
 * object onto the queue.
 */
public class Queue
{
  static class Node
  {
    public Object o;
    public Node next;

    public Node(Object obj) { o = obj; }
  }

  private Node _queue;
  private Node _end;
  private int  _size;

  public Queue()
  {
    _size = 0;
  }

  /** Push an object on the queue and notify any waiting threads.
   * @param o object to push onto the queue
   */
  public synchronized void push(Object o)
  {
    Node node = new Node(o);
    if(_queue == null)
      _queue = _end = node;
    else
    {
      _end.next = node;
      _end = node;
    }
    _size++;
    notify();
  }

  /** Returns the oldest object in the queue.
   * @return oldest object in the queue or null if no objects in the queue
   * @see popWait()
   * @see popWait(int)
   */
  public synchronized Object pop()
  {
    if(_queue == null)
      return null;

    Object ret = _queue.o;
    _queue = _queue.next;
    _size--;
    return ret;
  }

  /** Returns the oldest object in the queue.  If the queue is empty, then this method
   * will block until another thread pushes an object onto the queue.
   * @return oldest object in the queue
   * @see pop()
   * @see popWait(int)
   */
  public synchronized Object popWait()
  {
    try
    {
      while(_queue == null)
      {
        wait();
      }
    }
    catch(Exception ie)
    {
    }

    return pop();
  }

  /** Returns the oldest object in the queue.  If there are no object in the queue,
   * then this method will block until an object is placed in the queue from another
   * thread, or the specified time expires.  If the time expires and no object has
   * been placed in the queue, this method will return null.
   * @param maxTime maximum amount of time, in ms, to wait for an object to placed in the queue
   * @see pop()
   * @see popWait()
   * @return oldest object in the queue, or null if the specified time expires and no objects
   * have been placed in queue.
   */
  public synchronized Object popWait(int maxTime)
  {
    long begin;

    try
    {
      begin = System.currentTimeMillis();
      long timeLeft;
      while(_queue == null && (timeLeft = (maxTime - (System.currentTimeMillis() - begin))) > 0)
      {
        wait(timeLeft);
      }
    }
    catch(Exception ie)
    {
    }

    return pop();
  }

  public synchronized int getSize()
  {
    return _size;
  }
}
