/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>                         *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.fx;

import waba.sys.*;
/**
 * Color represents a color.
 * <p>
 * A color is defined as a mixture of red, green and blue color values.
 * Each value is in the range of 0 to 255 where 0 is darkest and 255 is brightest.
 * For example, Color(255, 0, 0) is the color red.
 * <p>
 * Here are some more examples:
 * <ul>
 * <li>Color(0, 0, 0) is black
 * <li>Color(255, 255, 255) is white
 * <li>Color(255, 0, 0 ) is red
 * <li>Color(0, 255, 0) is green
 * <li>Color(0, 0, 255) is blue
 * </ul>
 */

public final class Color
{
   protected int red;
   protected int green;
   protected int blue;
   protected int rgb;
   protected int gray;
   /** for fast comparisions. 2 colors are equal if their equ have the same value */
   public int equ;
   protected Color bright;
   protected Color dark;

   private static boolean isGray = !Settings.isColor;
   private static int colorDiv = Settings.maxColors==4 ? 6 : 4;
   public static int fadeStep = isGray?(Settings.maxColors==4?1:7):96;
   public static int smallerFadeStep = isGray?(Settings.maxColors==4?1:4):32;

   public static final Color WHITE  = new Color(255,255,255);
   public static final Color BRIGHT = new Color(191,191,191);/*Settings.maxColors==4 ? new Color(191,191,191) : *///new Color(230,230,230);
   public static final Color DARK   = new Color(127,127,127);
   public static final Color BLACK  = new Color(0,0,0);
   public static Color defaultForeColor = BLACK;
   public static Color defaultBackColor = BRIGHT;

   /**
   * Constructs a color object with the given red, green and blue values.
   * @param red the red value in the range of 0 to 255
   * @param green the green value in the range of 0 to 255
   * @param blue the blue value in the range of 0 to 255
   */
   public Color(int red, int green, int blue)
   {
      this.red = red;
      this.green = green;
      this.blue = blue;
      this.rgb = (red << 16) | (green << 8) | blue;
      if (isGray)
         this.equ = this.gray = (((red << 5) + (green << 6) + (blue << 2)) / 100) >> colorDiv;
      else
         this.equ = this.rgb;
   }
   /**
   * Constructs a color object with the given rgb
   * @param rgb a color in the format 0x00RRGGBB
   */
   public Color(int rgb)
   {
      red = (rgb >> 16) & 0xFF;
      green = (rgb >> 8) & 0xFF;
      blue = rgb & 0xFF;
      this.rgb = rgb;
      if (isGray)
         this.equ = this.gray = (((red << 5) + (green << 6) + (blue << 2)) / 100) >> colorDiv;
      else
         this.equ = this.rgb;
   }

   /**
   * Constructs a color object with the given rgb
   * @param rgb a string in the format RRGGBB (colors in hex).
   */
   public Color(String rrggbb) // guich@300_9
   {
      this((int)Convert.toLong(rrggbb,16));
   }

   /** Returns the alpha channel of this Color in a value from 0 to 255 */
   public int getAlpha()
   {
      return ((red << 5) + (green << 6) + (blue << 2)) / 100;
   }

   /** Returns the blue value of the color. */
   public int getBlue()
   {
      return blue;
   }

   /** Returns the green value of the color. */
   public int getGreen()
   {
      return green;
   }

   /** Returns the red value of the color. */
   public int getRed()
   {
      return red;
   }

   /** Returns this index of the color in gray scale mode. Since gray scale mode range from 4 to 16 colors, it may return indexes from 0 to 15, depending on the device.
   * @since SuperWaba2.0
   */
   public int getGray()
   {
      return isGray?gray:getAlpha(); // guich@230_21
   }
   /** Returns the current color as 0x00RRGGBB */
   public int getRGB()
   {
      return rgb;
   }

   /** returns the best hash code for this color (equ) */
   public int hashCode()
   {
      return equ;
   }

   /** Returns true if this color equals to the other color. if in grayScale mode, returns true if both grayValues are equal. */
   public boolean equals(Object obj)
   {
      if (this == obj)
         return true;
      if (obj instanceof Color)
         return ((Color)obj).equ == equ;
      return false;
   }

   private final int min(int i1, int i2)
   {
      return i1<i2?i1:i2;
   }
   private final int max(int i1, int i2)
   {
      return i1>i2?i1:i2;
   }

   /** return a color that is brighter than the current one. if grayscale, returns the following brighter color */
   public Color brighter()
   {
      if (isGray)
      {
         if (bright == null) bright = Palette.SystemPalette.getColor(gray+fadeStep); //c = GS_PAL[gray-1];
         //Vm.debug("brighter of "+Convert.toString(rgb,16)+"("+equ+") = "+Convert.toString(bright.rgb,16)+"("+bright.equ+")");
         return bright;
      }
      else
         return new Color(min(red+fadeStep,255),min(green+fadeStep,255),min(blue+fadeStep,255));
   }
   /** return a color that is darker than the current one. if grayscale, returns the following darker color */
   public Color darker()
   {
      if (isGray)
      {
         if (dark == null) dark = Palette.SystemPalette.getColor(gray-fadeStep); //c = GS_PAL[gray+1];
         //Vm.debug("darker of "+Convert.toString(rgb,16)+"("+equ+") = "+Convert.toString(dark.rgb,16)+"("+dark.equ+")");
         return dark;
      }
      else
         return new Color(max(red-fadeStep,0),max(green-fadeStep,0),max(blue-fadeStep,0));
   }
   /** return a color that is brighter than the current one. if grayscale, returns the following brighter color */
   public Color brighter(int step)
   {
      if (isGray)
         return Palette.SystemPalette.getColor(gray+step);
      else
         return new Color(min(red+step,255),min(green+step,255),min(blue+step,255));
   }
   /** return a color that is darker than the current one. if grayscale, returns the following darker color */
   public Color darker(int step)
   {
      if (isGray)
         return Palette.SystemPalette.getColor(gray-step);
      else
         return new Color(max(red-step,0),max(green-step,0),max(blue-step,0));
   }
   /** Returns the string representation of this color: the rgb in hexadecimal */
   public String toString()
   {
      return isGray?Convert.toString(gray):Convert.unsigned2hex(rgb,6); // never change this!
   }

   /** Returns the best cursor color depending on the color */
   public Color getCursorColor()
   {
      int maxColors = Settings.maxColors;
      if (maxColors==4)
         return Palette.SystemPalette.getColor(1); // guich@230_1
      return getAlpha() > 128 ? darker(maxColors>=256?fadeStep:smallerFadeStep) : brighter(maxColors>=256?fadeStep:smallerFadeStep); // guich@220_45 - guich@340_51: in color devices, use normal step
   }
}