/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>                         *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.fx;

/**
 * Font is the character font used when drawing text on a surface.
 * <p>
 * Fonts have a name, such as "Helvetica", a style and a point size (usually
 * around 10). It's important to note that many devices have an extremely
 * limited number of fonts. For example, most PalmPilot devices have only
 * two fonts: plain and bold. If the font specified can't be found during
 * drawing, the closest matching font will be used.
 * <p>
 * Here is an example showing text being drawn in a given font:
 *
 * <pre>
 * ...
 * Font font = new Font("Tiny", Font.BOLD, 10);
 *
 * public void onPaint(Graphics g)
 * {
 * g.setFont(font);
 * g.drawText("Hello", 10, 10);
 * ...
 * </pre>
 * <b>SuperWaba Notes:</b>
 * <br><br>1. If font size is >= 14, it uses the large font set.
 * <br><br>2. Since PalmOS 2.0 dont have large fonts, you must install in the device the font files
 * LSW.pdb located under SuperwabaSDK/lib. If this file
 * is not present, in PalmOS 2.0, the large font set will not be used.<br>
 * <br><br>3. Starting from SuperWaba 2.0 beta 2, custom fonts are supported. There are lots of custom
 * fonts under <a href=http://www.palmgear.com/software/answer_category.cfm?categoryIDs=198>PalmGear Developer Fonts</a>
 * and at <a href=http://www.symbioforge.com/fe/developer.html>Handy Pilot Font Editor</a>
 * <br><br>4. To make the fonts usable under SuperWaba, you must change their names to match a
 * pattern: fontName|Small/Large|Bold. Eg: suppose you download or create the font
 * myFont. If you create the 4 supported types: small plain large plain, small bold
 * and large bold, then you must set their name (see note 8.) to: myFontSmall, myFontSmallBold,
 * myFontLarge, myFontLargeBold. By doing that, when creating a font under SuperWaba, you
 * just specify its name: <code>g.setFont(new Font("myFont", Font.PLAIN or Font.BOLD,12 or 14))</code>.
 * <br><br>5. If you dont follow the pattern, you must specify the full name of the font at each
 * time it is used. Eg: <code>myLabel.setFont(new Font("SystemNarrow",Font.PLAIN,12));</code>
 * In this case, the style and size parameters are not used.
 * <br><br>6. If you specify a font under PalmOS that dont exist and isnt installed, the default
 * fonts from Rom will be used.
 * <br><br>7. Under the SuperwabaSDK/lib there are 4 pdb files: LSW.pdb, MSW.pdb, HSW.pdb and 5SW.pdb.
 * These font files are identical to the palm internal fonts, and are used to emulate the devices and
 * are selected by the given width/height parameters.
 * <br><br>8. To see if the font you created is installed in the target device, query its name after
 * the creation. If the font is not found, its name is changed to match the default font.
 * <br><br>9. <b>Important!</b> To change a font name, only renaming isnt enough. The font name is also
 * stored internally in the file. Please use the utility RenameFont located at superwaba/utils.
 * <br><br>10. You may create new fonts based on the TrueType fonts using the utility superwaba/utils/xplat/TTF2PDBs
 */

public final class Font
{
   private String name;
   private int style;
   private int size;
   // HOOK VARIABLES - can't be private - guich@350_20
   Object hvUserFont;

   /** A plain font style. */
   public static final int PLAIN  = 0;

   /** A bold font style. */
   public static final int BOLD   = 1;

   /**
   * Creates a font of the given name, style and size. Font styles are defined
   * in this class.
   * @see #PLAIN
   * @see #BOLD
   * @see Graphics
   */
   public Font(String name, int style, int size)
   {
	   this.name = name;
	   this.style = style;
	   this.size = size;
	   NativeMethods4JDK.fontCreate(this);
   }

   /** Returns the name of the font. */
   public String getName()
   {
	   return name;
   }

   /** Returns the size of the font. */
   public int getSize()
   {
	   return size;
   }

   /**
   * Returns the style of the font. Font styles are defined in this class.
   * @see #PLAIN
   * @see #BOLD
   */
   public int getStyle()
   {
	   return style;
   }

   /** returns this font as Bold */
   public Font asBold()
   {
      return new Font(name,BOLD,size);
   }
}