/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.fx;

/**
 * FontMetrics computes font metrics including string width and height.
 * <p>
 * FontMetrics are usually used to obtain information about the widths and
 * heights of characters and strings when drawing text on a surface.
 * A FontMetrics object references a font and surface since fonts may have
 * different metrics on different surfaces.
 * <p>
 * Here is an example that uses FontMetrics to get the width of a string:
 *
 * <pre>
 * ...
 * Font font = new Font("Helvetica", Font.BOLD, 10);
 * FontMetrics fm = this.getFontMetrics(font);
 * String s = "This is a line of text.";
 * int stringWidth = fm.getTextWidth(s);
 * ...
 * </pre>
 */

public final class FontMetrics
{
   // Note! These fields are accessed directly from the VM!
   protected Font font;
   private ISurface surface;
   protected int ascent;
   protected int descent;
   protected int leading;

   /**
   * Constructs a font metrics object referencing the given font and surface.
   * <p>
   * If you are trying to create a font metrics object in a Control subclass,
   * use the getFontMetrics() method in the Control class.
   * @see waba.ui.Control#getFontMetrics(waba.fx.Font font)
   */
   public FontMetrics(Font font, ISurface surface)
   {
	   this.font = font;
	   this.surface = surface;
	   NativeMethods4JDK.fontMetricsCreate(this);
   }

   /**
   * Returns the ascent of the font. This is the distance from the baseline
   * of a character to its top.
   */
   public int getAscent()
   {
	   return ascent;
   }

   /**
   * Returns the width of the given character in pixels.
   */
   public int getCharWidth(char c)
   {
      return NativeMethods4JDK.fontMetricsGetCharWidth(this,c);
   }

   /**
   * Returns the descent of a font. This is the distance from the baseline
   * of a character to the bottom of the character.
   */
   public int getDescent()
   {
	   return descent;
   }

   /**
   * Returns the height of the referenced font. This is equal to the font's
   * ascent plus its descent. This does not include leading (the space between lines).
   */
   public int getHeight()
   {
	   return ascent + descent;
   }

   /**
   * Returns the external leading which is the space between lines.
   */
   public int getLeading()
   {
	   return leading;
   }

   /**
   * Returns the width of the given text string in pixels.
   */
   public int getTextWidth(String s)
   {
      return NativeMethods4JDK.fontMetricsGetTextWidth(this,s);
   }

   /**
   * Returns the width of the given text in pixels.
   * @param chars the text character array
   * @param start the start position in array
   * @param count the number of characters
   */
   public int getTextWidth(char chars[], int start, int count)
   {
      return NativeMethods4JDK.fontMetricsGetTextWidth(this,chars,start,count);
   }
}