/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>                         *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.fx;

import waba.sys.*;
import waba.applet.*;
import waba.ui.*;
import waba.util.*;

//////////   IMPORTANT! ANY CHANGES HERE MUST BE SYNC'ED WITH GRAPHICS.JAV ///////

/**
 * Graphics draws on a surface.
 * <p>
 * Since SuperWaba 2.0: now the Graphics class is much more improved, with lots of new methods.
 * It also supports Grayscale in all Palm OS operating system versions.
 * In grayscale mode, color can be from 0 to 3, where 0 = no color and 3 = darker color.
 * <p>
 * Surfaces are objects that implement the ISurface interface.
 * MainWindow and Image are both examples of surfaces.
 * <p>
 * Here is an example that uses Graphics to draw a line:
 *
 * <pre>
 * public class MyProgram extends MainWindow
 * {
 * public void onPaint(Graphics g)
 *  {
 *  g.setForeColor(new Color(0, 0, 255));
 *  g.drawLine(0, 0, 10, 10);
 *  g.setBackColor(new Color(255,0,255));
 *  g.fillCircle(50,50,20);
 * ...
 * </pre>
 * @see waba.fx.Color
 */

// This class is used in Desktop.

public class Graphics
{
   public Font lastFont; // 1
   protected int drawOp; // 2
   protected ISurface surface; // 3 dont change this order

   /** used in the draw3dRect method */
   static public final byte R3D_EDIT=1;
   /** used in the draw3dRect method */
   static public final byte R3D_LOWERED=2;
   /** used in the draw3dRect method */
   static public final byte R3D_RAISED=3;
   /** used in the draw3dRect method */
   static public final byte R3D_CHECK=4;
   /** used in the draw3dRect method. Note that the <i>simple</i> parameter acts in anothe way: if is true, a menu shaded is drawed; if is false, a listbox shaded is drawed. Also note that the fore color must have been set. */
   static public final byte R3D_SHADED=5;
   /** used in the drawArrow method */
   static public final byte ARROW_UP = 1;
   /** used in the drawArrow method */
   static public final byte ARROW_DOWN = 2;
   /** used in the drawArrow method */
   static public final byte ARROW_LEFT = 3;
   /** used in the drawArrow method */
   static public final byte ARROW_RIGHT = 4;

   /** The constant for a draw operation where the destination is replaced with source pixels (copy mode). */
   public static final int DRAW_PAINT         = 0;


   /** The constant for a draw operation where the destination is cleared where source pixels are off (AND mode). */
   public static final int DRAW_ERASE         = 1;


   /** The constant for a draw operation where the destination is cleared where source pixels are on (AND NOT mode). */
   public static final int DRAW_MASK          = 2;


   /** The constant for a draw operation where the destination is inverted where source pixels are on (XOR mode). */
   public static final int DRAW_INVERT        = 3;


   /** The constant for a draw operation where the destination set only where source pixels are on (OR mode). */
   public static final int DRAW_OVERLAY       = 4;


   /** The constant for a draw operation where the destination is replaced with inverted source (copy NOT mode). */
   public static final int DRAW_PAINT_INVERSE = 5;


   /** The constant for a draw operation where the destination is replaced if source pixels != <i>background color</i> (simulating transparent color). */
   public static final int DRAW_SPRITE        = 6;


   /**
   * The constant for a draw operation where the destination is replaced with <i>foreground color</i> if source pixels different of <i>background color</i>.
   * Note that in WindowsCE this operation can be very slow for big images.
   */
   public static final int DRAW_REPLACE_COLOR = 7;


   /** The constant for a draw operation where the <i>foreground color</i> is swaped with the <i>background color</i>.
   * Note that in WindowsCE this operation can be very slow for big images.
   */
   public static final int DRAW_SWAP_COLORS   = 8;


   protected static int[] currentPalette = Settings.maxColors >= 256 ? NativeMethods4JDK.default256Palette :
                                           Settings.maxColors ==  16 ? Palette.default16PaletteDev2Scr
                                                                     : Palette.default4PaletteDev2Scr;
   protected static Palette ScreenPalette = Settings.maxColors >= 256 ? new Palette(NativeMethods4JDK.default256Palette) :
                                            Settings.maxColors ==  16 ? new Palette(Palette.default16PaletteScr2Dev)
                                                                      : new Palette(Palette.default4PaletteScr2Dev);

   private int MaxScreenX = Settings.screenWidth;
   private int MaxScreenY = Settings.screenHeight;
   private int TRANSX = 0, TRANSY = 0;
   private int CLIPX1 = 0, CLIPY1 = 0, CLIPX2 = MaxScreenX, CLIPY2 = MaxScreenY; // ps: java purists, this up case is to maintain the code equal to the original c code.
   private int maxX=MaxScreenX, maxY=MaxScreenY;
   private Font font;

   private int [][]pixels; // this array will store the pixels [y][x]
   private int foreColor=Color.BLACK.getRGB(),backColor=Color.BRIGHT.getRGB();
   private static boolean firstBadParam = true;
   private static Rect rtemp = new Rect();
   private static boolean isWinCE;

   /**
    * Constructs a graphics object which can be used to draw on the given
    * surface. For the sake of the methods in this class, the given surface
    * is known as the "current surface".
    * <p>
    * If you are trying to create a graphics object for drawing in a subclass
    * of control, use the createGraphics() method in the Control class. It
    * creates a graphics object and translated the origin to the origin of
    * the control.
    */
   public Graphics(ISurface surface)
   {
      this.isWinCE = Settings.screenWidth == 240 || Settings.screenWidth > 320;
      this.surface = surface;
      pixels = getPixels(surface);
      font = MainWindow.defaultFont;
      if (surface instanceof waba.fx.Image) // guich@200b4_37
      {
         /*maxScreenX = */maxX = CLIPX2 = ((waba.fx.Image)surface).width;
         /*maxScreenY = */maxY = CLIPY2 = ((waba.fx.Image)surface).height;
      }
   }

   /** Draws an Ellipse, using the current foreground color as the outline color.
    * @param xc x coordinate of the center of the ellipse
    * @param yc y coordinate of the center of the ellipse
    * @param rx radix x of the ellipse
    * @param ry radix y of the ellipse
    */
   public void drawEllipse(int xc, int yc, int rx, int ry)
   {
      ellipseDrawAndFill(xc,yc,rx,ry,foreColor,false);
   }

   /** Fills an Ellipse, using the current background color as the fill color.
    * @param xc x coordinate of the center of the ellipse
    * @param yc y coordinate of the center of the ellipse
    * @param rx radix x of the ellipse
    * @param ry radix y of the ellipse
    */
   public void fillEllipse(int xc, int yc, int rx, int ry)
   {
      ellipseDrawAndFill(xc,yc,rx,ry,backColor,true);
   }

   /**
    * Draws an arc, using the current foreground color as the outline color. If startAngle < 0 and endAngle > 359 a whole circle is drawn.
    * A pie differs from an arc so the pie connects the end points of the arc to the center, thus making a closed figure.
    * @param xc x coordinate of the center of the circle that contains the arc.
    * @param yc y coordinate of the center of the circle that contains the arc.
    * @param r radix of the circle that contains the arc
    * @param startAngle starting angle of the arc. It must be between 0 and 360 (degrees). 0 is at 3 o'clock.
    * @param endAngle ending angle of the arc. It must be between 0 and 360 (degrees). 0 is at 3 o'clock.
    */
   public void drawArc(int xc, int yc, int r, float startAngle, float endAngle)
   {
      preArcPiePointDrawAndFill(xc,yc,r,r,startAngle,endAngle,foreColor,foreColor,false,false);
   }

   /**
    * Draws a pie slice, using the current foreground color as the outline color. If startAngle < 0 and endAngle > 359 a whole circle is drawn.
    * A pie differs from an arc so the pie connects the end points of the arc to the center, thus making a closed figure.
    * @param xc x coordinate of the center of the circle that contains the arc.
    * @param yc y coordinate of the center of the circle that contains the arc.
    * @param r radix of the circle that contains the pie
    * @param startAngle starting angle of the arc. It must be between 0 and 360 (degrees). 0 is at 3 o'clock.
    * @param endAngle ending angle of the arc. It must be between 0 and 360 (degrees). 0 is at 3 o'clock.
    */
   public void drawPie(int xc, int yc, int r, float startAngle, float endAngle)
   {
      preArcPiePointDrawAndFill(xc,yc,r,r,startAngle,endAngle,foreColor,foreColor,false,true);
   }

   /**
    * Fills a pie slice, using the current background color as the fill color and the
    * current foreground color as the outline color.
    * If startAngle < 0 and endAngle > 359 a whole circle is filled.
    * @param xc x coordinate of the center of the circle that contains the arc.
    * @param yc y coordinate of the center of the circle that contains the arc.
    * @param r radix of the circle that contains the pie
    * @param startAngle starting angle of the arc. It must be between 0 and 360 (degrees). 0 is at 3 o'clock.
    * @param endAngle ending angle of the arc. It must be between 0 and 360 (degrees). 0 is at 3 o'clock.
    */
   public void fillPie(int xc, int yc, int r, float startAngle, float endAngle)
   {
      preArcPiePointDrawAndFill(xc,yc,r,r,startAngle,endAngle,foreColor,backColor,true,true);
   }

   /**
    * Draws an elliptical arc, using the current foreground color as the outline color.
    * If startAngle < 0 and endAngle > 359 a whole circle is drawn.
    * A pie differs from an arc so the pie connects the end points of the arc to the center, thus making a closed figure.
    * @param xc x coordinate of the center of the circle that contains the arc.
    * @param yc y coordinate of the center of the circle that contains the arc.
    * @param rx x radix of the circle that contains the pie
    * @param ry y radix of the circle that contains the pie
    * @param startAngle starting angle of the arc. It must be between 0 and 360 (degrees). 0 is at 3 o'clock.
    * @param endAngle ending angle of the arc. It must be between 0 and 360 (degrees). 0 is at 3 o'clock.
    */
   public void drawEllipticalArc(int xc, int yc, int rx, int ry, float startAngle, float endAngle)
   {
      preArcPiePointDrawAndFill(xc,yc,rx,ry,startAngle,endAngle,foreColor,0,false,false);
   }

   /**
    * Draws an elliptical pie slice, using the current foreground color as the outline color.
    * If startAngle < 0 and endAngle > 359 a whole circle is drawn.
    * A pie differs from an arc so the pie connects the end points of the arc to the center, thus making a closed figure.
    * @param xc x coordinate of the center of the circle that contains the arc.
    * @param yc y coordinate of the center of the circle that contains the arc.
    * @param rx x radix of the circle that contains the pie
    * @param ry y radix of the circle that contains the pie
    * @param startAngle starting angle of the arc. It must be between 0 and 360 (degrees). 0 is at 3 o'clock.
    * @param endAngle ending angle of the arc. It must be between 0 and 360 (degrees). 0 is at 3 o'clock.
    */
   public void drawEllipticalPie(int xc, int yc, int rx, int ry, float startAngle, float endAngle)
   {
      preArcPiePointDrawAndFill(xc,yc,rx,ry,startAngle,endAngle,foreColor,0,true,false);
   }

   /**
    * Fills an elliptical pie slice, using the current background color as the fill color and the
    * current foreground color as the outline color.
    * If startAngle < 0 and endAngle > 359 a whole circle is filled.
    * @param xc x coordinate of the center of the circle that contains the arc.
    * @param yc y coordinate of the center of the circle that contains the arc.
    * @param rx x radix of the circle that contains the pie
    * @param ry y radix of the circle that contains the pie
    * @param startAngle starting angle of the arc. It must be between 0 and 360 (degrees). 0 is at 3 o'clock.
    * @param endAngle ending angle of the arc. It must be between 0 and 360 (degrees). 0 is at 3 o'clock.
    */
   public void fillEllipticalPie(int xc, int yc, int rx, int ry, float startAngle, float endAngle)
   {
      preArcPiePointDrawAndFill(xc,yc,rx,ry,startAngle,endAngle,foreColor,backColor,true,false);
   }

   /** Draws a circle, using the current foreground color as the outline color.
    * @param xc x coordinate of the center of the circle
    * @param yc y coordinate of the center of the circle
    * @param r radix of the circle
    */
   public void drawCircle(int xc, int yc, int r)
   {
      ellipseDrawAndFill(xc,yc,r,r,foreColor,false);
   }

   /** Fills a circle, using the current background color as the fill color.
    * @param xc x coordinate of the center of the circle
    * @param yc y coordinate of the center of the circle
    * @param r radix of the circle
    */
   public void fillCircle(int xc, int yc, int r)
   {
      ellipseDrawAndFill(xc,yc,r,r,backColor,true);
   }

   /** Gets the pixel color <b>index</b> at given position
    * @param x x coordinate of the pixel
    * @param y y coordinate of the pixel
    * @return the color <b>index</b> of the pixel or -1 if the pixel lies outside the clip bounds.
    * To get the corresponding color, you can use <code>Palette.getColor(g.getPixel(x,y))</code> for color devices,
    * and (Settings.maxColor-1-g.getPixel(x,y)) for gray scale devices.
    * Note that if the pixel is from an image, you must use the image's palette like here:
    * <code>img.getPalette().getColor(g.getPixel(x,y))</code> (be sure to read Image.getPalette docs).
    */
   public int getPixel(int x, int y)
   {
      x += TRANSX;
      y += TRANSY;
      if (CLIPX1 <= x && x < CLIPX2 && CLIPY1 <= y && y < CLIPY2)
         return pixels[y][x];
      return -1;
   }

   /** Sets the pixel at the specified position, using the current foreground color as the pixel color.
    * @param x x coordinate of the pixel
    * @param y y coordinate of the pixel
    */
   public void setPixel(int x, int y)
   {
      setPixel(x,y,foreColor);
   }

   /** Changes all the pixels with the current foreground color to the current background color */
   public void eraseRect(int x, int y, int w, int h)
   {
      int xx;
      int []xywh = new int[]{x,y,w,h};
      if (translateAndClip(xywh))
      {
         x = xywh[0]; y = xywh[1]; w = xywh[2]; h = xywh[3];
         int x2 = x + w;
         for (xx = x; xx < x2; xx++)
            eraseVLineNOCT(xx,y,h,foreColor,backColor);
      }
   }

   /** Draws a line at any direction, using the current foreground color as the line color.
    * The drawLine method is smart, so it calls
    * the fastest methods drawHLine/drawVLine if the line is horizontal/vertical.
    * @param Ax x coordinate of the start of the line.
    * @param Ay y coordinate of the start of the line.
    * @param Bx x coordinate of the end of the line.
    * @param By y coordinate of the end of the line.
    */
   public void drawLine(int Ax, int Ay, int Bx, int By)
   {
      drawLine(Ax,Ay,Bx,By,foreColor);
   }

   /** Clears the screen bounded by this graphics object using the background color as
    * the clear color.
    */
   public void clearScreen()
   {
      for (int y = 0; y < maxY; y++)
      {
         int w = maxX;
         JavaBridge.GraphicsLine(surface,0,y,w-1,y,backColor); // guich@400_47: ignore clippings
         while (--w >= 0)
            pixels[y][w] = backColor;
      }
   }

   /** Draws a dotted line in any direction, using the current foreground color
    * as the 1st, 3rd, ... pixels and the background color as the 2nd, 4th, ... pixels.
    * @param Ax x coordinate of the start of the line.
    * @param Ay y coordinate of the start of the line.
    * @param Bx x coordinate of the end of the line.
    * @param By y coordinate of the end of the line.
    */
   public void drawDots(int Ax, int Ay, int Bx, int By)
   {
      drawDottedLine(Ax,Ay,Bx,By,foreColor,backColor);
   }

   /** Draws a rectangle using the xor operation. So, if drawed twice,
    * the original colors are placed back.
    * @param x left coordinate of the rectangle
    * @param y top coordinate of the rectangle
    * @param w width of the rectangle. the rectangle is drawn from x to x+w-1.
    * @param h height of the rectangle. the rectangle is drawn from y to y+h-1.
    */
   public void drawCursor(int x, int y, int w, int h)
   {
      int xx;
      int []xywh = new int[]{x,y,w,h};
      if (translateAndClip(xywh))
      {
         x = xywh[0]; y = xywh[1]; w = xywh[2]; h = xywh[3];
         int x2 = x + w;
         for (xx = x; xx < x2; xx++)
            invVLineNOCT(xx,y,h);
      }
   }

   /** Draws a rectangle outline using the xor operation. So, if drawed twice,
    * the original colors are placed back.
    * @param x left coordinate of the rectangle
    * @param y top coordinate of the rectangle
    * @param w width of the rectangle. the rectangle is drawn from x to x+w-1.
    * @param h height of the rectangle. the rectangle is drawn from y to y+h-1.
    * @since SuperWaba2.0beta4
    */
   public void drawCursorOutline(int x, int y, int w, int h)
   {
      int []xywh = new int[]{x,y,w,h};
      if (translateAndClip(xywh))
      {
         x = xywh[0]; y = xywh[1]; w = xywh[2]; h = xywh[3];
         int y2 = y + h -1;
         invVLineNOCT(x,y,h);
         invVLineNOCT(x+w-1,y,h);
         while (w-- > 0)
         {
            invPixelNOCT(x,y);
            invPixelNOCT(x,y2);
            x++;
         }
      }
   }

   /** Draws a rectangle, using the current foreground color as the outline color.
    * @param x left coordinate of the rectangle
    * @param y top coordinate of the rectangle
    * @param w width of the rectangle. the rectangle is drawn from x to x+w-1.
    * @param h height of the rectangle. the rectangle is drawn from y to y+h-1.
    */
   public void drawRect(int x, int y, int w, int h)
   {
      drawHLine(x,y,w,foreColor);
      drawHLine(x,y+h-1,w,foreColor);
      drawVLine(x,y,h,foreColor);
      drawVLine(x+w-1,y,h,foreColor);
   }

   /** Fills a rectangle, using the current background color as the outline
    * and fill colors. If the coords are equal to the device screen size,
    * it uses the faster clearScreen.
    * @param x left coordinate of the rectangle
    * @param y top coordinate of the rectangle
    * @param w width of the rectangle. the rectangle is filled from x to x+w-1.
    * @param h height of the rectangle. the rectangle is filled from y to y+h-1.
    */
   public void fillRect(int x, int y, int w, int h)
   {
      if (h > 10000 || h < 0)
         badParam("fillRect("+x+','+y+','+w+','+h+')');
      else
         for (; h > 0; h--)
            drawHLine(x,y++,w,backColor);
   }

   /** Draws a dotted rectangle, using the current foreground color
    * as the 1st, 3rd, ... pixels and the background color as the 2nd, 4th, ... pixels.
    * @param x left coordinate of the rectangle
    * @param y top coordinate of the rectangle
    * @param w width of the rectangle. the rectangle is drawed from x to x+w-1.
    * @param h height of the rectangle. the rectangle is drawed from y to y+h-1.
    */
   public void drawDottedRect(int x, int y, int w, int h)
   {
      int c1 = foreColor;
      int c2 = backColor;
      int x2 = x+w-1;
      int y2 = y+h-1;
      drawDottedLine(x, y, x2,y, c1,c2);
      drawDottedLine(x, y, x ,y2,c1,c2);
      drawDottedLine(x2,y, x2,y2,c1,c2);
      drawDottedLine(x, y2,x2,y2,c1,c2);
   }

   /** Fills a polygon, using the current background color as the fill color.
    * The polygon can be convex or concave.
    * @param xPoints the array with the x coordinate points
    * @param yPoints the array with the y coordinate points
    * @param nPoints the total number of point to be drawn.
    */
   public void fillPolygon(int []xPoints, int []yPoints, int nPoints)
   {
      if (nPoints > xPoints.length || nPoints > yPoints.length)
         Vm.debug("array index out of range: "+nPoints);
      else
         fillPolygon(xPoints,yPoints,nPoints,backColor);
   }

   /** Draws a polygon, using the current foreground color as the outline color.
    * The polygon is automatically closed.
    * @param xPoints the array with the x coordinate points
    * @param yPoints the array with the y coordinate points
    * @param nPoints the total number of point to be drawn.
    */
   public void drawPolygon(int []xPoints, int []yPoints, int nPoints)
   {
      if (nPoints > xPoints.length || nPoints > yPoints.length)
         Vm.debug("array index out of range: "+nPoints);
      else
         drawPolygon(xPoints,yPoints,nPoints,foreColor,true);
   }

   /** Draws a sequence of connected lines, using the current foreground color as the outline color.
    * The lines are <b>not</b> closed.
    * @param xPoints the array with the x coordinate points
    * @param yPoints the array with the y coordinate points
    * @param nPoints the total number of point to be drawn.
    */
   public void drawPolyline(int []xPoints, int []yPoints, int nPoints) // guich@330_19
   {
      if (nPoints > xPoints.length || nPoints > yPoints.length)
         Vm.debug("array index out of range: "+nPoints);
      else
         drawPolygon(xPoints,yPoints,nPoints,foreColor,false);
   }

   /** Gets the coordinates where the specified point lies.
    * This is useful if you're drawing a pie chart and want to know where to draw the label.
    * @param xc x coordinate of the center of the circle
    * @param yc y coordinate of the center of the circle
    * @param rx x radix of the circle that contains the point
    * @param ry y radix of the circle that contains the point
    * @param angle angle of the point. It must be between 0 and 360 (degrees). 0 is at 3 o'clock.
    * @return the Coord
    * @see waba.fx.Coord
    */
   public waba.fx.Coord getAnglePoint(int xc, int yc, int rx, int ry, float angle) // guich@300_41: fixed method signature
   {
      int []twoInts = new int[2];
      while (angle >= 360) angle -= 360;
      while (angle < 0) angle += 360;
      arcPiePointDrawAndFill(xc,yc,rx,ry,angle,angle,0,0,false,false,twoInts);
      return new waba.fx.Coord(twoInts[0],twoInts[1]);
   }

   /** Draws a text with the current font and the current foregrount color.
     * This method is slower than drawText(String,x,y) because it convert the
     * chars to a string.
     * @param chars the character array to display
     * @param start the start position in array
     * @param count the number of characters to display
     * @param x the left coordinate of the text's bounding box
     * @param y the top coordinate of the text's bounding box
     */
   public void drawText(char chars[], int start, int count, int x, int y)
   {
      drawText(new String(chars,start,count),x,y);
   }

   /** Draws a text with the current font and the current foregrount color.
    * @param text the text to be drawn
    * @param x x coordinate of the text
    * @param y y coordinate of the text
    */
   public void drawText(String text, int x, int y)
   {
      int ands8[] = {0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01};
      int offset,index;
      int width,w;
      int start,current;
      int startBit,currentBit;
      // sppedup
      waba.applet.UserFont font = JavaBridge.getFont(this.font);
      int height = (byte)font.fRectHeight,h = height;
      int firstChar = font.firstChar;
      int lastChar = font.lastChar-firstChar; // guich@340_46

      // guich@400_43: if emulating wince, erase the background, since this will happen in the device too.
      if (isWinCE)
         fillRect(x,y,font.stringWidth(text),font.ascent+font.descent+font.leading);

      if (text == null || text.length() == 0) return; // guich@200b3: corrected bug if text is null
      char []chars = text.toCharArray();
      for (int i =0; i < chars.length; i++)
      {
         char ch = chars[i];
         if (ch == 13 || ch == 10) // guich@341_9: fixed <enter> problems with MultiEdit - guich@350_6: added test for char(10)
            continue;
         index = (short)ch - firstChar;
         if (index < 0 || index > lastChar)
         {
            // guich@320_50: fix support for invalid chars - index = numberOfChars; // MCS index
            index = font.numberOfChars;//(byte)('?' - firstChar);
         }
         // Get the source x coordinate and width of the character
         offset = font.bitIndexTable[index];
         width  = (byte)(font.bitIndexTable[index+1] - offset);
         if (width == 0) // guich@340_46: show the MCS char if no width (glyph not present)
         {
            index  = font.numberOfChars;
            offset = font.bitIndexTable[index];
            width  = (byte)(font.bitIndexTable[index+1] - offset);
         }
         start     = (offset >> 3);
         startBit  = offset & 7;

         // draws the char
         for (h = height; h-- > 0; start += font.rowWidthInBytes) // draw each row
         {
            current = start;
            currentBit = startBit;
            for (w = width; w-- > 0;x++)  // draw each pixel
            {
               if ((font.bitmapTable[current] & ands8[currentBit]) != 0)
                  setPixel(x,y,foreColor);
               if (++currentBit == 8) // finished this byte?
               {
                  currentBit = 0; // reset counter
                  current++;      // inc current byte
               }
            }
            x -= width;
            y++;
         }
         x += width;
         y -= height;
      }
   }

   /** Draws a rectangle with rounded corners, using the current foreground color as the outline color.
    * @param x left coordinate of the rectangle
    * @param y top coordinate of the rectangle
    * @param w width of the rectangle. the rectangle is drawn from x to x+w-1.
    * @param h height of the rectangle. the rectangle is drawn from y to y+h-1.
    * @param r radix of the circle at the corners. If its greater than width/2 or greater than height/2, it will be adjusted to the minimum of both values.
    */
   public void drawRoundRect(int x, int y, int width, int height, int r)
   {
      if (r > (width >> 1) || r > (height >> 1)) r = Math.min(width >> 1,height >> 1); // guich@200b4_6: correct bug that crashed the device.
      int w = width - (r<<1);
      int h = height - (r<<1);
      int x1 = x+r,y1=y+r,x2=x+width-r-1,y2=y+height-r-1;
      int dec = 3-(r<<1),xx,yy;
      int c = foreColor;

      drawHLine(x+r,y,w,c); // top
      drawHLine(x+r,y+height-1,w,c); // bottom
      drawVLine(x,y+r,h,c); // left
      drawVLine(x+width-1,y+r,h,c); // right

      for (xx = 0, yy = r; xx <= yy; xx++)
      {
         setPixel(x2+xx,y2+yy,c);
         setPixel(x2+xx,y1-yy,c);
         setPixel(x1-xx,y2+yy,c);
         setPixel(x1-xx,y1-yy,c);

         setPixel(x2+yy,y2+xx,c);
         setPixel(x2+yy,y1-xx,c);
         setPixel(x1-yy,y2+xx,c);
         setPixel(x1-yy,y1-xx,c);
         if (dec >= 0)
            dec += -4*(yy--)+4;
         dec += 4*xx+6;
      }
   }

   /** Fills a rectangle with rounded corners, using the current background color as the fill color.
    * @param x left coordinate of the rectangle
    * @param y top coordinate of the rectangle
    * @param w width of the rectangle. the rectangle is filled from x to x+w-1.
    * @param h height of the rectangle. the rectangle is filled from y to y+h-1.
    * @param r radix of the circle at the corners. If its greater than width/2 or greater than height/2, it will be adjusted to the minimum of both values.
    */
   public void fillRoundRect(int x, int y, int width, int height, int r)
   {
      if (r > (width >> 1) || r > (height >> 1)) r = Math.min(width >> 1,height >> 1); // guich@200b4_6: correct bug that crashed the device.
      int x1 = x+r,y1=y+r,x2=x+width-r-1,y2=y+height-r-1;
      int dec = 3-(r<<1),xx,yy;
      int c = backColor;

      height -= (r<<1);
      y += r;
      while (height-- > 0)
         drawHLine(x,y++,width,c);

      for (xx = 0, yy = r; xx <= yy; xx++)
      {
         drawLine(x1-xx,y1-yy,x2+xx,y1-yy,c);
         drawLine(x1-xx,y2+yy,x2+xx,y2+yy,c);

         drawLine(x1-yy,y1-xx,x2+yy,y1-xx,c);
         drawLine(x1-yy,y2+xx,x2+yy,y2+xx,c);

         if (dec >= 0)
            dec += -4*(yy--)+4;
         dec += 4*xx+6;
      }
   }

   /** Sets the clipping rectangle, translated to the current translated origin.
    * Anything drawn outside of the rectangular area specified will be clipped.
    * Setting a clip overrides any previous clip. This clipping rectangle affects
    * all the drawing operations.
    * @param x left coordinate of the rectangle
    * @param y top coordinate of the rectangle
    * @param w width of the rectangle. the rectangle is filled from x to x+w-1.
    * @param h height of the rectangle. the rectangle is filled from y to y+h-1.
    */
   public void setClip(int x, int y, int w, int h)
   {
      int clipX1 = x+TRANSX;
      int clipY1 = y+TRANSY;
      int clipX2 = clipX1+w;
      int clipY2 = clipY1+h;

      if (clipX1 < 0) clipX1 = 0;
      if (clipY1 < 0) clipY1 = 0;
      if (clipX1 > maxX) clipX1 = maxX;
      if (clipY1 > maxY) clipY1 = maxY;

      if (clipX2 < 0) clipX2 = 0;
      if (clipY2 < 0) clipY2 = 0;
      if (clipX2 > maxX) clipX2 = maxX;
      if (clipY2 > maxY) clipY2 = maxY;

      CLIPX1 = clipX1;
      CLIPY1 = clipY1;
      CLIPX2 = clipX2;
      CLIPY2 = clipY2;
   }

   /** Sets the clipping rectangle, translated to the current translated origin.
    * Anything drawn outside of the rectangular area specified will be clipped.
    * Setting a clip overrides any previous clip. This clipping rectangle affects
    * all the drawing operations.
    * @param r the rectangle with the clipping coordinates
    */
   public void setClip(Rect r)
   {
      setClip(r.x, r.y, r.width, r.height);
   }

   /** Gets the current clipping rectangle.
    * @param r a Rect object where the clip coordinates will be stored and returned.
    */
   public Rect getClip(Rect r)
   {
      r.x = CLIPX1-TRANSX; // guich@200b4_38: fixed getClip - and guich@241_5: changed + to -
      r.y = CLIPY1-TRANSY;
      r.width  = CLIPX2-CLIPX1; // guich@300_36: removed +1 here and in line below
      r.height = CLIPY2-CLIPY1;
      return r;
   }

   /** Clips the specified rectangle to the clipping bounds.
    * @param r The coordinates of the rectangle that are to be adjusted to match the clipping bounds.
    * @return false if the specified rectangle does not intersects the clipping bounds
    * or the new coordinates of the clipped rectangle.
    */
   public boolean clip(Rect r)
   {
      int []xywh = new int[]{r.x,r.y,r.width,r.height};
      if (translateAndClip(xywh))
      {
         r.x = xywh[0] - TRANSX;
         r.y = xywh[1] - TRANSY;
         r.width = xywh[2];
         r.height = xywh[3];
         return true;
      }
      return false;
   }

   /**
   * Clears the current clipping rectangle. This allows drawing to occur
   * anywhere on the current surface.
   */
   public void clearClip()
   {
      setClip(0,0,maxX,maxY);
   }

   /** Translates the origin of the of the current coordinate system by the given dx and dy.
    * The final translation coords cannot be lesser than 0.
    * @param dx the delta x of the x coordinate translation. the new x origin coordinate will be x + dx
    * @param dy the delta y of the y coordinate translation. the new y origin coordinate will be y + dy
    */
   public void translate(int dx, int dy)
   {
      TRANSX += dx;
      TRANSY += dy;
   }

   /**
   * Copies a rectangular area from a surface to the given coordinates on
   * the current surface. The copy operation is performed by combining
   * pixels according to the setting of the current drawing operation.
   * The same surface can be used as a source and destination to
   * implement quick scrolling.
   * <p>
   * Not all combinations of surfaces are supported on all platforms.
   * Java doesn't allow copying from an Window surface to an Image.
   * <p>
   * If the bitmap surface being transferred has the width of the screen and
   * starts at x=0, the drawing occurs almost instantaneously.
   *
   * @param surface the surface to copy from
   * @param x the source x coordinate
   * @param y the source y coordinate
   * @param width the width of the area on the source surface
   * @param height the height of the area on the source surface
   * @param dstX the destination x location on the current surface
   * @param dstY the destination y location on the current surface
   * @see #setDrawOp
   */
   public void copyRect(ISurface surface, int x, int y, int width, int height, int dstX, int dstY)
   {
      int [][]srcPixels = getPixels(surface);
      if (srcPixels != null)
         drawBitmap(srcPixels, x,y,width,height, dstX,dstY, drawOp, backColor, foreColor, true);
   }

   /**
   * Frees any system resources (native device contexts) associated with the
   * graphics object. After calling this function, the graphics object can no
   * longer be used to draw. Calling this method is not required since any
   * system resources allocated will be freed when the object is garbage
   * collected. However, if a program uses many graphics objects, free()
   * should be called whenever one is no longer needed to prevent allocating
   * too many system resources before garbage collection can occur.
   * Note that calling free on Graphics created from Image objects is useless.
   */
   public void free()
   {
   }

   /** For backward compatibility only.
     * This method is slower
     * @deprecated Use setForeColor and setBackColor instead.
     */
   public void setColor(int r, int g, int b) // guich@330_29
   {
      foreColor = ScreenPalette.getColor(ScreenPalette.getColorIndex(new Color(r,g,b))).rgb;
      backColor = ScreenPalette.getColor(ScreenPalette.getColorIndex(new Color(r,g,b))).rgb;
   }

   /**
   * Sets the current color for border and drawing operations.
   * @param c the Color object with the r,g,b values
   */
   public void setForeColor(Color c)
   {
      foreColor = ScreenPalette.getColor(ScreenPalette.getColorIndex(c)).rgb;
      //Vm.debug("setForeColor("+Convert.toString(c.rgb,16)+") = "+Convert.toString(foreColor,16));
   }

   /**
   * Sets the current color for background and 2nd color operations.
   * @param c the Color object with the r,g,b values
   */
   public void setBackColor(Color c)
   {
      backColor = ScreenPalette.getColor(ScreenPalette.getColorIndex(c)).rgb;
      //Vm.debug("setBackColor("+Convert.toString(c.rgb,16)+") = "+Convert.toString(backColor,16));
   }

   /**
   * Sets the current color for border and drawing operations.
   * @param r the red value (0..255)
   * @param g the green value (0..255)
   * @param b the blue value (0..255)
   */
   public void setForeColor(int r, int g, int b)
   {
      foreColor = ScreenPalette.getColor(ScreenPalette.getColorIndex(new Color(r,g,b))).rgb;
   }

   /**
   * Sets the current color for background and 2nd color operations.
   * @param r the red value (0..255)
   * @param g the green value (0..255)
   * @param b the blue value (0..255)
   */
   public void setBackColor(int r, int g, int b)
   {
      backColor = ScreenPalette.getColor(ScreenPalette.getColorIndex(new Color(r,g,b))).rgb;
   }

   /** Sets the current font for operations that draw text. */
   public void setFont(Font font)
   {
      if (font == null)
         badParam("setFont("+font+')');
      this.lastFont = this.font = font;
   }

   /** Tests if a surface collides with an image. Note that the surface *can* also be an image.
   * The coordinates of both images are tested relative to the same origin.
   * @param surface one image or Window to test for collision
   * @param x1 the left position where this image is "drawn"
   * @param y1 the top position where this Image is "drawn"
   * @param img2 the other image to test for collision
   * @param x2 the left position where the other image is "drawn"
   * @param y2 the top position where the other Image is "drawn"
   * @return true if both image bounds intersects and if there are pixels whose color is different of the <b>current background color</b> in both images
   */
   public boolean collisionDetect(ISurface surface, int x1, int y1, Image img, int x2, int y2)
   {
      int [][]pixels1 = getPixels(surface);
      int [][]pixels2 = getPixels(img);

      Rect rect1 = new Rect(x1,y1,pixels1[0].length,pixels1.length); // guich@200b4_208: changed as Daniel Tauchke asked, don't know if there will be side effects.
      Rect rect2 = new Rect(x2,y2,pixels2[0].length,pixels2.length);

      if (!(surface instanceof Image))
      {
         rect1.width -= rect1.x;
         rect1.height -= rect1.y;
      }

      int i,j,w,h;
      // first and simplier test: if rects dont intersect, no colision could occur.
      if (!intersects(rect1,rect2)) return false;
      Rect rect = intersection(rect1,rect2);
      w = rect.width;
      h = rect.height;
      x1 = rect.x - rect1.x;
      y1 = rect.y - rect1.y;
      x2 = rect.x - rect2.x;
      y2 = rect.y - rect2.y;
      // now scans for pixels != color in both images
      for (j = 0; j < h; j++)
         for (i = 0; i < w; i++)
            if (pixels1[y1+j][x1+i] != backColor && pixels2[y2+j][x2+i] != backColor)
               return true;

      return false;
   }

   /** Returns the current translation from the origin for this Graphics object */
   public waba.fx.Coord getTranslation()
   {
      return new Coord(TRANSX,TRANSY);
   }

   /** apply the palette. since theres only one applied palette at any time,
    * this method is static. If you had modified only few colors in the palette,
    * specify its indexes or specify -1,-1 to apply all indexes of the palette.
    * To get the current applied palette, create a palette with the default constructor,
    * eg: new Palette(). */
   public static void applyPalette(Palette pal, int startIndex, int endIndex)
   {
      if (Settings.isColor)
      {
         if (startIndex < 0) startIndex = 0;
         if (endIndex < 0) endIndex = pal.size()-1;
         _applyPalette(pal.toRGBArray(startIndex,endIndex),startIndex,endIndex);
      }
   }

   /** Draws a rectangle with hatched corners, using the current foreground color as the outline color.
    * If both top and bottom are false, nothing is drawn.
    * @param x left coordinate of the rectangle
    * @param y top coordinate of the rectangle
    * @param w width of the rectangle. the rectangle is drawn from x to x+w-1.
    * @param h height of the rectangle. the rectangle is drawn from y to y+h-1.
    * @param top true if the top corners will be hatched
    * @param bottom true if the bottom corners will be hatched
    */
   public void drawHatchedRect(int x, int y, int width, int height, boolean top, boolean bottom)
   {
      int x2 = x+width-1;
      int y2 = y+height-1;
      if (top && bottom)
      {
         drawLine(x,y+2,x,y2-2,foreColor); // left
         drawLine(x+2,y2,x2-2,y2,foreColor); // bottom
         drawLine(x2,y+2,x2,y2-2,foreColor); // right
         drawLine(x+2,y,x2-2,y,foreColor); // top
         setPixel(x+1,y+1,foreColor); // top left
         setPixel(x2-1,y+1,foreColor); // top right
         setPixel(x+1,y2-1,foreColor); // bottom left
         setPixel(x2-1,y2-1,foreColor); // bottom right
      }
      else
      if (top && !bottom)
      {
         drawLine(x,y+2,x,y2,foreColor); // left
         drawLine(x,y2,x2,y2,foreColor); // bottom
         drawLine(x2,y+2,x2,y2,foreColor); // right
         drawLine(x+2,y,x2-2,y,foreColor); // top
         setPixel(x+1,y+1,foreColor); // top left
         setPixel(x2-1,y+1,foreColor); // top right
      }
      else
      if (!top && bottom)
      {
         drawLine(x,y,x,y2-2,foreColor); // left
         drawLine(x+2,y2,x2-2,y2,foreColor); // bottom
         drawLine(x2,y,x2,y2-2,foreColor); // right
         drawLine(x,y,x2,y,foreColor); // top
         setPixel(x+1,y2-1,foreColor); // bottom left
         setPixel(x2-1,y2-1,foreColor); // bottom right
      }
   }

   /** Fills a rectangle with hatched corners, using the current background color as the fill color.
    * If both top and bottom are false, nothing is drawn.
    * @param x left coordinate of the rectangle
    * @param y top coordinate of the rectangle
    * @param w width of the rectangle. the rectangle is drawn from x to x+w-1.
    * @param h height of the rectangle. the rectangle is drawn from y to y+h-1.
    * @param top true if the top corners will be hatched
    * @param bottom true if the bottom corners will be hatched
    */
   public void fillHatchedRect(int x, int y, int width, int height, boolean top, boolean bottom)
   {
      int x2 = x+width-1;
      int y2 = y+height-1;
      if (top && bottom)
      {
         fillRect(x,y+2,width,height-4); // middle
         drawLine(x+2,y,x2-2,y,backColor); // 1st line
         drawLine(x+1,y+1,x2-1,y+1,backColor); // 2nd line
         drawLine(x+1,y2-1,x2-1,y2-1,backColor); // last-1 line
         drawLine(x+2,y2,x2-2,y2,backColor); // last line
      }
      else
      if (top && !bottom)
      {
         drawLine(x+2,y,x2-2,y,backColor); // 1st line
         drawLine(x+1,y+1,x2-1,y+1,backColor); // 2nd line
         fillRect(x,y+2,width,height-2); // middle
      }
      else
      if (!top && bottom)
      {
         fillRect(x,y,width,height-2); // middle
         drawLine(x+1,y2-1,x2-1,y2-1,backColor); // last-1 line
         drawLine(x+2,y2,x2-2,y2,backColor); // last line
      }
   }

   /** used to draw the topleft lines of a rect with one color and the bottomright with another color. if yMirror, draws the borders mirrowed.*/
   public void drawHighLightFrame(int x, int y, int w, int h, Color topLeft, Color bottomRight, boolean yMirror)
   {
      int x2 = x+w-1;
      int y2 = y+h-1;
      if (topLeft != null)
      {
         int c = (topLeft.red << 16) | (topLeft.green << 8) | topLeft.blue;
         drawLine(x,y,x,y2,c);
         if (!yMirror)
            drawLine(x,y,x2,y,c);
         else
            drawLine(x,y2,x2,y2,c);
      }
      if (bottomRight != null)
      {
         int c = (bottomRight.red << 16) | (bottomRight.green << 8) | bottomRight.blue;
         drawLine(x2,y2,x2,y,c);
         if (!yMirror)
            drawLine(x2,y2,x,y2,c);
         else
            drawLine(x2,y,x,y,c);
      }
   }

   /** Sets the origin of the current coordinate system to x and y.
    * The final translation coords cannot be lesser than 0.
    * @param dx the delta x of the x coordinate translation. the new x origin coordinate will be x + dx
    * @param dy the delta y of the y coordinate translation. the new y origin coordinate will be y + dy
    */
   public void translateTo(int x, int y)
   {
      waba.fx.Coord c = getTranslation();
      if (c.x != 0 || c.y != 0) // if already translated, move back to origin
         translate(-c.x,-c.y);
      if (x != 0 || y != 0)
         translate(x,y);
   }

   static Color lastColor;
   static int lastColorValue;
   /** Draws the image in the given position with the given draw operation
     * and back color. Note that the current draw operation is not changed,
     * neither the current back color.
     * A cache is made comparing the last back color with the current one.
     * This comparision is by reference, so, be sure to use the same
     * color instead of creating a new one with "new Color(xxx)"; otherwise,
     * the cache will never be used.
     * If drawOp is DRAW_PAINT, backGround can be null. If drawOp is DRAW_SPRITE,
     * you must provide the background color.
     * The parameter doClip must be used with caution. If false, it will not do any
     * clipping checks (what may improve drawings in 8%). Set it to true
     * only when you're sure that the image will not pass the current surface
     * bounds. Otherwise, be prepared to even <b>hard-reset</b> your device!
     * @since SuperWaba 3.3
     */
   public void drawImage(waba.fx.Image image, int x, int y, int drawOp, waba.fx.Color backColor, boolean doClip)
   {
      if (drawOp != DRAW_PAINT && backColor != lastColor) // this same cache is used in the pda
      {
         lastColor = backColor;
         lastColorValue = ScreenPalette.getColor(ScreenPalette.getColorIndex(backColor)).rgb;
      }
      int [][]srcPixels = getPixels(image);
      if (srcPixels != null)
         drawBitmap(srcPixels, 0,0, image.width, image.height, x,y, drawOp, lastColorValue, this.foreColor, doClip);
   }

   public void copyImageRect(waba.fx.Image image, int x, int y, int width, int height, int drawOp, waba.fx.Color backColor, boolean doClip)
   {
      if (drawOp != DRAW_PAINT && backColor != lastColor) // this same cache is used in the pda
      {
         lastColor = backColor;
         lastColorValue = ScreenPalette.getColor(ScreenPalette.getColorIndex(backColor)).rgb;
      }
      int [][]srcPixels = getPixels(image);
      if (srcPixels != null)
         drawBitmap(srcPixels, x,y, width, height, 0,0, drawOp, lastColorValue, this.foreColor, doClip);
   }

   /** Copies the given surface to this surface. This will ignore any clippings and
     * also will use the fastest transfer method. It will transfer the areas
     * <code>0, ySrc, Settings.screenWidth, height</code> to <code>0, yDst, Settings.screenWidth, height</code>.
     * The draw operation will be <code>DRAW_PAINT</code>.
     */
   public void copyScreen(waba.fx.ISurface src, int ySrc, int yDst, int height)
   {
      if (ySrc < 0 || yDst < 0 || ySrc+height > MaxScreenY || yDst+height > MaxScreenY)
         badParam("copyScreen("+src+','+ySrc+','+yDst+','+height+")");
      int [][]srcPixels = getPixels(src);
      if (srcPixels != null)
      {
         if (srcPixels[0].length != MaxScreenX || this.pixels[0].length != MaxScreenX)
            badParam("copyScreen surfaces must have the screen width ("+srcPixels[0].length+" or "+this.pixels[0].length+")!");
         while (height-- > 0)
            drawScanLine(srcPixels[ySrc++], 0, 0, yDst++,MaxScreenX);
      }
   }
   private static final int RANGE = 4000;
   /** Draws an image at the given x and y coordinates.
     * Note: you can pass waba.ui.Control.CENTER as parameter to x and/or y. You can also
     * pass waba.ui.Control.CENTER+delta, where delta is a number greater or lower than 0 to
     * place the image relative from the center place. EG: CENTER+2 (place at the center + 2
     * pixels to right or bottom, depending on which parameter you're using).
     * You can also use LEFT,RIGHT,TOP and BOTTOM.
     * @see #copyRect
     */
   public void drawImage(waba.fx.Image image, int x, int y)
   {
      int w = image.width;
      int h = image.height;
      if (x > 50000)
      {
         if ((waba.ui.Control.LEFT      -RANGE) <= x && x <= (waba.ui.Control.LEFT   +RANGE)) x = x-waba.ui.Control.LEFT; else
         if ((waba.ui.Control.RIGHT     -RANGE) <= x && x <= (waba.ui.Control.RIGHT  +RANGE)) x = Settings.screenWidth-w + x-waba.ui.Control.RIGHT; else
         if ((waba.ui.Control.CENTER    -RANGE) <= x && x <= (waba.ui.Control.CENTER +RANGE)) x = ((Settings.screenWidth-w) >> 1) + x-waba.ui.Control.CENTER;
      }
      if (y > 50000)
      {
         if ((waba.ui.Control.TOP       -RANGE) <= y && y <= (waba.ui.Control.TOP    +RANGE)) y = y-waba.ui.Control.TOP; else
         if ((waba.ui.Control.BOTTOM    -RANGE) <= y && y <= (waba.ui.Control.BOTTOM +RANGE)) y = Settings.screenHeight-h + y-waba.ui.Control.BOTTOM; else
         if ((waba.ui.Control.CENTER    -RANGE) <= y && y <= (waba.ui.Control.CENTER +RANGE)) y = ((Settings.screenHeight-h) >> 1) + y-waba.ui.Control.CENTER;
      }
      copyRect(image, 0, 0, w,h, x, y);
   }


   /**
   * Sets the current color as the foreground and background colors.
   * This method has the same effect as calling setForeColor(c)
   * and setBackColor(c).
   * @param c the Color object with the r,g,b values
   */
   public void setColors(Color c)
   {
      setBackColor(c);
      setForeColor(c);
   }


   /**
   * Sets the drawing operation. The setting determines the raster
   * operation to use when drawing lines, rectangles, text and
   * images on the current surface. It also determines how pixels are
   * combined when copying one surface to another. The setting of
   * DRAW_OVER, where any drawing simply draws over the pixels on
   * a surface, is the default.
   * <p>
   * Not all operations are supported on all platforms. When used with
   * Java, DRAW_OVER is supported for all types of drawing and DRAW_XOR
   * is supported for drawing lines, rectangles, text and images.
   * However, DRAW_XOR is not supported when copying surface areas and
   * the DRAW_AND and DRAW_OR operations aren't supported at all under
   * Java.
   * <p>
   * PalmOS platforms supports all the drawing operations when drawing
   * images and copying surface regions. However, only the DRAW_OVER
   * operation is supported when drawing lines, rectangles and text.
   * If you need to use the XOR drawing operation for drawing lines
   * under PalmOS, you can draw the line into an image and then draw
   * the image with an XOR drawing operation.
   * <p>
   * Win32 and Windows CE platforms support all the drawing operations
   * except when drawing text. Only DRAW_OVER is supported when drawing
   * text. If you need to draw XOR'ed text, you can draw the text into
   * an image and then draw the image with an XOR draw operation.
   * <p>
   * When calculating the result of XOR, AND and OR drawing, the value
   * of the color black is all 1's (fully set) in binary and white is
   * all 0's (fully unset).
   *
   * @param op drawing operation
   * @see #DRAW_PAINT
   * @see #DRAW_ERASE
   * @see #DRAW_MASK
   * @see #DRAW_INVERT
   * @see #DRAW_OVERLAY
   * @see #DRAW_PAINT_INVERSE
   * @see #DRAW_SPRITE
   * @see #DRAW_REPLACE_COLOR
   * @see #DRAW_SWAP_COLORS
   */
   public void setDrawOp(int drawOp)
   {
      this.drawOp = drawOp;
   }

   /** Returns the last font set with setFont */
   public Font getFont() // guich@200b4_164
   {
      return lastFont;
   }

   private static boolean moreThan4 = Settings.maxColors > 4;

   /** Prior to use draw3dRect, you must compute the colors. This must be done
     * everytime you change the enabled state or the fore/back colors. This can
     * be easily achieved if you extend the onColorsChanged method
     * (but don't forget to call super.onColorsChanged).
     */
   public static void compute3dColors(boolean enabled, Color backColor, Color foreColor, Color fourColors[])
   {
      fourColors[0] = backColor; // outside topLeft - BRIGHT
      fourColors[1] = backColor.brighter(Color.smallerFadeStep); // inside topLeft - WHITE
      fourColors[2] = foreColor.brighter(Color.smallerFadeStep); // inside bottomRight - DARK
      fourColors[3] = foreColor; // outside bottomRight - BLACK
      if (fourColors[0].equ == fourColors[1].equ) // both WHITE?
         fourColors[1] = fourColors[0].darker(Color.smallerFadeStep);
      if (moreThan4)
      {
         fourColors[0] = fourColors[0].darker(Color.fadeStep >> 2); // in 16 colors, make the buttons more 3d
         if (!enabled)
         {
            fourColors[3] = fourColors[2];
            fourColors[2] = fourColors[2].brighter(Color.smallerFadeStep);
         }
      }
      else
         fourColors[0] = fourColors[0].darker();
      if (enabled && !Settings.isColor && Settings.uiStyle == Settings.PalmOS)
         fourColors[2] = Color.BLACK; // guich@220_43
   }

   /** draws a 3d rect, defined by the R3D_xxx constants. Note that if oldStyle is true, the 3d rect turns back to the control style */
   public void draw3dRect(int x, int y, int width, int height, byte type, boolean yMirror, boolean simple, Color []fourColors)
   {
      if (type == R3D_SHADED)
      {
         boolean menu = simple; // is menu?
         drawLine(menu?0:1,0,menu?width-1:width-3,0);
         drawLine(0,1,0,height-3);
         drawLine(width-2,1,width-2,height-3);
         drawLine(width-1,menu?1:2,width-1,height-3);
         drawLine(1,height-2,width-2,height-2);
         drawLine(2,height-1,width-3,height-1);
      } else
      if (Settings.uiStyle == Settings.WinCE)
         switch (type)
         {
            case R3D_EDIT:
               drawHighLightFrame(x,y,width,height,fourColors[2],fourColors[1],yMirror);
               if (!simple) drawHighLightFrame(x+1,y+1,width-2,height-2,fourColors[3],fourColors[0],yMirror);
               break;
            case R3D_LOWERED:
               drawHighLightFrame(x,y,width,height,fourColors[3],fourColors[0],yMirror);
               if (!simple) drawHighLightFrame(x+1,y+1,width-2,height-2,fourColors[2],fourColors[1],yMirror);
               break;
            case R3D_RAISED:
               drawHighLightFrame(x,y,width,height,fourColors[0],fourColors[3],yMirror);
               if (!simple) drawHighLightFrame(x+1,y+1,width-2,height-2,fourColors[1],fourColors[2],yMirror);
               break;
            case R3D_CHECK:
               drawHighLightFrame(x,y,width,height,fourColors[2],fourColors[1],yMirror);
               if (!simple) drawHighLightFrame(x+1,y+1,width-2,height-2,fourColors[3],fourColors[0],yMirror);
               break;
         }
      else
      {
         setForeColor(fourColors[2]);
         switch (type)
         {
            case R3D_CHECK:
               drawRect(x,y,width,height);
               break;
            case R3D_EDIT:
               int h = height-1;
               drawDots(x,y+h,x+width-1,y+h);
               break;
            case R3D_LOWERED:
               setBackColor(fourColors[1]); // dont move it from here!
               if (simple)
                  fillRect(x,y,width,height);
               else
                  fillHatchedRect(x,y,width,height,true,true); // no break; here!
            case R3D_RAISED:
               if (simple)
                  drawRect(x,y,width,height);
               else
                  drawHatchedRect(x,y,width,height,true,true);
               break;
         }
      }
   }

   public String toString()
   {
      Rect r = new Rect();
      getClip(r);
      return super.toString()+", Clip:"+r+", translation from origin: "+getTranslation();
   }

   /** Draws an arrow. use the ARROW_xxx type. h is the height (or width,
     * if left/right arrows) in pixels. 1st line has 1 pixel, 2nd line has 3,
     * 3rd line has 5, etc. h must be >= 2 or no arrow is drawn.
     * The fore and back colors are changed after this method is called.
     * <p>Example:<pre>
     * public void onPaint(Graphics g)
     * {
     *    g.drawArrow(30,30,8, g.ARROW_LEFT, false,true, Color.BLACK );
     * ...
     * </pre>
     */
   public void drawArrow(int x, int y, int h, byte type, boolean pressed, boolean enabled, Color fore)
   {
      setForeColor(enabled?fore:fore.getCursorColor()); // guich@300_11: now use getCursorColor so it can behave correctly in bright foregrounds
      if (pressed) {x++; y++;}
      int step = 1;
      if (type == ARROW_RIGHT || type == ARROW_LEFT)
      {
         if (type == ARROW_LEFT) {x += h-1; step = -1;}
         h--;
         while (h >= 0)
         {
            drawLine(x,y,x,y+(h<<1));
            x += step;
            y++;
            h--;
         }
      }
      else
      {
         if (type == ARROW_UP) {y += h-1; step = -1;}
         h--;
         while (h >= 0)
         {
            drawLine(x,y,x+(h<<1),y);
            y += step;
            x++;
            h--;
         }
      }
   }

   /////////////////////////////////////////////////////////////////////////////////
   //                      I N T E R N A L      R O U T I N E S                  //
   /////////////////////////////////////////////////////////////////////////////////

   private void badParam(String s)
   {
      if (firstBadParam)
      {
         firstBadParam = false;
         Vm.debug("\n\nAttention: a bad parameter was found when calling a Graphics routine. Please debug your application to find the bug. If you don't do it, next time you run your program in the device, the VM will probably CRASH.\n");
      }
      try {throw new Exception("Bad parameter in "+s);} catch (Exception e) {e.printStackTrace();}
   }

   protected static void _applyPalette(int []palrgb, int startIndex, int endIndex)
   {
      // put the pixels that are being changed in a hashtable to speedup the process
      IntHashtable ht = new IntHashtable(300);
      for (int i = startIndex; i <= endIndex; i++)
         ht.put(currentPalette[i],palrgb[i]);
      // change all the pixels of the current image that are being affected by this palette change
      Graphics g = MainWindow.getMainWindow().createGraphics();
      int width = g.pixels[0].length;
      int height = g.pixels.length;
      int oldPt,newPt;
      for (int j = 0; j < height; j++)
         for (int i = 0; i < width; i++)
         {
            oldPt = g.getPixelNOCT(i,j);
            newPt = ht.get(oldPt);
            if (newPt != IntHashtable.INVALID)
               g.setPixelNOCT(i,j,newPt);
         }
      // and copy the new palette indexes to the current one.
      Vm.copyArray(palrgb, startIndex, currentPalette, startIndex, endIndex-startIndex+1);
      ScreenPalette = new Palette(currentPalette); // guich@310_27: update ScreenPalette
   }

   private Coord getDimension(ISurface surface)
   {
      if (surface instanceof waba.fx.Image)
      {
         Image img = (waba.fx.Image)surface;
         return new Coord(img.getWidth(),img.getHeight());
      }
      else // Control
      {
         Control c = (waba.ui.Control)surface;
         return new Coord(c.getPos().x,c.getPos().y);
      }
   }

   private boolean intersects(Rect r1, Rect r2)
   {
      return !((r1.x + r1.width <= r2.x) ||
      (r1.y + r1.height <= r2.y) ||
      (r1.x >= r2.x + r2.width) ||
      (r1.y >= r2.y + r2.height));
   }
   private Rect intersection(Rect r1, Rect r2)
   {
      int x1 = Math.max(r1.x, r2.x);
      int x2 = Math.min(r1.x + r1.width, r2.x + r2.width);
      int y1 = Math.max(r1.y, r2.y);
      int y2 = Math.min(r1.y + r1.height, r2.y + r2.height);
      return new Rect(x1, y1, x2 - x1, y2 - y1);
   }
   private void intersectWith(Rect r1, Rect r2)
   {
      int x1 = Math.max(r1.x, r2.x);
      int x2 = Math.min(r1.x + r1.width, r2.x + r2.width);
      int y1 = Math.max(r1.y, r2.y);
      int y2 = Math.min(r1.y + r1.height, r2.y + r2.height);
      r1.x = x1;
      r1.y = y1;
      r1.width  = x2 - x1;
      r1.height = y2 - y1;
   }

   private int [][]getPixels(ISurface surface)
   {
      if (surface == null)
      {
         badParam("getPixels("+surface+')');
         return null;
      }

      if (surface instanceof waba.fx.Image)
         return ((waba.fx.Image)surface).pixels;
      else
      if (!(surface instanceof waba.applet.Applet))
         if (surface instanceof waba.ui.Window && waba.applet.Applet.mainG == null) // initializing?
            return new int[Settings.screenHeight][Settings.screenWidth]; // guich@220_29
      return waba.applet.Applet.mainG.pixels;
   }
   ////////////////////   METHODS TAKED FROM THE SUPERWABA VIRTUAL MACHINE //////////
   private void drawScanLine(int []pixels, int x, int dstX, int dstY, int width)
   {
      if (pixels == null || x < 0 || dstX < 0 || dstY < 0 || width <= 0)
         badParam("drawScanLine("+pixels+','+x+','+dstX+','+dstY+','+width+')');
      int lastColor = pixels[x];
      int lastI=0, i=1;
      for (; i < width; i++)
         if (pixels[x+i] != lastColor)
         {
            drawHLineNOCT(dstX+lastI,dstY,i-lastI,lastColor);
            lastI = i;
            lastColor = pixels[x+i];
         }
      drawHLineNOCT(dstX+lastI,dstY,i-lastI,lastColor);
   }
   // copy the area x,y,width,height of the bitmap bmp with dimensions bmpW,bmpH to the (current active) screen location dstX,dstY
   private void drawBitmap(int [][]pixels, int x, int y, int width, int height, int dstX, int dstY, int drawOp, int backColor, int foreColor, boolean doClip)
   {
      int originalWidth = width;
      int originalHeight = height;
      try // guich@200b4_63
      {
         int bmpPt,screenPt;
         Rect r1=new Rect(), r2=new Rect();
         int bmpW = pixels[0].length;
         int bmpH = pixels.length;
         int i,j,k;

         // Note: any change in this routine requires test in the Commander, ScrollImage, TestClip, and popup Window panning

         if (doClip) // guich@340_14
         {
            // guich@200b4_65: remade this drawBitmap routine
            // 1. make sure that the area of the bitmap that we want to copy IS inside its area
            r1.set(x,y,width,height);
            r2.set(0,0,bmpW,bmpH);
            if (!intersects(r1,r2)) return;

            // 2 test if we're panning a bitmap - guich@240_1
            boolean panX = bmpW > CLIPX2 && dstX <= CLIPX1;
            boolean panY = bmpH > CLIPY2 && dstY <= CLIPY1;
            if (panX || panY)
            {
               if (panX)
               {
                  x -= dstX;
                  dstX = 0;
                  width = Math.min(CLIPX2-CLIPX1,bmpW-x);
               }
               if (panY)
               {
                  y -= dstY;
                  dstY = 0;
                  height = Math.min(CLIPY2-CLIPY1,bmpH-y);
               }
            }
            // 3. translate the dstXY
            dstX += TRANSX;
            dstY += TRANSY;

            // 4. if not panning, intersect the source bitmap with the screen (Window saveBehind)
            if ((!panX && !panY) || x < 0 || y < 0 // guich@300_15: if moving a popup window to the left of the screen, it gets here still with x < 0 (or y < 0).
                || (x+width) > MaxScreenX || (y+height) > MaxScreenY) // guich@341_8: fixed problem when a MessageBox has width > than screenWidth, and it was dragged to the right and the saveBehind method would make this fail bc it was trying to move more pixels at the right than possible
            {
               if (dstX == 0 && dstY == 0 // note: without this test, if you create a bitmap greater than the screen widths, it will not be displayed correctly.
                 && (x < 0 || y < 0 || (dstX+width) > MaxScreenX || (dstY+height) > MaxScreenY)) // guich@350_27: without this test, if you create a bitmap greater than the screen widths, but the destination still fits, it will be truncated. first 2 are to avoid problems when dragging a window to the left.
               {
                  r2.set(0,0,MaxScreenX,MaxScreenY); // r1 already set
                  intersectWith(r1,r2);

                  k = r1.width - width;
                  if (k < 0 && r1.x == r2.x)
                     dstX -= k;
                  k = r1.height - height;
                  if (k < 0 && r1.y == r2.y)
                     dstY -= k;

                  x = r1.x;
                  y = r1.y;
                  width = r1.width;
                  height = r1.height;
               }
            }
            // 5. intersect the destination bitmap with the clip area (Commander/Sprite)
            r1.set(dstX,dstY,width,height);
            r2.set(CLIPX1,CLIPY1,CLIPX2-CLIPX1,CLIPY2-CLIPY1);
            intersectWith(r1,r2);

            if (r1.width <= 0 || r1.height <= 0) // nothing to draw?
               return;

            k = r1.width - width;
            if (!panX && k < 0 && r1.x == r2.x)
               x -= k;
            k = r1.height - height;
            if (!panY && k < 0 && r1.y == r2.y)
               y -= k;

            dstX   = r1.x;
            dstY   = r1.y;
            width  = r1.width;
            height = r1.height;
            if (originalWidth < width) // guich@341_15: fix a problem when drawing sub-images of an animation image
               width = originalWidth;
            if (originalHeight < height)
               height = originalHeight;
         }
         else
         {
            dstX += TRANSX;
            dstY += TRANSY;
         }

         //Vm.debug("drawBitmap("+x+','+y+','+width+','+height+" -> "+dstX+','+dstY+")");
         //Vm.debug("clip: "+CLIPX1+","+CLIPY1+","+CLIPX2+","+CLIPY2+" - trans: "+TRANSX+","+TRANSY);

         if (drawOp == DRAW_PAINT)
            for (j=0; j < height; j++)
               drawScanLine(pixels[y+j], x, dstX,dstY+j,width);
         else
         if (drawOp == DRAW_ERASE   || drawOp == DRAW_MASK
          || drawOp == DRAW_INVERT
          || drawOp == DRAW_OVERLAY || drawOp == DRAW_PAINT_INVERSE) // guich@230_20
         {
            // index manipulation
            int sub = Settings.maxColors-1;
            for (j=0; j < height; j++)
            {
               for (i=0; i < width; i++)
               {
                  screenPt = getPixelNOCT(dstX+i,dstY+j);
                  bmpPt = pixels[y+j][x+i];
                  // guich@230_20: convert to screen index
                  screenPt = sub - new Color(screenPt).getAlpha();
                  if (sub >= 255)
                     bmpPt = sub - new Color(bmpPt).getAlpha();
                  //
                  switch (drawOp)
                  {
                     case Graphics.DRAW_ERASE:        screenPt &=  bmpPt; break;
                     case Graphics.DRAW_MASK:         screenPt &= ~bmpPt; break;
                     case Graphics.DRAW_INVERT:       screenPt ^=  bmpPt; break;
                     case Graphics.DRAW_OVERLAY:
                        screenPt |=  bmpPt;
                        break;
                     case Graphics.DRAW_PAINT_INVERSE: screenPt  = ~bmpPt; break;
                  }
                  screenPt = Palette.SystemPalette.getColor(screenPt).rgb; // guich@230_20
                  setPixelNOCT(dstX+i,dstY+j,screenPt); // note: there is no performance gain by storing the results in a array and calling drawScanLine instead.
               }
            }
         }
         else
         {
            // color manipulation
            for (j=0; j < height; j++)
               for (i=0; i < width; i++)
               {
                  screenPt = getPixelNOCT(dstX+i,dstY+j);
                  bmpPt = pixels[y+j][x+i];
                  switch (drawOp)
                  {
                     case Graphics.DRAW_SPRITE:        if (bmpPt != backColor) screenPt = bmpPt;     break;
                     case Graphics.DRAW_REPLACE_COLOR: if (bmpPt != backColor) screenPt = foreColor; break;
                     case Graphics.DRAW_SWAP_COLORS:   if (bmpPt == backColor) screenPt = foreColor;
                                                       else
                                                       if (bmpPt == foreColor) screenPt = backColor; break;
                  }
                  setPixelNOCT(dstX+i,dstY+j,screenPt); // note: there is no performance gain by storing the results in a array and calling drawScanLine instead.
               }
         }
      }
      catch (Exception e)
      {
         Vm.debug("Exception in drawBitmap");
         Vm.debug("drawBitmap("+x+','+y+','+width+','+height+" -> "+dstX+','+dstY+','+doClip+")");
         Vm.debug("clip: "+CLIPX1+","+CLIPY1+","+CLIPX2+","+CLIPY2+" - trans: "+TRANSX+","+TRANSY);
         e.printStackTrace();
      }
   }
   ////////////////////////////////////////////////////////////////////////////
   // taked from gd 2.0.1 beta (http://www.boutell.com/gd/manual2.0.html) from Thomas Boutell
   private void fillPolygon(int []xPoints, int []yPoints, int nPoints, int c)
   {
      if (xPoints == null || yPoints == null || nPoints <= 0)
         badParam("fillPolygon("+xPoints+','+yPoints+','+nPoints+')');
      waba.util.IntVector ints = new waba.util.IntVector(10);
      int x1, y1, x2, y2;
      int ind1, ind2;
      if (nPoints < 2)
         return;
      int miny = yPoints[0];
      int maxy = yPoints[0];
      for (int i = 1; i < nPoints; i++)
      {
         if (yPoints[i] < miny)
            miny = yPoints[i];
         if (yPoints[i] > maxy)
            maxy = yPoints[i];
      }
      for (int y = miny; y <= maxy; y++)
      {
         ints.clear();
         for (int i = 0; i < nPoints; i++)
         {
            if (i == 0)
            {
               ind1 = nPoints - 1;
               ind2 = 0;
            }
            else
            {
               ind1 = i - 1;
               ind2 = i;
            }
            y1 = yPoints[ind1];
            y2 = yPoints[ind2];
            if (y1 < y2)
            {
               x1 = xPoints[ind1];
               x2 = xPoints[ind2];
            }
            else
            if (y1 > y2)
            {
               y2 = yPoints[ind1];
               y1 = yPoints[ind2];
               x2 = xPoints[ind1];
               x1 = xPoints[ind2];
            }
            else continue;
            // compute next x point
            if ( (y1 <= y && y < y2) || (y == maxy && y1 < y && y <= y2) )
               ints.add((y - y1) * (x2 - x1) / (y2 - y1) + x1);
         }
         // sort the ints
         ints.qsort();
         // draw the lines
         for (int i = 0; i < ints.getCount(); i += 2)
            drawHLine(ints.items[i],y,ints.items[i+1]-ints.items[i]+1,c);
      }
   }
   ////////////////////////////////////////////////////////////////////////////
   // draws a polygon. if the polygon is not closed, close it
   private void drawPolygon(int []xPoints, int []yPoints, int nPoints, int c, boolean close)
   {
      int i;
      if (nPoints < 2) return;
      if (xPoints == null || yPoints == null)
         badParam("drawPolygon("+xPoints+','+yPoints+')');
      for (i=1; i < nPoints; i++)
         drawLine(xPoints[i-1],yPoints[i-1],xPoints[i],yPoints[i],c);
      if (close)
      {
         nPoints--;
         if (xPoints[0] != xPoints[nPoints] || yPoints[0] != yPoints[nPoints])
            drawLine(xPoints[0],yPoints[0],xPoints[nPoints],yPoints[nPoints],c);
      }
   }
   ////////////////////////////////////////////////////////////////////////////
   // gets the bit value of the pixel, but does not translates or clip rect
   private int getPixelNOCT (int x, int y)
   {
      if (x < 0 || y < 0)
         badParam("getPixelNOCT("+x+','+y+')');
      return pixels[y][x];
   }
   ////////////////////////////////////////////////////////////////////////////
   // inverts the pixel, but does not translates or clip rect
   private void invPixelNOCT (int x, int y)
   {
      if (x < 0 || y < 0)
         badParam("invPixelNOCT("+x+','+y+')');
      setPixelNOCT(x,y,~pixels[y][x]);
   }
   ////////////////////////////////////////////////////////////////////////////
   // inverts the vertical line, but does not translates or clip rect
   private void invVLineNOCT (int x, int y, int h)
   {
      if (x < 0 || y < 0 || h < 0)
         badParam("invVLineNOCT("+x+','+y+','+h+')');
      while (h-- > 0)
      {
         setPixelNOCT(x,y,~pixels[y][x]);
         y++;
      }
   }
   ////////////////////////////////////////////////////////////////////////////
   // erase one color to the other, but does not translates or clip rect
   private void eraseVLineNOCT (int x, int y, int h, int cfrom, int cto)
   {
      if (x < 0 || y < 0 || h < 0)
         badParam("eraseVLineNOCT("+x+','+y+','+h+')');
      while (h-- > 0)
      {
         if (pixels[y][x] == cfrom)
            setPixelNOCT(x,y,cto);
         y++;
      }
   }
   ////////////////////////////////////////////////////////////////////////////
   // sets the bit value, but does not translates or clip rect
   private void setPixelNOCT (int x, int y, int color)
   {
      if (x < 0 || y < 0)
         badParam("setPixelNOCT("+x+','+y+')');
      JavaBridge.GraphicsLine(surface,x,y,x,y,color);
      pixels[y][x] = color; // keep our internal status of all pixels
   }
   ////////////////////////////////////////////////////////////////////////////
   // sets the bit value
   private void setPixel(int x, int y, int color)
   {
      x += TRANSX;
      y += TRANSY;
      if (CLIPX1 <= x && x < CLIPX2 && CLIPY1 <= y && y < CLIPY2)
      {
         if (x < 0 || y < 0)
            badParam("setPixel("+x+','+y+')');
         JavaBridge.GraphicsLine(surface,x,y,x,y,color);
         pixels[y][x] = color; // keep our internal status of all pixels
      }
   }
   ////////////////////////////////////////////////////////////////////////////
   // draws a line from x to x+w-1, with no clip or translate
   private void drawHLineNOCT(int x, int y, int w, int color)
   {
      if (x < 0 || y < 0 || w < 0)
         badParam("drawHLineNOCT("+x+','+y+','+w+')');
      JavaBridge.GraphicsLine(surface,x,y,x+w-1,y,color);
      while (w-- > 0)
         pixels[y][x++] = color;
   }
   ////////////////////////////////////////////////////////////////////////////
   // draws a line from x to x+w-1
   private void drawHLine (int x, int y, int w, int color)
   {
      x += TRANSX;
      y += TRANSY;
      if (CLIPY1 <= y && y < CLIPY2 && CLIPX1 <= (x+w) && x < CLIPX2) // line must lie inside y clip bounds, must not end before clip x1 and must not start after clip x2
      {
         if (x < CLIPX1) {w -= CLIPX1-x; x = CLIPX1;} // line start before clip x1
         if ((x+w) > CLIPX2) w = CLIPX2-x; // line stops after clip x2
         if (x < 0 || y < 0 || w < 0)
            badParam("drawHLine("+x+','+y+','+w+')');
         JavaBridge.GraphicsLine(surface,x,y,x+w-1,y,color);
         while (w-- > 0)
            pixels[y][x++] = color;
      }
   }
   ////////////////////////////////////////////////////////////////////////////
   // draws a line from y to y+h-1
   private void drawVLine (int x, int y, int h, int color)
   {
      x += TRANSX;
      y += TRANSY;
      if (CLIPX1 <= x && x < CLIPX2 && CLIPY1 <= (y+h) && y < CLIPY2) // line must lie inside x clip bounds, must not end before clip y1 and must not start after clip y2
      {
         if (y < CLIPY1) {h -= CLIPY1-y; y = CLIPY1;} // line start before clip y1
         if ((y+h) > CLIPY2) h = CLIPY2-y; // line stops after clip y2
         if (x < 0 || y < 0 || h < 0)
            badParam("drawVLine("+x+','+y+','+h+')');
         JavaBridge.GraphicsLine(surface,x,y,x,y+h-1,color);
         while (h-- > 0)
            pixels[y++][x] = color;
      }
   }
   ////////////////////////////////////////////////////////////////////////////
   // Bresenham algorithm to draw lines (modified by guich to improve performance in vertical and horizontal lines)
   private void drawLine(int Ax, int Ay, int Bx, int By, int c)
   {
      // INITIALIZE THE COMPONENTS OF THE ALGORITHM THAT ARE NOT AFFECTED BY THE SLOPE OR DIRECTION OF THE LINE
      int dX = Math.abs(Bx-Ax);    // store the change in X and Y of the line endpoints
      int dY = Math.abs(By-Ay);

      int CurrentX = Ax;      // store the starting point (just point A)
      int CurrentY = Ay;

      // DETERMINE "DIRECTIONS" TO INCREMENT X AND Y (REGARDLESS OF DECISION)
      int Xincr, Yincr;

      // guich: if its a pixel, draw only the pixel
      if (dX == 0 && dY == 0)
         setPixel(Ax,Ay,c); else
      // guich: if they are vertical or horizontal lines, use specific methods
      if (dY == 0) // horizontal line?
         drawHLine(Math.min(Ax,Bx),Math.min(Ay,By),dX+1,c); else
      if (dX == 0) // vertical line?
         drawVLine(Math.min(Ax,Bx),Math.min(Ay,By),dY+1,c);
      else // diagonal
      {
         if (Ax > Bx) { Xincr=-1; } else { Xincr=1; }    // which direction in X?
         if (Ay > By) { Yincr=-1; } else { Yincr=1; }    // which direction in Y?

         // DETERMINE INDEPENDENT VARIABLE (ONE THAT ALWAYS INCREMENTS BY 1 (OR -1) )
         // AND INITIATE APPROPRIATE LINE DRAWING ROUTINE (BASED ON FIRST OCTANT
         // ALWAYS). THE X AND Y'S MAY BE FLIPPED IF Y IS THE INDEPENDENT VARIABLE.
         if (dX >= dY)   // if X is the independent variable
         {
               int dPr     = dY<<1;                        // amount to increment decision if right is chosen (always)
               int dPru    = dPr - (dX<<1);                // amount to increment decision if up is chosen
               int P       = dPr - dX;                     // decision variable start value

               for (; dX>=0; dX--)                         // process each point in the line one at a time (just use dX)
               {
                  setPixel(CurrentX, CurrentY, c);        // plot the pixel
                  CurrentX+=Xincr;                        // increment independent variable
                  if (P > 0)                              // is the pixel going right AND up?
                  {
                     CurrentY+=Yincr;                    // increment dependent variable
                     P+=dPru;                            // increment decision (for up)
                  }
                  else                                    // is the pixel just going right?
                     P+=dPr;                             // increment decision (for right)
               }
         }
         else            // if Y is the independent variable
         {
               int dPr     = dX<<1;                        // amount to increment decision if right is chosen (always)
               int dPru    = dPr - (dY<<1);                    // amount to increment decision if up is chosen
               int P       = dPr - dY;                     // decision variable start value

               for (; dY>=0; dY--)                         // process each point in the line one at a time (just use dY)
               {
                  setPixel(CurrentX, CurrentY, c);        // plot the pixel
                  CurrentY+=Yincr;                        // increment independent variable
                  if (P > 0)                              // is the pixel going up AND right?
                  {
                     CurrentX+=Xincr;                    // increment dependent variable
                     P+=dPru;                            // increment decision (for up)
                  }
                  else                                    // is the pixel just going up?
                     P+=dPr;                             // increment decision (for right)
               }
         }
      }
   }
   ////////////////////////////////////////////////////////////////////////////
   // Bresenham algorithm to draw lines. draws one pixel with color c1 and one pixel with color c2
   private void drawDottedLine(int Ax, int Ay, int Bx, int By, int c1, int c2)
   {
      boolean on=true;
      // INITIALIZE THE COMPONENTS OF THE ALGORITHM THAT ARE NOT AFFECTED BY THE SLOPE OR DIRECTION OF THE LINE
      int dX = Math.abs(Bx-Ax);    // store the change in X and Y of the line endpoints
      int dY = Math.abs(By-Ay);

      int CurrentX = Ax;      // store the starting point (just point A)
      int CurrentY = Ay;

      // DETERMINE "DIRECTIONS" TO INCREMENT X AND Y (REGARDLESS OF DECISION)
      int Xincr, Yincr;

      if (Ax > Bx) { Xincr=-1; } else { Xincr=1; }    // which direction in X?
      if (Ay > By) { Yincr=-1; } else { Yincr=1; }    // which direction in Y?

      // DETERMINE INDEPENDENT VARIABLE (ONE THAT ALWAYS INCREMENTS BY 1 (OR -1) )
      // AND INITIATE APPROPRIATE LINE DRAWING ROUTINE (BASED ON FIRST OCTANT
      // ALWAYS). THE X AND Y'S MAY BE FLIPPED IF Y IS THE INDEPENDENT VARIABLE.
      if (dX >= dY)   // if X is the independent variable
      {
         int dPr     = dY<<1;                        // amount to increment decision if right is chosen (always)
         int dPru    = dPr - (dX<<1);                // amount to increment decision if up is chosen
         int P       = dPr - dX;                     // decision variable start value

         for (; dX>=0; dX--)                         // process each point in the line one at a time (just use dX)
         {
               setPixel(CurrentX, CurrentY, on?c1:c2); // plot the pixel
               on = !on;
               CurrentX+=Xincr;                        // increment independent variable
               if (P > 0)                              // is the pixel going right AND up?
               {
                  CurrentY+=Yincr;                    // increment dependent variable
                  P+=dPru;                            // increment decision (for up)
               }
               else                                    // is the pixel just going right?
                  P+=dPr;                             // increment decision (for right)
         }
      }
      else            // if Y is the independent variable
      {
         int dPr     = dX<<1;                        // amount to increment decision if right is chosen (always)
         int dPru    = dPr - (dY<<1);                    // amount to increment decision if up is chosen
         int P       = dPr - dY;                     // decision variable start value

         for (; dY>=0; dY--)                         // process each point in the line one at a time (just use dY)
         {
               setPixel(CurrentX, CurrentY, on?c1:c2); // plot the pixel
               on = !on;
               CurrentY+=Yincr;                        // increment independent variable
               if (P > 0)                              // is the pixel going up AND right?
               {
                  CurrentX+=Xincr;                    // increment dependent variable
                  P+=dPru;                            // increment decision (for up)
               }
               else                                    // is the pixel just going up?
                  P+=dPr;                             // increment decision (for right)
         }
      }
   }
   // returns 1 if the coords intersects the clip rectangle
   private boolean intersectsClipRect(int x1,int y1,int x2,int y2)
   {
      return !((x2 <= CLIPX1) || (y2 <= CLIPY1) || (x1 >= CLIPX2) || (y1 >= CLIPY2));
   }
   // transltates the given coords and returns the intersection between the clip rect and the coords passed. returns 0 if the coords are outside the clip rect
   private boolean translateAndClip(int []xywh)
   {
      int x = xywh[0], y = xywh[1], width = xywh[2], height = xywh[3];
      x += TRANSX;
      y += TRANSY;
      if (!intersectsClipRect(x,y,x+width,y+height)) return false;
      if (x < CLIPX1) {width  -= CLIPX1-x; x = CLIPX1;}
      if (y < CLIPY1) {height -= CLIPY1-y; y = CLIPY1;}
      if ((x+width)  > CLIPX2) width  = CLIPX2 - x;
      if ((y+height) > CLIPY2) height = CLIPY2 - y;
      xywh[0] = x;
      xywh[1] = y;
      xywh[2] = width;
      xywh[3] = height;
      return true;
   }
   private void quadPixel(int xc, int yc, int x, int y, int c)
   {
      // draw 4 points using symmetry
      setPixel(xc + x, yc + y, c);
      setPixel(xc + x, yc - y, c);
      setPixel(xc - x, yc + y, c);
      setPixel(xc - x, yc - y, c);
   }
   private void quadLine(int xc, int yc, int x, int y, int c)
   {
      int w = x+x+1; // plus 1 for the drawHLine (draws to width-1)
      // draw 2 lines using symmetry
      drawHLine(xc - x, yc - y, w, c);
      drawHLine(xc - x, yc + y, w, c);
   }
   // draws an ellipse incrementally
   private void ellipseDrawAndFill(int xc, int yc, int rx, int ry, int c, boolean fill)
   {
      if (xc < 0 || yc < 0 || rx < 0 || ry < 0)
         badParam("ellipseDrawAndFill("+xc+','+yc+','+rx+','+ry+')');
      // intermediate terms to speed up loop
      int t1 = rx*rx, t2 = t1<<1, t3 = t2<<1;
      int t4 = ry*ry, t5 = t4<<1, t6 = t5<<1;
      int t7 = rx*t5, t8 = t7<<1, t9 = 0;
      int d1 = t2 - t7 + (t4>>1);    // error terms
      int d2 = (t1>>1) - t8 + t5;
      int x = rx, y = 0; // ellipse points

      while (d2 < 0)          // til slope = -1
      {
         if (fill)
               quadLine(xc,yc,x,y,c);
         else
               quadPixel(xc,yc,x,y,c);
         y++;        // always move up here
         t9 += t3;
         if (d1 < 0) // move straight up
         {
               d1 += t9 + t2;
               d2 += t9;
         }
         else        // move up and left
         {
               x--;
               t8 -= t6;
               d1 += t9 + t2 - t8;
               d2 += t9 + t5 - t8;
         }
      }

      do              // rest of top right quadrant
      {
         // draw 4 points using symmetry
         if (fill)
               quadLine(xc,yc,x,y,c);
         else
               quadPixel(xc,yc,x,y,c);
         x--;        // always move left here
         t8 -= t6;
         if (d2 < 0) // move up and left
         {
               y++;
               t9 += t3;
               d2 += t9 + t5 - t8;
         }
         else        // move straight left
               d2 += t5 - t8;
      } while (x >= 0);
   }
   private void preArcPiePointDrawAndFill(int xc, int yc, int rx, int ry, float startAngle, float endAngle, int c, int c2, boolean fill, boolean pie)
   {
      // make sure the values are -359 <= x <= 359
      while (startAngle <= -360) startAngle += 360;
      while (endAngle   <= -360) endAngle   += 360;
      while (startAngle >   360) startAngle -= 360;
      while (endAngle   >   360) endAngle   -= 360;

      if (startAngle > endAngle) // eg 235 to 45
         startAngle -= 360; // set to -45 to 45 so we can handle it correctly
      if (startAngle >= 0 && endAngle <= 0) // eg 135 to -135
         endAngle += 360; // set to 135 to 225

      if (startAngle >= 0 && endAngle >= 0)
         arcPiePointDrawAndFill(xc, yc, rx, ry, startAngle, endAngle, c, c2, fill, pie, null);
      else
      if (startAngle <= 0 && endAngle >= 0) // eg -45 to 45
      {
         startAngle += 360;
         arcPiePointDrawAndFill(xc, yc, rx, ry, startAngle, 360, c, c2, fill, pie, null);
         arcPiePointDrawAndFill(xc, yc, rx, ry, 0, endAngle, c, c2, fill, pie, null);
      } else badParam("preArcPiePointDrawAndFill("+xc+','+yc+','+rx+','+ry+','+startAngle+','+endAngle+')');
   }
   ////////////////////////////////////////////////////////////////////////////
   private int doClipX(int x)
   {
      return ((x+TRANSX) < CLIPX1) ? (CLIPX1-TRANSX) : ((x+TRANSX) >= CLIPX2) ? (CLIPX2-TRANSX) : x; // guich@401_2: added TRANSX/Y
   }
   private int doClipY(int y)
   {
      return ((y+TRANSY) < CLIPY1) ? (CLIPY1-TRANSY) : ((y+TRANSY) >= CLIPY2) ? (CLIPY2-TRANSY) : y;
   }
   // draw an elliptical arc from startAngle to endAngle. c is the fill color and c2 is the outline color (if in fill mode - otherwise, c = outline color)
   private void arcPiePointDrawAndFill(int xc, int yc, int rx, int ry, float startAngle, float endAngle, int c, int c2, boolean fill, boolean pie, int []startAnglePoint)
   {
      int []xPoints=null;
      int []yPoints=null;
      int nPoints=0;
      int lastRX=-1,lastRY=-1,lastXC = -1, lastYC = -1,lastSize=0;
      float lastPPD=0;
      // this algorithm was created by Guilherme Campos Hazan
      float ppd;
      int startIndex,endIndex,index,i;
      int nq,size=0;
      int oldX=0,oldY=0;
      boolean checkClipX = (xc+rx+TRANSX) > CLIPX2 || (xc-rx+TRANSX) < CLIPX1; // guich@340_3 - guich@401_2: added TRANSX/Y
      boolean checkClipY = (yc+ry+TRANSY) > CLIPY2 || (yc-ry+TRANSY) < CLIPY1;
      Vm.debug(checkClipX+","+checkClipY);
      // step 0: correct angle values
      if (startAngle < 0.1 && endAngle > 359.9) // full circle? use the fastest routine instead
      {
         if (fill)
            ellipseDrawAndFill(xc,yc,rx,ry,c2,true); // guich@201_2: corrected colors. - guich@401_3: changed c->c2 and vice versa (line below)
         ellipseDrawAndFill(xc,yc,rx,ry,c,false);
         return;
      }
      if (xc < 0 || yc < 0 || rx < 0 || ry < 0)
         badParam("arcPiePointDrawAndFill("+xc+','+yc+','+rx+','+ry+')');
      // step 0: if possible, use cached results
      if (true)//xc != lastXC || yc != lastYC || rx != lastRX || ry != lastRY)
      {
         // step 1: computes how many points the circle has (computes only 45 degrees and mirrors the rest)
         // intermediate terms to speed up loop
         int t1 = rx*rx, t2 = t1<<1, t3 = t2<<1;
         int t4 = ry*ry, t5 = t4<<1, t6 = t5<<1;
         int t7 = rx*t5, t8 = t7<<1, t9 = 0;
         int d1 = t2 - t7 + (t4>>1);    // error terms
         int d2 = (t1>>1) - t8 + t5;

         int x = rx, y = 0; // ellipse points

         while (d2 < 0)          // til slope = -1
         {
            y++;        // always move up here
            t9 += t3;
            if (d1 < 0) // move straight up
            {
               d1 += t9 + t2;
               d2 += t9;
            }
            else        // move up and left
            {
               x--;
               t8 -= t6;
               d1 += t9 + t2 - t8;
               d2 += t9 + t5 - t8;
            }
            size++;
         }

         do              // rest of top right quadrant
         {
            x--;        // always move left here
            t8 -= t6;
            if (d2 < 0) // move up and left
            {
               y++;
               t9 += t3;
               d2 += t9 + t5 - t8;
            }
            else        // move straight left
               d2 += t5 - t8;
            size++;

         } while (x >= 0);
         nq = size;
         size *= 4;
         // step 2: computes how many points per degree
         ppd = (float)size / 360.0f;
         // step 3: create space in the buffer so it can save all the circle
         size+=2;
         if (nPoints < size)
         {
            /*if (xPoints) in java we get GC'd
            {
               xfree(xPoints);
               xfree(yPoints);
            }*/
            xPoints = new int[size];
            yPoints = new int[size];
            if (xPoints == null || yPoints == null)
               return;
         }
         nPoints = size;
         // step 4: stores all the circle in the array. the odd arcs are drawn in reverse order
         // intermediate terms to speed up loop
         t2 = t1<<1; t3 = t2<<1;
         t8 = t7<<1; t9 = 0;
         d1 = t2 - t7 + (t4>>1); // error terms
         d2 = (t1>>1) - t8 + t5;

         x = rx;
         y = 0;  // ellipse points

         i=0;
         while (d2 < 0)          // til slope = -1
         {
            // save 4 points using symmetry
            // guich@340_3: added clipping
            index = nq*0+i;      // 0/3
            xPoints[index]=checkClipX?doClipX(xc+x):(xc+x);
            yPoints[index]=checkClipY?doClipY(yc-y):(yc-y);

            index = nq*2-i-1;    // 1/3
            xPoints[index]=checkClipX?doClipX(xc-x):(xc-x);
            yPoints[index]=checkClipY?doClipY(yc-y):(yc-y);

            index = nq*2+i;      // 2/3
            xPoints[index]=checkClipX?doClipX(xc-x):(xc-x);
            yPoints[index]=checkClipY?doClipY(yc+y):(yc+y);

            index = nq*4-i-1;    // 3/3
            xPoints[index]=checkClipX?doClipX(xc+x):(xc+x);
            yPoints[index]=checkClipY?doClipY(yc+y):(yc+y);
            i++;

            y++;        // always move up here
            t9 += t3;
            if (d1 < 0) // move straight up
            {
               d1 += t9 + t2;
               d2 += t9;
            }
            else        // move up and left
            {
               x--;
               t8 -= t6;
               d1 += t9 + t2 - t8;
               d2 += t9 + t5 - t8;
            }
         }

         do              // rest of top right quadrant
         {
            // save 4 points using symmetry
            // guich@340_3: added clipping
            index = nq*0+i;    // 0/3
            xPoints[index]=checkClipX?doClipX(xc+x):(xc+x);
            yPoints[index]=checkClipY?doClipY(yc-y):(yc-y);

            index = nq*2-i-1;  // 1/3
            xPoints[index]=checkClipX?doClipX(xc-x):(xc-x);
            yPoints[index]=checkClipY?doClipY(yc-y):(yc-y);

            index = nq*2+i;    // 2/3
            xPoints[index]=checkClipX?doClipX(xc-x):(xc-x);
            yPoints[index]=checkClipY?doClipY(yc+y):(yc+y);

            index = nq*4-i-1;  // 3/3
            xPoints[index]=checkClipX?doClipX(xc+x):(xc+x);
            yPoints[index]=checkClipY?doClipY(yc+y):(yc+y);
            i++;

            x--;        // always move left here
            t8 -= t6;
            if (d2 < 0) // move up and left
            {
               y++;
               t9 += t3;
               d2 += t9 + t5 - t8;
            }
            else        // move straight left
               d2 += t5 - t8;
         } while (x >= 0);
         // save last arguments
         lastXC = xc;
         lastYC = yc;
         lastRX = rx;
         lastRY = ry;
         lastPPD = ppd;
         lastSize = size;
      }
      else ppd = lastPPD;
      // step 5: computes the start and end indexes that will become part of the arc
      startIndex = (int)(ppd * startAngle);
      endIndex = (int)(ppd * endAngle);
      // step 5 1/2: if only computing the point, return it
      if (startAnglePoint != null)
      {
         startAnglePoint[0] = xPoints[startIndex];
         startAnglePoint[1] = yPoints[startIndex];
         return;
      }
      if (endIndex == (lastSize-2)) // 360?
         endIndex--;
      // step 6: fill or draw the polygons
      endIndex++;
      if (pie)
      {
         oldX = xPoints[endIndex];
         oldY = yPoints[endIndex];
         xPoints[endIndex] = xc;
         yPoints[endIndex] = yc;
         endIndex++;
      }
      int n = endIndex-startIndex;
      int []newX = new int[n+1];
      int []newY = new int[n+1];
      Vm.copyArray(xPoints,startIndex,newX,0,n);
      Vm.copyArray(yPoints,startIndex,newY,0,n);
      if (fill)
         fillPolygon(newX,newY,n,c2); // guich@401_3: changed c to c2
      //if (!fill || c != c2) always draw border
         drawPolygon(newX,newY,n,fill?c2:c,fill);
      if (pie) // restore saved points
      {
         xPoints[endIndex-1] = oldX;
         yPoints[endIndex-1] = oldY;
      }
   }
}
