/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>                         *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.fx;

import waba.io.*;
import waba.sys.*;

/**
 * Image is a rectangular image.
 * <p>
 * You can draw into an image and copy an image to a surface using a Graphics
 * object. Before a image is drawed, its palette is set.
 * @see Graphics
 */
public class Image implements ISurface
{
   // note: if any other variable is added here, the WOBJ_Imagexxxx defines must be redefined in superwaba.h
   protected int width;
   protected int height;
   /** Sets the transparent color of this image. Used in many controls.
     * Note that you can set it to null to make the image use the background
     * color.
     * Note also that this value is not used by the system, but the controls
     * (E.g.: waba.ui.Button)
     * use the value stored here to set the background color when going to
     * draw a transparent image.
     * <b>Important!</b>: If your image doesnt use the web safe palette,
     * you <b>must</b> call <code>myImage.applyPalette</code> before calling
     * setTransparentColor (or before the first time the image is drawn),
     * otherwise the color may not be found in the active palette.
     */
   public Color transparentColor = Color.WHITE; // default if the bitmap is monochromatic.
   protected int[] palRGBs;

   // HOOK VARIABLES - can't be private - guich@350_20

   /** On desktop, contains the pixels of this image. On device, holds the GrHandle structure */
   protected int pixels[][];
   /** Used on desktop only to match the colors that are being saved with the colors that are being getted when saving a bitmap */
   protected Palette savingPal;

   protected Image()
   {
   }

   /**
   * Creates an image of the specified width and height. The image has
   * a color depth (number of bitplanes) and color map that matches the
   * default drawing surface.
   * Here is an example of use:
   * <pre>
   * Image img = new Image(100,100);
   * Graphics g = img.getGraphics();
   * g.setBackColor(Color.WHITE);
   * g.fillRect(25,25,50,50);
   * ...
   * Graphics screenG = createGraphics();
   * screenG.drawImage(img,CENTER,CENTER);
   * </pre>
   */
   public Image(int width, int height)
   {
      this.width = width;
      this.height = height;
      NativeMethods4JDK.imageCreate(this);
   }

   /**
   * Loads and constructs an image from a file. The path given is the path to the
   * image file. The file must be in 2, 16 or 256 color uncompressed BMP bitmap
   * format. If the image cannot be loaded, the image constructed will have
   * a width and height of 0.
   */
   public Image(String path)
   {
      NativeMethods4JDK.imageLoad(this,path);
   }

   /** Parses an image from the given byte array. Note that the byte array must
     * specify the full bitmap, with headers, palette and the data. The bitmap
     * can be in compressed rle4 or rle8.
     * Here is a code example: <pre>
     * // create the image and fill it with something
     * Image img = new Image(160,160);
     * Graphics g = img.getGraphics();
     * for (int i =0; i < 16; i++)
     * {
     *    g.setBackColor(new Color(10*i,10*i,10*i));
     *    g.fillRect(i*10,0,10,160);
     * }
     * // save the bmp in a byte stream
     * ByteArrayStream bas = new ByteArrayStream(4096);
     * DataStream ds = new DataStream(bas);
     * int totalBytesWritten = img.createBmp(ds);
     * // parse the saved bmp
     * Image im = new Image(bas.getBuffer()); // Caution! the buffer may be greater than totalBytesWritten, but when parsing theres no problem.
     * if (im.getWidth() > 0) // successfully parsed?
     * {
     *    createGraphics().drawImage(im,CENTER,CENTER);
     *    Vm.sleep(2000);
     * }
     * </pre>
     */
   public Image(byte []fullBmpDescription) // guich@220_17
   {
      NativeMethods4JDK.imageParse(this,fullBmpDescription);
   }

   /**
   * Sets the image width and height to 0 and frees any systems resources
   * associated with the image. The image will be no longer valid and will be GC'd.
   */
   public void free()
   {
      NativeMethods4JDK.imageFree(this);
   }

   /** Returns the height of the image. You can check if the image is ok comparing this with zero. */
   public int getHeight()
   {
      return height;
   }

   /** Returns the width of the image. You can check if the image is ok comparing this with zero. */
   public int getWidth()
   {
      return width;
   }

   /** Returns a new Graphics instance that can be used to drawing in this image. */
   public Graphics getGraphics()
   {
      return new Graphics(this);
   }

   /** Sets the transparent color of this image. Used in many controls.
     * Note that you can set it to null to make the image use the background
     * color.
     * Note also that this value is not used by the system, but the controls
     * (E.g.: waba.ui.Button)
     * use the value stored here to set the background color when going to
     * draw a transparent image.
     * <b>Important!</b>: If your image doesnt use the web safe palette,
     * you <b>must</b> call <code>myImage.applyPalette</code> before calling
     * setTransparentColor (or before the first time the image is drawn),
     * otherwise the color may not be found in the active palette.
     * @deprecated Access the public member transparentColor directly
     */
   public void setTransparentColor(Color c)
   {
      this.transparentColor = c;
   }

   /** Gets the transparent color of this image.
     * @deprecated Access the public member transparentColor directly
     */
   public Color getTransparentColor()
   {
      return this.transparentColor;
   }

   /** Changes all the pixels of the image from one color to the other.
   * Note that the current value of the transparent color is not changed.
   * When you load a mono bmp (consumes less memory), it is stored internally
   * as grayScale (or color).
   * Using this routine, you can change the colors to any other you want.
   */
   public void changeColors(Color from, Color to)
   {
      NativeMethods4JDK.imageChangeColors(this, from.red,from.green,from.blue, to.red,to.green,to.blue);
   }

   /** returns the palette associated with this image. If this method return null, the default SystemPalette is used. */
   public Palette getPalette()
   {
      return palRGBs != null ? new Palette(palRGBs) : null;
   }
   /** substitutes the original palette that came with this image with the
     * given one.
     */
   public void setPalette(Palette pal)
   {
      this.palRGBs = pal.toRGBArray(-1,-1);
   }

   /** saves this image as a Windows .bmp file format to the given Catalog.
     * Important note: the catalog can save multiple images, but the record must
     * be in the SuperWaba warp record format (the image name comes before the data)
     * and must be sorted.
     * This method finds the exact place where to insert the bmp and put it there.
     * If you want to create a bmp to be transfered by a stream to serial or socket
     * then you must use the method createBmp instead.
     * Important: when creating the catalog, the type must be 'SWAX' (SuperWAba
     * eXtension library); if not, the library will not be found by the VM.
     * The catalog name can not be the same name of the image. EG: you can have a
     * catalog named "images" and containing lots of images, one with each name.
     * if a record with this name already exists, it will be replaced.
     * The name is always converted to lowercase and the method makes sure that
     * .bmp is appended to it.
     * To get the list of images in a Catalog, just do a readString at the beginning of
     * each record.
     * <p>Here is an usage example: <pre>
     * // create the image and paint over it
     * Image img = new Image(100,100);
     * Graphics g = img.getGraphics();
     * g.setBackColor(new Color(100,150,200));
     * g.fillRect(25,25,50,50);
     * g.setForeColor(new Color(255,255,255));
     * g.drawCircle(50,50,20);
     * // create the catalog to save the image. You must change CRTR to match your apps creator ID
     * Catalog cat = new Catalog("images.CRTR.SWAX", Catalog.CREATE);
     * img.saveTo(cat, "boxcircle.bmp");
     * cat.close();
     * // load the previously created image
     * if (Vm.attachLibrary("images.CRTR")) // must add the creator so it can be used in desktop. The type is always SWAX
     * {
     *    Image im = new Image("boxcircle.bmp");
     *    if (im.getWidth() > 0) // successfully loaded?
     *       createGraphics().drawImage(im,CENTER,CENTER);
     * }
     * </pre>
     * @see #createBmp(waba.io.DataStream)
     * @return the numbers of bytes writted.
     */
   public int saveTo(Catalog cat, String name)
   {
      name = waba.sys.Convert.toLowerCase(name);
      if (!name.endsWith(".bmp"))
         name += ".bmp";
      int n = cat.getRecordCount();
      int index = -1;
      // guich@200b4_45: fixed the insert_in_order routine
      for (int i =0; i < n; i++) // find the correct position to insert the record. the records must be sorted
      {
         cat.setRecordPos(i);
         DataStream ds = new DataStream(cat);
         String recName = ds.readString();
         if (recName.compareTo(name) >= 0) // is recName greater than name
         {
            if (name.equals(recName)) // same name? delete it
               cat.deleteRecord();
            index = i;
            break;
         }
      }
      if (index == -1) index = n;
      ResizeStream rs = new ResizeStream(cat,width*height+1200);
      DataStream ds = new DataStream(rs);
      rs.startRecord(index);
      ds.writeString(name); // write the name
      n = createBmp(ds);
      rs.endRecord();
      return n;
   }

   /** saves this image as a Windows .bmp file format, with the same bpp of the screen
     * using the current palette of this image, to the given DataStream.
     * If you're sending the bmp through a stream but not saving to a catalog,
     * you can use this method. If you're goind to save it to a catalog, then
     * you must use the saveTo method.
     * Note: if you're saving an image over a previously loaded one, the palette
     * will be preserved only if the device is color. Otherwise,
     * the colors will be converted to the current System palette or to the device's palette.
     * @see #saveTo(waba.io.Catalog, java.lang.String)
     * @return the number of bytes writted.
     */
   public int createBmp(DataStream ds)
   {
      int bpp = Settings.maxColors >= 256 ? 8 : Settings.maxColors == 16 ? 4 : 2;
      int div = 32 / bpp;
      int rowW = ((width+div-1) / div)*div; // guich@230_9

      // compute the total size of this bitmap
      int headerSize = 54;
      // only preserve palette if the device is color and the image is being written from a previously opened one.
      int []rgbs = bpp >= 8 && palRGBs != null ? palRGBs : Palette.getDeviceRGBs();
      int ncolors = rgbs.length;
      int paletteSize = ncolors*4;
      int dataSize = (rowW/(8/bpp)) * height;
      int totalSize = headerSize + paletteSize + dataSize;
      // struct bitmapfileheader
      ds.writeByte('B');
      ds.writeByte('M');                     // int8 type[2];          // File type - should be "BM" for BMP files
      ds.writeIntLE(totalSize);              // 02 int32 size;            // File size
      ds.writeIntLE(0);                      // 06 int8 reserved1[4];     // Reserved
      ds.writeIntLE(headerSize+paletteSize); // 10 int32 offbits;         // offset from beginning of file to bitmap data
      // struct bitmapinfoheader
      ds.writeIntLE(40);                     // 14 int32 size;            // Size of BITMAPINFOHEADER structure
      ds.writeIntLE(width);                  // 18 int32 width;           // Width of image
      ds.writeIntLE(height);                 // 22 int32 height;          // Height of image
      ds.writeShortLE(1);                    // 26 int16 planes;          // No. of planes
      ds.writeShortLE(bpp);                  // 28 int16 bitcount;        // No. of bpi
      ds.writeIntLE(0);                      // 30 int32 compression;     // Type of compression
      ds.writeIntLE(dataSize);               // 34 int32 sizeimage;       // Size of image data in bytes
      ds.writeIntLE(0x0B13);                 // 38 int32 xpelspermeter;   // Horizontal pixels per meter
      ds.writeIntLE(0x0B13);                 // 42 int32 ypelspermeter;   // Vertical pixels per meter
      ds.writeIntLE(rgbs.length);            // 46 int32 clrused;         // No. of colors used
      ds.writeIntLE(rgbs.length);            // 50 int32 clrimportant;    // No. of important colors
      // palette                                54
      if (!waba.sys.Settings.onDevice)
         savingPal = new Palette(rgbs);
      for (int i =0; i < rgbs.length; i++)
         ds.writeIntLE(rgbs[i]);
      // data
      byte row[] = new byte[rowW];
      dataSize = 0;
      switch (ncolors)
      {
         case 16:
            for (int y = height-1; y >= 0; y--) // rows are stored upside down.
            {
               getPixelRow(row,y);
               for (int i = 0; i < rowW; i+=2)
                  row[i >> 1] = (byte)((row[i] << 4) | row[i+1]);
               ds.writeBytes(row,0,rowW >> 1);
            }
            break;
         case 4:
            for (int y = height-1; y >= 0; y--) // rows are stored upside down.
            {
               getPixelRow(row,y);
               for (int i =0; i < rowW; i+=4)
                  row[i >> 2] = (byte)((row[i] << 6) | (row[i+1] << 4) | (row[i+2] << 2) | row[i+3]);
               ds.writeBytes(row,0,rowW >> 2);
            }
            break;
         default: // 256 or greater
            for (int y = height-1; y >= 0; y--) // rows are stored upside down.
            {
               getPixelRow(row,y);
               dataSize += ds.writeBytes(row,0,rowW);
            }
            break;
      }
      return totalSize;
   }
   /** used in saveTo method. fill in the y row into the fillIn array.
     * there must be enough space for the full line be filled. */
   private void getPixelRow(byte []fillIn, int y)
   {
      NativeMethods4JDK.imageGetPixelRow(this, fillIn, y);
   }

   /** apply the palette used by this bitmap to the display */
   public void applyPalette()
   {
      if (palRGBs != null)
         Graphics._applyPalette(palRGBs,0,palRGBs.length-1); // guich@310_24
   }

   /** Returns the scaled instance for this picture. The algorithm used is
     * the replicate scale, and it is very memory hungry, although fast.
     * @since SuperWaba 3.5
     */
   public Image getScaledInstance(int newWidth, int newHeight) // guich@350_22
   {
      return (newWidth==width && newHeight==height)?this:NativeMethods4JDK.imageGetScaledInstance(this, newWidth, newHeight); // guich@400_23: now test if the width/height are the same, what returns the original image
   }

}
