/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>                         *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.fx;

/**
 * Sound is used to play sounds such as beeps and tones.
 * <p>
 * Playing beeps is supported under all platforms but tones are only supported
 * where the underlying platform supports generating tones. Tones aren't supported
 * under Java or Windows CE.
 * <p>
 * Here is an example that beeps the speaker and plays a tone:
 *
 * <pre>
 * Sound.beep();
 * Sound.tone(4000, 300);
 * </pre>
 */
public final class Sound
{
   private Sound()
   {
   }

   /** Plays the device's default beep sound. */
   public static void beep()
   {
      NativeMethods4JDK.soundBeep();
   }


   /**
   * Plays a tone of the specified frequency for the specified
   * duration. Tones will only play under Win32 and PalmOS, they won't
   * play under Java or Windows CE due to underlying platform limitations
   * (in these cases, use waba.fx.SoundClip).
   * @param freq frequency in hertz from 32 to 32767
   * @param duration duration in milliseconds
   */
   public static void tone(int freq, int duration)
   {
      NativeMethods4JDK.soundTone(freq,duration);
   }

   /**
   * Sets the sound of the device on/off. Note that in PalmOS this turns off
   * the system sound only and, in this case, the Sound.beep goes off but the
   * Sound.tone will still work. The original volume is restored at the vm exit.
   * <p>Important: If the user had set its device sound to off, this method will
   * not turn it on, ie, it will keep the device in silence. Otherwise, it will
   * set the volume to its original configuration.
   * @param on if true enables the sound, if false disable it.
   */
   public static void setEnabled(boolean on)
   {
      NativeMethods4JDK.soundSetEnabled(on);
   }

   /**
   * Plays a tone of the specified Midi Note Number for the specified
   * duration. Tones will only play under Win32 and PalmOS, they won't
   * play under Java or Windows CE due to underlying platform limitations
   * (in these cases, use waba.fx.SoundClip).
   * <p>Note: The smaller the midi number, the greater the error in the frequency,
   * due to the fact that the frequency calculated must be an integer.
   * @param midiNoteNumber number of the note according to MIDI standard
   * (eg. A (440hz) = 69), from 24 to 143
   * @param duration duration in milliseconds
   */
   public static void midiTone(int midiNoteNumber, int duration) // guich@300_33
   {
      int freq = (int)Math.round(440.0 * Math.pow (2, ((double)midiNoteNumber-69.0) / 12));
      NativeMethods4JDK.soundTone(freq,duration);
   }

}