/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>                         *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.io;

/**
 * Catalog is a collection of records commonly referred to as a database
 * on small devices.
 * <p>
 * Here is an example showing data being read from records in a catalog:
 *
 * <pre>
 * Catalog c = new Catalog("MyCatalog", Catalog.READ_ONLY);
 * if (!c.isOpen())
 *   return;
 * DataStream ds = new DataStream(c);
 * int count = c.getRecordCount();
 * for (int i = 0; i < count; i++)
 *    if (c.setRecordPos(i))
 *    {
 *       String name = ds.readString();
 *       int age = ds.readShort();
 *       double salary = ds.readDouble();
 *       ...
 *    }
 * c.close();
 * </pre>
 * <p>
 * Note: a Catalog cannot be opened twice. You must close one instance before
 * opening the other or share the same instance.
 * <p>
 * In Windows CE, you can specify the path of the pdb file, either absolute
 * ("\\My Documents\\Test") or relative ("..\\..\\Test").
 */

public class Catalog extends Stream
{
   // HOOK VARIABLES - can't be private - guich@350_20
   /** On desktop, stores the bag that represents this Catalog.
     * On device, stores the DmOpenRef.
     */
   Object bag;
   int hvRecordPos;
   int hvRecordHandle;
   int hvRecordPtr;
   int hvRecordOffset; // used in device and desktop
   int hvRecordLength;
   boolean hvRecordChanged;
   boolean hvDoBackup;

   /** Read-only open mode. This flag has no effect in PalmOS (same as READ_WRITE), only in Windows CE. */
   public static final int READ_ONLY  = 1;
   /** Write-only open mode. This flag has no effect in PalmOS (same as READ_WRITE), only in Windows CE. */
   public static final int WRITE_ONLY = 2;
   /** Read-write open mode. */
   public static final int READ_WRITE = 3; // READ | WRITE
   /** Create open mode. If the database don't exists, it is created and stays in READ_WRITE mode. Otherwise, if it does exists, it is <b>not</b> erased; it just stays in READ_WRITE mode. */
   public static final int CREATE = 4;

   /** Read Only database */
   public static final int DB_ATTR_READ_ONLY = 0x0002;
   /** Set if Application Info block is dirty. Optionally supported by an App's conduit */
   public static final int DB_ATTR_APPINFODIRTY = 0x0004;
   /** Set if database should be backed up to PC if no app-specific synchronization conduit has been supplied. */
   public static final int DB_ATTR_BACKUP = 0x0008;
   /** This tells the backup conduit that it's OK for it to install a newer version of this database with a different name if the current database is open. This mechanism is used to update the Graffiti Shortcuts database, for example. */
   public static final int DB_ATTR_OK_TO_INSTALL_NEWER = 0x0010;
   /** Device requires a reset after this database is installed. */
   public static final int DB_ATTR_RESET_AFTER_INSTALL = 0x0020;
   /** This database should not be copied to */
   public static final int DB_ATTR_COPY_PREVENTION = 0x0040;
   /** This database is used for file stream implementation. This attribute may not be used. */
   public static final int DB_ATTR_STREAM = 0x0080;

   /** Record atribute: Deleted. Note that after a record has it delete attribute set you don't have access to it anymore. */
   public static final byte REC_ATTR_DELETE = (byte)0x80;
   /** Record atribute: Dirty (has been modified since last sync) */
   public static final byte REC_ATTR_DIRTY = (byte)0x40;
   /** Record atribute: Private */
   public static final byte REC_ATTR_SECRET = (byte)0x10;

   /**
   * Opens a catalog with the given name and mode. If mode is CREATE, the
   * catalog will be created if it does not exist.
   * <p>
   * For PalmOS: A PalmOS creator id and type can be specified by appending
   * a 4 character creator id and 4 character type to the name seperated
   * by periods. For example:
   * <pre>
   * Catalog c = new Catalog("MyCatalog.CRTR.TYPE", Catalog.CREATE);
   * </pre>
   * Will create a PalmOS database with the name "MyCatalog", creator id
   * of "CRTR" and type of "TYPE".
   * <p>
   * If no creator id and type is specified, the creator id will default
   * to the creator id of current SuperWaba program and the type will default
   * to "DATA".
   * Note that in desktop the complete form is required.
   * <p>
   * Under PalmOS, the name of the catalog must be 31 characters or less,
   * not including the creator id and type. Windows CE supports a 32
   * character catalog name but to maintain compatibility with PalmOS,
   * you should use 31 characters maximum for the name of the catalog.
   * Every time the database is modified the backup attribute is set.
   * @param name catalog name. Don't use accentuated characters in the catalog name.
   * @param mode one of READ_ONLY, WRITE_ONLY, READ_WRITE or CREATE
   */
   public Catalog(String name, int mode)
   {
      NativeMethods4JDK.catalogCreate(this, name, mode);
   }

   /** this constructor was created to let users use the toPdb and fromPdb methods if running in the desktop.
     * NEVER use it to deal with normal Catalogs! If you do so, your program (or the vm) will crash.
     */
   public Catalog()
   {
      NativeMethods4JDK.catalogCreate(this);
   }

   /** Renames the currently open catalog to the given name. In PalmOS, the name
     * must be in the form "MyNewCatalogName.CRTR.TYPE".
     * <b>Note</b>: in desktop only, the old file <b>may</b> remain if
     * its deletion is not possible (in other words, in
     * desktop the rename operation may work like a "copy oldName newName").
     * <p>Here is an example of use:<pre>
     * Catalog c = new Catalog("guich.Crtr.Type",Catalog.READ_ONLY);
     * if (c.isOpen())
     *    c.rename("flor.CrTr.TyPe");
     * c.close();
     *
     * </pre>
     */
   public boolean rename(String newName) // guich@230_11
   {
      return NativeMethods4JDK.catalogRename(this,newName);
   }

   /**
   * Adds a record to the end of the catalog. If this operation is successful,
   * the position of the new record is returned and the current position is
   * set to the new record. If it is unsuccessful the current position is
   * unset and -1 is returned.
   * @param size the size in bytes of the record to add
   */
   public int addRecord(int size)
   {
      return NativeMethods4JDK.catalogAddRecord(this, size);
   }

   /**
   * Adds a record to the <pos> position of the catalog. If this operation is successful,
   * the position of the new record is returned and the current position is
   * set to the new record. If it is unsuccessful the current position is
   * unset and -1 is returned.
   * implemented by guich (guich@email.com) in 06/30/2000.
   * @param size the size in bytes of the record to add
   */
   public int addRecord(int size, int pos)
   {
      return NativeMethods4JDK.catalogAddRecord(this, size, pos);
   }

   /**
   * Resizes a record. This method changes the size (in bytes) of the current record.
   * The contents of the existing record are preserved if the new size is larger
   * than the existing size. If the new size is less than the existing size, the
   * contents of the record are also preserved but truncated to the new size.
   * Returns true if the operation is successful and false otherwise.
   * @param size the new size of the record
   */
   public boolean resizeRecord(int size)
   {
      return NativeMethods4JDK.catalogResizeRecord(this, size);
   }


   /**
   * Closes the catalog. Returns true if the operation is successful and false
   * otherwise.
   * If you don't close the Catalog, it will be closed when it gets garbage
   * collected. Note that in desktop the catalog is only written to disk
   * after it is closed.
   */
   public boolean close()
   {
      return NativeMethods4JDK.catalogClose(this);
   }


   /**
   * Deletes the catalog. Returns true if the operation is successful and false
   * otherwise.
   * <p>
   * After this, the catalog is closed (of course).
   * <p>
   * Note that this behavior is different between running on desktop and running
   * on palm os, because in applets we cannot erase files at the server, in this
   * case only records are all deleted, but file still exists, with palm-info-header.
   * PS: since 3.0, the catalog may be deleted in applet too.
   */
   public boolean delete()
   {
      return NativeMethods4JDK.catalogDelete(this);
   }


   /**
   * Returns the complete list of existing catalogs. <b>If no catalogs exist, this
   * method returns null.</b>
   * In the desktop, it searches in the already opened files and also the pdb
   * files in the current "." directory.
   * Be careful! The catalog may return null strings in the array.
   */
   public static String []listCatalogs()
   {
      return listCatalogs(0,0);
   }


   /**
   * Returns the list of existing catalogs with the given creator id and/or type.
   * If no catalogs exist, this method returns null.
   * In the desktop, it searches in the already opened files and also the pdb
   * files in the current "." directory.
   * The creator id and type works like a wildcard. You may pass 0 to ignore them.
   * You can use the function Convert.chars2int(String fourChars) to convert a
   * String like "Wrp2" to an integer to pass in this method.
   * Be careful! The catalog may return null strings in the array.
   * @since SuperWaba 3.3
   */
   public static String []listCatalogs(int creatorId, int type) // guich@330_17
   {
      return NativeMethods4JDK.catalogListCatalogs(creatorId, type);
   }


   /**
   * Deletes the current record and sets the current record position to -1.
   * The record is immediately removed from the catalog and all subsequent
   * records are moved up one position.
   */
   public boolean deleteRecord()
   {
      return NativeMethods4JDK.catalogDeleteRecord(this);
   }


   /**
   * Returns the number of records in the catalog or -1 if the catalog is not open.
   */
   public int getRecordCount()
   {
      return NativeMethods4JDK.catalogGetRecordCount(this);
   }


   /**
   * Returns the size of the current record in bytes or -1 if there is no
   * current record.
   */
   public int getRecordSize()
   {
      return NativeMethods4JDK.catalogGetRecordSize(this);
   }


   /**
   * Returns true if the catalog is open and false otherwise. This can
   * be used to check if opening or creating a catalog was successful.
   */
   public boolean isOpen()
   {
      return NativeMethods4JDK.catalogIsOpen(this);
   }


   /**
   * Sets the current record position and locks the given record. The value
   * -1 can be passed to unset and unlock the current record. If the operation
   * is succesful, true is returned and the read/write cursor is set to
   * the beginning of the record. Otherwise, false is returned.
   */
   public boolean setRecordPos(int pos)
   {
      return NativeMethods4JDK.catalogSetRecordPos(this, pos);
   }


   /**
   * Reads bytes from the current record into a byte array. Returns the
   * number of bytes actually read or -1 if an error prevented the
   * read operation from occurring. After the read is complete, the location of
   * the cursor in the current record (where read and write operations start from)
   * is advanced the number of bytes read.
   * @param buf the byte array to read data into
   * @param start the start position in the array
   * @param count the number of bytes to read
   */
   public int readBytes(byte buf[], int start, int count)
   {
      return NativeMethods4JDK.catalogReadBytes(this, buf, start, count);
   }


   /**
   * Offsets the cursor in the current record a number of bytes. The cursor
   * defines where read and write operations start from in the record. Returns
   * the number of bytes actually skipped or -1 if an error occurs.
   * @param count the number of bytes to skip
   */
   public int skipBytes(int count)
   {
      return NativeMethods4JDK.catalogSkipBytes(this, count);
   }


   /**
   * Writes to the current record. Returns the number of bytes written or -1
   * if an error prevented the write operation from occurring.
   * After the write is complete, the location of the cursor in the current record
   * (where read and write operations start from) is advanced the number of bytes
   * written.
   * @param buf the byte array to write data from
   * @param start the start position in the byte array
   * @param count the number of bytes to write
   */
   public int writeBytes(byte buf[], int start, int count)
   {
      return NativeMethods4JDK.catalogWriteBytes(this, buf, start, count);
   }

   /** Inspects a record. use this method with careful, none of the params are checked for validity.
   * the cursor is not advanced, neither the current record position. this method must be used
   * only for a fast way of viewing the contents of a record,
   * like searching for a specific header or filling a grid of data.
   * <i>buf.length</i> bytes (at maximum) are read from the record into <i>buf</i>.
   * Returns the number of bytes read (can be different of buf.length if buf.length is greater
   * than the record size) or -1 if an error prevented the read operation from occurring.
   * @since SuperWaba 1.1
   */
   public int inspectRecord(byte buf[], int recPosition)
   {
      return NativeMethods4JDK.catalogInspectRecord(this, buf, recPosition);
   }

   /** Retrieves the attributes of the current record. Use the REC_ATTR_xxx masks to set/retrieve its states.
     * @return 0 if db dont exists
     * @deprecated use getRecordAttributes(recordPos) instead.
     */
   public byte getRecordAttributes()
   {
      return getRecordAttributes(getRecordPos());
   }

   /** Sets the attributes of the current record. Use the REC_ATTR_xxx masks to set/retrieve its states.
     * @deprecated use setRecordAttributes(recordPos, attr) instead.
     */
   public void setRecordAttributes(byte attr)
   {
      setRecordAttributes(getRecordPos(), attr);
   }

   /** Retrieves the attributes of the current record. Use the REC_ATTR_xxx masks to set/retrieve its states.
     * Note that the current record, if modified, will have its attributes changed
     * after you call this method.
     * @return 0 if db dont exists
     */
   public byte getRecordAttributes(int recordPos)
   {
      return NativeMethods4JDK.catalogGetRecordAttributes(this, recordPos);
   }

   /** Sets the attributes of the current record. Use the REC_ATTR_xxx masks to set/retrieve its states.
     * Don't try to change the current record's attribute, because it may change again (by the VM)
     * after the record is released.
     */
   public void setRecordAttributes(int recordPos, byte attr)
   {
      NativeMethods4JDK.catalogSetRecordAttributes(this, recordPos, attr);
   }

   /** Retrieves this catalog's attributes. Use the DB_ATTR_xxx masks to retrieve its states. */
   public int getAttributes()
   {
      return NativeMethods4JDK.catalogGetAttributes(this);
   }

   /** Sets this catalog's attributes, defined by the DB_ATTR_xxx constants. The original attributes must be getted prior applying this value. If not, you can loose your database and your app will crash. */
   public void setAttributes(int i)
   {
      NativeMethods4JDK.catalogSetAttributes(this, i);
   }

   /** Returns the current record position or -1 if there is no current record. */
   public int getRecordPos()
   {
      return NativeMethods4JDK.catalogGetRecordPos(this);
   }

   /** Returns the internal record offset. You may use this value in conjunction with
     * the skipBytes method. Note that setRecordPos resets the offset to 0.
     * @since SuperWaba 4.0
     */
   public int getRecordOffset()
   {
      return hvRecordOffset;
   }
}