/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  Copyright (C) 1999 Rob Nielsen                                               *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.io;

import waba.sys.*;

/**
 * DataStream is a wrapper you can place around any Stream such as a
 * SerialPort, Catalog, or BufferStream which lets you read and write
 * standard SuperWaba data types like ints, floats, and Strings in a simple
 * manner.  Here's an example
 * <pre>
 *   SerialPort port=new SerialPort(9600,0);
 *   DataStream ds=new DataStream(port);
 *   ds.writeString("Hello");
 *   int status=ds.readUnsignedByte();
 *   if (status==1)
 *   {
 *     ds.writeString("Pi");
 *     ds.writeFloat(3.14);
 *   }
 *   port.close();
 * </pre>
 * <br>ps: this class was created to the Waba Extras package, but SW added a few methods
 * for storing and retrieving strings.
 */
public class DataStream extends Stream
{
    /** the underlying stream */
    protected Stream stream;

    /** a four byte array for reading and writing numbers */
    protected byte[] b=new byte[4];

    static protected byte bytes[] = new byte[128]; // buffer

   /**
    * Constructs a new DataStream which sits upon the given stream using
    * big endian notation for multibyte values.
    * @param stream the base stream
    */
    public DataStream(Stream stream)
    {
       this.stream = stream;
    }
    /** closes the stream */
    public boolean close()
    {
       if (stream != null)
          return stream.close();
       return true;
    }
    /** pads the stream writting n bytes. all bytes will be 0. */
    final public int pad(int n)
    {
       if (bytes.length < n) bytes = new byte[n+16]; // guich@200 grows the buffer if necessary - corrected by mike mcroberts
       for (int i =0; i < n; i++)
          bytes[i] = 0;
       writeBytes(bytes,0,n);
       return n;
    }
    /**
     * Skips reading the next n bytes in the stream (only forward!)
     * @param n the number of bytes to skip, must be > 0
     */
    final public void skip(int n)
    {
       byte[] b=new byte[n];
       stream.readBytes(b,0,n);
    }
    /**
     * Reads a boolean from the stream as a byte.  True is returned if
     * the byte is not zero, false if it is.
     * @return the boolean value read.
     */
    final public boolean readBoolean()
    {
       stream.readBytes(b,0,1);
       return b[0]!=0;
    }
    /**
     * Reads a single byte from the stream.  The returned value will range
     * from -128 to 127.
     * @return the read byte
     */
    final public byte readByte()
    {
       stream.readBytes(b,0,1);
       return b[0];
    }
    /**
     * Reads bytes from the stream. Returns the
     * number of bytes actually read or -1 if an error prevented the
     * read operation from occurring.
     * @param buf the byte array to read data into
     * @param start the start position in the array
     * @param count the number of bytes to read
     */
    final public int readBytes(byte buf[], int start, int count)
    {
       return stream.readBytes(buf,start,count);
    }
    /**
     * Reads bytes from the stream. Returns the
     * number of bytes actually read or -1 if an error prevented the
     * read operation from occurring.
     * @param buf the byte array to read data into
     */
    final public int readBytes(byte buf[])
    {
       return stream.readBytes(buf,0,buf.length);
    }
    /**
     * Reads a float value from the stream as four bytes in IEEE 754 format.
     * @return the float value
     */
    final public float readFloat()
    {
       return Convert.toFloatBitwise(readInt());
    }
    /**
     * Reads an integer using little endian from the stream as four bytes.  The returned value
     * will range from -2147483648 to 2147483647.
     * @return the integer
     */
    final public int readIntLE()
    {
       stream.readBytes(b,0,4);
       return (((b[3]&0xFF) << 24) | ((b[2]&0xFF) << 16) | ((b[1]&0xFF) << 8) | (b[0]&0xFF));
    }

    /**
     * Reads a short from the stream as two bytes in the little endian format.  The returned value will
     * range from -32768 to 32767.
     * @return the short
     */
    final public short readShortLE()
    {
       stream.readBytes(b,0,2);
       return (short)(((b[1]&0xFF) << 8) | (b[0]&0xFF));
    }

    /**
     * Reads an integer from the stream as four bytes.  The returned value
     * will range from -2147483648 to 2147483647.
     * @return the integer
     */
    final public int readInt()
    {
       stream.readBytes(b,0,4);
       return (((b[0]&0xFF) << 24) | ((b[1]&0xFF) << 16) | ((b[2]&0xFF) << 8) | (b[3]&0xFF));
    }
    /**
     * Reads a short from the stream as two bytes.  The returned value will
     * range from -32768 to 32767.
     * @return the short
     */
    final public short readShort()
    {
       stream.readBytes(b,0,2);
       return (short)(((b[0]&0xFF) << 8) | (b[1]&0xFF));
    }

     /** Reads a double.
     @since 2.0 */
     final public double readDouble()
     {
        return Convert.longBitsToDouble(readLong());
     }
     /** Reads a long.
     @since 2.0 */
     final public long readLong()
     {
       long l1 = (long)readInt();
       long l2 = (long)readInt();
       long ll = (l1 << 32) + l2;
       if (l2 < 0) ll += 0x100000000L; // guich@241_19
       return ll;
     }
    /**
     * Reads a single unsigned byte from the stream.  The returned value will
     * range from 0 to 255.
     * @return the read byte.
     */
    final public int readUnsignedByte()
    {
       stream.readBytes(b,0,1);
       return b[0]&0xFF;
    }
    /**
     * Reads an unsigned short from the stream as two bytes.  The returned
     * value will range from 0 to 65535.
     * @return the short
     */
    final public int readUnsignedShort()
    {
       stream.readBytes(b,0,2);
       return (((b[0]&0xFF) << 8) | (b[1]&0xFF));
    }
    /**
     * Writes an integer to the stream as four bytes.
     * @param i the integer to write
     * @return the number of bytes written
     */
    final public int writeIntLE(int i)
    {
       b[0] = (byte)(i & 0xFF); i >>= 8; // guich@300_40
       b[1] = (byte)(i & 0xFF); i >>= 8;
       b[2] = (byte)(i & 0xFF); i >>= 8;
       b[3] = (byte)(i & 0xFF);
       stream.writeBytes(b,0,4);
       return 4;
    }

    /**
     * Writes a short to the stream as two bytes.  As there is no
     * short type in SuperWaba but we often only want to only use two bytes in
     * storage.
     * An int is used but the upper two bytes are ignored, so this method
     * can be used to store a signed or an unsigned short.
     * @param i the short to write
     * @return the number of bytes written
     */
    final public int writeShortLE(int i)
    {
       b[0] = (byte)(i & 0xFF); i >>= 8; // guich@300_40
       b[1] = (byte)(i & 0xFF);
       stream.writeBytes(b,0,2);
       return 2;
    }
    /**
     * Writes a boolean to the stream as a byte.  True values are written as
     * 1 and false values as 0.
     * @param b the boolean to write
     * @return the number of bytes written
     */
    final public int writeBoolean(boolean bool)
    {
       b[0]=(bool?(byte)1:(byte)0);
       stream.writeBytes(b,0,1);
       return 1;
    }
    /**
     * Writes a single byte to the stream.
     * @param b the byte to write
     * @return the number of bytes written
     */
    final public int writeByte(byte by)
    {
       b[0]=by;
       stream.writeBytes(b,0,1);
       return 1;
    }
    /**
     * Writes a single byte to the stream.
     * @param b the byte to write (only least significant byte is written)
     * @return the number of bytes written
     */
    final public int writeByte(int by)
    {
       b[0]=(byte)(by&0xFF);
       stream.writeBytes(b,0,1);
       return 1;
    }
    /**
     * Writes bytes to the stream. Returns the
     * number of bytes actually written or -1 if an error prevented the
     * write operation from occurring.
     * @param buf the byte array to write data from
     * @param start the start position in the byte array
     * @param count the number of bytes to write
     */
    final public int writeBytes(byte buf[], int start, int count)
    {
       return stream.writeBytes(buf,start,count);
    }
    /**
     * Writes bytes to the stream. Returns the
     * number of bytes actually written or -1 if an error prevented the
     * write operation from occurring.
     * @param buf the byte array to write data from
     */
    final public int writeBytes(byte buf[])
    {
       return stream.writeBytes(buf,0,buf.length);
    }
    /**
     * Writes a float value to the stream as four bytes in IEEE 754 format
     * @param f the float to write
     * @return the number of bytes written
     */
    final public int writeFloat(float f)
    {
       writeInt(Convert.toIntBitwise(f));
       return 4;
    }
    /**
     * Writes an integer to the stream as four bytes.
     * @param i the integer to write
     * @return the number of bytes written
     */
    final public int writeInt(int i)
    {
       b[3] = (byte)(i & 0xFF); i >>= 8; // guich@300_40
       b[2] = (byte)(i & 0xFF); i >>= 8;
       b[1] = (byte)(i & 0xFF); i >>= 8;
       b[0] = (byte)(i & 0xFF);
       stream.writeBytes(b,0,4);
       return 4;
    }
    /**
     * Writes an short to the stream as two bytes.  As there is no
     * short type in SuperWaba but we often only want to only use two bytes in
     * storage.
     * an int is used but the upper two bytes are ignored.
     * @param i the short to write
     * @return the number of bytes written
     */
    final public int writeShort(int i)
    {
       b[1] = (byte)(i & 0xFF); i >>= 8; // guich@300_40
       b[0] = (byte)(i & 0xFF);
       stream.writeBytes(b,0,2);
       return 2;
    }
    /** reads a string, converting from palm to java format.
    @return a zero or more length string. null is never returned.
    */
    final public String readString()
    {
       int len = readUnsignedShort();  // guich@341_2
       if (len == 0) return "";

       if (bytes.length < len) bytes = new byte[len+16]; // guich@200 grows the buffer if necessary - corrected by mike mcroberts
       readBytes(bytes,0,len);

       return new String(bytes, 0, len);
    }
    /** reads an array of string.
       @return a zero or more length array. null is never returned.
     */
    final public String[] readStringArray()
    {
       int size = (int)readUnsignedShort();
       String []a = new String[size];
       for (int i =0; i < size; i++)
          a[i] = readString();
       return a;
    }
     /** writes the string into the stream, converting it from international format to palm format */
    final public int writeString(String s)
     {
        int n = 0;
        if (s != null && s.length() > 0)
        {
           byte[] c = s.getBytes();
           n += writeShort(c.length);
           n += writeBytes(c,0,c.length);
        }
        else n += writeShort(0);
        return n;
     }
     /** writes the string array into the stream */
     final public int writeStringArray(String []v)
     {
        int n = 0;
        if (v == null || v.length == 0)
           n += writeShort(0);
        else
        {
           n += writeShort(v.length);
           for (int i=0; i < v.length; i++)
              n += writeString(v[i]);
        }
        return n;
     }
     /** writes a double.
     @since 2.0 */
     final public int writeDouble(double d)
     {
        return writeLong(Convert.doubleToLongBits(d));
     }

     /** writes a long.
     @since 2.0 */
     final public int writeLong(long l)
     {
        writeInt((int)((l >> 32) & 0xFFFFFFFF));
        writeInt((int)(l & 0xFFFFFFFF));
        return 8;
     }
    /**
    * Reads a C-style string from the stream.  This is a NUL (0) terminated
    * series of characters.  This format is commonly used by other Palm
    * applications. Note that if you're creating your own stream, choose
    * readString instead of readCString, because readCString is *much* slower.
    * Also, this method does not handle correctly international characters.
    * @return the loaded String
    */
    public String readCString()
    {
       // we have to read one character at a time.
       int size=0;
       while (true)
       {
          stream.readBytes(b,0,1);
          if (b[0]==0)
             break;
          if (size == bytes.length) // grow the array if necessary
          {
             byte[] temp = new byte[bytes.length+16];
             Vm.copyArray(bytes, 0, temp, 0, size);
             bytes = temp;
          }
          bytes[size++] = b[0];
       }
       return new String(bytes,0,size);
    }

   /**
    * Writes a C-style string to the stream.  This means that all the
    * characters of the string are written out, followed by a NUL (0)
    * character.  This format is commonly used by other Palm applications.
    * Note that if you're creating your own stream, choose
    * writeString instead of writeCString, because writeCString is *much* slower.
    * Also, this method does not handle correctly international characters.
    * @param s the string to write
    * @return the number of bytes written.
    */
    final public int writeCString(String s)
    {
       if (s == null)
          return writeByte(0);

       int n = 0;
       byte[] c = s.getBytes();
       n += writeBytes(c,0,c.length);
       n += writeByte(0);
       return n;
   }

   public boolean isOpen() // guich@330_40
   {
      return stream.isOpen();
   }
   /** Reads an array of chars, where its length is stored in the first two bytes
     * as an unsigned short
     * @since SuperWaba 3.5
     */
   final public char[] readChars()
   {
      int len = readUnsignedShort();
      char []chars = new char[len];
      for (int i =0; --len >= 0; i++)
      {
         stream.readBytes(b,0,2);
         chars[i] = (char)(((b[0]&0xFF) << 8) | (b[1]&0xFF));
      }
      return chars;
   }
   /** Writes an array of chars, placing its length in the first two bytes, as
     * an unsigned short.
     * @param len the length to be written or -1 if it is to write the whole char array
     * @param chars the char array to be written.
     * @param start the starting index (in most cases: 0).
     * @since SuperWaba 3.5
     */
   final public int writeChars(char []chars, int start, int len)
   {
      int c;
      int i;
      if (len < 0)
         len = chars == null ? 0 : chars.length;
      writeShort(len-start);
      for (i =start; i < len; i++)
      {
         c = chars[i];
         b[1] = (byte)(c & 0xFF); c >>= 8;
         b[0] = (byte)(c & 0xFF);
         stream.writeBytes(b,0,2);
      }
      return ((len-start) << 1) + 2;
   }
}