/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>                         *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.io;

/**
 * File is a file or directory.
 * <p>
 * The File class will work in Windows CE or in a Palm OS >= 3.5, containing
 * an expansion card. Use the <code>isAvailable</code> method to determine if the device
 * has or not support for file systems.
 * <p>
 * Note that in Palm OS, writting to a storage card can be slower than writting
 * to the main memory.
 * <p>
 * Here is an example showing data being read from a file:
 *
 * <pre>
 * File file = new File("/temp/tempfile", File.READ_ONLY);
 * if (!file.isOpen())
 *   return;
 * byte b[] = new byte[10];
 * file.readBytes(b, 0, 10);
 * file.close();
 * file = new File("/temp/tempfile", File.DONT_OPEN);
 * file.delete();
 * </pre>
 */

public class File extends Stream
{
   /** The path that represents this file */
   protected String path;

   // HOOK VARIABLES - can't be private - guich@350_20

   /** On desktop, stores the java.io.File. On device, stores the FileRef. */
   Object jfile;

   /** On desktop, stores the java.io.RandomAccessFile. */
   Object jraf;

   /** On device, stores the last error.
     * For WinCE, you must check the WINERROR.H file.<p>
     * For Palm OS, the list of possible errors are described below:
     * <table border="1" cellpadding="3">
     * <tr><td>BufferOverflow      </td><td>10753</td></tr>
     * <tr><td>FileGeneric         </td><td>10754</td></tr>
     * <tr><td>FileBadRef          </td><td>10755</td></tr>
     * <tr><td>FileStillOpen       </td><td>10756</td></tr>
     * <tr><td>FilePermissionDenied</td><td>10757</td></tr>
     * <tr><td>FileAlreadyExists   </td><td>10758</td></tr>
     * <tr><td>FileEOF             </td><td>10759</td></tr>
     * <tr><td>FileNotFound        </td><td>10760</td></tr>
     * <tr><td>VolumeBadRef        </td><td>10761</td></tr>
     * <tr><td>VolumeStillMounted  </td><td>10762</td></tr>
     * <tr><td>NoFileSystem        </td><td>10763</td></tr>
     * <tr><td>BadData             </td><td>10764</td></tr>
     * <tr><td>DirNotEmpty         </td><td>10765</td></tr>
     * <tr><td>BadName             </td><td>10766</td></tr>
     * <tr><td>VolumeFull          </td><td>10767</td></tr>
     * <tr><td>Unimplemented       </td><td>10768</td></tr>
     * <tr><td>NotADirectory       </td><td>10769</td></tr>
     * <tr><td>IsADirectory        </td><td>10770</td></tr>
     * <tr><td>DirectoryNotFound   </td><td>10771</td></tr>
     * <tr><td>NameShortened       </td><td>10772</td></tr>
     */
   public int lastError;

   /** Don't open the file mode. */
   public static final int DONT_OPEN = 0;
   /** Read-only open mode. */
   public static final int READ_ONLY  = 1;
   /** Write-only open mode. */
   public static final int WRITE_ONLY = 2;
   /** Read-write open mode. */
   public static final int READ_WRITE = 3; // READ | WRITE
   /** Create open mode. Used to create a file if one does not exist. */
   public static final int CREATE = 4;

   /** Used in the setTime method, in parameter whichTime. This sets all times */
   public static final byte TIME_ALL = (byte)0xF;
   /** Used in the getAttributes and setAttributes method. These values are platform independent, and can be ORed together in the setTime method. */
   public static final byte TIME_CREATED = (byte)1;
   /** Used in the getAttributes and setAttributes method. These values are platform independent, and can be ORed together in the setTime method. */
   public static final byte TIME_MODIFIED = (byte)2;
   /** Used in the getAttributes and setAttributes method. These values are platform independent, and can be ORed together in the setTime method. */
   public static final byte TIME_ACCESSED = (byte)4;

   /** Used in the getAttributes and setAttributes method. These values are platform independent. */
   public static final int ATTR_ARCHIVE = 1;
   /** Used in the getAttributes and setAttributes method. These values are platform independent. */
   public static final int ATTR_HIDDEN = 2;
   /** Used in the getAttributes and setAttributes method. These values are platform independent. */
   public static final int ATTR_READ_ONLY = 4;
   /** Used in the getAttributes and setAttributes method. These values are platform independent. */
   public static final int ATTR_SYSTEM = 8;

   /**
   * Opens a file with the given name and mode. If mode is CREATE, the
   * file will be created if it does not exist. The DONT_OPEN mode
   * allows the exists(), rename(), delete(), listDir(), createDir()
   * and isDir() methods to be called without requiring the file to be
   * open for reading or writing.<p>Note that the filename must not contain
   * accentuated characters. Also, the slash / MUST be the path separator.
   * @param path the file's path
   * @param mode one of DONT_OPEN, READ_ONLY, WRITE_ONLY, READ_WRITE or CREATE
   */
   public File(String path, int mode)
   {
      this.path = path;
      if (mode != DONT_OPEN)
         NativeMethods4JDK.fileCreate(this, path, mode);
   }

   /** Opens a file in DONT_OPEN mode. This constructor is useful for directory
     * manipulation and/or check if file exists. No read/write operation can be
     * done with a file created in this mode. */
   public File(String path)
   {
      this.path = path;
   }

   /** Returns true if a File System is available in the current PDA.
     * In Windows CE and Desktop, this always returns true. In Palm OS,
     * it will depend if the file has a external card (Secure Digital, Smart Media,
     * etc). Using other methods in this class without first checking if this class
     * is available in the target device will simply reset it.
     */
   public static boolean isAvailable()
   {
      return NativeMethods4JDK.fileIsAvailable();
   }

   /**
   * Closes the file. Returns true if the operation is successful and false
   * otherwise.
   */

   public boolean close()
   {
      return NativeMethods4JDK.fileClose(this);
   }

   /**
   * Creates a directory. Returns true if the operation is successful and false
   * otherwise.
   */

   public boolean createDir()
   {
      return NativeMethods4JDK.fileCreateDir(this);
   }
   /**
   * Deletes the file or directory. Returns true if the operation is
   * successful and false otherwise. After this operation, this File object is invalid.
   */

   public boolean delete()
   {
      return NativeMethods4JDK.fileDelete(this);
   }
   /** Returns true if the file exists and false otherwise. */

   public boolean exists()
   {
      return NativeMethods4JDK.fileExists(this);
   }

   /** Returns the size of the file in bytes. If the file is not opened
   * 0 will be returned.
   */
   public int getSize()
   {
      return NativeMethods4JDK.fileGetSize(this);
   }

   /** Return the file's path. */
   public String getPath()
   {
      return path;
   }
   /** Returns true if the file is a directory and false otherwise. */

   public boolean isDir()
   {
      return NativeMethods4JDK.fileIsDir(this);
   }
   /**
   * Returns true if the file is open for reading or writing and
   * false otherwise. This can be used to check if opening or
   * creating a file was successful.
   */

   public boolean isOpen()
   {
      return NativeMethods4JDK.fileIsOpen(this);
   }
   /**
   * Lists the files contained in a directory. The strings returned are the
   * names of the files and directories contained within this directory.
   * This method returns null if the directory can't be read or if the
   * operation fails. Volume labels are returned between [], and paths have
   * a / after the name.
   */

   public String []listFiles()
   {
      return NativeMethods4JDK.fileListFiles(this);
   }

   /**
   * Reads bytes from the file into a byte array. Returns the
   * number of bytes actually read or -1 if an error prevented the
   * read operation from occurring. After the read is complete, the location of
   * the file pointer (where read and write operations start from)
   * is advanced the number of bytes read.
   * @param buf the byte array to read data into
   * @param start the start position in the array
   * @param count the number of bytes to read
   */
   public int readBytes(byte b[], int off, int len)
   {
      return NativeMethods4JDK.fileReadBytes(this, b,off,len);
   }
   /**
   * Renames the file. You must give the full directory specification for the file,
   * to keep compatibility between all platforms. WinCE platform lets you move a file
   * using rename, while Palm OS does not let you move the file.
   * Returns true if the renaming was successful and false otherwise.
   * After this operation, this File object is invalid.
   */

   public boolean rename(String path)
   {
      return NativeMethods4JDK.fileRename(this, path);
   }
   /**
   * Sets the file pointer for read and write operations to the given
   * position. The position passed is an absolute position, in bytes,
   * from the beginning of the file. To set the position to just after
   * the end of the file, you can call:
   * <pre>
   * file.setPos(file.getSize());
   * </pre>
   * True is returned if the operation is successful and false otherwise.
   */

   public boolean setPos(int pos)
   {
      return NativeMethods4JDK.fileSetPos(this, pos);
   }
   /**
   * Writes to the file. Returns the number of bytes written or -1
   * if an error prevented the write operation from occurring.
   * After the write is complete, the file pointer (where read and
   * write operations start from) is advanced the number of bytes
   * written.
   * @param buf the byte array to write data from
   * @param start the start position in the byte array
   * @param count the number of bytes to write
   */

   public int writeBytes(byte b[], int off, int len)
   {
      return NativeMethods4JDK.fileWriteBytes(this, b,off,len);
   }

   /** Sets this file attributes. This method does not work on desktop. The file must be opened in a mode different of DONT_OPEN.
     * @param attr one ore more ATTR_xxx constants ORed together. Values different from the ATTR_xxx constants are simply ignored. */
   public void setAttributes(int attr)
   {
      NativeMethods4JDK.fileSetAttributes(this, attr);
   }

   /** Gets this file attributes. See the ATTR_xxx constants, which can be ORed together. This method does not work on desktop. The file must be opened in a mode different of DONT_OPEN.
     * @returns -1 if failure, otherwise the attributes ORed together.
     */
   public int getAttributes()
   {
      return NativeMethods4JDK.fileGetAttributes(this);
   }

   /** Sets the time for the given time type. See the TIME_xxx constants. This method does not work on desktop.  The file must be opened in a mode different of DONT_OPEN.*/
   public void setTime(byte whichTime, waba.sys.Time time)
   {
      NativeMethods4JDK.fileSetTime(this, whichTime, time);
   }

   /** Gets the time for the given time type. See the TIME_xxx constants. This method does not work on desktop. Note that the constants in this method CANNOT be ORed together; in other words, this method only returns one time in each call.  The file must be opened in a mode different of DONT_OPEN. */
   public waba.sys.Time getTime(byte whichTime)
   {
      return NativeMethods4JDK.fileGetTime(this, whichTime);
   }
}