/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.io;

/** this class is used to create an resizable record. you can use it with DataStream. example:
   <pre>
      Catalog cat = new Catalog(type+"."+creator+"."+type,Catalog.READ_WRITE);
      ResizeStream rs = new ResizeStream(cat,512);
      DataStream ds = new DataStream(rs);
      rs.startRecord();
      ds.writeStringArray(aStringArray);
      rs.endRecord();
      ds.close();
   </pre>
   <br>ps: if you dont call startRecord, writeBytes will simply call catalog.writeBytes and will not resize the record.
*/
public class ResizeStream extends Stream
{
   /** the associated catalog */
   private Catalog cat;
   /** the intial size of the record */
   private int initialSize;
   /** how many bytes were written */
   private int written;
   /** current record size */
   private int size;

   /** constructs the resize stream.
   @param cat the catalog associated
   @param initialSize the initial size of the record. CANNOT BE 0!
   */
   public ResizeStream(Catalog cat, int initialSize)
   {
      this.cat = cat;
      this.initialSize = initialSize;
   }

   /** inserts the record at the specified index in the catalog.
       @return true if operation was successful, false otherwise.
       @since SuperWaba 1.21. */
   public boolean startRecord(int pos)
   {
   	boolean ok; // guich@241_14
      if (pos >= cat.getRecordCount()) // append?
         ok = cat.addRecord(initialSize) != -1;
      else
         ok = cat.addRecord(initialSize, pos) != -1;
      written = 0;
      size = initialSize;
      return ok;
   }

   /** add a new record to the catalog.
       @return true if operation was successful, false otherwise. */
   public boolean startRecord()
   {
      boolean ok = cat.addRecord(initialSize) != -1; // guich@200b4_96
      written = 0;
      size = initialSize;
      return ok;
   }

   /** Restart writing to the given record pos. If pos is < 0 or greater than the
     * number of records of the Catalog, the record is appended.
     */
   public boolean restartRecord(int pos)
   {
      if (pos < 0 || pos >= cat.getRecordCount()) // append?  guich@310_4: check if < 0
         return startRecord();
      written = 0;
      size = initialSize;
      if (cat.setRecordPos(pos)) // guich@241_14
      {
         size = cat.getRecordSize(); // guich@300_35: initialize to the current record size
         return true;
      }
      return false;
   }

   /** must be called after the record is finished so it can be resized */
   public void endRecord()
   {
      if (written > 0) // guich@400_22
         cat.resizeRecord(written);
      else
         cat.deleteRecord();
      size = 0;
   }

   /** simply read the bytes from the associated catalog */
   public int readBytes(byte buf[], int start, int count)
   {
      return cat.readBytes(buf,start,count);
   }
   /** writes to the buffer, growing the record if necessary. */
   public int writeBytes(byte buf[], int start, int count)
   {
      if (size == 0)
      	return cat.writeBytes(buf,start,count);
      if (count - start <= 0) return 0;

      int total = written + (count-start);

      while (total > size) // no more space?
      {
         size += initialSize;
         if (!cat.resizeRecord(size)) // expand
            return -1;
      }
      int n = cat.writeBytes(buf,start,count);
      if (n >= 0) written += n;
      return n;
   }
   /** closes the catalog */
   public boolean close()
   {
      return cat.close();
   }

   public boolean isOpen() // guich@330_40
   {
      return cat.isOpen();
   }
}