/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2001-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *                      and Matthias Ringwald                                    *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>                         *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.io;

/**
 * SerialPort accesses a device's serial port.
 * <p>
 * This works under the devices and under JDK (in this case, it uses javax.comm)
 * <p>
 * When a serial port is created, an attempt is made to open the port.
 * If the open attempt is successful, a call to isOpen()
 * will return true and the port will remain open until close() is called.
 * If close() is never called, the port will be closed when the object
 * is garbage collected.
 * <p>
 *
 * Here is an example showing data being written and read from a serial port:
 *
 * <pre>
 * SerialPort port = new SerialPort(0, 9600);
 * if (!port.isOpen())
 *   return;
 * byte buf[] = new byte[10];
 * buf[0] = 3;
 * buf[1] = 7;
 * port.writeBytes(buf, 0, 2);
 * int count = port.readBytes(buf, 0, 10);
 * if (count == 10)
 *   ...
 * port.close();
 * </pre>
 * <p>
 * Note on HandSpring users: the Keyboard thread is disabled after open
 * and enabled again after close.
 */

public class SerialPort extends Stream
{
   /** Returns the last error when calling a method. Below you can see a list of error codes from SerialMgr.h.
   <pre>
   SW Custom codes (See nmpalm_c.c for details):
   65525 (-11): NO_SERIAL_LIBRARY_AVAILABLE
   65522 (-14): ARRAY_RANGE_CHECK

   </pre>
   WindowsCE error codes: see winerror.h in the Windows SDK.

   Palm OS error codes.
   <table border="1" cellpadding="3">
   <tr><td>serErrBadParam           </td><td>769</td></tr>
   <tr><td>BadPort                  </td><td>770</td></tr>
   <tr><td>NoMem                    </td><td>771</td></tr>
   <tr><td>BadConnID                </td><td>772</td></tr>
   <tr><td>TimeOut                  </td><td>773</td></tr>
   <tr><td>LineErr                  </td><td>774</td></tr>
   <tr><td>AlreadyOpen              </td><td>775</td></tr>
   <tr><td>StillOpen                </td><td>776</td></tr>
   <tr><td>NotOpen                  </td><td>777</td></tr>
   <tr><td>NotSupported             </td><td>778</td></tr>
   <tr><td>NoDevicesAvail           </td><td>779</td></tr>
   <tr><td>USBConfigurationFailed   </td><td>780</td></tr>
   </table>
   */
   public int lastError; // guich@230_27

   // HOOK VARIABLES - can't be private - guich@350_20

   /** On desktop, stores the javax.comm.SerialPort. On device, the SerialRef. */
   Object thisSerialPort;
   /** On desktop, stores the java.io.InputStream. On device, the read timeout */
   Object thisInputStream;
   /** On desktop, stores the java.io.OutputStream. On device, the port number */
   Object thisOutputStream;

   boolean hvIsNewSerialMgr;
   boolean hvIsDirectHardware;


   // Logical Ports used in SuperWaba
   // NOTE: The SuperWaba VM indexes directly to these values

   /** Default Port (cradle) */
   public static final int DEFAULT    = 0;

   /** IrCOMM Port (Serial Connection on top of IrDA Stack) */
   public static final int IRCOMM     = 0x1000;

   /** SIR Port (Physical Layer of IrDA Stack). Important! In order
     * to access the SIR port on the Pocket PCs the following has to be
     * UNCHECKED: Settings/Connections/Beam/Receive all incoming beams and select
     * discoverable mode.
     */
   public static final int SIR         = 0x1001;

   /** USB Endpoint 2 Port */
   public static final int USB      = 0x1002;

   /** Bluetooth: open the built-in Bluetooth discovery dialog,
     * then establish a Serial Port Profile (just serial emulation across
     * Bluetooth, using the RFComm BT layer) virtual serial port connection.
     * Note: in WinCE, this maps to the Serial (port 0). You must explicitly
     * map your BlueTooth outside SW.
     */
   public static final int BLUETOOTH = 0x1003; // guich@330_34


   /**
   * Opens a serial port. The number passed is the number of the
   * serial port for devices with multiple serial ports. Port number
   * 0 defines the "default port" of a given type. For Windows CE
   * and PalmPilot devices, you should pass 0 as the port number to
   * access the device's single serial port.
   * <p>
   * On Windows devices, port numbers map to COM port numbers.
   * For example, serial port 2 maps to "COM2:".
   * <p>
   * Here is an example showing how to open the serial port of a
   * PalmPilot device at 9600 baud with settings of 8 bits,
   * no partity and one stop bit (8/N/1):
   * <pre>
   * SerialPort port = new SerialPort(0, 9600, 8, false, 1);
   * </pre>
   * Here is an example of opening serial port COM2: on a Windows device:
   * <pre>
   * SerialPort port = new SerialPort(2, 57600, 8, false, 1);
   * </pre>
   * No serial XON/XOFF flow control (commonly called software flow control)
   * is used and RTS/CTS flow control (commonly called hardware flow control)
   * is turn on by default on all platforms but Windows CE. The parity setting
   * is a boolean. If false, no parity is used. If true, "even" parity is used.
   *
   * @param number port number. In Windows, this is the number of the COM port.
   * @param baudRate baud rate
   * @param bits bits per char [5 to 8]
   * @param parity true for even parity, false for no parity
   * @param stopBits number of stop bits
   * @see #setFlowControl
   */
   public SerialPort(int number, int baudRate, int bits, boolean parity, int stopBits)
   {
      NativeMethods4JDK.serialPortCreate(this, number, baudRate, bits, parity, stopBits);
   }

   /**
   * Open a serial port with settings of 8 bits, no parity and 1 stop bit.
   * These are the most commonly used serial port settings.
   * @param number port number. In Windows, this is the number of the COM port.
   * @param baudRate baud rate
   */
   public SerialPort(int number, int baudRate)
   {
      this(number, baudRate, 8, false, 1);
   }

   /**
   * Closes the port. Returns true if the operation is successful
   * and false otherwise.
   */
   public boolean close()
   {
      return NativeMethods4JDK.serialPortClose(this);
   }


   /**
   * Returns true if the port is open and false otherwise. This can
   * be used to check if opening the serial port was successful.
   * This method does not clear the lastError flag.
   */
   public boolean isOpen()
   {
      return NativeMethods4JDK.serialPortIsOpen(this);
   }


   /**
   * Turns RTS/CTS flow control (hardware flow control) on or off.
   * @param on pass true to set flow control on and false to set it off
   */
   public boolean setFlowControl(boolean on)
   {
      return NativeMethods4JDK.serialPortSetFlowControl(this,on);
   }


   /**
   * Sets the timeout value for read operations. The value specifies
   * the number of milliseconds to wait from the time of last activity
   * before timing out a read operation. Passing a value of 0 sets
   * no timeout causing any read operation to return immediately with
   * or without data. The default timeout is 100 milliseconds. This
   * method returns true if successful and false if the value passed
   * is negative or the port is not open.
   * @param millis timeout in milliseconds
   */
   public boolean setReadTimeout(int millis)
   {
      return NativeMethods4JDK.serialPortSetReadTimeout(this, millis);
   }


   /**
   * Reads bytes from the port into a byte array. Returns the
   * number of bytes actually read or -1 if an error prevented the read
   * operation from occurring. The read will timeout if no activity
   * takes place within the timeout value for the port.
   * @param buf the byte array to read data into
   * @param start the start position in the byte array
   * @param count the number of bytes to read
   * @see #setReadTimeout
   */
   public int readBytes(byte buf[], int start, int count)
   {
      return NativeMethods4JDK.serialPortReadBytes(this, buf, start, count);
   }


   /**
   * Returns the number of bytes currently available to be read from the
   * serial port's queue. */
   public int readCheck()
   {
      return NativeMethods4JDK.serialPortReadCheck(this);
   }


   /**
   * Writes to the port. Returns the number of bytes written or -1
   * if an error prevented the write operation from occurring. If data
   * can't be written to the port and flow control is on, the write
   * operation will time out and fail after approximately 2 seconds.
   * @param buf the byte array to write data from
   * @param start the start position in the byte array
   * @param count the number of bytes to write
   */
   public int writeBytes(byte buf[], int start, int count)
   {
      return NativeMethods4JDK.serialPortWriteBytes(this, buf, start, count);
   }
}