/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>                         *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.io;

/**
 * Socket is a TCP/IP network socket.
 * <p>
 * Under Java and Windows CE, if no network is present, the socket
 * constructor may hang for an extended period of time due to the
 * implementation of sockets in the underlying OS. This is a known
 * problem.
 * <p>
 * Here is an example showing data being written and read from a socket:
 *
 * <pre>
 * Socket socket = new Socket("http://www.superwaba.org", 80);
 * if (!socket.isOpen())
 *   return;
 * byte []bytes = "GET / HTTP/1.0\n\n".getBytes();
 * socket.writeBytes(bytes,0,bytes.length);
 * byte buf[] = new byte[10];
 * int count = socket.readBytes(buf, 0, buf.length);
 * if (count == buf.length)
 *    ...
 * socket.close();
 * </pre>
 * Important: you cannot open a socket before the main event loop. In other
 * words, you cannot open a socket in the app's constructor, or even in the
 * onStart method. The socket can only be openned after the onStart method
 * finishes. You may setup a timer in the onStart that, when called, opens
 * the socket. Or simply open the socket after a button press.
 */

public class Socket extends Stream
{
   /** Contains the last error caused by a method call <b>on the PalmOS device only</b>. Below is the errors that this can return (from NetMgr.h):
   <pre>

   SW Custom codes (See nmpalm_wabaio.c for details):
   65526 (-10): INVALID_HOST
   65525 (-11): NO_NET_LIBRARY_AVAILABLE
   65524 (-12): NET_NOT_READY_FOR_OPEN
   65523 (-13): NET_NOT_OPENED_PROPERLY
   65522 (-14): ARRAY_RANGE_CHECK

   </pre>
   WindowsCE error codes: see winerror.h in the Windows SDK.

   Palm OS error codes.
   <p>
   Modem Error codes:
   <table border="1" cellpadding="3">
   <tr><td>NoTone           </td><td>4353</td></tr>
   <tr><td>NoCarrier        </td><td>4354</td></tr>
   <tr><td>LineBusy         </td><td>4355</td></tr>
   <tr><td>UserCancelled    </td><td>4356</td></tr>
   <tr><td>CmdError         </td><td>4357</td></tr>
   <tr><td>NoModem          </td><td>4358</td></tr>
   <tr><td>OutOfMemory      </td><td>4359</td></tr>
   <tr><td>PrefsNotSetup    </td><td>4360</td></tr>
   <tr><td>DialStringErr    </td><td>4361</td></tr>
   <tr><td>mdmErrNoPhoneNum </td><td>4362</td></tr>
   </table>

   <p>Net Error codes:
   <table border="1" cellpadding="3">
   <tr><td>AlreadyOpen                </td><td>4609</td></tr>
   <tr><td>NotOpen                    </td><td>4610</td></tr>
   <tr><td>StillOpen                  </td><td>4611</td></tr>
   <tr><td>ParamErr                   </td><td>4612</td></tr>
   <tr><td>NoMoreSockets              </td><td>4613</td></tr>
   <tr><td>OutOfResources             </td><td>4614</td></tr>
   <tr><td>OutOfMemory                </td><td>4615</td></tr>
   <tr><td>SocketNotOpen              </td><td>4616</td></tr>
   <tr><td>SocketBusy                 </td><td>4617</td></tr>
   <tr><td>MessageTooBig              </td><td>4618</td></tr>
   <tr><td>SocketNotConnected         </td><td>4619</td></tr>
   <tr><td>NoInterfaces               </td><td>4620</td></tr>
   <tr><td>BufTooSmall                </td><td>4621</td></tr>
   <tr><td>Unimplemented              </td><td>4622</td></tr>
   <tr><td>PortInUse                  </td><td>4623</td></tr>
   <tr><td>QuietTimeNotElapsed        </td><td>4624</td></tr>
   <tr><td>Internal                   </td><td>4625</td></tr>
   <tr><td>Timeout                    </td><td>4626</td></tr>
   <tr><td>SocketAlreadyConnected     </td><td>4627</td></tr>
   <tr><td>SocketClosedByRemote       </td><td>4628</td></tr>
   <tr><td>OutOfCmdBlocks             </td><td>4629</td></tr>
   <tr><td>WrongSocketType            </td><td>4630</td></tr>
   <tr><td>SocketNotListening         </td><td>4631</td></tr>
   <tr><td>UnknownSetting             </td><td>4632</td></tr>
   <tr><td>InvalidSettingSize         </td><td>4633</td></tr>
   <tr><td>PrefNotFound               </td><td>4634</td></tr>
   <tr><td>InvalidInterface           </td><td>4635</td></tr>
   <tr><td>InterfaceNotFound          </td><td>4636</td></tr>
   <tr><td>TooManyInterfaces          </td><td>4637</td></tr>
   <tr><td>BufWrongSize               </td><td>4638</td></tr>
   <tr><td>UserCancel                 </td><td>4639</td></tr>
   <tr><td>BadScript                  </td><td>4640</td></tr>
   <tr><td>NoSocket                   </td><td>4641</td></tr>
   <tr><td>SocketRcvBufFull           </td><td>4642</td></tr>
   <tr><td>NoPendingConnect           </td><td>4643</td></tr>
   <tr><td>UnexpectedCmd              </td><td>4644</td></tr>
   <tr><td>NoTCB                      </td><td>4645</td></tr>
   <tr><td>NilRemoteWindowSize        </td><td>4646</td></tr>
   <tr><td>NoTimerProc                </td><td>4647</td></tr>
   <tr><td>SocketInputShutdown        </td><td>4648</td></tr>
   <tr><td>CmdBlockNotCheckedOut      </td><td>4649</td></tr>
   <tr><td>CmdNotDone                 </td><td>4650</td></tr>
   <tr><td>UnknownProtocol            </td><td>4651</td></tr>
   <tr><td>UnknownService             </td><td>4652</td></tr>
   <tr><td>UnreachableDest            </td><td>4653</td></tr>
   <tr><td>ReadOnlySetting            </td><td>4654</td></tr>
   <tr><td>WouldBlock                 </td><td>4655</td></tr>
   <tr><td>AlreadyInProgress          </td><td>4656</td></tr>
   <tr><td>PPPTimeout                 </td><td>4657</td></tr>
   <tr><td>PPPBroughtDown             </td><td>4658</td></tr>
   <tr><td>AuthFailure                </td><td>4659</td></tr>
   <tr><td>PPPAddressRefused          </td><td>4660</td></tr>
   <tr><td>DNSNameTooLong             </td><td>4661</td></tr>
   <tr><td>DNSBadName                 </td><td>4662</td></tr>
   <tr><td>DNSBadArgs                 </td><td>4663</td></tr>
   <tr><td>DNSLabelTooLong            </td><td>4664</td></tr>
   <tr><td>DNSAllocationFailure       </td><td>4665</td></tr>
   <tr><td>DNSTimeout                 </td><td>4666</td></tr>
   <tr><td>DNSUnreachable             </td><td>4667</td></tr>
   <tr><td>DNSFormat                  </td><td>4668</td></tr>
   <tr><td>DNSServerFailure           </td><td>4669</td></tr>
   <tr><td>DNSNonexistantName         </td><td>4670</td></tr>
   <tr><td>DNSNIY                     </td><td>4671</td></tr>
   <tr><td>DNSRefused                 </td><td>4672</td></tr>
   <tr><td>DNSImpossible              </td><td>4673</td></tr>
   <tr><td>DNSNoRRS                   </td><td>4674</td></tr>
   <tr><td>DNSAborted                 </td><td>4675</td></tr>
   <tr><td>DNSBadProtocol             </td><td>4676</td></tr>
   <tr><td>DNSTruncated               </td><td>4677</td></tr>
   <tr><td>DNSNoRecursion             </td><td>4678</td></tr>
   <tr><td>DNSIrrelevant              </td><td>4679</td></tr>
   <tr><td>DNSNotInLocalCache         </td><td>4680</td></tr>
   <tr><td>DNSNoPort                  </td><td>4681</td></tr>
   <tr><td>IPCantFragment             </td><td>4682</td></tr>
   <tr><td>IPNoRoute                  </td><td>4683</td></tr>
   <tr><td>IPNoSrc                    </td><td>4684</td></tr>
   <tr><td>IPNoDst                    </td><td>4685</td></tr>
   <tr><td>IPktOverflow               </td><td>4686</td></tr>
   <tr><td>TooManyTCPConnections      </td><td>4687</td></tr>
   <tr><td>NoDNSServers               </td><td>4688</td></tr>
   <tr><td>InterfaceDown              </td><td>4689</td></tr>
   <tr><td>NoChannel                  </td><td>4690</td></tr>
   <tr><td>DieState                   </td><td>4691</td></tr>
   <tr><td>ReturnedInMail             </td><td>4692</td></tr>
   <tr><td>ReturnedNoTransfer         </td><td>4693</td></tr>
   <tr><td>ReturnedIllegal            </td><td>4694</td></tr>
   <tr><td>ReturnedCongest            </td><td>4695</td></tr>
   <tr><td>ReturnedError              </td><td>4696</td></tr>
   <tr><td>ReturnedBusy               </td><td>4697</td></tr>
   <tr><td>GMANState                  </td><td>4698</td></tr>
   <tr><td>QuitOnTxFail               </td><td>4699</td></tr>
   <tr><td>FlexListFull               </td><td>4700</td></tr>
   <tr><td>SenderMAN                  </td><td>4701</td></tr>
   <tr><td>IllegalType                </td><td>4702</td></tr>
   <tr><td>IllegalState               </td><td>4703</td></tr>
   <tr><td>IllegalFlags               </td><td>4704</td></tr>
   <tr><td>IllegalSendlist            </td><td>4705</td></tr>
   <tr><td>IllegalMPAKLength          </td><td>4706</td></tr>
   <tr><td>IllegalAddressee           </td><td>4707</td></tr>
   <tr><td>IllegalPacketClass         </td><td>4708</td></tr>
   <tr><td>BufferLength               </td><td>4709</td></tr>
   <tr><td>NiCdLowBattery             </td><td>4710</td></tr>
   <tr><td>RFinterfaceFatal           </td><td>4711</td></tr>
   <tr><td>IllegalLogout              </td><td>4712</td></tr>
   <tr><td>AAARadioLoad               </td><td>4713</td></tr>
   <tr><td>AntennaDown                </td><td>4714</td></tr>
   <tr><td>NiCdCharging               </td><td>4715</td></tr>
   <tr><td>AntennaWentDown            </td><td>4716</td></tr>
   <tr><td>NotActivated               </td><td>4717</td></tr>
   <tr><td>RadioTemp                  </td><td>4718</td></tr>
   <tr><td>NiCdChargeError            </td><td>4729</td></tr>
   <tr><td>NiCdSag                    </td><td>4720</td></tr>
   <tr><td>NiCdChargeSuspend          </td><td>4721</td></tr>
   <tr><td>reserved                   </td><td>4722</td></tr>
   <tr><td>ConfigNotFound             </td><td>4723</td></tr>
   <tr><td>ConfigCantDelete           </td><td>4724</td></tr>
   <tr><td>ConfigTooMany              </td><td>4725</td></tr>
   <tr><td>ConfigBadName              </td><td>4726</td></tr>
   <tr><td>ConfigNotAlias             </td><td>4727</td></tr>
   <tr><td>ConfigCantPointToAlias     </td><td>4728</td></tr>
   <tr><td>ConfigEmpty                </td><td>4729</td></tr>
   <tr><td>AlreadyOpenWithOtherConfig </td><td>4730</td></tr>
   <tr><td>ConfigAliasErr             </td><td>4731</td></tr>
   <tr><td>NoMultiPktAddr             </td><td>4732</td></tr>
   <tr><td>OutOfPackets               </td><td>4733 </td></tr>
   <tr><td>MultiPktAddrReset          </td><td>4734</td></tr>
   <tr><td>StaleMultiPktAddr          </td><td>4735</td></tr>
   <tr><td>ScptPluginMissing          </td><td>4736</td></tr>
   <tr><td>ScptPluginLaunchFail       </td><td>4737</td></tr>
   <tr><td>ScptPluginCmdFail          </td><td>4738</td></tr>
   <tr><td>ScptPluginInvalidCmd       </td><td>4739</td></tr>
   <tr><td>TelMissingComponent        </td><td>4740</td></tr>
   <tr><td>TelErrorNotHandled         </td><td>4741</td></tr>
   </table>
   */
   public int lastError; // guich@230_27
   /** If set to true, the function NetLibConnectionRefresh will be called
     * before each read/write. Note that in some tests this makes the
     * read/write operation 3 times slower!
     */
   public boolean refreshBeforeEachRead=false; // guich@340_37

   // HOOK VARIABLES - can't be private - guich@350_20

   /** On desktop, true if is open. On device, stores the SocketRef */
   boolean _isOpen;
   /** On desktop, stores the socket itself. On device, stores the read timeout */
   Object _socket;
   /** Used just in the desktop side */
   Object _in;
   /** Used just in the desktop side */
   Object _out;

   /** For internal use only */
   protected Socket()
   {
   }

   /**
   * Opens a socket. This method establishes a socket connection by
   * looking up the given host and performing the 3 way TCP/IP handshake.
   * The default timeout for opening the connection is 1.5 seconds and the default noLinger is false.
   * @param host the host name or IP address to connect to
   * @param port the port number to connect to
   */
   public Socket(String host, int port)
   {
      this(host, port, 1500, false);
   }

   /**
   * Opens a socket. This method establishes a socket connection by
   * looking up the given host and performing the 3 way TCP/IP handshake.
   * The default noLinger is false;
   * @param host the host name or IP address to connect to
   * @param port the port number to connect to
   * @param timeout the timeout (in ms) for opening the socket.
   */
   public Socket(String host, int port, int timeout)
   {
      NativeMethods4JDK.socketCreate(this, host, port, timeout, false);
   }

   /**
   * Opens a socket. This method establishes a socket connection by
   * looking up the given host and performing the 3 way TCP/IP handshake.
   * @param host the host name or IP address to connect to
   * @param port the port number to connect to
   * @param timeout the timeout (in ms) for opening the socket. This should be 1500.
   * @param noLinger if true, the socket is closed immediately, and no ack is waited from the server. Note that this must be done for the very first socket creation per application.
   */
   public Socket(String host, int port, int timeout, boolean noLinger) // guich@330_18
   {
      NativeMethods4JDK.socketCreate(this, host, port, timeout, noLinger);
   }

   /**
   * Closes the socket. Returns true if the operation is successful
   * and false otherwise.
   */
   public boolean close()
   {
      return NativeMethods4JDK.socketClose(this);
   }


   /**
   * Returns true if the socket is open and false otherwise. This can
   * be used to check if opening the socket was successful.
   * This method does not clear the lastError flag.
   */
   public boolean isOpen()
   {
      return NativeMethods4JDK.socketIsOpen(this);
   }


   /**
   * Sets the timeout value for read operations. The value specifies
   * the number of milliseconds to wait from the time of last activity
   * before timing out a read operation. Passing a value of 0 sets
   * no timeout causing any read operation to return immediately with
   * or without data. The default timeout is 5000 milliseconds. This
   * method returns true if successful and false if the value passed
   * is negative or the socket is not open. The read timeout will have
   * effect in the next read instruction.
   * @param millis timeout in milliseconds.
   */
   public boolean setReadTimeout(int millis)
   {
      return NativeMethods4JDK.socketSetReadTimeout(this, millis);
   }


   /**
   * Reads bytes from the socket into a byte array. Returns the
   * number of bytes actually read or -1 if the server closed
   * the connection or an error prevented the read operation from
   * occurring. Note that in the device it may return with less bytes
   * than requested. Note also that, if -1 is returned and the lastError
   * is 4626, please insist 3 or 4 more times, because it may work.
   * @param buf the byte array to read data into
   * @param start the start position in the byte array
   * @param count the number of bytes to read
   */
   public int readBytes(byte buf[], int start, int count)
   {
      return NativeMethods4JDK.socketReadBytes(this, buf, start, count);
   }


   /**
   * Writes to the socket. Returns the number of bytes written or -1
   * if an error prevented the write operation from occurring. If data
   * can't be written to the socket for approximately 2 seconds, the
   * write operation will time out.
   * @param buf the byte array to write data from
   * @param start the start position in the byte array
   * @param count the number of bytes to write
   */
   public int writeBytes(byte buf[], int start, int count)
   {
      return NativeMethods4JDK.socketWriteBytes(this, buf, start, count);
   }

   /** Disconects the socket. This may hangup the modem or disconect the
     * attached device */
   static public boolean disconnect() // guich@300_63
   {
      return NativeMethods4JDK.socketDisconnect();
   }
}
