/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>                         *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.ui;

import waba.fx.*;
import waba.sys.*;

/**
 * Button is a push button control.
 * <p>
 * Here is an example showing a push button being used:
 *
 * <pre>
 * public class MyProgram extends MainWindow
 * {
 * Button pushB;
 *
 * public void onStart()
 * {
 *    add(pushB = new Button("Push me"), CENTER,TOP); // PS: you can also create an image button with pushB = new Button(new Image("myimage.bmp"));
 * }
 *
 * public void onEvent(Event event)
 * {
 *   if (event.type == ControlEvent.PRESSED &&
 *      event.target == pushB)
 *   {
 *   ... handle pushB being pressed
 * </pre>
 */

public class Button extends Control
{
   public static final byte BORDER_NONE = 0;
   public static final byte BORDER_SIMPLE = 1;
   public static final byte BORDER_3D = 2;
   protected String text;
   protected Image img;
   protected boolean armed;
   protected boolean isText;
   protected int gap;
   protected byte border = BORDER_3D;
   protected int ww,hh;
   protected Color fColor,imgColor,aColor;
   protected Graphics myg;
   protected Color fourColors[] = new Color[4];
   protected boolean ignoreEvents;
   /** Set commonGap to a value to make all further buttons with the same internal
     * gap. This is useful when you're creating the buttons but don't want to
     * keep calling the <code>setGap</code> method.
     * Note that this value will be added to the value of <code>gap</code>.
     * Remember to save the current value and restore it when done.
     */
   public static int commonGap; // guich@300_3

   /** Creates a button displaying the given text. */
   public Button(String text)
   {
      this.text = text;
      isText = true;
      img = null; // release any setted image
   }

   /** Sets the text that is displayed in the button. */
   public void setText(String text)
   {
      this.text = text;
      isText = true;
      img = null;
      repaint();
   }

   /** Creates a button with the given image. The transparentColor property of the Image must be set before calling this constructor. */
   public Button(Image img)
   {
      this.img = img;
      imgColor = img.transparentColor;
      isText = false;
      text = null;
   }

   /** Make this button act like an image control. The button will have no border and will ignore all events */
   public void onlyShowImage()
   {
      this.border = BORDER_NONE;
      ignoreEvents = true;
   }

   /** Sets the style of the border. use the constants BORDER_xxx */
   public void setBorder(byte border)
   {
      this.border = border;
   }
   /** Sets the gap between the text/image and the border. The default is 1. used when the size is set using the PREFERRED constant. */
   public void setGap(int gap)
   {
      this.gap = gap;
   }

   /** Gets the text displayed in the button. */
   public String getText()
   {
      return isText?text:"image";
   }

   /** returns the preffered width of this control. added by guich */
   public int getPreferredWidth()
   {
      return (isText?fm.getTextWidth(text):img.getWidth()) + ((border + gap + commonGap) << 1);
   }

   /** returns the preffered height of this control. added by guich */
   public int getPreferredHeight()
   {
      return isText?(fmH+((commonGap+gap) << 1)):(img.getHeight()+((border + gap + commonGap) << 1));
   }

   /** Called by the system to pass events to the button. */
   public void onEvent(Event event)
   {
      if (ignoreEvents) return;
      PenEvent pe;
      switch (event.type)
      {
         case ControlEvent.WINDOW_MOVED:
            if (myg != null) myg.free();
            myg = createGraphics();
            break;
         case PenEvent.PEN_DOWN:
            armed = true;
            repaint();
            break;
         case PenEvent.PEN_UP:
            armed = false;
            repaint();
            pe = (PenEvent)event;
            if (0 <= pe.x && pe.x < this.width && 0 <= pe.y && pe.y < this.height)
               postEvent(new ControlEvent(ControlEvent.PRESSED, this));
            break;
         case PenEvent.PEN_DRAG:
            pe = (PenEvent)event;
            boolean lArmed = false;
            if (0 <= pe.x && pe.x < this.width && 0 <= pe.y && pe.y < this.height)
               lArmed = true;
            if (armed != lArmed)
            {
               armed = lArmed;
               repaint();
            }
            break;
         case ControlEvent.FOCUS_OUT:
            armed = false;
      }
   }

   /** simulate the press or release of this button. does not generate events */
   public void press(boolean pressed)
   {
      myg.clearClip();
      paintButton(myg,pressed);
   }

   /** Called by the system to draw the button. it cuts the text if the button is too small. */
   public void onPaint(Graphics g)
   {
      if (myg == null) myg = createGraphics();
      paintButton(g,armed);
   }

   protected void onFontChanged()
   {
      onBoundsChanged();
   }

   protected void onBoundsChanged()
   {
      if (isText)
      {
         if (text != null)
         {
            ww = (width - fm.getTextWidth(text)) >> 1;
            hh = (height - fmH) >> 1;
         }
      }
      else
      {
         if (img != null)
         {
            ww = (width-img.getWidth()) >> 1;
            hh = (height-img.getHeight()) >> 1;
         }
      }
      myg = createGraphics(); // guich@350_25: create a new myg
   }

   protected void onColorsChanged(boolean colorsChanged)
   {
      fColor = getForeColor();
      if (Settings.uiStyle == Settings.PalmOS) aColor = backColor.getCursorColor();
      Graphics.compute3dColors(enabled,backColor,foreColor,fourColors);
   }

   protected void paintButton(Graphics g, boolean armed)
   {
      int xx=ww;
      int yy=hh;
      if (!(ignoreEvents && img != null)) // guich@310_25: changed to support transparent color
      {
         if (Settings.uiStyle == Settings.WinCE)
         {
            g.setBackColor(backColor);
            g.fillRect(0,0,width,height);
         }
         else
         {
            g.setBackColor(armed?aColor:backColor);
            if (border == BORDER_NONE)
               g.fillRect(0,0,width,height);
            else
               g.fillHatchedRect(0,0,width,height,true,true);
         }
         if (border != BORDER_NONE)
            g.draw3dRect(0,0,width,height,armed?Graphics.R3D_LOWERED:Graphics.R3D_RAISED,false,border == BORDER_SIMPLE,fourColors);

         g.setClip(border,border,width-(border<<1),height-(border<<1)); // guich@101: cut text if button is too small
         if (armed && Settings.uiStyle == Settings.WinCE)
         {
            xx++;
            yy++;
         }

         g.setForeColor(fColor);
         if (isText)
            g.drawText(text, xx, yy); // draw label
      }
      if (!isText)
      {
         if (imgColor != null) // guich@310_26: allow imgColor be null
         {
            g.setBackColor(imgColor); // the transparent color is our background color
            g.setDrawOp(enabled?Graphics.DRAW_SPRITE:Graphics.DRAW_REPLACE_COLOR);  // draw the image transparent _OR_ draw the image making all pixels != background equal to our disabled color
         }
         else g.setDrawOp(enabled?Graphics.DRAW_PAINT:Graphics.DRAW_REPLACE_COLOR);
         g.drawImage(img,xx,yy);
         g.setDrawOp(Graphics.DRAW_PAINT);
      }
   }

   /** just a handy method to create buttons with an arrow drawed.
    * @param byte Graphics.ARROW_UP or Graphics.ARROW_DOWN or Graphics.ARROW_LEFT or Graphics.ARROW_RIGHT
    * @param h same of Graphics.drawArrow
    * @param fore the foreground color for the arrow. the background color will be the required transparent color.
    * @see waba.fx.Graphics#drawArrow(int, int, int, byte, boolean, boolean, Color) */
   public static Button createArrowButton(byte type, int h, Color fore)
   {
      boolean rotate = type == Graphics.ARROW_RIGHT || type == Graphics.ARROW_LEFT;
      int w = (h<<1)-1;
      Image img = rotate?new Image(h,w):new Image(w,h);
      Graphics g = img.getGraphics();
      g.setBackColor(img.transparentColor);
      if (rotate)
         g.fillRect(0,0,h,w);
      else
         g.fillRect(0,0,w,h);
      g.drawArrow(0,0,h,type,false,true,fore); // here is h regardless the case
      return new Button(img);
   }

   /** Returns the image that is assigned to this Button */
   public Image getImage()
   {
      return isText?null:img;
   }
}
