/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>                         *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.ui;

import waba.fx.*;

/**
 * Container is a control that contains child controls.
 */

public class Container extends Control
{
   /** The children of the container. */
   protected Control children;
   /** The tail of the children list. */
   protected Control tail;
   /** The type of border of this Container */
   private byte borderStyle = BORDER_NONE;
   private Color []fourColors = new Color[4];

   /** used in the setBorderStyle method */
   static public final byte BORDER_NONE=0;
   /** used in the setBorderStyle method */
   static public final byte BORDER_LOWERED=2;
   /** used in the setBorderStyle method */
   static public final byte BORDER_RAISED=3;
   /** used in the setBorderStyle method */
   static public final byte BORDER_SIMPLE=5;

   protected Window parentWindow;

   protected int lastX=-1,lastY=-1,lastW,lastH; // guich@200b4_100
   private int numChildren;
   boolean started; // guich@340_15

   /** Creates a container with the default colors.
     * Important note: this container has no default size. If adding
     * to the MainWindow as another container, you may set its size as
     * <code>MainWindow.getMainWindow().getClientRect()</code>.
     * <p>
     * If you're extending the Container class and adding controls in its
     * constructor, you may come into problems if you don't set the bounds
     * as the first thing. This is easy if you're just adding the controls
     * to a MainWindow. But if you're adding to a TabPanel, for example,
     * there is no way to know the TabPanel's size. The solution for this
     * problem is to pass in the constructor of your Container class the
     * Rect with the desired bounds, as demonstrated in the examples/ui/ContainerSwitch
     */ // guich@300_58
   public Container()
   {
      foreColor = Color.defaultForeColor;
      backColor = Color.defaultBackColor;
      foreDis = foreColor.brighter();
      backDis = backColor.darker();
      asContainer = this;
   }

   /** Add the control to this container and set its rect
     * to be the given x,y and PREFERRED as width/height
     * relative to the last control added
     */
   public void add(Control control, int x, int y)
   {
      add(control);
      control.setRect(x,y,PREFERRED,PREFERRED, null);
   }

   /** Add the control to this container and set its rect
     * to be the given x,y and PREFERRED as width/height
     * relative to the given control
     */
   public void add(Control control, int x, int y, Control relative) // guich@200b4_138
   {
      add(control);
      control.setRect(x,y,PREFERRED,PREFERRED, relative);
   }

   /**
   * Adds a child control to this container.
   * <b>Important</b>: If you're swapping containers from the MainWindow, be sure
   * that you set the focus on the new container after calling this add method.
   * Otherwise, a MenuBar will not work. Or, use the handy method Window.swap
   */
   public void add(Control control)
   {
      if (control.parent != null)
         control.parent.remove(control);
      // set children, next, prev, tail and parent
      control.next = null;
      if (children == null)
         children = control;
      else
         tail.next = control;
      control.prev = tail;
      tail = control;
      control.parent = this;
      if (control.backColor == null) {control.backColor = this.backColor; control.backDis = this.backDis;} // guich@200b4_125
      if (control.foreColor == null) {control.foreColor = this.foreColor; control.foreDis = this.foreDis;} // guich@200b4_125
      numChildren++;
      if (control.asContainer != null)
         control.asContainer.parentWindow = parentWindow;
      control.onColorsChanged(true);
      control.repaint();
   }

   /**
   * Removes a child control from the container.
   */
   public void remove(Control control)
   {
      if (control.parent != this)
         return;
      // set children, next, prev, tail and parent
      Control prev = control.prev;
      Control next = control.next;
      if (prev == null)
         children = next;
      else
         prev.next = next;
      if (next != null)
         next.prev = prev;
      if (tail == control)
         tail = prev;
      control.next = null;
      control.prev = null;
      numChildren--;
      control.repaint(); // invalidate the control's area
      repaint(); // guich@200b4_16: invalidate the hole container's area
      control.parent = null;
      if (getParentWindow().getFocus() == control) requestFocus(); // guich200b4_16: set the focus to this container so that if a control is removed, it also looses focus
   }

   /** Returns the child located at the given x and y coordinates. */
   public Control findChild(int x, int y)
   {
      Container container;
      Control child;

      container = this;
      while (true)
      {
         // search tail to head since paint goes head to tail
         child = container.tail;
         while (child != null && (!child.contains(x, y) || !child.visible)) // guich@240_8: fixed when adding two controls in the same location but one of them was not visible (!child.visible)
            child = child.prev;
         if (child == null)
            return container;
         if (child.asContainer == null)
            return child;
         x -= child.x;
         y -= child.y;
         container = child.asContainer;
      }
   }

   /** return an array of Controls that are added to this Container. if theres no Controls, returns null. */
   public Control []getChildren()
   {
      if (numChildren == 0) return null;
      Control []ac = new Control[numChildren];
      Control child = this.tail;
      for (int i=0; child != null; i++,child = child.prev)
         ac[i] = child;
      return ac;
   }
   /** sets if this container and all childrens can or not accept events */
   public void setEnabled(boolean enabled)
   {
      if (enabled != this.enabled)
      {
         this.enabled = enabled;
         onColorsChanged(false);
         Control child = children;
         while (child != null)
         {
            child.setEnabled(enabled);
            child = child.next;
         }
         repaint(); // now the controls have different l&f for disabled states
      }
   }

   /** posts a event to the children of this container and to all containers inside this containers; recursively.
   @since SuperWaba2.0beta4 */
   public void broadcastEvent(Event e)
   {
      onEvent(e); // guich@200b4_110: make sure this container receive this event.
      Control child = children;
      while (child != null)
      {
         child.onEvent(e);
         if (child.asContainer != null)
            child.asContainer.broadcastEvent(e);
         child = child.next;
      }
   }

   /** Called by the system to draw the children of the container. */
   public void paintChildren(Graphics g, int x, int y, int width, int height)
   {
      Control child = children;
      int xx2 = x + width - 1;
      int yy2 = y + height - 1;
      int x1,y1,x2,y2,cx1,cy1,cx2,cy2;
      while (child != null)
      {
         if (child.visible) // guich@200: ignore hidden controls - note: a window added to a container may not be painted correctly
         {
            // trivial clip
            if (!((x2 = xx2) < (cx1 = child.x) || (x1 = x) > (cx2 = child.x2)
               || (y2 = yy2) < (cy1 = child.y) || (y1 = y) > (cy2 = child.y2)))
            {
               if (x1 < cx1) x1 = cx1;
               if (y1 < cy1) y1 = cy1;
               if (x2 > cx2) x2 = cx2;
               if (y2 > cy2) y2 = cy2;
               g.setClip(x1, y1, x2 - x1 + 1, y2 - y1 + 1);
               g.translate(cx1, cy1);
               if (child.font != g.lastFont) // the correct should be use !child.font.equals(g.lastFont), but this comparision is faster
                  g.setFont(child.font);
               child.onPaint(g);
               if (child.asContainer != null)
                  child.asContainer.paintChildren(g, x1 - cx1, y1 - cy1, x2 - x1 + 1, y2 - y1 + 1);
               g.translate(- cx1, - cy1);
            }
         }
         child = child.next;
      }
      g.clearClip();
   }

   /** Sets the border for this container. Use BORDER_NONE,BORDER_LOWERED,BORDER_RAISED,BORDER_SIMPLE. */
   public void setBorderStyle(byte border) // guich@200final_16
   {
      this.borderStyle = border;
      onColorsChanged(false);
   }

   /** returns the client rect for this Container, in relative coords */
   public Rect getClientRect() // guich@200final_15
   {
      int gap = borderStyle == BORDER_NONE ? 0 :
                borderStyle == BORDER_SIMPLE ? 1 : 2;
      return new Rect(gap,gap,width-gap-gap,height-gap-gap);
   }

   protected void onColorsChanged(boolean colorsChanged)
   {
      if (borderStyle != BORDER_NONE && borderStyle != BORDER_SIMPLE)
         Graphics.compute3dColors(enabled, backColor, foreColor, fourColors);
   }

   /** Draws the border (if any). If you override this method, be sure to call
     * <code>super.onPaint(g);</code>, or the border will not be drawn.
     */
   public void onPaint(Graphics g)
   {
      if (parent != null && backColor != parent.backColor) // guich@300_6
      {
         g.setBackColor(backColor);
         g.fillRect(0,0,width,height);
      }
      switch (borderStyle)
      {
         case BORDER_NONE:
            break;

         case BORDER_SIMPLE:
            g.setForeColor(getForeColor());
            g.drawRect(0,0,width,height);
            break;

         default:
            g.draw3dRect(0,0,width,height,borderStyle,false,false,fourColors);
      }
   }

   /** Called when this container can be initialized
     * @since SuperWaba 3.4
     */
   protected void onStart() // guich@340_15
   {
   }
}
