/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>                         *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.ui;

import waba.fx.*;
import waba.sys.*;
import waba.util.*;

/**
 * Edit is a text entry control.
 * <p>
 * Here is an example showing an edit control being used:
 *
 * <pre>
 * public class MyProgram extends MainWindow
 * {
 * Edit edit;
 *
 * public void onStart()
 *  {
 *  edit = new Edit(""); // mask used to compute preferred width - empty = fill width
 *  add(edit,LEFT,TOP);
 *  }
 * </pre>
 */

public class Edit extends Control
{
   /** Defines if the virtual keyboard can be moved in Windows CE devices.   
     * Some devices have problems with this, so setting this to false will 
     * make the keyboard always appear at the default position. 
     * The default is true.
     * @since SuperWaba 4.01
     */
   public static boolean canMoveVirtualKeyboard = true;
   private Graphics drawg; // only valid while the edit has focus
   private Timer blinkTimer; // only valid while the edit has focus

   private boolean hasFocus = false;
   private boolean cursorShowing = false;
   /** Specifies if the control accepts input from the user. */
   public boolean editable = true;
   
   private Rect clipRect=new Rect();
   private char chars[] = new char[4];
   private byte charWidths[] = new byte[4];
   private int totalCharWidth = 0;
   private int len = 0;
   private int xMax,xMin,gap;
   private Color fColor,back0,back1;
   private Color fourColors[] = new Color[4];

   private int insertPos;
   private int startSelectPos;
   private int xOffset;

   private String validChars; // guich
   public static Keyboard keyboard; // guich
   public static Calendar calendar; // guich
   public static Calculator calculator; // guich@200
   byte mode; // guich
   private int maxLength; // guich@200b4
   /** used only to compute the preferred width of this edit. If the mask is empty, the edit fills to width. Can be accessed directly */
   public String mask;
   private byte capitalise; // guich@320_26

   private byte kbdType=KBD_DEFAULT;

   /** Set to false if you don't want the cursor to blink when the edit is not editable */
   public boolean hasCursorWhenNotEditable = true; // guich@340_23

   /** No keyboard will be popped up for this Edit */
   public static final byte KBD_NONE = 0;
   /** The default keyboard for the current mode will be used for this Edit */
   public static final byte KBD_DEFAULT = 1;
   /** The Keyboard class (or the internal virtual keyboard) will be used for this Edit */
   public static final byte KBD_KEYBOARD = 2;
   /** The Calculator will be used for this Edit */
   public static final byte KBD_CALCULATOR = 3;
   /** The Calendar will be used for this Edit */
   public static final byte KBD_CALENDAR = 4;

   /** to be used in the setValidChars method */
   public static final String numbersSet = "0123456789";
   /** to be used in the setValidChars method */
   public static final String currencyCharsSet = "0123456789.+-";
   /** to be used in the setValidChars method */
   public static final String dateSet = numbersSet+Settings.dateSeparator;

   /** to be used in the setMode method */
   public static final byte NORMAL   = 0;
   /** to be used in the setMode method */
   public static final byte DATE     = 1;
   /** to be used in the setMode method */
   public static final byte CURRENCY = 2;
   /** to be used in the setMode method. The last char will be always shown */
   public static final byte PASSWORD = 3;
   /** to be used in the setMode method. All chars are replaced by '*' */
   public static final byte PASSWORD_ALL = 4;

   /** to be used in the setCapitaliseMode method */
   public static final byte ALL_NORMAL = 0;

   /** to be used in the setCapitaliseMode method */
   public static final byte ALL_UPPER = 1;

   /** to be used in the setCapitaliseMode method */
   public static final byte ALL_LOWER = 2;

   private boolean isCommand;

   /** Construct an Edit with a default width of 10 chars. */
   public Edit()
   {
      clearPosState();
   }
   /** Construct an Edit with the default width computed based in the specified
     * mask and in the control's font. Note that the mask does not <i>masks</i>
     * the input. If mask is "", the FILL width is choosen.
     */
   public Edit(String mask)
   {
      this();
      this.mask = mask;
   }

   /** Sets the capitalise settings for this Edit. Use ALL_NORMAL,
     * ALL_UPPER and ALL_LOWER. Text entered will made as is, uppercase or
     * lowercase
     */
   public void setCapitaliseMode(byte cap)
   {
      this.capitalise = cap;
   }

   /** Used to change the default keyboard to be used with this Edit control.
     * Use the constants KBD_xxxx.
     * Note that setMode calls setKeyboard(KBD_DEFAULT), so be sure to set the mode before calling setKeyboard. */
   public void setKeyboard(byte kbd) // guich@310_19
   {
      this.kbdType = kbd;
      if (kbd == KBD_DEFAULT)
         switch (mode)
         {
            case DATE:     kbdType = KBD_CALENDAR;   break;
            case CURRENCY: kbdType = KBD_CALCULATOR; break;
            default:       kbdType = KBD_KEYBOARD;   break;
         }
   }

   /** Used to set the valid characters that can be entered by using one of the constants DATE, CURRENCY and NORMAL.
     * Note that setMode calls setKeyboard(KBD_DEFAULT), so be sure to set the mode before calling setKeyboard. */
   public void setMode(byte mode)
   {
      this.mode = mode;
      switch (mode)
      {
         case DATE:      setValidChars(dateSet);          break;
         case CURRENCY:  setValidChars(currencyCharsSet); break;
         default:        setValidChars(null);             break;
      }
      setKeyboard(KBD_DEFAULT);
   }

   /** Sets the valid chars that can be entered in this edit (they are converted to uppercase to make the verification easy). if null is passed, any char can be entered. (case insensitive). added by guich */
   public void setValidChars(String validCharsString)
   {
      if (validCharsString != null)
         validChars = Convert.toUpperCase(validCharsString);
      else
         validChars = null;
   }

   /** Return true if the given char exists in the set of valid characters for this Edit */
   protected boolean isCharValid(char c)
   {
      if (validChars == null) return true;
      return validChars.indexOf(Convert.toUpperCase(c)) != -1;
   }

   /** Sets the desired maximum length for text entered in the Edit.
   @since SuperWaba2.0beta4 */
   public void setMaxLength(int length)
   {
      this.maxLength = length;
   }

   private void clearPosState()
   {
      insertPos = 0;
      startSelectPos = -1;
      xOffset = (Settings.uiStyle == Settings.PalmOS)?1:4;
   }

   protected int pushedInsertPos;
   protected int pushedStartSelectPos;
   protected int pushedxOffset;
   protected void pushPosState()
   {
      pushedInsertPos = insertPos;
      pushedStartSelectPos = startSelectPos;
      pushedxOffset = xOffset;
   }
   protected void popPosState()
   {
      if (cursorShowing)
         draw(drawg,true);
      insertPos = pushedInsertPos;
      startSelectPos = pushedStartSelectPos;
      xOffset = pushedxOffset;
   }
   protected int x2charPos(int x)
   {
      int cx = xOffset;
      for (int i = 0; i < len; i++)
      {
         int cw = charWidths[i];
         if (x <= cx + (cw >> 1)) // cw / 2
            return i;
         cx += cw;
      }
      return len;
   }

   protected int charPos2x(int n)
   {
      int cx;
      if (n > len)
         n = len;

      cx = xOffset;
      for (int i = 0; i < n; i++)
         cx += charWidths[i];

      return cx;
   }

   /**
   * Returns the text displayed in the edit control.
   */
   public String getText()
   {
      return new String(chars, 0, len);
   }

   /**
   * Sets the text displayed in the edit control.
   */
   public void setText(String s)
   {
      if (s != null)
      {
         chars = s.toCharArray();
         len = chars.length;
      }
      else
      {
         chars = null;
         len = 0;
      }
      charWidths = null;
      totalCharWidth = 0;
      clearPosState();
      repaint();
   }

   /** Sets if the control accepts input from the user 
     * deprecated Access the editable member directly
     */
   public void setEditable(boolean on)
   {
      editable = on;
   }
   /** Gets if the control accepts input from the user 
     * deprecated Access the editable member directly
     */
   public boolean isEditable()
   {
      return editable;
   }

   protected void onBoundsChanged() // guich
   {
      drawg = null;
      gap =  (Settings.uiStyle == Settings.PalmOS)?0:2;
      xMin = (Settings.uiStyle == Settings.PalmOS)?1:4;
      xMax = this.width - xMin;
      xMax -= 2;
   }

   public int getPreferredWidth()
   {
      return mask==null?(waba.sys.Settings.screenWidth>>2):(mask.length()==0)?FILL:(fm.getTextWidth(mask) + ((Settings.uiStyle == Settings.PalmOS)?4:10)); // guich@200b4_202: from 2 -> 4 is PalmOS style - guic@300_52: empty mask means FILL
   }

   public int getPreferredHeight()
   {
      return fmH+((Settings.uiStyle == Settings.PalmOS)?0:4); // old style dont have borders
   }

   protected void onColorsChanged(boolean colorsChanged)
   {
      fColor = getForeColor();
      back0  = getBackColor().brighter();
      back1  = back0.equ != Color.WHITE.equ?backColor:back0.getCursorColor();//guich@300_20: use backColor instead of: back0.getCursorColor();
      Graphics.compute3dColors(enabled,backColor,foreColor,fourColors);
   }

   protected void draw(Graphics g, boolean cursorOnly)
   {
      if (g == null) return;
      int y = this.height - fmH - gap;

      if (!cursorOnly) // guich@200b4_23: optimized when cursorOnly
      {
         char []showchars = chars;
         if (mode >= PASSWORD && chars.length > 0 && len > 0) // in password mode, only the last char is visible
         {
            showchars = new char[chars.length];
            boolean includeLast = mode == PASSWORD_ALL; // guich@310_20
            int n = Math.min(includeLast?len:(len-1), chars.length);   // clim@200b4_176: another fix, just ensure the loop cannot get beyond chars.length, not sure if needed
            int w = fm.getCharWidth('*');
            for (int i =0; i < n; i++)
               showchars[i] = '*';
            // clim@200b4_176: next 3 lines replace the original that set:  charWidths[i] =(bytes) w in the above for loop (2 lines up)
            if (charWidths != null)
               for (int i=Math.min(charWidths.length, n)-1; i >= 0; i--)
                    charWidths[i] = (byte) w;
            if (includeLast)
            {
               if (charWidths != null && charWidths.length > n)   // change the if condition, added check for length > n, just in case
                  totalCharWidth = n*w;
            }
            else
            {
               showchars[n] = chars[n];
               if (charWidths != null && charWidths.length > n)   // change the if condition, added check for length > n, just in case
               {
                  charWidths[n] = (byte)fm.getCharWidth(chars[n]);
                  totalCharWidth = n*w+charWidths[n];
               }
            }
         }
         if (charWidths == null)
         {
            charWidths = new byte[len];
            totalCharWidth = 0;
            for (int i = 0; i < len; i++)
            {
               int charWidth = fm.getCharWidth(showchars[i]);
               charWidths[i] = (byte)charWidth;
               totalCharWidth += charWidth;
            }
         }

         int h = this.height - ((Settings.uiStyle == Settings.PalmOS)?0:3);
         // get current clip rect and intersect with edit rect to set
         // a new clip to draw in
         clipRect.set(0,0,this.width,this.height);//xMin,y,xMax,h); // avoid keep recreating the rect.
         g.setClip(clipRect); // intersect the graphics clipping bounds with this one. - guich@400_78: changed clip to setClip. clip was not intercecting the clip correctly
         g.setBackColor(back0);
         g.fillRect(gap,gap, this.width - (gap << 1), this.height - (gap << 1));
         // draw the text and/or the selection
         if (len > 0)
         {
            g.setForeColor(fColor);
            g.drawText(showchars, 0, len, xOffset, y);
            if (startSelectPos != -1)
            {
               // character regions are:
               // 0 to (sel1-1) .. sel1 to (sel2-1) .. sel2 to last_char
               int sel1 = Math.min(startSelectPos,insertPos);
               int sel2 = Math.max(startSelectPos,insertPos);
               int sel1X = charPos2x(sel1);
               int sel2X = charPos2x(sel2);

               g.setForeColor(back0);
               g.setBackColor(back1);
               g.eraseRect(sel1X,y,sel2X-sel1X,fmH);
            }
         }
         g.clearClip(); // guich@400_40: moved to here bc when drawing PalmOS edit control, the dotted line is erased. //setClip(clipRect);
         g.draw3dRect(0,0,this.width,this.height,Graphics.R3D_EDIT,false,false,fourColors); // draw the border and erase the rect
      }
      if (hasFocus && (editable || hasCursorWhenNotEditable))
      {
         // draw cursor
         int xx = charPos2x(insertPos);
         if (xMin <= xx && xx <= xMax) // guich@200b4_155
         {
            g.clearClip();
            g.drawCursor(xx - 1, y, 1, fmH);
         }
         cursorShowing = cursorOnly ? !cursorShowing : true;
      }
      else
         cursorShowing = false;
   }

   /** Sets the selected text of this Edit (if start != end). Can be used to set the cursor position,
     * if start equals end. Start must be less or equal to end, and both must be >= 0.
     * It can also be used to clear the selectedText, calling <code>setCursorPos(-1,0)</code>.
     * Note: if you're setting the cursor position before the edit is drawn for the first
     * time, the edit will not be scrolled if the end position goes beyond the limits.
     * Important! No bounds checking is made. Be sure to not call this method with invalid positions!
     * Example:
     * <pre>
     * ed.setText("1234567890123456");
     * ed.setCursorPos(3,14);
     * ed.requestFocus();
     * </pre>
     */
   public void setCursorPos(int start, int end) // guich@400_18
   {
      startSelectPos = (start != end)?start:-1;
      insertPos = end;
      if (charWidths != null)
      {
         int x = charPos2x(end);
         if (x + 3 > xMax)
         {
            // characters hidden on right - jump
            xOffset -= (x - xMax);
            if (xOffset < xMax - totalCharWidth)
               xOffset = xMax - totalCharWidth;
         }
      }
      repaint();
   }

   /** Returns an array with the cursor positions. You can use it with getText
     * to find the selected text String.
     * E.g.:
     * <pre>
     * int []cursorPos = ed.getCursorPos();
     * int start = cursorPos[0];
     * int end = cursorPos[1];
     * String text = ed.getText();
     * if (start != -1) // is the text selected?
     * {
     *    String selectedText = text.substring(start,end);
     *    ...
     * </pre>
     */
   public int[] getCursorPos() // guich@400_18
   {
      int []i = new int[2];
      i[0] = startSelectPos;
      i[1] = insertPos;
      return i;
   }
   
   /** user method to popup the keyboard/calendar/calculator for this edit. */
   public void popupKCC()
   {
      if (kbdType == KBD_NONE) return;

      Sound.beep(); // guich@200b4_119: add a click
      if (cursorShowing)
         draw(drawg, true); // pierre@400_93: erase cursor at old insert position
      getParentWindow().swapFocus(this);//requestFocus(); // guich@200b4: bring focus back -  guich@401_15: changed to swapFocus
   
      switch (kbdType)
      {
         case KBD_CALENDAR:
            if (calendar == null) calendar = new Calendar();
            calendar.setSelectedDate(new Date(getText()));
            Window.topMost.popupModal(calendar);            
            break;
            
         case KBD_CALCULATOR:
            if (calculator == null) calculator = new Calculator();
            Window.topMost.popupModal(calculator);
            break;
            
         default:
            if (Settings.virtualKeyboard)
               getParentWindow().setStatePosition(0, !canMoveVirtualKeyboard ? Window.VK_SHOW : (getAbsoluteRect().y < (Settings.screenHeight>>1)) ? Window.VK_BOTTOM : Window.VK_TOP); // if running on a PocketPC device, set the bounds of Sip in a way to not cover the edit
            else
            {
               if (keyboard == null) keyboard = new Keyboard();
               pushPosState();
               if (removeTimer(blinkTimer)) // guich@200b4_167
                  blinkTimer = null;
               Window.topMost.popupModal(keyboard);
               popPosState();
               requestFocus();
            }
      }
   }

   /** Called by the system to pass events to the edit control. */
   public void onEvent(Event event)
   {
      if (charWidths == null)
         return; // widths have not been initialized - not added to parent
      if (calendar != null && event.type == ControlEvent.WINDOW_CLOSED && event.target == calendar) // called from the keyboard and from the calendar
      {
         Date d = calendar.getSelectedDate();
         if (d != null)
            setText(d.toString());
         else
         if (!calendar.canceled)
            setText("");
         return;
      }
      boolean redraw = false;
      boolean extendSelect = false;
      boolean clearSelect = false;
      int newInsertPos = insertPos;
      switch (event.type)
      {
         case ControlEvent.WINDOW_MOVED: // guich@200b4: renew the drawg
            drawg = createGraphics();
            repaint();
            break;
         case ControlEvent.TIMER:
            if (parent != null)
               draw(drawg, true);
            event.consumed=true;     //astein@230_5: prevent blinking cursor event from propagating
            return;
         case ControlEvent.FOCUS_IN:
            drawg = createGraphics();
            hasFocus = true;
            redraw = true;
            if (blinkTimer == null)
               blinkTimer = addTimer(350);
            break;
         case ControlEvent.FOCUS_OUT:
            if (Settings.virtualKeyboard && mode != DATE && mode != CURRENCY) // if running on a PocketPC device, set the bounds of Sip in a way to not cover the edit
               getParentWindow().setStatePosition(Window.VK_HIDE,Window.VK_HIDE);
            hasFocus = false;
            clearPosState();
            newInsertPos = 0;
            redraw = true;
            if (removeTimer(blinkTimer)) // guich@200b4_167
               blinkTimer = null;
            break;
         case KeyEvent.KEY_PRESS:
            if (editable)
            {
               KeyEvent ke = (KeyEvent)event;
               boolean isPrintable;
               if (ke.key < 65536 && (ke.modifiers & IKeys.ALT) == 0 &&
                  (ke.modifiers & IKeys.CONTROL) == 0)
                  isPrintable = true;
               else
                  isPrintable = false;
               // print state
               if (ke.key == IKeys.COMMAND)
               {
                  isCommand = !isCommand;
                  break; // command char? break
               }
               if ((ke.key == IKeys.KEYBOARD_ABC || ke.key == IKeys.KEYBOARD_123) && (keyboard == null || !keyboard.isVisible()) && (calendar == null || !calendar.isVisible()) && (calculator == null || !calculator.isVisible()) ) // guich@102: pop's up the keyboard
               {
                  popupKCC();
                  break;
               }
               boolean isControl = (ke.modifiers & IKeys.CONTROL) != 0; // guich@320_46
               boolean isDelete = (ke.key == IKeys.DELETE);
               boolean isBackspace = (ke.key == IKeys.BACKSPACE);
               int del1 = -1;
               int del2 = -1;
               int sel1 = startSelectPos;
               int sel2 = insertPos;
               if (sel1 > sel2)
               {
                  int temp = sel1;
                  sel1 = sel2;
                  sel2 = temp;
               }
               // clipboard
               if (isCommand || isControl)
               {
                  isCommand = false;
                  if (isControl)
                  {
                     ke.key += 64;
                     ke.modifiers &= ~IKeys.CONTROL; // remove control
                  }
                  int key = Convert.toUpperCase((char)ke.key);
                  switch (key)
                  {
                     case ' ': // guich@320_47
                        setText("");
                        break;
                     case 'X':
                     case 'C':
                        if (sel1 != -1) // cut/copy
                        {
                           Vm.clipboardCopy(new String(chars).substring(sel1,sel2));
                           if (key == 'X') // cut
                           {
                              ke.key = IKeys.BACKSPACE;
                              onEvent(ke);
                           } else Sound.beep();
                        }
                        break;
                     case 'P':
                     case 'V':
                        String pasted = Vm.clipboardPaste();
                        if (pasted == null || pasted.length() == 0)
                           Sound.beep();
                        else
                        {
                           char ch[] = pasted.toCharArray();
                           for (int i =0; i < ch.length; i++)
                           {
                              ke.key = ch[i];
                              onEvent(ke);
                           }
                           newInsertPos += ch.length;
                        }
                        break;
                     default:
                        Sound.beep();
                  }
                  break;
               }
               if (isPrintable)
               {
                  if (capitalise == ALL_NORMAL)
                     ;
                  else
                  if (capitalise == ALL_UPPER)
                     ke.key = Convert.toUpperCase((char)ke.key);
                  else
                  if (capitalise == ALL_LOWER)
                     ke.key = Convert.toLowerCase((char)ke.key);

                  if (!isCharValid((char)ke.key)) // guich@101: tests if the key is in the valid char set - moved to here because an valid clipboard char can be an invalid edit char
                  {
                     Sound.beep();
                     break;
                  }
               }
               if (sel1 != -1 && (isPrintable || isDelete || isBackspace))
               {
                  del1 = sel1;
                  del2 = sel2 - 1;
               }
               else if (isDelete)
               {
                  del1 = insertPos;
                  del2 = insertPos;
               }
               else if (isBackspace)
               {
                  del1 = insertPos - 1;
                  del2 = insertPos - 1;
               }
               if (del1 >= 0 && del2 < len)
               {
                  int deleteCount = del2 - del1 + 1;
                  int numOnRight = len - del2 - 1;
                  for (int i = del1; i <= del2; i++)
                     totalCharWidth -= charWidths[i];
                  if (numOnRight > 0)
                  {
                     Vm.copyArray(chars, del2 + 1, chars, del1, numOnRight);
                     Vm.copyArray(charWidths, del2 + 1, charWidths, del1, numOnRight);
                  }
                  len -= deleteCount;
                  newInsertPos = del1;
                  redraw = true;
                  clearSelect = true;
               }
               if (isPrintable)
                  if (maxLength == 0 || len < maxLength)
                  {
                     // grow the array if required (grows by 8)
                     if (len == chars.length)
                     {
                        char newChars[] = new char[len + 8];
                        Vm.copyArray(chars, 0, newChars, 0, len);
                        chars = newChars;
                        byte newCharWidths[] = new byte[len + 8];
                        Vm.copyArray(charWidths, 0, newCharWidths, 0, len);
                        charWidths = newCharWidths;
                     }
                     char c = (char)ke.key;
                     int charWidth = fm.getCharWidth(c);
                     if (newInsertPos != len)
                     {
                        int i = newInsertPos;
                        int l = len - newInsertPos;
                        Vm.copyArray(chars, i, chars, i + 1, l);
                        Vm.copyArray(charWidths, i, charWidths, i + 1, l);
                     }
                     chars[newInsertPos] = c;
                     charWidths[newInsertPos] = (byte)charWidth;
                     len++;
                     newInsertPos++;
                     totalCharWidth += charWidth;
                     redraw = true;
                     clearSelect = true;
                  }
                  else Sound.beep();
               boolean isMove = true;
               switch (ke.key)
               {
                  case IKeys.HOME  : newInsertPos = 0; break;
                  case IKeys.END   : newInsertPos = len; break;
                  case IKeys.LEFT  :
                  case IKeys.UP    : newInsertPos--; break;
                  case IKeys.RIGHT :
                  case IKeys.DOWN  : newInsertPos++; break;
                  default: isMove = false;
               }
               if (isMove && newInsertPos != insertPos)
               {
                  if ((ke.modifiers & IKeys.SHIFT) > 0)
                     extendSelect = true;
                  else
                     clearSelect = true;
               }
            }
            break;
         case PenEvent.PEN_DOWN:
         {
            PenEvent pe = (PenEvent)event;
            newInsertPos = x2charPos(pe.x);
            if ((pe.modifiers & IKeys.SHIFT) > 0) // shift
               extendSelect = true;
            else
               clearSelect = true;
            break;
         }
         case PenEvent.PEN_DRAG:
         {
            PenEvent pe = (PenEvent)event;
            newInsertPos = x2charPos(pe.x);
            if (newInsertPos != insertPos)
               extendSelect = true;
            break;
         }
         case PenEvent.PEN_UP:
            // guich@300_43: ignoreFocus is needed bc when popupKCC is called, the focus comes back to here; also, when the popped up window is closed, the focus comes back again, so we could enter in an infinite loop
            if (kbdType != KBD_NONE && Settings.virtualKeyboard)
               popupKCC();
            break;         
         default:
            if (event.type == Keyboard.KEYBOARD_ON_UNPOP) // guich@320_34
               pushPosState();
            else
            if (event.type == Keyboard.KEYBOARD_POST_UNPOP)
               popPosState();
            return;
      }
      if (extendSelect)
      {
         if (startSelectPos == -1)
            startSelectPos = insertPos;
         else if (newInsertPos == startSelectPos)
            startSelectPos = -1;
         redraw = true;
      }
      if (clearSelect && startSelectPos != -1)
      {
         startSelectPos = -1;
         redraw = true;
      }
      if (newInsertPos > len)
         newInsertPos = len;
      if (newInsertPos < 0)
         newInsertPos = 0;
      boolean insertChanged = (newInsertPos != insertPos);
      if (insertChanged && !redraw && cursorShowing)
         draw(drawg, true); // erase cursor at old insert position
      if (insertChanged)
      {
         int x = charPos2x(newInsertPos);
         if (x - 3 < xMin)
         {
            // characters hidden on left - jump
            xOffset += (xMin - x) + 20;
            if (xOffset > xMin)
               xOffset = xMin;
            redraw = true;
         }
         if (x + 3 > xMax)
         {
            // characters hidden on right - jump
            xOffset -= (x - xMax) + 20;
            if (xOffset < xMax - totalCharWidth)
               xOffset = xMax - totalCharWidth;
            redraw = true;
         }
         if (totalCharWidth < xMax - xMin && xOffset != xMin)
         {
            xOffset = xMin;
            redraw = true;
         }
      }
      insertPos = newInsertPos;
      if (redraw)
         draw(drawg, false);
      else
      if (insertChanged)
         draw(drawg, true); // draw cursor at new insert position
      if (event.type == ControlEvent.FOCUS_OUT)
      {
         if (drawg != null) drawg.free();
         drawg = null;
      }
   }

   protected void onWindowPaintFinished()
   {
      if (!hasFocus) onEvent(new Event(ControlEvent.FOCUS_IN,this,0)); // this event is called on the focused control of the parent window. so, if we are not in FOCUS state, set it now.
   }

   /** Called by the system to draw the edit control. */
   public void onPaint(Graphics g)
   {
      draw(g, false);
   }

   /** Gets the current mask. Note that the mask is currently only used to compute the preferred width of this Edit.
   @deprecated access the mask member directly
   */
   public String getMask() // guich@300_10
   {
      return mask;
   }

   /** Sets the current mask to the given one. Note that the mask is currently only used to compute the preferred width of this Edit.
     @deprecated access the mask member directly
     */
   public void setMask(String mask) // guich@300_10
   {
      this.mask = mask;
   }
}
