/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.ui;

import waba.fx.*;
import waba.sys.*;

/** This class pops up a Window with a Label, an Edit and some buttons. Good
    to input some text from the user.
    Here is an example (taken from examples/apps/GuiBuilder):
    <pre>
    // on some event...
      if (renameDialog == null)
         renameDialog = new InputDialog("Project Rename","Enter the new name:","");
      renameDialog.setValue(projectName);
      popupBlockingModal(renameDialog);
      // when window closes...
      if (renameDialog.getPressedButtonIndex() == 0) // ok?
      {
         projectName = renameDialog.getValue();
      }
    </pre>
  */

public class InputDialog extends Window
{
   private Label msg;
   private PushButtonGroup btns;
   private Edit ed;
   private int selected = -1;

   /** Creates a new InputDialog with the given window Title,
     * the given label Text, the given Default value for the Edit,
     * and two buttons "Ok" and "Cancel".
     * The text used in a Label must be one line only!
     */
   public InputDialog(String title, String text, String defaultValue)
   {
      this(title, text, defaultValue, new String[]{"Ok","Cancel"});
   }

   /** Creates a new InputDialog with the given window Title,
     * the given label Text, the given Default value for the Edit
     * and with the given buttons.
     * The text used in a Label must be one line only!
     */
   public InputDialog(String title, String text, String defaultValue, String[] buttonCaptions)
   {
      super(title,ROUND_BORDER);
      highResPrepared = true;
      setDoubleBuffer(true);
      btns = new PushButtonGroup(buttonCaptions,false,-1,4,6,1,false,PushButtonGroup.BUTTON);
      msg = new Label(text,Label.LEFT);
      ed = new Edit("@@@@@@@@@@");
      if (defaultValue != null) ed.setText(defaultValue);

      int wb = btns.getPreferredWidth();
      int hb = btns.getPreferredHeight();
      int wm = Math.min(msg.getPreferredWidth()+1,Settings.screenWidth-6);
      int hm = msg.getPreferredHeight();
      int we = ed.getPreferredWidth();
      int he = ed.getPreferredHeight();
      FontMetrics fm2 = getFontMetrics(titleFont); // guich@220_28
      int captionH = fm2.getHeight()+10;

      int h = captionH + hb + hm + he;
      int w = Math.max(Math.max(Math.max(wb,wm),we),fm2.getTextWidth(title))+6; // guich@200b4_29
      w = Math.min(w,Settings.screenWidth); // guich@200b4_28: dont let the window be greater than the screen size
      setRect(CENTER,CENTER,w,h);
      add(msg);
      add(btns);
      add(ed);
      msg.setRect(4,TOP,wm,hm);
      ed.setRect(CENTER,AFTER+2,we,he);
      btns.setRect(CENTER,AFTER+2,wb,hb);
   }

   protected void onColorsChanged(boolean colorsChanged) // guich@220_47
   {
      if (colorsChanged)
      {
         msg.setForeColor(foreColor);
         msg.setBackColor(backColor);
      }
   }
   protected void postPopup()
   {
   	ed.requestFocus();
   }

   /** handle scroll buttons and normal buttons */
   public void onEvent(Event e)
   {
      switch (e.type)
      {
         case ControlEvent.PRESSED:
            if (e.target == btns && (selected=btns.getSelected()) != -1)
            {
               btns.setSelected(-1);
               unpop();
            }
            break;
      }
   }

   /** Returns the pressed button index, starting from 0 */
   public int getPressedButtonIndex()
   {
      return selected;
   }

   /** Returns the value entered */
   public String getValue()
   {
   	return ed.getText();
   }
   /** Sets the default value on the Edit field */
   public void setValue(String value)
   {
   	ed.setText(value);
   }
   /** Returns the Edit so you can set its properties. */
   public Edit getEdit() // guich@310_23
   {
      return ed;
   }
}