/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>                         *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.ui;

import waba.fx.*;
import waba.sys.*;

/**
 * Label is a text label control. It supports multiline,
 * but you need to separate the text with |.
 * <p>
 * Note: this class does not do automatic parse of the text, because this is a slow operation.
 * But, you can use a handy method to parse the text that you want to display: see the
 * waba.sys.Convert.insertLineBreak method.
 * <p>
 * Here is an example showing a label being used:
 *
 * <pre>
 * public class MyProgram extends MainWindow
 * {
 * public void onStart()
 *  {
 *  add(new Label("Value:"), LEFT, TOP);
 *  }
 * </pre>
 */

public class Label extends Control
{
   private String text;
   //Font font; removed by guich
   public int align; // guich@400_71: made public instead of protected
   private static final String[] emptyStringArray = {""};
   private String []lines = emptyStringArray;
   private int []linesW;
   private int linesPerPage,currentLine;
   private boolean invert;
   private boolean is3d;
   private int hh;
   /** You can set this to 4 so that the labels will have the same height of a WinCE Edit.
     * Don't forget to reset it to the default, zero.
     * Since SuperWaba 4.01
     */
   public static int commonVGap = 0;

   /** Creates a label displaying the given text with left alignment. supports inverted text, multiple lines and is scrollable */
   public Label(String text)
   {
      this(text, LEFT);
   }

   /**
   * Creates a label displaying the given text with the given alignment.
   * @param text the text displayed
   * @param align the alignment
   * @see #LEFT
   * @see #RIGHT
   * @see #CENTER
   */
   public Label(String text, int align)
   {
      this.align = align;
      setText(text);
   }

   /** if 3d is true, draws the label with a 3d effect. turns off invert. */
   public void set3d(boolean on)
   {
      is3d = on;
      if (on) invert = false;
   }
   /** if invert is true, the back and fore color are swaped. turns off 3d. */
   public void setInvert(boolean on)
   {
      invert = on;
      if (on) is3d = false;
      //onFontChanged();
   }
   /** Sets the text that is displayed in the label. */
   public void setText(String text)
   {
      this.text = text;
      lines = Convert.tokenizeString(text,'|');
      linesW = new int[lines.length];
      int inv = (invert && align==RIGHT)?1:0; // guich@400_88
      for (int i =0; i < lines.length; i++)
         linesW[i] = fm.getTextWidth(lines[i]) + inv;
      currentLine = 0;
      repaint();
   }

   /** Gets the text that is displayed in the label. */
   public String getText()
   {
      return text;
   }

   /** returns the preffered width of this control. added by guich */
   public int getPreferredWidth()
   {
      if (text.length() == 0) // guich@330_5
         return FILL;
      int w = 0;
      for (int i =0; i < lines.length; i++)
         w = Math.max(w, fm.getTextWidth(lines[i]));
      return w+(invert?2:0);
   }

   /** returns the preffered width of this control. added by guich */
   public int getPreferredHeight()
   {
      return fmH*lines.length + (invert?1:0) + commonVGap; // if inverted, make sure the string is surrounded by the black box - guich@401_18: added commonVGap
   }

   protected void onFontChanged()
   {
      int inv = (invert && align==RIGHT)?1:0; // guich@400_88
      for (int i =0; i < lines.length; i++)
         linesW[i] = fm.getTextWidth(lines[i]) + inv;
      hh = fmH*linesPerPage;
   }

   protected void onBoundsChanged()
   {
      linesPerPage = height / fmH;
      if (linesPerPage < 1) linesPerPage = 1;
      hh = fmH*linesPerPage;
   }

   /** returns if the label can scroll in the given direction */
   public boolean canScroll(boolean down) // guich@200b4_142
   {
      if (lines.length > linesPerPage)
      {
         if (down)
            return currentLine+linesPerPage < lines.length;
         else
            return currentLine - linesPerPage >= 0;
      }
      return false;
   }

   /** scroll one page. returns true if success, false if no scroll possible */
   public boolean scroll(boolean down)
   {
      if (lines.length > linesPerPage)
      {
         int lastLine = currentLine;
         if (down)
         {
            if (currentLine+linesPerPage < lines.length)
               currentLine += linesPerPage;
         }
         else
         {
            currentLine -= linesPerPage;
            if (currentLine < 0)
               currentLine = 0;
         }
         if (lastLine != currentLine)
         {
            repaint();
            return true;
         }
      }
      return false;
   }

   /** Called by the system to draw the button. */
   public void onPaint(Graphics g)
   {
      // draw label
      if (invert)
      {
         g.setForeColor(backColor);
         g.setBackColor(foreColor);
      }
      else
      {
         g.setForeColor(foreColor);
         g.setBackColor(backColor);
      }
      // guich@200b4_126: repaint the background always.
      g.fillRect(0,0,width,height); // guich@200b4_120: make sure the label is painted with the correct color
      if (text.length() > 0)
      {
         int y = (this.height - hh) >> 1; // center on y (if necessary)
         int n = Math.min(currentLine+linesPerPage, lines.length);
         int xx = invert?1:0;
         for (int i =currentLine; i < n; i++,y+=fmH)
         {
            int x = 0;
            if (align != LEFT)
            {
               if (align == CENTER)
                  x = (width - linesW[i]) >> 1;
               else
               if (align == RIGHT)
                  x = width - linesW[i];
            }

            if (is3d) // if 3d, invert = false
            {
               g.setForeColor(backColor.darker());
               g.drawText(lines[i], xx+x+1, y+1);
               g.setForeColor(foreColor);
            }
            g.drawText(lines[i], xx+x, y);
         }
      }
   }
}
