/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>                         *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.ui;

import waba.fx.*;
import waba.sys.*;
import waba.util.*;
import waba.sys.Thread;

/**
 * MainWindow is the main window of a UI based application.
 * <p>
 * All SuperWaba programs with a user-interface must have <b>one and only one</b> main window.
 * <p>
 * Here is an example showing a basic application:
 *
 * <pre>
 * public class MyProgram extends MainWindow
 * {
 *    Edit edName;
 *    public void onStart()
 *    {
 *       ... initialization code ...
 *       add(new Label("Name:"), LEFT,TOP+2);
 *       add(edName = new Edit(""), AFTER,SAME-2);
 *    }
 * }
 * </pre>
 */

public class MainWindow extends Window
{
   private Vector vThreads = new Vector(5);
   private Timer timers;
   private Font _cachedFont;
   private FontMetrics _cachedFontMetrics;
   private int currentRunningThread;
   private int threadCount, highThreadCount;
   static MainWindow _mainWindow;
   static Graphics myg0; // never translated

   /** Default font used in all controls created. To change the default font, assign it to this member in the MainWindow constructor. */
   public static Font defaultFont = new Font("SW", Font.PLAIN, 12);

   /** Constructs a main window with no title and no border. */
   public MainWindow()
   {
      this(null,NO_BORDER);
   }

   /** Constructs a main window with the given title and border style.
   @since SuperWaba2.0b4 */
   public MainWindow(String title, byte style) // guich@112
   {
      super(title,style);
      // update some settings
      highResPrepared = true; // main window is always prepared

      if (Settings.isColor)
         setBackColor(new Color(160,216,216)); // guich@200b4_39

	   _mainWindow = this;
      NativeMethods4JDK.mainWindowCreate(this);
	   zStack.push(this); // guich

	   canDrag = false; // we can't drag main windows.
	   x2 = width-1;
	   y2 = height-1;
	   topMost = this;
   }

   /**
   * Notifies the application that it should stop executing and exit. It will
   * exit after executing any pending events. If the underlying system supports
   * it, the exitCode passed is returned to the program that started the app.
   * Note: in PalmOS and in AppletViewer/Browser the exitCode is useless.
   * If you want your code to be called when the VM exits, extend the onExit method.
   */
   public final void exit(int exitCode)
   {
      NativeMethods4JDK.mainWindowExit(exitCode);
   }

   /** Returns the MainWindow of the current application. */
   public static MainWindow getMainWindow()
   {
	   return _mainWindow;
   }

   /** Returns the font metrics for a given font. */
   public FontMetrics getFontMetrics(Font font)
   {
	   if (font == _cachedFont)
		   return _cachedFontMetrics;
	   _cachedFont = font;
	   _cachedFontMetrics = new FontMetrics(font, this);
	   return _cachedFontMetrics;
   }

   /**
   * Adds a timer to a control. This method is protected, the public
   * method to add a timer to a control is the addTimer() method in
   * the Control class.
   */
   protected Timer addTimer(Control target, int millis)
   {
	   Timer t = new Timer();
	   t.target = target;
	   t.millis = millis;
	   t.lastTick = Vm.getTimeStamp();
	   t.next = timers;
	   timers = t;
	   _onTimerTick();
	   return t;
   }

   /**
   * Removes a timer. This method returns true if the timer was found
   * and removed and false if the given timer could not be found.
   */
   public boolean removeTimer(Timer timer)
   {
	   if (timer == null)
		   return false;
	   Timer t = timers;
	   Timer prev = null;
	   while (t != timer)
	   {
		   if (t == null)
			   return false;
		   prev = t;
		   t = t.next;
	   }
	   if (prev == null)
		   timers = t.next;
	   else
		   prev.next = t.next;
	   _onTimerTick();
	   return true;
   }

   /** Called by the VM when the application is starting. Calls onStart.
     * Never call this method directly; this method is not private
     * to prevent the compiler from removing it during optimization.
     */
   final public void appStarting() // guich@200b4_107
   {
      started = true;
      myg0 = createGraphics();
      onStart();
      if (zStack.getCount() > 1) // was a Window poped up during initialization?
         repaintLastPopup = true;
   }

   /** Called by the system so we can finish things correctly.
     * Never call this method directly; this method is not private
     * to prevent the compiler from removing it during optimization.
    */
   final public void appEnding() // guich@200final_11: fixed when switching apps not calling killThreads.
   {
      killThreads(); // guich@200b4_85: stop all threads
      MainWindow._mainWindow.onExit(); // guich@200b4_85
   }

   /**
   * Called just before an application exits.
   * When this is called, all threads are already killed.
   */
   public void onExit()
   {
   }

   /**
   * Called when an application starts. Initialization code is usually either placed
   * in this method or simply in the app's constructor. This method is called
   * just after the app's constructor is called.
   */
   public void onStart()
   {
   }

   /**
   * Called by the VM to process timer interrupts. This method is not private
   * to prevent the compiler from removing it during optimization.
   */
   public void _onTimerTick()
   {
	   int minInterval = 0;
	   int now = Vm.getTimeStamp();
	   Timer timer = timers;
	   while (timer != null)
	   {
		   int diff = now - timer.lastTick;
		   if (diff < 0)
			   diff += (1 << 30); // wrap around - max stamp is (1 << 30)
		   int interval;
		   if (diff >= timer.millis)
		   {
			   // post TIMER event
			   Control c = timer.target;
			   _controlEvent.type = ControlEvent.TIMER;
			   _controlEvent.target = c;
			   timer.triggered = true; // guich@220_39
			   c.postEvent(_controlEvent);
			   timer.triggered = false;
			   timer.lastTick = now;
			   interval = timer.millis;
		   }
		   else
			   interval = timer.millis - diff;
		   if (interval < minInterval || minInterval == 0)
			   minInterval = interval;
		   timer = timer.next;
	   }
	   NativeMethods4JDK.mainWindowSetTimerInterval(minInterval);
	   if (topMost.needsPaint) // guich@200b4_1: corrected the infinit repaint on popup windows
		   _doPaint();
   }

   /** Returns the command line passed by the application that called us in the Vm.exec method */
   public String getCommandLine()
   {
      return NativeMethods4JDK.mainWindowGetCommandLine();
   }

   /** Add the given thread to the active list of threads. Note that a thread
     * cannot be added twice; you must first remove it by calling <code>removeThread</code>.
     * @param t the thread to be enqueued
     * @param highPriority High priority threads (when parameter is true) are run
     * all at once each time the MainWindow._runThreads method is called by the
     * VM; low priority threads (when parameter is false) are run sequencially one
     * at a time in each call of _runThreads by the VM. Making too much high
     * priority threads isnt a good choice because it blocks normal events. After
     * each call of MainWindow.runThreads, the VM dispatch the normal events.
     */
   public void addThread(Thread t, boolean highPriority)
   {
      if (vThreads.find(t) == -1)
      {
         if (highPriority)
         {
            vThreads.insert(0,t);
            highThreadCount++;
            if (currentRunningThread > 0)
               currentRunningThread++;
         }
         else
            vThreads.add(t);
         setThreadCount(vThreads.size());
         t.started();
      }
   }
   /** Remove the given thread from the active list of threads.
     * Calls the <code>stopped</code> method of the thread.
     */
   public void removeThread(Thread t)
   {
      int index = vThreads.indexOf(t);
      if (index >= 0)
      {
         vThreads.del(t);
         if (index < highThreadCount) // was it a high priority thread?
         {
            highThreadCount--;
            if (currentRunningThread > 0)
               currentRunningThread--;
         }
         t.stopped();
         setThreadCount(vThreads.size());
      }
   }

   /** Called by the VM to process threads. This method is not private
   * to prevent the compiler from removing it during optimization.
   */
   public void _runThreads()
   {
      // High priority threads: run all threads per call
      if (highThreadCount > 0)
         for (int i =0; i < highThreadCount; i++)
            ((Thread)vThreads.items[i]).run();
      // run one thread per call
      if (threadCount - highThreadCount > 0) // are there any low priority threads?
      {
         if (currentRunningThread >= threadCount || currentRunningThread < highThreadCount)
            currentRunningThread = highThreadCount;
         ((Thread)vThreads.items[currentRunningThread++]).run();
      }
   }

   /** Kill all running threads. This method is for internal use only */
   public void killThreads()
   {
      while (--threadCount >= 0)
         ((Thread)vThreads.pop()).stopped();
      highThreadCount = 0; // guich@200final_10: fixed bug when exiting with high priority threads active.
   }

   private void setThreadCount(int count)
   {
      this.threadCount = count;
      NativeMethods4JDK.mainWindowSetThreadCount(count);
   }
}
