/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.ui;

import waba.fx.*;
import waba.sys.*;

/** Constructs a Menu with the given items. The Menu supports disabled and checked items.
   * Here is an example of how to build the menu for the DateBook:<br>
   * <pre>
   * String col0[] =
   * {
   *   "Record",         // caption for the MenuBar
   *   "?NewEvent",      // checked item
   *   "Delete Event...",
   *   "Attach Note",
   *   "Delete Note...",
   *   "Purge...",
   *   "*Beam Event",    // disabled item
   * };
   * String col1[] =
   * {
   *   "Edit",           // caption for the MenuBar
   *   "Undo",
   *   "Cut",
   *   "Copy",
   *   "Paste",
   *   "Select All",
   *   "-",              // a dotted line
   *   "Keyboard",
   *   "Graffiti Help",
   * };
   * String col2[] =
   * {
   *   "Options",        // caption for the MenuBar
   *   "Font...",
   *   "Preferences...",
   *   "Display Options...",
   *   "Phone Lookup",
   *   "About Date Book",
   * };
   * setMenuBar(new MenuBar(new String[][]{col0,col1,col2}));
   * </pre>
   * An WINDOW_CLOSED event will be thrown when the menu is closed.
   * The menu can be closed by a click on a valid item or clicking outside of its bounds.
   * To discover why it was closed, see method getSelectedMenuItem.
   * <p>
   * Note that the separator dotted line don't generate events.
   * <p>
   * I know that storing things in String is very odd; I should create a class MenuItem and store
   * things there; but this way is faster and uses less memory.
   */

public class MenuBar extends Window
{
   // changed drawing of menu border to look more native
   // PMD 25Oct2001
   // I have removed Rect[] rCaps because only the x component is used
   // replace it with int[] xpos
   private String [][]items;
   private int []xpos;
   private int xmin,xmax;
   /** Note: if you want to change the spacement between the menu items (maybe to make more items fit in the row), change this gap value. Default is 3 (3 pixels at left and 3 at right). */
   public int gap = 3; // guich@200b4_89
   private PopupMenu pop;
   private int selected=0;
   private Graphics gr;
   private int menuItemSelected;
   private boolean switching;
   public static final char CHECKED = '!';
   public static final char UNCHECKED = '?';
   public static final char DISABLED = '*';
   private Image wholeScreen;
   private Color eColor,dColor,bColor,cursorColor;
   private Color fourColors[] = new Color[4];
   private Color popFore,popBack,popCursor;

   /** Create a MenuBar with the given items. */
   public MenuBar(String [][]items)
   {
      this.items = items;
      highResPrepared = true;
      started = true; // avoid calling the onStart method
      canDrag = false;
      setFont(getFont().asBold()); // use a bold font.
      setBackColor(Color.WHITE);
      dontSaveBehind(false);
      borderStyle = -1; // guich@220_48
   }

   /** Sets the current Menu items. */
   public void setItems(String [][]items) // guich@400_19
   {
      this.items = items;
   }

   /** Set the colors for the popup windows. You can pass <code>null</code>
     * to any parameter to keep the current settings.
     */
   public void setPopColors(Color back, Color fore, Color cursor)
   {
      this.popFore = fore;
      this.popBack = back;
      this.popCursor = cursor;
   }

   /** Sets the cursor color. By default, it is based in the background color */
   public void setCursorColor(Color c) // guich@220_49
   {
      cursorColor = c;
   }

   /** Called by the Window class to popup this MenuBar */
   public void setVisible(boolean b)
   {
      if (b)
      {
         try
         {
            // guich@200b4_69: improve menu performance
            if (wholeScreen == null) // guich@200b4_70: now only gets memory one time
               wholeScreen = new Image(Settings.screenWidth,Settings.screenHeight);
            wholeScreen.getGraphics().copyScreen(MainWindow._mainWindow, 0, 0, wholeScreen.getHeight());
         } catch (OutOfMemoryError oome) {wholeScreen = null;} // guich@340_52

         popupModal(this);
      }
   }

   /** Change the font and recompute some parameters */
   protected void onFontChanged() // guich@200b4_153
   {
      // all this is recalculated because the font had changed
      setRect(0,0,Settings.screenWidth,fmH+4);
      xpos = new int[items.length+1];
      xpos[0] = xmax = xmin = 4;
      gr = createGraphics();
      int temp = (gap<<1)-1;
      for (int i=0; i < items.length; i++)
      {
         xmax += fm.getTextWidth(items[i][0])+temp;
         xpos[i+1] = xmax;
      }
      xmax -= 1;
   }

   /** Returns the current menu item selected.
     * @return -1 if no item was selected, or a number with format <i>xyy</i>, where x is the index of the item
     * selected at the MenuBar and yy is the index+1 ( = the index of the String array that defines the items)
     * of the item selected in the PopupMenu.
     * EG: suppose that the col1/"Cut" of the
     * example at the top of this page was clicked, then 102 is returned.
     */
   public int getSelectedMenuItem()
   {
      return menuItemSelected;
   }
   /** Returns true if the given menu item in the xyy format is checked. To make an item checkable at the constructor,
     * prefix it with a ?. When it is selected, the ? is replaced by a !.
     */
   public boolean isChecked(int menuItem)
   {
      int col = menuItem / 100;
      int row = menuItem % 100;
      try
      {
         int index = items[col][row].indexOf(CHECKED); // guich@321_5
         return index == 0 || index == 1;
      } catch (/*ArrayIndexOutOfBounds*/Exception e) {}
      return false;
   }
   /** Returns true if a menu item is checkable */
   public boolean isCheckable(int menuItem)
   {
      int col = menuItem / 100;
      int row = menuItem % 100;
      try
      {
         char c = items[col][row].charAt(0);
         return c == CHECKED || c == UNCHECKED;
      } catch (/*ArrayIndexOutOfBounds*/Exception e) {}
      return false;
   }
   /** Returns the enabled state of the given menu item. To disable an item,
     * use method setEnabled or prefix it with a * at the constructor.
     */
   public boolean isEnabled(int menuItem)
   {
      int col = menuItem / 100;
      int row = menuItem % 100;
      try // guich@321_6: and all tries below
      {
         return items[col][row].charAt(0) != DISABLED;
      } catch (/*ArrayIndexOutOfBounds*/Exception e) {}
      return false;
   }
   /** Put a check in the given menu item in the xyy format. To make an item checkable, at the constructor
     * you must prefix it with a ?. When a checkable item is selected, the ? is replaced by a !.
     * <p>Note: the items at the MenuBar cannot be checkable.
     */
   public void setChecked(int menuItem, boolean check)
   {
      int col = menuItem / 100;
      int row = menuItem % 100;
      try
      {
         items[col][row] = items[col][row].replace(check?UNCHECKED:CHECKED, check?CHECKED:UNCHECKED); // guich@321_5
      } catch (/*ArrayIndexOutOfBounds*/Exception e) {}
   }
   /** Enable the given menu item.
     */
   public void setEnabled(int menuItem, boolean enable)
   {
      int col = menuItem / 100;
      int row = menuItem % 100;
      try
      {
         boolean isEnabled = items[col][row].charAt(0) != DISABLED;
         if (isEnabled && !enable)
            items[col][row] = DISABLED+items[col][row];
         else
         if (!isEnabled && enable)
            items[col][row] = items[col][row].substring(1);
      } catch (/*ArrayIndexOutOfBounds*/Exception e) {}
   }

   /** Setup some important variables */
   protected void onPopup()
   {
      menuItemSelected = -1;
   }

   /** Close the popup list with a click outside its bounds */
   protected boolean onClickedOutside(int x,int y)
   {
      close();
      return true;
   }

   void close()
   {
      // guich@200b4_69: put back the saved screen
      if (wholeScreen != null)
         MainWindow.myg0.copyScreen(wholeScreen, 0, 0, wholeScreen.getHeight());
      else
         MainWindow.repaintActiveWindows(pop != null ? -2 : -1); // guich@340_52: caused by an out-of-memory - tomasf@400_76: fixed when pop==null

      if (pop != null)
         pop.unpop();
      unpop();
   }

   private int getItemAt(int x)
   {
      int n = xpos.length-1;
      for (int i=0; i < n; i++)
         if (xpos[i] <= x && x < xpos[i+1] && items[i][0].charAt(0) != DISABLED)
            return i;
      return -1;
   }

   private void drawCursor(Graphics g, int index, boolean on)
   {
      Color f = cursorColor != null ? cursorColor : bColor.getCursorColor(); // guich@220_49
      g.setForeColor(on?bColor:f);
      g.setBackColor(on?f:bColor);
      g.eraseRect(xpos[index],1,xpos[index+1]-xpos[index],height-2);
   }

   public void onEvent(Event event)
   {
      switch (event.type)
      {
         case KeyEvent.KEY_PRESS:
            if ( ((KeyEvent)event).key == IKeys.MENU)
               close();
            break;
         case ControlEvent.WINDOW_CLOSED:
            if (!switching && event.target == pop)
            {
               int row = pop.getSelected();
               pop = null;
               if (row != -1) // closed bc a item was clicked?
               {
                  menuItemSelected = selected * 100 + row;
                  if (isCheckable(menuItemSelected))
                     setChecked(menuItemSelected,!isChecked(menuItemSelected));
                  close();
               }
            }
            break;
         case PenEvent.PEN_DOWN:
         case PenEvent.PEN_MOVE:
         case PenEvent.PEN_DRAG:
            PenEvent pe = (PenEvent)event;
            if (xmin <= pe.x && pe.x <= xmax)
            {
               int newSelected = getItemAt(pe.x);
               if (newSelected != selected)
               {
                  if (selected != -1) drawCursor(gr,selected,false);
                  selected = newSelected;
                  if (selected != -1) drawCursor(gr,selected,true);
                  switchTo(selected);
               }
            }
            else
            if (selected != -1) // outside valid area?
            {
               drawCursor(gr,selected,false);
               selected = -1;
               switchTo(selected);
            }
            break;
      }
   }

   protected void postPopup()
   {
      if ( selected != -1 )
         switchTo( selected );
   }

   /** called from menu to tell that it is being closed */
   protected void popupClosed(PopupMenu pop)
   {
      //loadBehindLastPopup(pop.height); tomazf@400_76
      pop = null;
      drawCursor(gr,selected,false);
      //selected = -1; // guich@220_21: fixed problem when closing without selecting an item and in the next menu popup no popupMenu was opened.
   }

   protected void loadBehindLastPopup(int ph)
   {
      // guich@200b4_69: put back part of the saved screen
      if (wholeScreen != null)
         MainWindow.myg0.copyScreen(wholeScreen, this.height, this.height, ph); // guich@200b4_70
      else
         MainWindow.repaintActiveWindows(0); // guich@340_52: caused by an out-of-memory
      // guich@200b4_88: paint the bottom right of menu again, since the copyRect above erased it
      if (borderStyle == -1) // guich@220_48
      {
         gr.setForeColor(foreColor);
         gr.drawLine(2,height-1,width-3,height-1);
         gr.setPixel(width-2,height-2);
         gr.setForeColor(backColor);
         gr.setPixel(width-2,height-1);
         gr.setPixel(width-1,height-2);
         gr.setPixel(width-1,height-1);
      }
      else
      {
         gr.setForeColor(backColor);
         gr.drawLine(0,height-1,width,height-1);
         gr.drawLine(0,height-2,width,height-2);
      }
   }
   protected void switchTo(int index)
   {
      switching = true;
      if (pop != null)
      {
         int ph = pop.height;
         pop.unpop();
         loadBehindLastPopup(ph);
      }

      needsPaint = false; // avoid flickering
      if (index != -1)
      {
         pop = new PopupMenu(xpos[index],height-2,items[index]);
         pop.setFont(this.font); // guich@350_8: added a fix when the user changes the font
         if (borderStyle == NO_BORDER)  // guich@220_48
            pop.setBorderStyle(NO_BORDER);
         pop.setBackForeColors(popBack!=null?popBack:backColor, popFore!=null?popFore:foreColor);
         if (popCursor != null)
            pop.setCursorColor(popCursor); // guich@220_49
         pop.dontSaveBehind(false);
         popupModal(pop);
      }
      else
         pop = null;
      switching = false;
   }
   protected void onColorsChanged(boolean colorsChanged)
   {
      eColor = getForeColor();
      dColor = eColor.getAlpha() > 128 ? eColor.darker() : eColor.brighter();
      bColor = getBackColor();
      if (colorsChanged)
      {
         Graphics.compute3dColors(true,backColor,foreColor,fourColors);
         if (cursorColor == null && Settings.maxColors >= 256)
            cursorColor = new Color(0,0,240);
      }
   }

   public void onPaint(Graphics g)
   {
      // changed drawing of menu border to look more native
      // PMD 25Oct2001
      g.setForeColor(eColor);
      g.setBackColor(bColor);
      if (borderStyle == -1) // guich@220_48
         g.draw3dRect(0,0,width,height,Graphics.R3D_SHADED,false,false,fourColors);

      // paint border
      g.setForeColor(eColor);
      // paint captions
      for (int i =0; i < items.length; i++)
      {
         String cap = items[i][0];
         if (cap.charAt(0) != DISABLED)
            g.drawText(cap,xpos[i]+gap,1);
         else
         {
            g.setForeColor(dColor);
            g.drawText(cap.substring(1),xpos[i]+gap,1);
            g.setForeColor(eColor);
         }
      }
      // paint cursor
      if (selected != -1)
         drawCursor(g,selected,true);
   }
}