/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.ui;

import waba.fx.*;
import waba.sys.*;

/** This class implements a scrollable message box window with customized buttons, delayed
  * unpop and scroll text.
  * <br>for example, to create an automatic unpop after 5 seconds, do:
  * <pre>
  *   MessageBox mb = new MessageBox("SuperWaba","SuperWaba is an enhanced version|of the Waba Virtual Machine.|Programmed by:|Guilherme Campos Hazan",null);
  *   popupBlockingModal(mb);
  * </pre>
  * Note: this class does not do automatic parse of the text, because this is a slow operation.
  * But you can use the method waba.sys.Convert.insertLineBreak to do that.
  */

public class MessageBox extends Window
{
   private Label msg;
   private PushButtonGroup btns;
   private int selected = -1;
   private boolean hasScroll = false;
   private int xa,ya,wa=14,ha=6; // arrow coords
   private Timer unpopTimer;

   /** constructs a message box with the text and one "Ok" button */
   public MessageBox(String title, String msg)
   {
      this(title, msg, new String[]{"Ok"});
   }

   /** constructs a message box with the text and the specified button captions.
   the text may be separated by '|' as the line delimiters.
   if buttonCaptions is null, no buttons are displayed and you must dismiss the dialog by calling unpop or
   by setting the delay using setUnpopDelay method */
   public MessageBox(String title, String text, String[] buttonCaptions)
   {
      super(title,ROUND_BORDER);
      highResPrepared = true;
      setDoubleBuffer(true);
      if (buttonCaptions != null)
         btns = new PushButtonGroup(buttonCaptions,false,-1,4,6,1,false,PushButtonGroup.BUTTON);
      msg = new Label(text,Label.CENTER);

      int wb = btns==null?0:btns.getPreferredWidth();
      int hb = btns==null?0:btns.getPreferredHeight();
      int wm = Math.min(msg.getPreferredWidth()+1,Settings.screenWidth-6);
      int hm = msg.getPreferredHeight();
      FontMetrics fm2 = getFontMetrics(titleFont); // guich@220_28
      int captionH = fm2.getHeight()+8;
      if (captionH+hb+hm > Settings.screenHeight) // needs scroll?
      {
         if (hb == 0) hb = ha;
         hm = Settings.screenHeight - captionH - hb - ha;
         hasScroll = true;
      }
      int h = captionH + hb + hm;
      int w = Math.max(Math.max(wb,wm),fm2.getTextWidth(title))+6; // guich@200b4_29
      w = Math.min(w,Settings.screenWidth); // guich@200b4_28: dont let the window be greater than the screen size
      setRect(CENTER,CENTER,w,h);
      add(msg);
      if (btns != null) add(btns);
      msg.setRect(4,captionH-6,/*wm*/getClientRect().width,hm); // guich@350_17: replaced wm by client_rect.width
      if (btns != null) btns.setRect(CENTER,captionH-4+hm,wb,hb);
      Rect r = msg.getRect();
      xa = r.x+r.width-(wa << 1);
      ya = r.y+r.height+3;
      if (Settings.isColor)
         setBackForeColors(new Color(0xFF0000), new Color(0xFFFFFF)); // guich@300_17: add some colors to the messagebox
   }

   /** Sets the alignment for the text. Must be CENTER (default), LEFT or RIGHT */
   public void setTextAlignment(int align)
   {
   	msg.align = align; // guich@241_4: ugly, i know. :-p
   }

   /** sets a delay for the unpop of this dialog */
   public void setUnpopDelay(int unpopDelay)
   {
      if (unpopDelay > 0)
         unpopTimer = addTimer(unpopDelay);
   }

   public void onPaint(Graphics g)
   {
      if (hasScroll)
      {
         g.drawArrow(xa,ya,ha,Graphics.ARROW_UP,false,msg.canScroll(false),foreColor); // guich@200b4_143: msg.canScroll
         g.drawArrow(xa+wa,ya,ha,Graphics.ARROW_DOWN,false,msg.canScroll(true),foreColor);
      }
   }
   protected void onColorsChanged(boolean colorsChanged) // guich@220_47
   {
      if (colorsChanged)
      {
         msg.setBackForeColors(backColor, foreColor);
         if (btns != null)
            if (backColor.getAlpha() < foreColor.getAlpha()) // guich@300_12
               btns.setBackForeColors(foreColor, backColor);
            else
               btns.setBackForeColors(backColor, foreColor);
      }
   }

   /** handle scroll buttons and normal buttons */
   public void onEvent(Event e)
   {
      switch (e.type)
      {
         case ControlEvent.TIMER:
            if (e.target == this)
            {
               removeTimer(unpopTimer);
               unpop();
            }
            break;
         case PenEvent.PEN_DOWN:
            if (hasScroll)
            {
               int px=((PenEvent)e).x;
               int py=((PenEvent)e).y;

               if (ya <= py && py <= ya+ha && xa <= px && px < xa+(wa<<1) && msg.scroll((px-xa)/wa != 0)) // at the arrow points?
               {
                  repaint();
                  Sound.beep();
               }
            }
            break;
         case KeyEvent.KEY_PRESS: // guich@200b4_42
            int key = ((KeyEvent)e).key;
            if (key == IKeys.PAGE_UP || key == IKeys.UP || key == IKeys.JOG_UP) // guich@330_45
            {
               msg.scroll(false);
               repaint(); // guich@300_16: update the arrow's state
            }
            else
            if (key == IKeys.PAGE_DOWN || key == IKeys.DOWN || key == IKeys.JOG_DOWN) // guich@330_45
            {
               msg.scroll(true);
               repaint(); // guich@300_16: update the arrow's state
            }
            break;
         case ControlEvent.PRESSED:
            if (e.target == btns && (selected=btns.getSelected()) != -1)
            {
               btns.setSelected(-1);
               unpop();
            }
            break;
      }
   }

   /** returns the pressed button index, starting from 0 */
   public int getPressedButtonIndex()
   {
      return selected;
   }
}