/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.ui;

import waba.fx.*;
import waba.sys.*;

/** Constructs a PopupMenu with the given items.
   * This class is used in conjunction with the MenuBar. Although, you can also
   * use it to create a stand alone "right click" menu.
   * <p>It automatically saves and restores the covered area.
   * Here is an example of how to build the popupmenu for the DateBook:<br>
   * <pre>
   * String col1[] = // <b>note that the first String is always skipped</b>
   * {
   *   "Record",
   *   "NewEvent",
   *   "Delete Note...",
   *   "Purge...",
   *   "-",           // create a dotted line
   *   "Beam Event",
   * };
   * PopupMenu pop = new PopupMenu(10,10,col1);
   * popupModal(pop);
   * // a WINDOW_CLOSED event is generated when a item is selected. The Window
   * // is closed when the user clicks outside or selects an item. Use the getSelected
   * // method to discover the index of the String array that was selected (-1 if none).
   * </pre>
   * <b>Note</b>: the menu items must fit on screen. No clipping is applied. Also, the font and colors can be changed if desired.
   */

public class PopupMenu extends Window
{
   // changed drawing of menu border to look more native
   // PMD 25Oct2001
   // I have removed Rect[] rCaps because only the x component is used
   // replace it with int[] ypos
   private String []items;
   private int []ypos;
   private int selected=-1;
   private Graphics gr;
   private int popX,popY;
   private PenEvent pe = new PenEvent();
   private int lineHeight;
   private Color dColor,bColor,fColor,cursorColor;
   private Color fourColors[] = new Color[4];

   /** Create a MenuBar with the given items and that will be poped up at the given x,y coords. */
   public PopupMenu(int x, int y, String []items)
   {
      this.items = items;
      highResPrepared = true;
      canDrag = false;
      popX = x;
      popY = y;
      setFont(getFont().asBold()); // use a bold font.
      setBackColor(Color.WHITE);
      borderStyle = -1;  // guich@220_48
      started = true; // avoid calling the onStart method

      pe.target = this;
      pe.type = PenEvent.PEN_DOWN;
   }

   /** Change the font and recompute some parameters */
   protected void onFontChanged() // guich@200b4_153
   {
      // all this is recalculated because the font had changed
      int w = 0;
      lineHeight = fmH+1;
      ypos = new int[items.length];
      // compute the maximum width
      boolean haveCheckable = false;
      for (int i =1; i < items.length; i++)
      {
         String s = items[i];
         if (s.charAt(0) == MenuBar.DISABLED) s = s.substring(1);
         if (s.charAt(0) == MenuBar.CHECKED || s.charAt(0) == MenuBar.UNCHECKED)
         {
            haveCheckable = true;
            s = s.substring(1);
         }
         w = Math.max(w,fm.getTextWidth(s));
      }
      if (haveCheckable)
      {
         float factor = (float)fmH / 11f;
         w += (int)(factor*8)+1; // guich@300_1
      }
      // compute the rects
      int y=1;
      ypos[0] = y;
      for (int i=1; i < items.length; i++)
      {
         y += (items[i].charAt(0) == '-') ? 1 : lineHeight;
         ypos[i] = y;
      }
      // Check if we're not after the screen limits
      int screenW = Settings.screenWidth;
      w += 6;
      if (popX+w > screenW)
         popX = screenW-w;
      setRect(popX,popY,w,y+2);
   }

   /** Returns the selected index when this window was closed or -1 if non was selected */
   public int getSelected()
   {
      return selected;
   }

   /** Setup some important variables */
   protected void onPopup()
   {
      selected = -1;
      if (gr != null) gr.free();
      gr = createGraphics();
   }

   /** Close the popup list with a click outside its bounds */
   protected boolean onClickedOutside(int x,int y)
   {
      selected = -1;
      if (y < this.y) // clicked on MenuBar? propagate the event to it
      {
         Window w = (Window)zStack.items[zStack.getCount()-2];
         if (w instanceof waba.ui.MenuBar) // only do this if we was called by a MenuBar
         {
            pe.x = x;
            pe.y = y;
            w.onEvent(pe);
            if (topMost == this) requestFocus(); // we must get back the focus
            return true;
         }
      }
      if (x < this.x || x > this.x2 || y > this.y || y < this.y2) // guich@300_39: added the last condition to fix a problem when this is used not with a menubar.
      {
         unpop();
         if (Window.topMost instanceof MenuBar) // guich@200final_6: since in menu bars we don't load our behind, we ask the parent MenuBar to do it for us.
            ((MenuBar)Window.topMost).popupClosed(this);
      }
      return true;
   }

   private int getItemAt(int y)
   {
      for (int i =1; i < ypos.length; i++)
      {
         char c = items[i].charAt(0);
         if (ypos[i-1] <= y && y <= ypos[i] && c != MenuBar.DISABLED && c != '-')
            return i;
      }
      return -1;
   }

   private void drawCursor(Graphics g, int index, boolean on)
   {
      Color f = cursorColor != null ? cursorColor : bColor.getCursorColor(); // guich@220_49
      g.setForeColor(on?bColor:f);
      g.setBackColor(on?f:bColor);
      g.eraseRect(1,ypos[index-1],width-3,lineHeight);
   }

   public void onEvent(Event event)
   {
      switch (event.type)
      {
         case KeyEvent.KEY_PRESS:
            if ( ((KeyEvent)event).key == IKeys.MENU)
               unpop();
            Window w = Window.topMost;
            if (w instanceof MenuBar) ((MenuBar)w).close(); // close all menu windows
            break;
         case PenEvent.PEN_DOWN:
         case PenEvent.PEN_DRAG:
         case PenEvent.PEN_MOVE:
         case PenEvent.PEN_UP:
            PenEvent pe = (PenEvent)event;
            if (0 < pe.y && pe.y < height)
            {
               int newSelected = getItemAt(pe.y);
               if (newSelected != selected)
               {
                  if (selected != -1) drawCursor(gr,selected,false);
                  selected = newSelected;
                  if (selected != -1) drawCursor(gr,selected,true);
               }
               if (event.type == PenEvent.PEN_UP && selected != -1)
                  unpop();
            }
            else
            if (selected != -1) // outside valid area?
            {
               drawCursor(gr,selected,false);
               selected = -1;
            }
            break;
      }
   }

   /** Sets the cursor color. By default, it is based in the background color */
   public void setCursorColor(Color c) // guich@220_49
   {
      cursorColor = c;
   }

   protected void onColorsChanged(boolean colorsChanged)
   {
      fColor = getForeColor();
      bColor = getBackColor();
      dColor = fColor.getAlpha() > 128 ? fColor.darker() : fColor.brighter();
      if (colorsChanged)
      {
         Graphics.compute3dColors(true,backColor,foreColor,fourColors);
         if (cursorColor == null && Settings.maxColors >= 256)
            cursorColor = new Color(0,0,240);
      }
   }

   public void onPaint(Graphics g)
   {
      boolean checked;
      // paint border
      g.setForeColor(fColor);
      g.setBackColor(bColor);

      // changed drawing of popup menu border to look more native
      // PMD 25Oct2001
      if (borderStyle == -1) // guich@220_48
         g.draw3dRect(0,0,width,height,Graphics.R3D_SHADED,false,true,fourColors);

      // paing captions
      for (int i =1; i < items.length; i++)
      {
         String cap = items[i];
         char ch = cap.charAt(0);
         if (ch == '-') // dotted line?
            g.drawDots(0,ypos[i-1],width-3,ypos[i-1]);
         else
         {
            // is the menu item disabled?
            boolean dis = ch == MenuBar.DISABLED;
            if (dis)
            {
               g.setForeColor(dColor);
               cap = cap.substring(1);
               ch = cap.charAt(0);
            }
            // can the menu item be checkable?
            if ((checked = (ch == MenuBar.CHECKED)) || ch == MenuBar.UNCHECKED)
            {
               cap = cap.substring(1);
               if (checked)
               {
                  // draws the check
                  int x = width-10;
                  int y = ypos[i-1] + 5;
                  int my = 2;
                  int mx = 7;
                  if (fmH > 11)
                  {
                     float factor = (float)fmH / 11f;
                     int k = (int)(2*factor); // guich@300_1
                     x -= k;
                     y += k;
                     my += k>>1;
                     mx += k;
                  }
                  for (int j = 0; j < mx; j++)
                  {
                     g.drawLine(x, y, x, y + 2);
                     x++;
                     if (j < my)
                        y++;
                     else
                        y--;
                  }
               }
            }
            g.drawText(cap,3,ypos[i-1]);
            if (dis) g.setForeColor(fColor);
         }
      }
      // paint cursor
      if (selected != -1)
         drawCursor(g,selected,true);
   }
}