/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.ui;

import waba.fx.*;
import waba.sys.*;

/** Group or matrix of pushbuttons in just one control. Is one of the most versatiles controls of superwaba. Created by guich from the original
  * WExtras class Pushbutton of Stefan Kellner.
  * here is a example of constructor:
  * <pre>
  * new PushButtonGroup(new String[]{"Button1","Button2","Button3"},false,-1,-1,4,0,false,PushButtonGroup.NORMAL)
  * </pre>
  * <p>Note: the width of each button is calculated based on its caption size plus the insideGap.
  * Now, the height is calculated based on the font's size or on the height you specified.
  */

public class PushButtonGroup extends Control
{
   /** normal: only one selected at a time */
   public static final byte NORMAL = 0;
   /** the button will be selected and de-selected immediatly, acting like a real button */
   public static final byte BUTTON = 1;
   /** one click in the button will select it and another click will unselect it. However, only one button can be selected at a time */
   public static final byte CHECK = 2;
   protected String []names; // guich@401_36: made protected
   private int widths[];
   private int selectedIndex = -1;
   private int gap;
   private int insideGap;
   private int rows, cols;
   private boolean atLeastOne,actLikeButton,actLikeCheck;
   private int cellH,rowH;
   private int lastSel=-1;
   public int maxWidth=-1;
   private boolean simpleBorder = Settings.uiStyle == Settings.PalmOS;
   private Graphics myg;
   private Rect rects[];
   private int count;
   private boolean allSameWidth;
   private int[] tX;
   private Color dColor,fColor;
   private Color fourColors[] = new Color[4];
   private Color userCursorColor;

   /** create the buttons.
       @param names captions of the buttons. You can specify some names as null so the button is not displayed. This is good if you're creating a button matrix and want to hide some buttons.
       @param atLeastOne if true, at least one button must be selected
       @param selected default index to appear selected, or -1 if none
       @param gap space between the buttons, -1 glue them.
       @param insideGap space between the text and the button border. the ideal is 6.
       @param rows if > 1, creates a button matrix
       @param allSameWidth if true, all the buttons will have the width of the most large one.
       @param type can be NORMAL, BUTTON or CHECK
   */
   public PushButtonGroup(String[] names, boolean atLeastOne, int selected, int gap, int insideGap, int rows, boolean allSameWidth, byte type)
   {
      this.names = names;
      this.insideGap = insideGap;
      this.atLeastOne = atLeastOne;
      this.actLikeButton = type == BUTTON;
      this.actLikeCheck = type == CHECK;
      this.allSameWidth = allSameWidth;
      this.selectedIndex = selected;
      this.gap = gap;
      count = names.length;
      widths = new int[count];
      rects = new Rect[count];
      tX = new int[count];
      if (rows < 1) rows = 1;
      this.rows = rows;
      cols = count / rows;
      if (count % cols != 0) // guich@340_28: needs more rows than given
         this.rows++;
      onFontChanged();
   }
   /** uses the old style border */
   public void setSimpleBorder(boolean simple)
   {
      this.simpleBorder = simple || Settings.uiStyle == Settings.PalmOS;
   }
   /** returns the index of the selected button */
   public int getSelected()
   {
      return selectedIndex;
   }
   /** Returns the caption of the selected button, or <code>null</code> if no button is selected 
     * @since SuperWaba 4.01
     */
   public String getSelectedCaption() // guich@401_1
   {
      return selectedIndex == -1 ? null : names[selectedIndex];
   }
   /** Sets the cursor color for this PushButtonGroup. The default is equal to the background slightly darker. Make sure you tested it in 2,4 and 8bpp devices. */
   public void setCursorColor(Color color) // guich@210_19
   {
      this.userCursorColor = color;
      onColorsChanged(true); // jrissoto@220_26
   }
   public int getPreferredWidth()
   {
      int w = 0;
      int wc = 0;
      if (count == rows) // only one column?
      {
         for (int i =0; i < count; i++)
            w = Math.max(w,widths[i]);
      }
      else
      {
         for (int i =0; i < count; i++) // compute the maximum size of each row
         {
            wc += getControlWidth(i)+gap;
            if (i != 0 && ((i+1)%cols == 0))
            {
               wc -= gap;
               w = (wc > w)?wc:w;
               wc = 0;
            }
         }
         w = (wc > w)?wc:w;
      }
      return w;
   }

   public int getPreferredHeight()
   {
      return (fmH+(simpleBorder?0:2)+gap)*rows - gap; // remove the last gap
   }

   private int getControlWidth(int ind)
   {
      return (maxWidth == -1)?widths[ind] :maxWidth;
   }

   protected void onFontChanged()
   {
      // computes the best width for all controls
      int avg = 0;
      int nullNames = 0;
      maxWidth=-1;
      for (int i =0; i < count; i++)
      {
         if (names[i] == null)
            nullNames++;
         else
            avg += (widths[i] = fm.getTextWidth(names[i])+insideGap);
         if (allSameWidth) maxWidth = Math.max(maxWidth,widths[i]);
      }
      // guich@200b4: search again for possible null names and set it width as the average width
      if (nullNames > 0)
      {
         avg /= count-nullNames;
         for (int i =0; i < count; i++)
            if (names[i] == null)
               widths[i] = avg;
      }
   }

   protected void onColorsChanged(boolean colorsChanged)
   {
      if (colorsChanged) // backColor has changed
         dColor = userCursorColor!=null?userCursorColor:backColor.getCursorColor();
      fColor = getForeColor();;
      Graphics.compute3dColors(enabled,backColor,foreColor,fourColors);
   }

   public void onBoundsChanged()
   {
      rowH = (height+gap) / rows; // calculate through the given height
      cellH = rowH-gap; // guich@200b4: corrected if more than 1 row
      // compute the rects
      int i =0;
      int c = cols;
      int x = 0;
      int y = 0;
      while (true)
      {
         int w = getControlWidth(i);
         if (names[i] != null)
            rects[i] = new Rect(x,y,w,cellH);
         tX[i] = x+((w-widths[i]+insideGap) >> 1);
         if (++i >= count) break;
         if (--c == 0)
         {
            x = 0;
            y += rowH;
            c = cols;
         }
         else x += w + gap;
      }
      myg = null; // jrissoto@220_25
   }

   public void onPaint(Graphics g)
   {
      if (myg == null) myg = createGraphics();
      int ty = (cellH-fmH) / 2; // nopt
      g.setBackColor(backColor);
      g.setForeColor(fColor);
      boolean drawEachBack = gap > 0 && parent != null && backColor.equ != parent.backColor.equ; // guich@230_34
      if (!drawEachBack)
         g.fillRect(0,0,width,height);
      Rect r;
      for (int i=0; i < count; i++)
         if ((r = rects[i]) != null)
         {
            if (drawEachBack)
               g.fillRect(r.x,r.y,r.width,r.height);
            if (simpleBorder)
               g.drawRect(r.x,r.y,r.width,r.height);
            else
               g.draw3dRect(r.x,r.y,r.width,r.height,actLikeCheck?Graphics.R3D_CHECK:Graphics.R3D_RAISED,false,false,fourColors);
         }
      g.setForeColor(fColor);
      for (int i=0; i < count; i++)
         if ((r = rects[i]) != null)
         {
            g.setClip(r.x+1,r.y+1,r.width-2,r.height-2);
            g.drawText(names[i], tX[i], r.y+ty); // if allSameWidth, center the label in the button
         }
      g.clearClip();
      if (selectedIndex != -1) drawCursor(g,selectedIndex,true);
   }

   /** sets the selected button index. Note. if there are any null buttons, you must count them too to compute the correct index. */
   public void setSelected(int ind)
   {
      int min = atLeastOne?0:-1;
      if ((actLikeCheck || selectedIndex != ind) && min <= ind && ind < count)
      {
         if (myg != null && selectedIndex >= 0) drawCursor(myg,selectedIndex,false); // remove from the last
         selectedIndex = ind;
         if (myg != null && selectedIndex >= 0 && names[ind] != null) drawCursor(myg,ind,true); // draw on the new
      }
   }
   private void drawCursor(Graphics g, int ind, boolean on)
   {
      Rect r;
      if ((r = rects[ind]) != null)
      {
         g.setForeColor(on?backColor:dColor);
         g.setBackColor(on?dColor:backColor);
         int k = simpleBorder?1:2;
         g.eraseRect(r.x+k,r.y+k,r.width-(k<<1),r.height-(k<<1));
      }
   }

   private int findButtonAt(int px, int py)
   {
      if (0 <= px && px < width && 0 <= py && py < height)
      {
         Rect r;
         int s = cols*(py / rowH);
         int e = Math.min(count,s+cols);
         for (int i =s; i < e; i++)
            if ((r=rects[i]) != null && r.contains(px,py))
               return i;
      }
      return -1;
   }

   public void onEvent(Event event)
   {
      if (event instanceof PenEvent)
      {
         int sel = findButtonAt(((PenEvent)event).x,((PenEvent)event).y);
         switch (event.type)
         {
            case PenEvent.PEN_MOVE:
            case PenEvent.PEN_DRAG:
            case PenEvent.PEN_DOWN:
               if (sel != selectedIndex && (!atLeastOne || sel != -1))
                  setSelected(sel);
               break;
            case PenEvent.PEN_UP:
               if (!atLeastOne || sel != -1)
                  postEvent(new ControlEvent(ControlEvent.PRESSED,this));
               if (actLikeCheck)
               {
                  if (lastSel != selectedIndex)
                     lastSel = selectedIndex;
                  else
                  {
                     lastSel = -1;
                     setSelected(-1);
                  }
               } else
               if (actLikeButton || (sel == -1 && !atLeastOne)) {Vm.sleep(150); setSelected(-1);}
               break;
         }
      }
      else
      if (event.type == ControlEvent.WINDOW_MOVED)
      {
         if (myg != null) myg.free();
         myg = createGraphics();
      }
   }
}