/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2003 Guilherme Campos Hazan <support@superwaba.com.br>         *
 *  Copyright (C) 2001 Daniel Tauchke                                            *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

// ScrollBar.java written 2001 by Daniel Tauschke version
// and changed by Guich.
// http://www.tauschke.com
// E-mail: tauschke@ tauschke.com

package waba.ui;

import waba.fx.*;
import waba.sys.*;

/**
 * ScrollBar is a simple implementation of a Scrollbar.
 * The scroll bar orientation can be horizontal or vertical.
 * Here is an example of how to use it.
 * <pre>
 *  // declarations
 *  ScrollBar sb1;
 *  ScrollBar sb2;
 *  ScrollBar sb3;
 *  ScrollBar sb4;
 *  // init
 *  public void testScrollBars()
 *  {
 *     add(sb1 = new ScrollBar(ScrollBar.VERTICAL));
 *     sb1.setRect(RIGHT, CENTER, PREFERRED, waba.sys.Settings.screenHeight/2);
 *
 *     add(sb2 = new ScrollBar(ScrollBar.VERTICAL));
 *     sb2.setRect(BEFORE, SAME, PREFERRED, SAME);
 *     sb2.setLiveScrolling(true);
 *
 *     add(sb3 = new ScrollBar(ScrollBar.HORIZONTAL));
 *     sb3.setRect(LEFT,CENTER,waba.sys.Settings.screenWidth/2, PREFERRED);
 *
 *     add(sb4 = new ScrollBar(ScrollBar.HORIZONTAL));
 *     sb4.setRect(SAME, AFTER, SAME, PREFERRED);
 *  }
 *
 *
 *  public void onEvent(Event event)
 *  {
 *     if (event.type == ControlEvent.PRESSED && event.target == sb2)
 *     {
 *        int value = sb2.getValue();
 *        sb1.setValue(value);
 *        sb3.setValue(value);
 *        sb4.setValue(value);
 *     }
 *  }
 *
 */

public class ScrollBar extends Container
{
   /** To be passed in the constructor */
   public static final byte VERTICAL = 1;
   /** To be passed in the constructor */
   public static final byte HORIZONTAL = 2;

   private int maximum = 100;
   private int minimum = 0;
   private int blockIncrement = 50;
   private int unitIncrement = 1;
   private int value = 0;
   private int visibleItems = 50;

   private double valuesPerPixel;
   private int dragBarSize;
   private int btnWH;

   private Button btnInc, btnDec;
   private int startDragPos=-1;
   private boolean verticalBar;

   private int dragBarPos;
   private boolean liveScrolling;
   private int size;
   private int dragBarMax;
   private Color bColor,sbColor,sfColor;
   private Color fourColors[] = new Color[4];
   private static boolean isGray4;

   /** Creates Scrollbar with default values:<br>
   *  maximum = 100 <br>
   *  minimum = 0<br>
   *  orientation = VERTICAL<br>
   *  blockIncrement = 10<br>
   *  unitIncrement = 1<br>
   *  value = 0<br>
   *  visibleItems = 10<br>
   */
   public ScrollBar()
   {
      this(VERTICAL);
   }

   public ScrollBar(byte orientation)
   {
      this.verticalBar = orientation == VERTICAL;
      add(btnDec = Button.createArrowButton(verticalBar?Graphics.ARROW_UP  :Graphics.ARROW_LEFT ,fmH*3/11,Color.BLACK));
      add(btnInc = Button.createArrowButton(verticalBar?Graphics.ARROW_DOWN:Graphics.ARROW_RIGHT,fmH*3/11,Color.BLACK));
      btnDec.setBorder((Settings.uiStyle == Settings.PalmOS)?Button.BORDER_NONE:Button.BORDER_3D);
      btnInc.setBorder((Settings.uiStyle == Settings.PalmOS)?Button.BORDER_NONE:Button.BORDER_3D);
      isGray4 = Settings.maxColors == 4;
      started = true; // avoid calling the onStart method
   }

   /** Sets the value, visibleItems, minimum and maximum values */
   public void setValues(int newValue, int newVisibleItems, int newMinimum, int newMaximum)
   {
      maximum = newMaximum;
      minimum = newMinimum;
      visibleItems = Math.max(newVisibleItems,1);
      blockIncrement = visibleItems; // /  4
      if (visibleItems+newValue+minimum <= maximum)
         value = newValue;
      recomputeParams(false);
      repaint();
   }

   /** Set the maximum of elements */
   public void setMaximum(int i)
   {
      if (i != maximum) // guich@320_30
      {
         maximum = i;
         recomputeParams(false);
         repaint();
      }
   }

   /** Get the maximum of elements */
   public int getMaximum()
   {
      return maximum;
   }

   /** Set the minimum of elements */
   public void setMinimum(int i)
   {
      minimum = i;
      recomputeParams(false);
      repaint();
   }

   /** Get the minimum of elements */
   public int getMinimum()
   {
      return minimum;
   }

   /** Set the items to increment by clicking the above the scrollbar.
       This value is set as default to be equal to visibleItems. */
   public void setBlockIncrement(int i)
   {
      blockIncrement = i;
   }

   /** Get the items to increment by clicking the above the scrollbar*/
   public int getBlockIncrement()
   {
      return blockIncrement;
   }

   /** Set the items to increment by clicking the up button */
   public void setUnitIncrement(int i)
   {
      unitIncrement = i;
      repaint();
   }

   /** Get the items to increment by clicking the above the scrollbar*/
   public int getUnitIncrement()
   {
      return unitIncrement;
   }

   /** Set the value if visibleItems+i+minimum <= maximum. */
   public void setValue(int i)
   {
      //Vm.debug("i: "+i+", visibleItems: "+visibleItems+", maximum: "+maximum);
      if (i != value && visibleItems+i+minimum <= maximum) // guich@320_30
      {
         value = i;
         recomputeParams(true); // can't remove this line.
         repaint();
      }
   }

   /** Get the value. This is the value minus the visible items.*/
   public int getValue()
   {
      return value;
   }

   /** Set the count of visible items for the scrollbar. This value cannot be zero, neither can change (unless the control height changes). It also sets the blockIncrement to be equal to the given value. */
   public void setVisibleItems(int i)
   {
      visibleItems = i;
      if (visibleItems <= 0) visibleItems = 1;
      blockIncrement = visibleItems;
      recomputeParams(false);
      repaint();
   }

   /** Get the count of visible items for the scrollbar*/
   public int getVisibleItems()
   {
      return visibleItems;
   }

   /** Set the live scrolling. If "true" an event is thrown during dragging.*/
   public void setLiveScrolling(boolean liveScrolling)
   {
      this.liveScrolling = liveScrolling;
   }

   private void recomputeParams(boolean justValue)
   {
      if (size <= 0) return;
      if (!justValue)
      {
         // Calculate and draw the slider button
         int delta = Math.max(visibleItems, maximum-minimum);
         int barArea = size-(btnWH<<1);
         valuesPerPixel = (double)barArea / (double)delta;
         dragBarSize = Math.min(barArea,(int) (valuesPerPixel * visibleItems)+1); // guich@300_13
         if (dragBarSize < 3)
         {
            dragBarSize = 3;
            //valuesPerPixel = (double)visibleItems / (double)dragBarSize; guich@220_5: had to remove this bc adding 1000 items to a ListBox with this make the scroll bar work wrong
         }
         dragBarMax = size-btnWH-dragBarSize;
      }
      dragBarPos = Math.min(dragBarMax,btnWH+(int)(valuesPerPixel * value + 5e-1)); // round value

      //Vm.debug("valuesPerPixel: "+valuesPerPixel+", visibleItems: "+visibleItems+", delta: "+delta+", blockIncrement: "+blockIncrement);
      boolean b;
      Window win = getParentWindow();
      btnDec.setEnabled(b = (enabled && value > minimum));
      if (!b && win != null && win.getFocus() == btnDec) // if we're disabled now and we got the focus, move focus to other control
         requestFocus();
      btnInc.setEnabled(b = (enabled && value+visibleItems < maximum));
      if (!b && win != null && win.getFocus() == btnDec) // if we're disabled now and we got the focus, move focus to other control
         requestFocus();
   }

   public void onEvent(Event event)
   {
      int oldValue = value;
      boolean mustPostEvent = false;
      switch (event.type)
      {
         case ControlEvent.PRESSED:
            if (event.target == btnDec)
               {value=Math.max(minimum,value-unitIncrement); event.consumed = true; requestFocus();} // guich@220_12: unitIncrement now is being used
            else
            if (event.target == btnInc)
               {value=Math.min(maximum,value+unitIncrement); event.consumed = true; requestFocus();}
            //mustPostEvent = true; afarine@350_15: avoid extra 2 events when pressing a button - the event will be posted in the penUp event
            break;
         case PenEvent.PEN_DOWN:
            if (event.target == this) // can be the buttons
            {
               int pos = verticalBar?((PenEvent)event).y:((PenEvent)event).x;
               if (pos < dragBarPos)
                  value -= blockIncrement;
               else
               if (pos > dragBarPos+dragBarSize)
                  value += blockIncrement;
               else
                  startDragPos = pos-dragBarPos; // point inside drag bar
            }
            break;
         case PenEvent.PEN_DRAG:
            if (startDragPos != -1)
            {
               int pos = verticalBar?((PenEvent)event).y:((PenEvent)event).x;
               dragBarPos = pos - startDragPos;
               if (dragBarPos < btnWH) dragBarPos = btnWH; else
               if (dragBarPos > dragBarMax) dragBarPos = dragBarMax;
               if (dragBarPos == dragBarMax) // guich@240_12: make sure that at the maximum bar pos we get the maximum value
                  value = maximum;
               else
               {
                  value = (int) ((dragBarPos-btnWH)/valuesPerPixel);
                  if (unitIncrement != 1) // guich@340_45: snap the scrollbar to the nearest increment
                     value = ((int)value/unitIncrement)*unitIncrement;
               }
            }
            break;
         case PenEvent.PEN_UP:
            startDragPos = -1;
            mustPostEvent = btnDec.enabled || btnInc.enabled;
            break;
         case KeyEvent.KEY_PRESS:
            int key = ((KeyEvent)event).key;
            if (key == IKeys.PAGE_UP || key == IKeys.UP || key == IKeys.JOG_UP) // guich@330_45
            {
               if (btnDec.enabled)
                  value -= blockIncrement;
               event.consumed = true;
               mustPostEvent = true; // guich@240_21
            }
            else
            if (key == IKeys.PAGE_DOWN || key == IKeys.DOWN || key == IKeys.JOG_DOWN) // guich@330_45
            {
               if (btnInc.enabled)
                  value += blockIncrement;
               event.consumed = true;
               mustPostEvent = true; // guich@240_21
            }
            break;
      }

      // only repaint if necessary
      if (value != oldValue)
      {
         if (value > maximum-visibleItems)
            value = maximum-visibleItems;
         else
         if (value < minimum)
            value = minimum;

         if (value != oldValue)
         {
            recomputeParams(true);
            if (liveScrolling || mustPostEvent)
               postEvent(new ControlEvent(ControlEvent.PRESSED, this));
            repaint();
         }
      }
      else
      if (mustPostEvent)
      {
         parent.postEvent(new ControlEvent(ControlEvent.PRESSED, this));
         mustPostEvent = false;
      }
   }

   public void setEnabled(boolean enabled)
   {
      if (enabled != this.enabled)
      {
         this.enabled = enabled;
         btnDec.setEnabled(enabled && value > minimum);
         btnInc.setEnabled(enabled && value+visibleItems < maximum);
         onColorsChanged(false);
         repaint();
      }
   }

   protected void onColorsChanged(boolean colorsChanged)
   {
      if (colorsChanged)
      {
         btnDec.setBackForeColors(backColor,foreColor);
         btnInc.setBackForeColors(backColor,foreColor);
      }
      if (Settings.uiStyle == Settings.PalmOS)
      {
         bColor = backColor;
         sbColor = getForeColor();
         sfColor = enabled?sbColor.brighter():sbColor;
      }
      else bColor = getBackColor().brighter();
      if (parent != null && bColor.equ == parent.getBackColor().equ)
         bColor = bColor.getCursorColor();
      Graphics.compute3dColors(enabled,backColor,foreColor,fourColors);
   }

   public void onPaint(Graphics g)
   {
      // Draw background and borders
      if (Settings.uiStyle == Settings.PalmOS)
      {
         g.setBackColor(backColor);
         g.fillRect(0,0,width,height);
         g.setBackColor(sbColor);
         g.setForeColor(sfColor);
         int c = btnWH >> 1;
         if (verticalBar)
         {
            g.drawLine(c,btnWH,c,size-btnWH);
            g.fillRect(c-1,dragBarPos,3,dragBarSize);
         }
         else
         {
            g.drawLine(btnWH,c,size-btnWH,c);
            g.fillRect(dragBarPos,c-1,dragBarSize,3);
         }
      }
      else
      {
         g.setBackColor(bColor);
         int k = size - (btnWH<<1);
         if (k <= 0) return;
         if (verticalBar)
            g.fillRect(0,btnWH,btnWH,k);
         else
            g.fillRect(btnWH,0,k,btnWH);
         g.setBackColor(backColor);
         if (verticalBar)
         {
            g.fillRect(0,dragBarPos,width,dragBarSize);
            g.draw3dRect(0,dragBarPos,width,dragBarSize,Graphics.R3D_RAISED,false,false,fourColors);
         }
         else
         {
            g.fillRect(dragBarPos,0,dragBarSize,height);
            g.draw3dRect(dragBarPos,0,dragBarSize,height,Graphics.R3D_RAISED,false,false,fourColors);
         }
      }
   }

   /** If this is a vertical scroll bar, i strongly suggest you use PREFERRED in your control's width. */
   public int getPreferredWidth()
   {
      return verticalBar?btnDec.getPreferredWidth():(btnDec.getPreferredHeight()<<1); // guich@240_18 - guich@300_70
   }
   /** If this is a horizontal scroll bar, i strongly suggest you use PREFERRED in your control's height */
   public int getPreferredHeight()
   {
      return verticalBar?(btnDec.getPreferredWidth()<<1):btnDec.getPreferredHeight(); // guich@300_70: vertical bar always use width; horizontal always use height
   }

   /* this is needed to recalculate the box size for the selected item if the control is resized by the main application */
   protected void onBoundsChanged()
   {
      size = verticalBar?height:width;
      btnWH = verticalBar?width:height;
      btnDec.setRect(0,0,btnWH,btnWH);
      btnInc.setRect(verticalBar?0:(size-btnWH),verticalBar?(size-btnWH):0,btnWH,btnWH);
      dragBarPos = btnWH;
      recomputeParams(false);
   }
}