/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>                         *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.ui;

import waba.fx.*;
import waba.sys.*;

/**
 * Window is a "floating" top-level window. This class IS functional like a modal window.
 * <p>
 * Attention! Is not a good practice use this class as a Container, adding it to other Window.
 * Use this class only as a popup window, or strange things can happen!
 * <P>
 * Important: the <code>popupModal</code> method is <b>not blocking</b>, but the <code>popupBlockingModal</code> <b>is</b>
 * blocking. In other words, after you call popupModal, the process
 * continues immediately to the next line. You'll need to catch the <code>WINDOW_CLOSED</code> event
 * to do things after the user had really closed the window.
 * If you want to block until the Window is dismissed, just use <code>popupBlockingModal</code>
 * <b>Very important!</b> <code>popupBlockingModal</code> can't be used on the program startup, ie, in the
 * constructor, onStart or other application initialization methods. The main event loop must have been initialized
 * in order to use it. Otherwise, the whole application will just halt!
 * <p>
 * The following example creates an popup window class:
 *<pre>
 *  class TestWindow extends Window
 *  {
 *     Button btnHi;
 *     public TestWindow()
 *     {
 *        super("Test",RECT_BORDER); // with caption and borders
 *        setRect(CENTER,CENTER,Settings.screenWidth/2,Settings.screenHeight/4);
 *        add(btnHi=new Button("Hi!"),CENTER,CENTER);
 *     }
 *     public void onEvent(Event event)
 *     {
 *        if (event.type == ControlEvent.PRESSED && event.target == btnHi)
 *           unpop(); // a WINDOW_CLOSED event will be posted to this PARENT window by the Window class.
 *     }
 *  }
 *</pre>
 * To use it (<b>non-blocking</b>):
 *<pre>
 *  public class Main extends MainWindow
 *  {
 *     TestWindow tw;
 *     public void onStart()
 *     {
 *        tw = new TestWindow();
 *        tw.dontSaveBehind(true); // dont save behind (theres still nothing to be saved) and call repaint when window is closed - only needed when opening windows in onStart
 *        popupModal(tw);
 *        // this line is executed immediately
 *     }
 *     public void onEvent(Event event)
 *     {
 *        if (event.target == tw && event.type == ControlEvent.WINDOW_CLOSED)
 *        {
 *           // any stuff
 *        }
 *     }
 *  }
 *</pre>
 * To use it (<b>blocking</b>):
 *<pre>
 *  public class Main extends MainWindow
 *  {
 *     Button btn;
 *     public void onEvent(Event e)
 *     {
 *        if (e.target == btn)
 *        {
 *           TestWindow tw = new TestWindow();
 *           popupBlockingModal(tw);
 *           // this line is only executed after the window is closed.
 *        }
 *     }
 *  }
 *</pre>
 * PopupBlockingModal may be use in InputDialog/MessageBox classes, while PopupModal
 * is used in MenuBar and other classes.
 * <p> Important note: you can't use popupBlockingModal with a delay to unpop it.
 * In this case, the correct would be to use popupModal:
 * <pre>
 * popupModal(mb = new MessageBox(...));
 * Vm.sleep(5000);
 * mb.unpop;
 * </pre>
 * If you use popupBlockingModal in this specific case, the Vm may hang.
 */

public class Window extends Container
{
   protected boolean needsPaint;
   protected String title; // guich@102
   protected boolean beepIfOut = true; // guich@102: cancels the event if the user type outside the window area. to be used in a future window manager
   /** If true, the background is erased on each paint. If false, you must take care of erasing the background */
   protected boolean flicker = true;
   protected static Window topMost;
   protected Font titleFont = MainWindow.defaultFont.asBold(); // guich@110
   protected byte borderStyle = NO_BORDER;
   protected Control menubar; // guich@200
   protected boolean canDrag = true;
   public boolean eraseBackgroundNow; // guich@200b4_163

   private byte []borderGaps = {0,2,3,2,0}; // guich@200final_14 - guich@400_77
   private byte []topBorderGaps = {0,2,3,2,2}; // guich@200final_14
   private Image imgBuf; // guich@101
   private Graphics gbuf; // guich@101
   protected Image imgCovered;
   private Graphics _g;
   private int saveBehindWhenPopup = 1;
   private int paintX, paintY, paintWidth, paintHeight;
   private Control _focus;
   private boolean _inPenDrag;
   private boolean doubleBuf; // guich@101
   private Control focusOnPopup; // last control that had focus when popup was called.
   private Graphics gMoving;
   private int xMoving,yMoving;
   private int xDeltaMoving, yDeltaMoving;
   private Coord ptMoving;
   private Control tempFocus; // guich@320_31
   private FontMetrics fmTitle;
   private Rect rTitle; // guich@200b4_52: the area where the title is located
   /** Must set to true if your Window is prepared for 320x320 resolutions.
     If false (default), the Window is doubled size (and centered) to make controls fit */
   protected boolean highResPrepared = !Settings.platform.equals("PalmOS"); // guich@400_35: as default for WinCE, highres is true

   /** Used in popupBlockingModal */
   boolean popped;
   /** Used in swap */
   protected Container mainSwapContainer;
   /** Used in swap */
   protected Container lastSwappedContainer;

   public static final byte NO_BORDER = 0;
   public static final byte RECT_BORDER = 1;
   public static final byte ROUND_BORDER = 2;
   public static final byte TAB_BORDER = 3;
   public static final byte TAB_ONLY_BORDER = 4;
   /** used to hide the showing state */
   public static final int HIDE_STATE = 1000;

   /** used to hide the virtual keyboard */
   public static final int VK_HIDE = 10000;
   /** used to place the keyboard on top of screen */
   public static final int VK_TOP = 10001;
   /** used to place the keyboard on bottom of screen */
   public static final int VK_BOTTOM = 10002;
   /** used to show the virtual keyboard, without changing the position */
   public static final int VK_SHOW = 10003;

   /** Dont mess with this, unless you want to crash the VM! */
   public static waba.util.Vector zStack = new waba.util.Vector(5); // guich@102

   // event objects to minimize memory usage
   static KeyEvent _keyEvent = new KeyEvent();
   static PenEvent _penEvent = new PenEvent();
   static ControlEvent _controlEvent = new ControlEvent();
   static boolean repaintLastPopup = false;
   ////////////////////////////////////////////////////////////////////////////////////
   ////////////////////////////////////////////////////////////////////////////////////
   /** Constructs a window. */
   public Window()
   {
      NativeMethods4JDK.windowCreate(this);
      this.x2 = x+width-1; // kambizDarabi@200final_7: added changes to fix the behaviour of no events being catched if you don't call setRect in the Window.
      this.y2 = y+height-1;
      parentWindow = this;
      asWindow = this;
      fmTitle = getFontMetrics(titleFont);
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** Constructs a window with the given title and border borderStyle. */
   public Window(String title, byte borderStyle) // guich@112
   {
      this();
      this.title = title != null && title.length() > 0 ? title : null;
      this.borderStyle = borderStyle;
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** sets the title font */
   public void setTitleFont(Font titleFont)
   {
      this.titleFont = titleFont;
      fmTitle = getFontMetrics(titleFont);
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** sets the title and call repaint. if you want an imediate repaint, call paintTitle. */
   public void setTitle(String title) // guich@102
   {
      this.title = title;
      this.rTitle = null; // guich@400_38
      repaint();
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** sets the border borderStyle. use the constants XXXX_BORDER. sets the flag border accordingly to the borderStyle. */
   public void setBorderStyle(byte borderStyle)
   {
      this.borderStyle = borderStyle;
      repaint();
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** Sets true if we have to use an double buffer to paint the screen. */
   public void setDoubleBuffer(boolean doubleBuf)
   {
      this.doubleBuf = doubleBuf;
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** Returns the image that is used for the offscreen (double buffering).
    * if there is no double buffer or an out of memory problem occurs, returns null. Initializes the buffer if necessary.
    */
   public Image getOffScreen()
   {
      if (imgBuf == null && doubleBuf) // guich
      {
         try
         {
            imgBuf = new Image(this.width, this.height);
            gbuf = imgBuf.getGraphics();
         } catch (OutOfMemoryError oome) {doubleBuf = false; imgBuf = null;} // guich@340_52
      }
      return imgBuf;
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** Makes this window static, i.e., the user will not be able to move it around
     * the screen by dragging the title area. It just set the canDrag member to false.
     */
   public void makeUnmovable() // guich@200b4_194
   {
      canDrag = false;
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /**
    * Sets focus to the given control. When a user types a key, the control with
    * focus get the key event. At any given time, only one control in a window
    * can have focus. Calling this method will cause a FOCUS_OUT control event
    * to be posted to the window's current focus control (if one exists)
    * and will cause a FOCUS_IN control event to be posted to the new focus
    * control.
    */
   public void setFocus(Control c)
   {
      if (c != null && !c.focusLess) // guich@320_31
      {
         if (_focus != null)
         {
            _controlEvent.type = ControlEvent.FOCUS_OUT;
            _controlEvent.target = _focus;
            _controlEvent.touch();
            _focus.postEvent(_controlEvent);
         }
         _focus = c;
         if (c != null)
         {
            _controlEvent.type = ControlEvent.FOCUS_IN;
            _controlEvent.target = c;
            _controlEvent.touch();
            c.postEvent(_controlEvent);
         }
      }
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** Changes the focus to the desired control, without posting FOCUS_IN and FOCUS_OUT events 
     * @since SuperWaba 4.01
     */
   protected void swapFocus(Control c)
   {
      _focus = c;
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /**
    * Returns the focus control for this window.
    * @see waba.ui.Window#setFocus
    */
   public Control getFocus()
   {
      return _focus;
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /**
    * Adds a damage rectangle to the current list of areas that need
    * repainting.
    */
   protected void damageRect(int x, int y, int width, int height)
   {
      if (needsPaint)
      {
         int ax = x + width;
         int ay = y + height;
         int bx = paintX + paintWidth;
         int by = paintY + paintHeight;
         if (paintX < x)
            x = paintX;
         if (paintY < y)
            y = paintY;
         if (ax > bx)
            width = ax - x;
         else
            width = bx - x;
         if (ay > by)
            height = ay - y;
         else
            height = by - y;
      }
      // guich@102 : makes sure that the paint is inside the area of the control
      paintX = x;
      paintY = y;
      paintWidth = (width <= this.width)?width:this.width;
      paintHeight = (height <= this.height)?height:this.height;
      needsPaint = true;
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** called when the user clicks outside the bounds of this window. must return true
    * if the event was handled, false otherwise. If false is returned and <code>beepIfOut</code>
    * is true, then a beep is played and nothing more happens.
    * @since SuperWaba 1.2 */
   protected boolean onClickedOutside(int x, int y)
   {
      return false;
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** Calls _doPaint if the window needs painting.
     * @since SuperWaba 4.0
     */
   public void validate() // guich@400_44
   {
      if (needsPaint)
         _doPaint(paintX, paintY, paintWidth, paintHeight);
   }
   ////////////////////////////////////////////////////////////////////////////////////
   private static boolean ignoreNextEvent;
   /**
    * Called by the VM to post key and pen events.
    */
   public void _postEvent(int type, int key, int x, int y, int modifiers, int timeStamp)
   {      
      if (type == -6000) // guich@310_18: invalidate all stored mygs. called by the VM usualy after a popup window has been called.
      {
         _g = createGraphics();
         MainWindow.myg0 = createGraphics();
         int n = zStack.getCount();
         for (int i =0; i < n; i++) // redraws each window, from first to last parent
         {
            Window w = (Window)zStack.items[i];
            w.broadcastEvent(new ControlEvent(ControlEvent.WINDOW_MOVED, w));
         }
      }
      if (ignoreNextEvent) // set when the user pops up the Find window and he presses the close button; the event is propagated to here; not what we want.
      {
         ignoreNextEvent = false;
         return;
      }
      if (this != topMost) // if this isnt the top-most window, pass control to it. i must do this because the vm calls MainWindow._postEvent
      {
         topMost._postEvent(type, key, x, y, modifiers, timeStamp);
         return;
      }
      if (timeStamp == 0) timeStamp = Vm.getTimeStamp(); // guich@401_13: get the timestamp
      boolean invokeMenu = false;
      boolean isPenEvent = PenEvent.PEN_DOWN <= type && type <= PenEvent.PEN_DRAG;
      if (_focus == null) _focus = this; // guich@200b4: make sure that there is always one control with focus. this test was being made in // 1 and // 2

      // guich@200b4: code to move the window.
      if (gMoving != null || (isPenEvent && /*(this instanceof Window) && */rTitle != null && rTitle.contains(x-this.x,y - this.y)))
      {
         switch (type)
         {
            case PenEvent.PEN_DOWN:
               if (canDrag)
                  ptMoving = new Coord(x, y);
               break;
            case PenEvent.PEN_DRAG:
            case PenEvent.PEN_MOVE:
               if (gMoving == null && ptMoving != null && (Math.abs(ptMoving.x - x) > 2 || Math.abs(ptMoving.y - y) > 2))
               {
                  requestFocus(); // tell control: hey, control, time is up!
                  gMoving = MainWindow.myg0;
                  xMoving = this.x;
                  yMoving = this.y;
                  xDeltaMoving = ptMoving.x - xMoving; // guich@200b4_157
                  yDeltaMoving = ptMoving.y - yMoving;
               }
               else if (gMoving != null)
               {
                  gMoving.drawCursorOutline(xMoving, yMoving, this.width, this.height);
                  xMoving = x - xDeltaMoving;
                  yMoving = y - yDeltaMoving;
               }
               if (gMoving != null)
                  gMoving.drawCursorOutline(xMoving, yMoving, this.width, this.height);
               break;
            case PenEvent.PEN_UP:
               if (gMoving == null || this instanceof MainWindow) // didnt drag? invoke menu
                  invokeMenu = true;
               else
               {
                  gMoving.drawCursorOutline(xMoving, yMoving, this.width, this.height);
                  loadBehind(false);
                  this.x = xMoving;
                  this.y = yMoving;
                  this.x2 = this.x+this.width-1;
                  this.y2 = this.y+this.height-1;
                  broadcastEvent(new ControlEvent(ControlEvent.WINDOW_MOVED, this)); // tell other windows that their Graphics object now is invalid. If used, it will draw in a wrong location
                  saveBehind();
                  // repaint at the new position
                  if (doubleBuf)
                  {
                     gMoving.copyRect(imgBuf, 0, 0, width, height, xMoving, yMoving); // x,y -> 0,0
                     onWindowPaintFinished(); // guich@200b4_167: try to correct bug when the Keyboard window is moved and the edit is not repainted correctly
                  }
                  else
                     repaintNow();
                  gMoving = null;
               }
               ptMoving = null;
         }
         if (!invokeMenu) return;
      }

      if (isPenEvent && !contains(x, y)) // guich@102: user clicked outside the window?
      {
         if (!onClickedOutside(x, y)) // if clicked outside was not handled by this method...
            if (type == PenEvent.PEN_DOWN && beepIfOut) // alert him! - ds: i changed this accordingly to your comments about win32 problems
               Sound.beep();
         return;
      }
      if (type == KeyEvent.KEY_PRESS && Vm.actionEqualsMenu && key == IKeys.ACTION) // if the key was action (at WinCE), make it work as menu
         key = IKeys.MENU;
      if (menubar != null && (invokeMenu || (type == KeyEvent.KEY_PRESS && key == IKeys.MENU))) // guich@200b4: popup the menu if was have a title and it was clicked
      {
         popupMenuBar();
         return;
      }

      Event event;
      // guich@200b4_147: make sure that the focused control is an enabled one
      if (!_focus.enabled || _focus.parent == null) // guich@300_55: make always sure that the focused control is enabled; added 2nd condition
      {
         _controlEvent.type = ControlEvent.FOCUS_OUT;
         _controlEvent.target = _focus;
         _controlEvent.touch();
         _focus.postEvent(_controlEvent);

         if (_focus.parent == null)
            _focus = this;
         else
            while (!_focus.enabled && _focus.parent != null)
               _focus = _focus.parent;
      }

      if (type == KeyEvent.KEY_PRESS)
      {
         _keyEvent.type = type;
         _keyEvent.key = key;
         _keyEvent.modifiers = modifiers;
         _keyEvent.touch();
         event = _keyEvent;
      }
      else
      {
         // set focus to new control
         if (type == PenEvent.PEN_DOWN)
         {
            Control c = findChild(x - this.x, y - this.y);
            if (c != _focus)
               setFocus(c);
            tempFocus = c.focusLess ? c : null; // can't receive focus but yes events?
            _inPenDrag = true;
         }
         else if (type == PenEvent.PEN_MOVE && _inPenDrag)
            type = PenEvent.PEN_DRAG;
         else if (type == PenEvent.PEN_UP)
            _inPenDrag = false;
         _penEvent.type = type;
         _penEvent.x = x;
         _penEvent.y = y;

         // translate x, y to coordinate system of target
         Control c = tempFocus != null ? tempFocus : _focus;
         while (c != null)
         {
            _penEvent.x -= c.x;
            _penEvent.y -= c.y;
            c = c.parent;
         }

         _penEvent.modifiers = modifiers;
         _penEvent.touch();
         event = _penEvent;
      }
      event.target = tempFocus != null ? tempFocus : _focus;
      event.timeStamp = timeStamp;
      if (tempFocus != null) // guich@320_31
         tempFocus.postEvent(event);
      else
      if (_focus != null)
         _focus.postEvent(event);
      if (topMost.needsPaint) // guich@200b4_18: maybe the current event had poped up a Window.
         _doPaint();
      if (event.type == PenEvent.PEN_UP && tempFocus != null) // guich@320_31: release tempFocus
         tempFocus = null;
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** returns the client rect, ie, the rect minus the border and title area, in relative coords */
   public Rect getClientRect()
   {
      Rect r = getRect();
      r.x = r.y = 0;
      int my = topBorderGaps[borderStyle];
      int m = borderGaps[borderStyle];
      if (title != null)
      {
         int hh = my + fmH;
         r.modify(m, hh, -(m << 1), -(hh + m));
      }
      else
         r.modify(m, my, -(m << 1), -(m << 1));
      return r;
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** paints the title immediatly. */
   protected void paintTitle(String title, Graphics gg)
   {
      if (title != null || borderStyle > NO_BORDER) // guich@220_48: changed = NO_BORDER by > NO_BORDER to let MenuBar set borderStyle to -1 and thus we don't interfere with its paint
      {
         if (title == null) title = " ";
         int ww = fmTitle.getTextWidth(title);
         int hh = fmTitle.getHeight() + (borderStyle == ROUND_BORDER?2:0);
         int xx = (this.width - ww) >> 1, yy = 1;
         Color f = getForeColor();
         Color b = getBackColor();
         gg.setColors(f);
         if (borderStyle != NO_BORDER)
         {
            int y0 = borderStyle == RECT_BORDER?0:hh;
            if (borderStyle != TAB_ONLY_BORDER && borderStyle != ROUND_BORDER) gg.drawRect(0, y0, this.width, this.height - y0);
            switch (borderStyle)
            {
               case TAB_BORDER:
               case TAB_ONLY_BORDER:
                  gg.setForeColor(f);
                  gg.drawLine(1, 0, ww + 2, 0); // Draws the tab
                  gg.fillRect(0, 1, ww + 4, hh);
                  gg.fillRect(0, hh, width, 2); // Draws the line
                  xx = 3;
                  break;
               case ROUND_BORDER:
                  // guich@121 - uses the new round rect methods
                  gg.fillRoundRect(0, 0, width, height, 3);
                  gg.setBackColor(b);
                  gg.fillRect(2, hh, width - 4, height - hh - 2);
                  gg.setPixel(2, height - 3);
                  gg.setPixel(width - 3, height - 3);
                  break;
               case RECT_BORDER:
               default:
                  gg.fillRect(0, 0, this.width, hh + 2); // black border, white text
                  break;
            }
            gg.setForeColor(b); // draws the text with inversed color - guich@400_40: changed setForeColor to setColors
            gg.setBackColor(f);
         }
         gg.setFont(titleFont);
         gg.drawText(title, xx, yy);
         gg.setFont(font);
         if (rTitle == null)
            rTitle = new Rect(xx-2,0,ww+4,hh+1); // guich@200b4_52
      }
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /**
    * Called by the VM to repaint an area. If doubleBuf is true, paints the whole screen in a buffer and draws it on screen.
    */
   public void _doPaint(int x, int y, int width, int height) // x and y are coords relative to this window
   {
      if (x == -5000) // guich@230_28
      {
         x = 0;
         repaintActiveWindows(0);
         if (Settings.platform.equals("PalmOS"))
         {
            repaint();
            ignoreNextEvent = true; // the user probably clicked in another window, so we ignore that click - guich@320_10: if we do this in WinCE too, when the app starts it will loose the first event.
         }
         return;
      }
      if (x < 0)
         x = 0; // guich@300_21: fix problem when drawing a control before screen limits (try to add a control to -20,10 with doubleBuffer true).
      if (y < 0)
         y = 0;

      if (imgBuf == null && doubleBuf) // guich
         getOffScreen();

      if (_g == null) _g = createGraphics();
      _g.translateTo(this.x, this.y); // reset the translation bc the window could have been moved
      Graphics gg = doubleBuf?gbuf:_g;
      gg.setClip(x, y, width, height);
      // clear background
      gg.setBackColor(backColor); // disabled here?
//      if (flicker && (x != this.x || y != this.y || (width == this.width && height == this.height))) // guich@200b4_53: fix remove of containers not being correctly erased.
      if (eraseBackgroundNow || (flicker && x==0 && y==0 && width == this.width && height == this.height)) // guich@200b4_131
      {
         gg.fillRect(x, y, width, height); // guich@110
         eraseBackgroundNow = false;
      }
      // guich@102: if border or title, draw it
      paintTitle(title, gg);
      onPaint(gg);
      gg.clearClip();
      paintChildren(gg, x, y, width, height);
      _g.setDrawOp(Graphics.DRAW_PAINT);

      if (doubleBuf)
/*         guich@350_1: commented this bc a tabpanel with full width was being painted wrong.
         if (x == 0 && this.asWindow != null && width == Settings.screenWidth) // guich@340_13
            _g.copyScreen(imgBuf, y,y,height); // guich@341_7
         else
         */
            _g.copyRect(imgBuf, x, y, width, height, x, y);

      if (needsPaint)
      {
         if (paintX < 0) paintX = 0;
         if (paintY < 0) paintY = 0; // guich@300_59
         if (paintWidth > this.width) paintWidth = this.width;
         if (paintHeight > this.height) paintHeight = this.height;

         int ax = x + width;
         int ay = y + height;
         int bx = paintX + paintWidth;
         int by = paintY + paintHeight;
         if (x <= paintX && y <= paintY && ax >= bx && ay >= by)
         {
            needsPaint = false;
            onWindowPaintFinished();
            if (_focus != null && _focus.getParentWindow() == this) _focus.onWindowPaintFinished(); // guich@200b4: test if the last focused control belongs to this window; this corrects the painted control after a window is poped up
         }
      }
      if (repaintLastPopup) // guich@200b4_107: this special situation occurs when a window is poped up in the onStart method
      {
         repaintLastPopup = false;
         topMost.repaintNow();
      }
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** Popup a modal window, and make it child of this one. All events in the behind window are deactivated */
   final public void popupModal(Window newWin)
   {
      if (this != topMost) // kambiz@320_18: is the user calling in the wrong way?
         topMost.popupModal(newWin);
      else
      {
         newWin.popped = true;
         focusOnPopup = _focus;
         if (focusOnPopup instanceof waba.ui.PopupMenu || focusOnPopup == null) // guich@200b4_166: case 2 consecutive clicks on the MenuBar; in the 1st, it is the control; in the 2nd, it is the MenuBar
            focusOnPopup = this;

         // the order was : 1. saveBehind, 2. onPopup, 3. repaintNow. But if you click Welcome/Fonts/Uninstall button, the screen will be saved with the button in pressed state; so, we need to release the button first by repainting it
         if (needsPaint
            && focusOnPopup.getParentWindow().isVisible() // guich@200final_4: this condition is necessary to correct when a Button inside a popup window is pressed and this action makes the window close and pop another window, the button was being repainted (when it should not, since its window is no more visible)
            && !(focusOnPopup instanceof Window)) // guich@200final_5: without this line, pressing the button of a combobox and popping up a MessageBox in the on_closed event, makes the ListBox be painted again.
               focusOnPopup.repaintNow(); // example of case where this is needed: a button that calls a popup window. without this line, the button remains pressed
         newWin.saveBehind();
         newWin.onPopup();

         eventsEnabled = false; // disables this window
         zStack.push(topMost = newWin);
         topMost.needsPaint = false; // guich@401_19: fix window flickering that occurs when Edit calls removeTimer.
         setFocus(topMost);
         topMost.needsPaint = true; // guich@401_19: set again needsPaint.
         topMost.eventsEnabled = true; // enable the new window
         topMost.resetPaintArea();
         topMost._doPaint();
         topMost.postPopup();
      }
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** Popup a modal window, blocking the program execution, and make it child of this one. All events in the behind window are deactivated.
       Important! You can't use this method in the application's constructor or in the onStart method!
   */
   final public void popupBlockingModal(Window newWin) // anodos@320_9
   {
      popupModal(newWin);
      do
      {
         pumpEvents();
      } while (newWin.popped);
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** Hides this window. Calling unpop when only the MainWindow is active does nothing. */
   final public void unpop()
   {
      if (zStack.size() == 1) // guich@400_69
         return;
      onUnpop();
      eventsEnabled = false;
      zStack.pop();
      topMost = (Window)zStack.peek();
      if (topMost != null)
      {
         loadBehind(true); // guich@200b4: restore the saved window
         topMost.eventsEnabled = true;
         if (topMost.focusOnPopup instanceof waba.ui.MenuBar)
            topMost.focusOnPopup = topMost; // make sure that the focus is not on the closed menu bar
         if (topMost.focusOnPopup instanceof waba.ui.Window && !topMost.focusOnPopup.asWindow.isVisible()) // guich@200b4_207: fixed a bug when popup menu->select item->popup instance of Window class->popup MessageBox when win closed->select menu->events were wrongly dispatched to the focusOnPopup whom was the win previously popped up and nothing happens
            topMost.focusOnPopup = topMost;
         if (!topMost.focusOnPopup.isDisplayed()) // guich@300_61: if we popped up a MessageBox and swapped the container...
            topMost.focusOnPopup = topMost;
         if (!topMost.focusOnPopup.enabled) // guich@300_62: if the button that dispatched the event is not more enabled...
            topMost.focusOnPopup = topMost;
         Window lastTopMost = topMost;
         topMost.focusOnPopup.postEvent(new ControlEvent(ControlEvent.WINDOW_CLOSED, this)); // tell last control that we closed
         if (topMost == lastTopMost) // guich@240_23: if the postEvent before pops up another Window, we must not set the focus back because the topMost var has changed
            topMost.setFocus(topMost.focusOnPopup);
         postUnpop();
         if (doubleBuf) // guich@200b4_40: free up resources
         {
            //gbuf.free(); guich@200b4_165: this was causing an invalid chunk ptr bc the imgBuf.free below was accessing a freed gbuf (here)
            imgBuf.free();
            imgBuf = null;
            gbuf = null;
         }
         popped = false;
      }
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** Let the VM handle all the events in the event queue.
     * This method is used to implement a blocking Window.
     * Here is an example: <pre>
     * while(someCondition)
     * {
     *    Event.pumpEvents();
     * }
     * </pre>
     */
   public static void pumpEvents()
   {
      NativeMethods4JDK.windowPumpEvents();
   }
   ////////////////////////////////////////////////////////////////////////////////////
   private void resetPaintArea()
   {
      paintX = paintY = 0;
      paintWidth = width;
      paintHeight = height;
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** paints the top-level window. called by MainWindow */
   public void _doPaint()
   {
      if (this != topMost)
         topMost._doPaint();
      else
         _doPaint(paintX, paintY, paintWidth, paintHeight);
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** returns true if this window is the top-level window */
   public boolean isTopMost()
   {
      return this == topMost;
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** returns the topmost window */
   static public Window getTopMost()
   {
      return topMost;
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** Placeholder called imediatly before the popup began. the default implementation does nothing. */
   protected void onPopup()
   {
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** Placeholder called after the popup is done and after the repaint of this window. the default implementation does nothing. */
   protected void postPopup()
   {
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** Placeholder called imediatly before the unpop began. the default implementation does nothing. */
   protected void onUnpop()
   {
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** Placeholder called after the unpop is done and after the repaint of the other window. the default implementation does nothing. */
   protected void postUnpop()
   {
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** true if this is the topmost window */
   public boolean isVisible()
   {
      return this == topMost || parent != null;
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** used to set the position where the status chars will be displayed.
     * the chars are 8x11 pixels. this method must be called so the status can be displayed.
     * these positions must be relative to this window, because they are converted to absolute coordinates.
     * use setStatePosition(Window.HIDE_STATE,Window.HIDE_STATE) to remove the display of the state.
     * Note: in SuperWaba 2.0, this only works on PalmOS >= 3.5.
     */
   public void setStatePosition(int x, int y)
   {
      boolean isVK = x >= VK_HIDE || y >= VK_HIDE;
      if (!isVK) // guich@300_42
      {
         if (x < 0 || x > Settings.screenWidth)  x = HIDE_STATE;
         if (y < 0 || y > Settings.screenHeight) y = HIDE_STATE;
      }
      NativeMethods4JDK.windowSetStatePosition(this, isVK?x:(this.x + x), isVK?y:(this.y + y));
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** sets the menu bar for this window */
   public void setMenuBar(Control menubar)
   {
      this.menubar = menubar;
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** Open the menu bar that is associated with this window */
   protected void popupMenuBar() // guich@200b4_41
   {
      if (menubar != null)
         menubar.setVisible(true);
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** Sets the saveBehindWhenPopup flag to false. If repaintOnClose is true, the window
     * calls repaint on the parent window when it closes. This method is important if you
     * popup a window on the onStart method before the MainWindow is drawed.
     */
   public void dontSaveBehind(boolean repaintOnClose) // guich@200b4_107
   {
      saveBehindWhenPopup = repaintOnClose?2:0;
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** This method tries to save the area behind the window before it is popped up. It is
    * used in the popupModal method so the area can be restored when the window is unpopped.
    * If an out of memory error occurs, the window will be repainted instead of saved.
    * @since SuperWaba2.0beta4
    */
   protected void saveBehind()
   {
      // saves the covered area
      if (saveBehindWhenPopup == 1)
      {
         try
         {
            if (imgCovered == null) imgCovered = new Image(width, height);
            imgCovered.getGraphics().copyRect(MainWindow._mainWindow, x, y, width, height, 0, 0);
         } catch (OutOfMemoryError oome) {saveBehindWhenPopup = 2;} // guich@340_52
      }
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** This method loads the saved area behind of the current window. It is automaticaly used
    * in the unpop method.
    * @param release if true, the memory used is released
    * @since SuperWaba2.0beta4
    */
   protected void loadBehind(boolean release)
   {
      if (saveBehindWhenPopup == 2)
          repaintActiveWindows(release?0:-1); //was: MainWindow._mainWindow.repaintNow();
      else
      if (saveBehindWhenPopup == 1)
      {
         if (imgCovered != null)
            MainWindow.myg0.copyRect(imgCovered, 0, 0, width, height, x, y);
         if (release && imgCovered != null)
         {
            imgCovered.free(); // release memory
            imgCovered = null;
         }
      }
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** this method is for internal use only. calling it directly will cause the VM to crash. */
   public static void destroyZStack()
   {
      zStack.clear();
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** Used to swap containers from the <Main>Window. To restore the first
     * container, pass null as parameter. See samples/ui/ContainerSwitch.
     */
   public void swap(Container newContainer) // guich@300_57: make easier to swap containers
   {
      // store the first container added as the main one
      if (mainSwapContainer == null)
         mainSwapContainer = newContainer;
      // remove the last container
      if (lastSwappedContainer != null)
         remove(lastSwappedContainer);
      // returning back to the main one?
      if (newContainer == null)
         newContainer = mainSwapContainer;
      // add the new container.
      lastSwappedContainer = newContainer;
      add(newContainer);
      if (!newContainer.started) // guich@340_15: if the container did not start yet, set its size
         newContainer.setRect(getClientRect());
      newContainer.requestFocus();
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** Returns the size of the title if any plus the size of the border.
     * Note that the value returned here does not handle the controls inside the
     * window.
     */
   public int getPreferredWidth()
   {
      int wtitle = title == null ? 0 : fmTitle.getTextWidth(title);
      int wborder = (borderStyle == NO_BORDER) ? 0 : (borderStyle == ROUND_BORDER?4:2);
      return wtitle + wborder;
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** Returns the size of the title if any plus the size of the border.
     * Note that the value returned here does not handle the controls inside the
     * window.
     */
   public int getPreferredHeight()
   {
      int htitle = title == null ? 0 : fmTitle.getHeight();
      int hborder = (borderStyle == NO_BORDER) ? 0 : (borderStyle == ROUND_BORDER?4:2);
      return htitle + hborder;
   }
   ////////////////////////////////////////////////////////////////////////////////////
   /** Repaints the window stack from 0 to zStack.size()+upTo.
     * @param upTo needs to be <= 0
     */
   static void repaintActiveWindows(int upTo)
   {
      int n = zStack.getCount()+upTo;
      // guich@400_73 guich@400_76
      Rect mainWindowRect = MainWindow._mainWindow.getRect(); // size of the MainWindow
      int i;
      for (i=n-1; i > 0; i--) // search for the the most top window with the same size of MainWindow - 0=mainwindow, so we skip it
         if (((Window)zStack.items[i]).getRect().equals(mainWindowRect))
            break;
      while (i < n) // repaints every window, from the nearest with the MainWindow size to last parent
         ((Window)zStack.items[i++]).repaintNow();
   }
}
