/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2003 Guilherme Campos Hazan <support@superwaba.com.br>         *
 *  Copyright (C) 2000-2001 Arthur van Hoff                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.util;

/*
* This class is almost identical to java.util.Hashtable, with some
modifications.
*/

/**
* This class implements a hashtable, which maps keys to values. Any
* non-<code>null</code> object can be used as a key or as a value.
* <p>
* An instance of <code>Hashtable</code> has two parameters that
* affect its efficiency: its <i>capacity</i> and its <i>load
* factor</i>. The load factor should be between 0.0 and 1.0. When
* the number of entries in the hashtable exceeds the product of the
* load factor and the current capacity, the capacity is increased by
* calling the <code>rehash</code> method. Larger load factors use
* memory more efficiently, at the expense of larger expected time
* per lookup.
* <p>
* If many entries are to be made into a <code>Hashtable</code>,
* creating it with a sufficiently large capacity may allow the
* entries to be inserted more efficiently than letting it perform
* automatic rehashing as needed to grow the table.
* <p>
* This example creates a hashtable of numbers. It uses the names of
* the numbers as keys:
* <p><blockquote><pre>
*     Hashtable numbers = new Hashtable(10);
*     numbers.put("one", Convert.toString(1));
*     numbers.put("two", Convert.toString(2));
*     numbers.put("three", Convert.toString(3));
* </pre></blockquote>
* <p>
* To retrieve a number, use the following code:
* <p><blockquote><pre>
*     String n = (String)numbers.get("two");
*     if (n != null) {
*         Vm.debug("two = " + Convert.toInt(n));
*     }
* </pre></blockquote>
*
* <b>Important notes</b> (added by guich)
* <p>
* There are no default constructor to decrease memory usage. Also,
* <p>
* thanks to Arthur van Hoff
*/
public final class Hashtable
{
   /** Hashtable collision list. */
   private class Entry
   {
      int hash;
      Object key;
      Object value;
      Entry next;
   }
   /** The hash table data. */
   private Entry table[];
   /** The total number of entries in the hash table. */
   private transient int count;
   /** Rehashes the table when count exceeds this threshold. */
   private int threshold;
   /** The load factor for the hashtable. */
   private float loadFactor;

   /**
   * Constructs a new, empty hashtable with the specified initial capacity
   * and default load factor of 0.75f.
   *
   * @param   initialCapacity   the initial capacity of the hashtable.
   */
   public Hashtable(int initialCapacity)
   {
      this(initialCapacity, 0.75f);
   }
   /**
   * Constructs a new, empty hashtable with the specified initial
   * capacity and the specified load factor.
   * If initialCapacity is zero, it is changed to 5.
   *
   * @param      initialCapacity   the initial capacity of the hashtable.
   * @param      loadFactor        a number between 0.0 and 1.0.
   */
   public Hashtable(int initialCapacity, float loadFactor)
   {
      if (initialCapacity <= 0) initialCapacity = 5; // guich@310_6
      this.loadFactor = loadFactor;
      table = new Entry[initialCapacity];
      threshold = (int)(initialCapacity * loadFactor);
   }
   /**
   * Clears this hashtable so that it contains no keys.
   */
   public void clear()
   {
      Entry tab[] = table;
      for (int index = tab.length; --index >= 0; )
         tab[index] = null;
      count = 0;
   }
   /**
   * Returns the value to which the specified key is mapped in this hashtable.
   *
   * @param   key   a key in the hashtable.
   * @return  the value to which the key is mapped in this hashtable;
   *          <code>null</code> if the key is not mapped to any value in
   *          this hashtable.
   */
   public Object get(Object key)
   {
      int hash = key.hashCode();
      int index = (hash & 0x7FFFFFFF) % table.length;
      for (Entry e = table[index] ; e != null ; e = e.next)
         if ((e.hash == hash) && e.key.equals(key))
            return e.value;
      return null;
   }
   /**
   * Return a Vector of the keys in the Hashtable.
   * Added ds@120.
   * corrected by dgecawich@200
   */
   public Vector getKeys()
   {
      // dgecawich 5/16/01 - fix so that all keys are returned rather than just the last one
      // the sympton for this was that getCount() always returned 1 regardless of how many items were added
      Vector keys = new Vector(size());
      if ( table != null )
      {
         int len = table.length;
         Entry entry = null;
         for ( int i = 0; i < len; i++ )
         {
            entry = table[i];
            if ( entry != null )
            {
               keys.add(entry.key);
               // bagfors@200b3: get the stranglers. there's always one for some reason
               while(entry.next != null)
               {
                  entry = entry.next;
                  keys.add(entry.key);
               }
            }
         }
      }
      return keys;
   }

   /**
   * Maps the specified <code>key</code> to the specified
   * <code>value</code> in this hashtable. Neither the key nor the
   * value can be <code>null</code>.
   * <p>
   * The value can be retrieved by calling the <code>get</code> method
   * with a key that is equal to the original key.
   *
   * @param      key     the hashtable key.
   * @param      value   the value.
   * @return     the previous value of the specified key in this hashtable,
   *             or <code>null</code> if it did not have one.
   * @see     java.lang.Object#equals(java.lang.Object)
   */
   public Object put(Object key, Object value)
   {
      // Make sure the value is not null
      if (key == null || value == null) // guich@200b2: included the key in this test
         return null;
      // Makes sure the key is not already in the hashtable.
      Entry tab[] = table;
      int hash = key.hashCode();
      int index = (hash & 0x7FFFFFFF) % tab.length;
      for (Entry e = tab[index] ; e != null ; e = e.next)
         if ((e.hash == hash) && e.key.equals(key))
         {
            Object old = e.value;
            e.value = value;
            return old;
         }
      if (count >= threshold)
      {
         // Rehash the table if the threshold is exceeded
         rehash();
         return put(key, value);
      }

      // Creates the new entry.
      Entry e = new Entry();
      e.hash = hash;
      e.key = key;
      e.value = value;
      e.next = tab[index];
      tab[index] = e;
      count++;
      return null;
   }
   /**
   * Rehashes the contents of the hashtable into a hashtable with a
   * larger capacity. This method is called automatically when the
   * number of keys in the hashtable exceeds this hashtable's capacity
   * and load factor.
   */
   protected void rehash()
   {
      int oldCapacity = table.length;
      Entry oldTable[] = table;

      int newCapacity = (((oldCapacity << 1) + oldCapacity) >> 1) + 1; // guich@120 - grows 50% instead of 100% - guich@200b4_198: added Peter Dickerson and Andrew Chitty changes to correct the optimization i made with << and >>
      Entry newTable[] = new Entry[newCapacity];

      threshold = (int)(newCapacity * loadFactor);
      table = newTable;

      for (int i = oldCapacity ; i-- > 0 ;)
         for (Entry old = oldTable[i] ; old != null ; )
         {
            Entry e = old;
            old = old.next;

            int index = (e.hash & 0x7FFFFFFF) % newCapacity;
            e.next = newTable[index];
            newTable[index] = e;
         }
   }
   /**
   * Removes the key (and its corresponding value) from this
   * hashtable. This method does nothing if the key is not in the hashtable.
   *
   * @param   key   the key that needs to be removed.
   * @return  the value to which the key had been mapped in this hashtable,
   *          or <code>null</code> if the key did not have a mapping.
   */
   public Object remove(Object key)
   {
      Entry tab[] = table;
      int hash = key.hashCode();
      int index = (hash & 0x7FFFFFFF) % tab.length;
      for (Entry e = tab[index], prev = null ; e != null ; prev = e, e = e.next)
         if ((e.hash == hash) && e.key.equals(key))
         {
            if (prev != null)
               prev.next = e.next;
            else
               tab[index] = e.next;
            count--;
            return e.value;
         }
      return null;
   }
   /**
   * Returns the number of keys in this hashtable.
   *
   * @return  the number of keys in this hashtable.
   */
   public int size()
   {
      return count;
   }
   /**
   * Returns a rather long string representation of this hashtable.
   *
   * @return  a string representation of this hashtable.
   */
/*   public String toStrin g()
   {
      return "This hashtable has: "+count+" elements and threshold of "+threshold;
   }*/
}