/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2000-2003 Guilherme Campos Hazan <support@superwaba.com.br>    *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.util;

import waba.sys.*;
import waba.io.*;

/**
 * An int vector is an array of int's. The vector grows and shrinks
 * dynamically as ints are added and removed.
 * <p>
 * Here is an example showing a vector being used:
 *
 * <pre>
 * ...
 * IntVector vec = new IntVector();
 * vec.add(int1);
 * vec.add(int22);
 * ...
 * vec.insert(3, int3);
 * vec.del(2);
 * if (vec.getCount() > 5)
 * ...
 * </pre>
 * <br>For efficiency, get and set is made directly through the <items> public array. <b>please use add and del to insert and remove elements</b>
 * @author created by guich from waba.util.Vector.
 */
public final class IntVector
{
   public static final int INVALID = -2147483648;
   public int items[];
   private int count;

   /** Constructs an empty vector. */
   public IntVector()
   {
      this(20);
   }
   /** Constructs an vector read from the stream */
   public IntVector(DataStream in)
   {
      count = (int)in.readUnsignedShort(); // guich@400_5
      items = new int[count+5];
      for (int i =0; i < count; i++)
         items[i] = in.readInt();
   }

   /**
    * Constructs an empty vector with a given initial size. The size is
    * the initial size of the vector's internal int array. The vector
    * will grow as needed when ints are added. SIZE CANNOT BE 0!
    */
   public IntVector(int size)
   {
      items = new int[size];
   }

   /** Adds an int to the end of the vector. */
   public void add(int obj)
   {
      if (count < items.length)
         items[count++] = obj;
      else
         insert(count, obj);
   }

   /** Inserts an int at the given index. */
   public void insert(int index, int obj)
   {
      if (index < 0 || index > count) index = count; // guich@200b3: check if index is valid
      if (count == items.length)
      {
         // grows 20%
         int newItems[] = new int[(items.length * 12) / 10 +1];
         Vm.copyArray(items, 0, newItems, 0, count);
         items = newItems;
      }
      if (index != count)
         Vm.copyArray(items, index, items, index + 1, count - index);
      items[index] = obj;
      count++;
   }

   /** Deletes the int reference at the given index. */
   public void del(int index)
   {
      if (index != count - 1)
         Vm.copyArray(items, index + 1, items, index, count - index - 1);
      items[count - 1] = 0;
      count--;
   }

   /**
   * Finds the index of the given int. The list is searched using a O(n) linear
   * search through all the ints in the vector.
   * @return -1 if the object is not found.
   */
   public int find(int obj)
   {
      return find(obj,0);
   }
   /**
   * Finds the index of the given int. The list is searched using a O(n) linear
   * search starting in startIndex up through all the ints in the vector.
   * @return -1 if the object is not found.
   */
   public int find(int obj, int startIndex)
   {
      for (int i = startIndex; i < count; i++)
         if (items[i] == obj)
            return i;
      return -1;
   }
   /**
    * Finds the index of the given int. The list is searched using a O(n) linear
    * search through all the ints in the vector.
    */

   /** Returns the number of ints in the vector. */
   public int getCount()
   {
      return count;
   }

   /** writes this int vector to the stream */
   public void writeTo(DataStream out)
   {
      out.writeShort(count);
      for (int i=0; i < count; i++)
         out.writeInt(items[i]);
   }
   /** Useful method to use when this IntVector will act like a bit vector, through
     * the methods <code>isBitSet</code> and <code>setBit</code>. Just call it
     * with the maximum bit index what will be used (starting from 0); then
     * you can safely use the two methods. This must be done because those methods
     * does not check the bounds of the array.
     */
   public void ensureBit(int index) // guich@400_5
   {
      int newCount = (index >> 5)+1; // convert from bits to int
      if (newCount >= items.length)
      {
         if (count == 0) // no items yet? just assign a new vector
            items = new int[newCount];
         else
         {
            int newItems[] = new int[newCount];
            Vm.copyArray(items, 0, newItems, 0, count);
            items = newItems;
         }
      }
      count = newCount;
   }
   /** Used to let this int vector act like a bit vector. returns true if the bit specified is set.
   you must guarantee that the index exists in the vector. guich@102 */
   public boolean isBitSet(int index) // guich@400_5
   {
      return (items[index>>5] & ((int)1 << (index & 31))) != 0; // guich@321_7
   }

   /** Used to let this int vector act like a bit vector. returns true if the bit specified is set.
   you must guarantee that the index exists in the vector. guich@102 */
   public void setBit(int index, boolean on) // guich@400_5
   {
      if (on)
         items[index>>5] |= ((int)1 << (index & 31));  // set
      else
         items[index>>5] &= ~((int)1 << (index & 31)); // reset
   }

   // methods to let the vector act like a stack

   /** pushes a int. simply calls add. */ // guich@102
   public void push(int obj)
   {
      add(obj);
   }
   /** returns the last int, removing it. returns INVALID if no more elements. */ // guich@102
   public int pop()
   {
      int o=INVALID; // guich@200final_1: return invalid instead of 0
      if (count > 0)
         o = items[--count];
      return o;
   }
   /** returns the last int, without removing it. returns INVALID if no more elements. */ // guich@102
   public int peek()
   {
      return (count > 0)?items[count-1]:INVALID;     // guich@200final_1: return invalid instead of 0
   }

   /** clears all elements in this vector and sets its length to 0 */
   public void clear()
   {
      for (int i =0; i < count; i++)
         items[i] = 0;
      count = 0;
   }

   //// guich@200b2: ok, i hate to do this, but here they are! the methods to make the vector class compatible with JDK. This incompatibility is very annoying!
   /** same of getCount() */
   public int size()
   {
      return count;
   }
   /** same of find(int) */
   public int indexOf(int elem)
   {
      return find(elem, 0);
   }
   /** same of find(int, index) */
   public int indexOf(int elem, int index)
   {
      return find(elem, index);
   }
   /** same of items[index] */
   public int elementAt(int index)
   {
      return items[index];
   }
   /** Set the element at the given index. */
   public void setElementAt(int obj, int index)
   {
      items[index] = obj;
   }
   /** same of del(index) */
   public void removeElementAt(int index)
   {
      del(index);
   }
   /** same of insert(index, int) */
   public void insertElementAt(int obj, int index)
   {
      insert(index,obj);
   }
   /** same of add(int) */
   public void addElement(int obj)
   {
      add(obj);
   }
   /** same of del(int) */
   public void removeElement(int obj)
   {
      del(obj);
   }
   /** same of clear() */
   void removeAllElements()
   {
      clear();
   }

   /** Do a quick sort in the elements of this IntVector */
   public void qsort()
   {
      qsort(0, count-1);
   }
   // ps: to preserve memory, the "int items[]" is accessed directly
   private void qsort(/*int items[], */int first, int last)
   {
      int low = first;
      int high = last;
      if (first >= last)
         return;
      int mid = items[(first+last) >> 1];
      do
      {
         while (items[low] < mid)
            low++;
         while (items[high] > mid)
            high--;
         if (low <= high)
         {
            int temp = items[low];
            items[low++] = items[high];
            items[high--] = temp;
         }
      } while (low <= high);
      qsort(first,high);
      qsort(low,last);
   }
}