/*********************************************************************************
 *  SuperWaba Virtual Machine, version 4                                         *
 *  Copyright (C) 2003 Guilherme Campos Hazan <support@superwaba.com.br>         *
 *  Copyright (C) 2001 Sean Luke                                                 *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  For the purposes of the SuperWaba software we request that software using    *
 *  or linking to the SuperWaba virtual machine or its libraries display the     *
 *  following notice:                                                            *
 *                                                                               *
 *                   Created with SuperWaba                                      *
 *                  http://www.superwaba.org                                     *
 *                                                                               *
 *  Please see the software license located at SuperWabaSDK/license.txt          *
 *  for more details.                                                            *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     Guilherme Campos Hazan                                                    *
 *     Av. Nossa Senhora de Copacabana 728 apto 605 - Copacabana                 *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22050-000                                                            *
 *     E-mail: support@superwaba.com.br                                          *
 *                                                                               *
 *********************************************************************************/

package waba.util;

// Sean Luke <seanl@cs.umd.edu>  6/3/01

/** This is a simple Linear Congruential Generator which
    produces random numbers in the range [0,2^31), derived
    from ran0 in Numerical Recipies.  Note that ran0 isn't
    wonderfully random -- there are much better generators
    out there -- but it'll do the job, and it's fast and
    has low memory consumption.  */

public class Random
{
   // Do NOT modify IA, IM, IQ, or IR
   // You are free to set ZERO_SEED_REPLACEMENT to any integer
   // greater than 0 (but NOT including 0!!)
   public static final int IA = 16807;
   public static final int IM = 2147483647;
   public static final int IQ = 127773;
   public static final int IR = 2836;
   public static final int ZERO_SEED_REPLACEMENT = 234123;  // whatever

   private int seed;

   /** The only reasonable seeds are between 0 and 2^31 inclusive; if
       you're negative, it'll get the absolute value of it. */
   public Random(int _seed)
   {
      // strip out the minus-sign if any
      seed = (_seed << 1) >>> 1;
      // if seed is 0, it's invalid, we need another seed
      if (seed==0) seed = ZERO_SEED_REPLACEMENT;
   }

   /** Randomizes based on the current timestamp */
   public Random()
   {
      waba.sys.Time t = new waba.sys.Time(); // guich@300_30
      seed = t.hour*60*60 + t.minute*60 + t.second + 100*t.millis; // hope this value never gets < 0 !
   }

  /*  -- a useful debugging function
   static void printBits(int seed)
   {
      String s = "";
      for (int x=31; x>=0; x--)
         s = s + ((seed >>> x) % 2 == 1 ? "1" : "0");
      Vm.debug(s);
   }
  */

   /** bits should be <= 31 */
   protected int next(final int bits)
   {
      int k = seed/IQ;
      seed = IA*(seed-k*IQ)-IR*k;
      if (seed < 0) seed += IM;
      return seed >> (31-bits);
   }

   /** Returns a floating-point value in the half-open range [0,1). */
   public float nextFloat()
   {
      return next(24) / ((float)(1 << 24));
   }

   /** Returns an integer in the range [0,n).
       n must be > 0, else -1 is returned. */
   public int nextInt(int n)
   {
      if (n<=0) return -1;
      int bits, val;
      do
      {
         bits = next(31);
         val = bits % n;
      }
      while(bits - val + (n-1) < 0);
      return val;
   }

   // SW. for compatibility with earlier Random
   public int getRandom(int n)
   {
      return nextInt(n);
   }

   // guich@200b4_3
   /** Returns either random character in the given region.
     * Eg: rand('a', 'e') returns a random character in the range of 'a' to 'e'.
     * Added by Imran Zahid
     * @param s the start character in the range
     * @param e the ending character in the range
     */
	public char rand(char s, char e)
	{
		return (char)rand((int)(s), (int)(e));
	}

   // guich@200b4_3
   /** Returns either random integer in the given region.
     * Eg: rand(1, 10) returns a random integer in the range of 1 to 10.
     * Added by Imran Zahid
     * @param s the start number in the range
     * @param e the ending number in the range
     */
	public int rand(int s, int e)
	{
		if (s > e)
		{
			int t = e;
			e = s; s = t;
		} else
		if (s == e) e++;

		while(true)
		{
			int rnd = nextInt(e);
			if (s <= rnd && rnd <= e) return (char)rnd;
		}
	}
}