package superwaba.ext.xplat.ui;

import waba.ui.*;
import waba.fx.*;

/**
 * A light variant of the standard Grid.<br>
 * Demands a table model (GridModel) to work<br>
 * Supports editing (of course), selection by row, column or single cell and toggle of borders<br>
 * Note: this grid does not support scrollbars. If you add support for
 * them, please share your enhancements with the community: send the sources to classes@superwaba.org
 * <p>
 *
 * Ver 1.0.2 With a quicker update when selecting/editing<br>
 * Ver 1.0.3. Lots of bugfixes,including the keyboard error
 * @see superwaba.ext.xplat.ui.GridModel
 * @author Stein Rustad (stein@politikontroll.org)
 */

/* Hi,

I have written a light version of JTable for SuperWaba. I wrote it
cause I needed it for a project, but anyone who wants can use it.

It supports editing, headers, selection by cell, row or column and
toggle of borders. It does not support scrolling. It does, unlike the
original JTable, require a TableModel to be used.

Performance is acceptable but not exactly great, so if anyone wants to
improve or even take over development of the table, please contact me
at stein@politikontroll.org

The code can also be downloaded from: http://waba.politikontroll.org

regards
Stein Rustad
*/

public class Grid extends Container {

	/** The datamodel that this table uses to interrogate for values */
    private GridModel dataModel;
    private Graphics g;
    /** Specifies if the table has drawn borders. Default is true. */
    private boolean bordered = true;
    /** Spacing between cell content and borders (shown or not) */
    private int space = 2; // should be both vertical and horizontal

    /** Specifies the width of all columns */
    private int[] colWidth;
    /** Specifies the height of all rows */
    private int[] rowHeight;

    // Editing stuff
    /** The column index of the cell currently edited */
    private int editCol = -1;
    /** The row    index of the cell currently edited */
    private int editRow = -1;
    private boolean isEditing = false;
    /** The edit used in cell editing */
    private Edit edit = null;

    // Font stuff
    /** The font used in this table, the normal font */
    private Font orgFont;
    /** The font metrics used in this table */
    private FontMetrics fm;

    // Selecting stuff
    /** Constant used to specify selection by row */
    public static final int SELECT_BY_ROW = 0;
    /** Constant used to specify selection by column */
    public static final int SELECT_BY_COLUMN = 1;
    /** Constant used to specify selection by cell */
    public static final int SELECT_BY_CELL = 2;
    /** Specifies the current selection mode */
    private int selectBy = SELECT_BY_CELL;
    /** The row index of the currently selected cell */
    private int selRow = -1;
    /** The column index of the currently selected cell */
    private int selCol = -1;

    /** Indicates which cell should be repainted (contents changed) */
    boolean[][] dirtyCell;
    /** Indicates if we can do a simple repaint */
    boolean simpleRepaint = false;

    /** Lets the user edit a cell without selecting it first. */
    private boolean oneClickEdit=false;

/**
 * Constructs a Grid that is initialized with dm as the data model.
 *
 * @param dm The tablemodel
 */

public Grid(GridModel dm) {
    setTableModel(dm);
}
/**
 * Causes this table to lay out its rows and columns.
 * Quite complex, and time consuming
*/

private void doLayout() {
    int cols = dataModel.getColumnCount();
    int rows = dataModel.getRowCount();

    colWidth = new int[cols];
    rowHeight = new int[rows];

    // Set up colours
    g.setForeColor(foreColor);
    Color orgBackColor = backColor; // Save it for select use
    g.setBackColor(orgBackColor);

    if (!simpleRepaint)
        g.fillRect(0, 0, getRect().width, getRect().height);

    int currentX = 0; // Current horizontal position
    int currentY = 0; // Current vertical postition

    // The current font metrics
    orgFont = new Font(getFont().getName(), getFont().getStyle(), Font.PLAIN);
    Font boldFont = orgFont.asBold();
    FontMetrics pfm = getFontMetrics(orgFont);
    FontMetrics bfm = getFontMetrics(boldFont);
    fm = pfm;

    // Lay out the colums and rows
    for (int c = 0; c < cols; c++) {
        int prefColWidth = colWidth[c] = dataModel.getPreferredColumnWidth(c);

        for (int r = 0; r < rows; r++) { // First run through rows, to check sizes
            if (dataModel.isHeader(r)) {
                g.setFont(boldFont);
                fm = bfm;
            } else {
                g.setFont(orgFont);
                fm = pfm;
            }

            rowHeight[r] = dataModel.getPreferredRowHeight(r);

            // Getting the value
            String text = (String) dataModel.getValueAt(r, c);
            if (text == null)
                text = "";

            // Checking size
            if (!text.equals(""))
                text = shrinkText(prefColWidth, text); // Shrink the text if it is too long
            if (fm.getTextWidth(text) > colWidth[c])
                colWidth[c] = fm.getTextWidth(text);
            rowHeight[r] = dataModel.getPreferredRowHeight(r);
            if (rowHeight[r] < 0)
                rowHeight[r] = fm.getHeight();
        } // For each row

        for (int r = 0; r < rows; r++) { // Second run through the rows, paint background (if any) and texts
            drawValue : {
                if (!dirtyCell[c][r])
                    break drawValue;
                if (dataModel.isHeader(r)) {
                    g.setFont(boldFont);
                } else {
                    g.setFont(orgFont);
                }

                // Getting the value
                String text = (String) dataModel.getValueAt(r, c);
                if (text == null)
                    text = "";
                if (!text.equals(""))
                    text = shrinkText(prefColWidth, text); // Shrink the text if it is too long

                // Fill in background
                if (isSelected(r, c)) { // If the cell is selected, paint it darker
                    g.setBackColor(Color.DARK); // The step value seems to have a different effect on different systems. Why?
                    g.fillRect(currentX + 1, currentY + 1, colWidth[c] /*+ space * 2*/, rowHeight[r] + space * 2);
                    g.setBackColor(orgBackColor); // Not really necessary
                } else {
                    g.fillRect(currentX + 1, currentY + 1, colWidth[c] + space * 2, rowHeight[r] + space * 2);
                }

                // Draw horizontal lines
                if (bordered)
                    g.drawLine(currentX, currentY + rowHeight[r] + space * 2, currentX + colWidth[c] /*+space*2*/, currentY + rowHeight[r] + space * 2);

                // Draw the value
                g.drawText(text, currentX + space, currentY + space);

                // Cell painted
                dirtyCell[c][r] = false;
            }

            // Dropping down one row
            currentY += rowHeight[r] + space * 2;
        } // For each row

        int colHeight = currentY;
        currentY = 0; // resetting to start a new column
        currentX += colWidth[c] + /*space*2*/
        0; // Jumping forward one column

        // Draw lines
        if (bordered) {
            // First horizontal line
            if (c == cols - 1)
                g.drawLine(0, 0, currentX, 0);
            // Vertical lines
            if (c == 0)
                g.drawLine(0, 0, 0, colHeight);
            g.drawLine(currentX, currentY, currentX, colHeight);
        }

    } // For each column
}
  /**
   * Returns if the table is bordered or not
   *
   * @return true if the table is bordered
  */

  public boolean getBordered() {
	return bordered;
  }
/**
 * Returns the coordinates of the cell clicked.
 *
 * @param coordx The requested cell will contain this x (horizontal) coordinate
 * @param coordy The requested cell will contain this y (vertical) coordinate
 * @return the rectangle specifying the location of the cell
*/

private Rect getCellRect(int coordx, int coordy) {
    if (!contains(x, y)) {
        editCol = -1;
        editRow = -1;
        return null;
    } // The contains returns true for -1,-1, why??
    Rect cellRect = new Rect(0, 0, 0, 0);
    cellRect.x=space*2;
    for (int x = 0; x < colWidth.length; x++) {
        cellRect.x += colWidth[x] /*+space * 2*/;
        if (cellRect.x >= coordx) {

            cellRect.x -= colWidth[x] + space * 2;
            cellRect.width = colWidth[x] /*+ space * 2*/ + 1;
            editCol = x;
            break;
        }
    }
    for (int y = 0; y < rowHeight.length; y++) {
        cellRect.y += rowHeight[y] + space * 2;
        if (cellRect.y >= coordy) {
            cellRect.y -= rowHeight[y] + space * 2;
            cellRect.height = rowHeight[y] + space * 2 + 1;
            editRow = y;
            break;
        }
    }
    return cellRect;
}
/**
 * Indicates if a user is allowed to start editing a cell without selecting it first.
 * @return true if one click editing is allowed
 */
public boolean getOneClickEditMode() {
	return oneClickEdit;
}
  /**
   * Returns the currently selected item in the table
   *
   * @return the value that is associated with the currently selected cell
  */

  public Object getSelected() {
	if(selRow == -1 || selCol == -1) return null;
	return dataModel.getSelected(selRow,selCol);
  }
/**
 * Returns the currently selected row
 * @return the row that is selected
 */
public int getSelectedColumn() {
	return selCol;
}
/**
 * Returns the currently selected row
 * @return the row that is selected
 */
public int getSelectedRow() {
	return selRow;
}
  /**
   * Returns the mode of selection.
   * Value will be one of 3 legal values: SELECT_BY_CELL, SELECT_BY_ROW, SELECT_BY_COLUMN
   *
   * @return the mode of selection
   * @see #SELECT_BY_CELL
   * @see #SELECT_BY_ROW
   * @see #SELECT_BY_COLUMN
  */

  public int getSelectMode() {
	return selectBy;
  }
  /**
   * Returns the table model
   *
   * @return the tablemodel for this table
  */

  public GridModel getTableModel() {
	return dataModel;
  }
  /**
   * Returns if the cell is selected.
   *
   * @param rowIndex The row index of the cell
   * @param columnIndex The column index of the cell
   * @return true if the quired cell is selected
  */

  private boolean isSelected(int rowIndex, int columnIndex) {
	switch(getSelectMode()) {
		case SELECT_BY_ROW:
			return (rowIndex == selRow);
		case SELECT_BY_COLUMN:
			return (columnIndex == selCol);
		case SELECT_BY_CELL:
			return (rowIndex == selRow && columnIndex == selCol);
		default:
			return false;
	} // switch
  }
/**
 * Handels events passed to the table
 *
 * @param event the event
 */

public void onEvent(Event event) {
    if (event.target == edit) { // If we are loosing focus, stop editing
        if (event.type == ControlEvent.FOCUS_OUT) {
            dataModel.setValueAt(edit.getText(), editRow, editCol);
            remove(edit);
            //edit = null;
            isEditing = false;
            simpleRepaint = true;
            dirtyCell[editCol][editRow] = true;
            repaintNow(); // guich@330_12
        } else
            return;
    }

    if (event.type == PenEvent.PEN_UP) { // The user has done something
        if (isEditing) { // Leave editing and set data in the datamodel
            dataModel.setValueAt(edit.getText(), editRow, editCol);
            remove(edit);
            //edit = null;
            isEditing = false;
            simpleRepaint = true;
            dirtyCell[editCol][editRow] = true;
        } // if isEditing

        // Get the rectangle of the selected cell
        Rect cellRect = getCellRect(((PenEvent) event).x, ((PenEvent) event).y);
        if(oneClickEdit) {
			select(((PenEvent) event).x, ((PenEvent) event).y);
        }
        if (isSelected(editRow, editCol)) { // The cells row/and or column is selected, start editing
            // Set up an edit field and let the user edit it
            if (!dataModel.isCellEditable(editRow, editCol))
                return;
            if (edit == null)
            {
                edit = new Edit();
                edit.setKeyboard(Edit.KBD_NONE);
            }
            edit.setRect(cellRect);
            String validChars = dataModel.getValidChars(editRow, editCol);
            if (validChars != null && validChars != "")
                edit.setValidChars(validChars);
            String text = (String) dataModel.getValueAt(editRow, editCol);
            if (text == null)
                text = "";
            edit.setText(text);
            add(edit);
            isEditing = true;
            edit.requestFocus();
        } else
            select(((PenEvent) event).x, ((PenEvent) event).y);

    } // if PEN_UP
} // onEvent
/**
 * Called by the system to draw the table, will lay out the rows and columns.
 *
 * @param g The Graphic object to paint on
*/

public void onPaint(Graphics g) {
    if (dataModel == null)
        return; // Do nothing if the tablemodel aint set
    this.g = g;
    if (simpleRepaint) {
        doLayout();
        simpleRepaint = false;
    } else {
        setAll(true);
        doLayout();
    }
}
/**
 * Selects the cell at the current cordinates.
 *
 * @param coordx The cell that is to be selected will contain this x (horizontal) coordinate
 * @param coordy The cell that is to be selected will contain this y (vertical) coordinate
*/

private void select(int coordx, int coordy) {
	int oldSelRow=selRow;
	int oldSelCol=selCol;

	getCellRect(coordx, coordy); // Use the edit vars a bit
    selRow = editRow;
    selCol = editCol;

    if(oldSelRow==selRow && oldSelCol==selCol) return; // Same cell selected (not good this one)

    // Repaint the last cell selected, to make it the right backColor
    if (oldSelRow != -1 && oldSelCol != -1) { // If something is previously selected
        switch (getSelectMode()) {
            case SELECT_BY_ROW :
            	for(int i=0;i<dirtyCell.length;i++) { dirtyCell[i][oldSelRow]=true; }
                break;
            case SELECT_BY_COLUMN :
            	for(int i=0;i<dataModel.getRowCount();i++) { dirtyCell[oldSelCol][i]=true; }
                break;
            default :
            	dirtyCell[oldSelCol][oldSelRow]=true;
                break;

        } // switch
    }

    switch (getSelectMode()) {
        case SELECT_BY_ROW :
        	for(int i=0;i<dirtyCell.length;i++) { dirtyCell[i][selRow]=true; }
            break;
        case SELECT_BY_COLUMN :
        	for(int i=0;i<dirtyCell.length;i++) { dirtyCell[selCol][i]=true; }
            break;
        default :
        	dirtyCell[selCol][selRow]=true;
			break;
    } // switch

    simpleRepaint=true;
    repaint();
}
/**
 * Fills the dirtyCell table with a value (used for repainting the entire table)
 * Creation date: (15.04.02 10:50:03)
 * @param dirty boolean
 */
private void setAll(boolean dirty) {
	for(int i=0;i<dirtyCell.length;i++) {
		for(int j=0;j<dataModel.getRowCount();j++) {
			dirtyCell[i][j]=dirty;
		}
	}
}
  /**
   * Sets if the table should be bordered or not.
   *
   * @param bordered If true, the table is drawn with lines between cells
  */

  public void setBordered(boolean bordered) {
	this.bordered = bordered;
  }
/**
 * Sets if a user is allowed to start editing a cell without selecting it first.<br>
 * When the user clicks a cell, it is selected for her.<br>
 * Default is false.
 * @param oneClickEditMode if true, one click editing is allowed
 */
public void setOneClickEditMode(boolean oneClickEditMode) {
	oneClickEdit = oneClickEditMode;
}
  /**
   * Sets the table's selection mode to allow for selection by single cell, row or column.
   * Default mode is SELECT_BY_ROW
   *
   * @param selMode One of 3 legal values: SELECT_BY_CELL, SELECT_BY_ROW, SELECT_BY_COLUMN
   * @see #SELECT_BY_CELL
   * @see #SELECT_BY_ROW
   * @see #SELECT_BY_COLUMN
  */

  public void setSelectMode(int selMode) {
	if(selMode<0 || selMode>2) return;
	selectBy = selMode;
  }
  /**
   * Sets the spacing between texts.
   *
   * Note: Future versions should support seperate vertical and horizontal spacing attributes
   *
   * @param space The space in pixels beween the text and the cell border
  */

  public void setSpace(int space) {
	this.space = space;
  }
  /**
   * Set the table model
   * @param newModel GridModel
   */

  public void setTableModel(GridModel newModel) {
	if(newModel == dataModel) return; // Same model, no need to repaint
	dataModel = newModel;
	if(dataModel==null) return;
	dirtyCell = new boolean[dataModel.getColumnCount()][dataModel.getRowCount()];
	setAll(true); // All cells are dirty at first
  }
  /**
   * If the prefered size is shorter than the text size, shrink the text.
   * NOTE: Should maybe add something to the end to signify that the text has been cropped (...)
   *
   * @param prefWidth the prefered width for the cell
   * @param text the text that is to fit into the cell
   * @return the (cropped) text
  */

  private String shrinkText(int prefWidth, String text) {
	String dots="..";
	int dot_width = fm.getTextWidth(dots);
	 int temp=0;
	boolean text_shrunk=false;
	if(prefWidth>0) { while((temp=fm.getTextWidth(text))>prefWidth-(text_shrunk?dot_width:0)-space*2) { text_shrunk=true; text=text.substring(0,text.length()-1); } }
	if(text_shrunk) text+=dots;
	temp=fm.getTextWidth(text);
	return text;
  }
}

